#!/usr/bin/perl -w
######
#
#   UnSplicer, RNA-seq alignment with DNA sequence model assistance
#
#   Paul D. Burns (paul.burns@gatech.edu)
#   Yang Li (yangli9@illinois.edu)
#
#
######
use Getopt::Long ;
use Getopt::Long qw(:config no_ignore_case);
use POSIX;
use FindBin qw($Bin);
use Parallel::ForkManager;
use Cwd 'abs_path';

$clean = 1;
$class_options = "sldht";
$filter_pair = 1;

GetOptions(
	#MANDATORY
	'f=s{1,}' => \@Fq, 
	'r|bowtie-index=s'=>\$bwt,
	'mod|model-params=s' => \$mod,
	'genes|gene-predictions=s' => \$gff_ref,
	#OPTIONAL	
	'p|thread=i'=>\$P,
	'ft|feat=s' =>\$class_options,
	's|segment-length=i'=>\$segment_length, 
	'v|mismatch=i'=>\$V, 
	'e|filter-pair=i'=>\$filter_pair, 
	'm=i'=>\$M, 
	'C|clean_temp=i'=>\$clean,
	'h|help' =>\$help,
	'segment-name'=>\$report_read_name); ## template name or original read name

#################################
$VERSION = 0.9;
#################################


if(scalar @Fq == 0 || $help || $bwt eq "")
{
	if(defined $bwt && ! -e "$bwt.fa"){
		print "$bwt.fa does not exist\n";
	}
	print "UnSplicer paired-end v$VERSION\nUsage:\n
   -f                        [MANDATORY] input file name, fq_1 fq_2; if there are multiple files for fq_1/fq_2, they should be seperated by comma; All reads should have the same length and >= 36bp
   -r/--bowtie-index         [MANDATORY] bowtie reference dir/name, thus bowtie index should be dir/name.*.ebwt, reference genome should also be located here, e.g. dir/name.fa; chromosome line for dir/name.fa should be clean and there is no space within it, such as \">XXXXX\\n\"
   -mod/--model-params       [MANDATORY] folder containing model parameter file(s) (created by GeneMark-ES)
   -genes/--gene-predictions [MANDATORY] gene predictions (GFF3 format, can be created by GeneMark-ES)
   -p/--thread               number of cores
   -v/--mismatch             mismatches, 0-3 (default = 2)
   -e/--filter-pair <0|1>    filter out paired ends which do not map on opposite strands (default = 1)
   -s/--segment-length       segment length, 18-25 (default = 25)
   -h/--help
   --segment-name
   -ft/--features <str>      list of features to use for discrimination (default: sldht)
       choose any non-empty subset:
            s: splice site score (donor and acceptor PSFM)
            l: intron length distribution
            d: mapping depth
            h: frame shift error
            t: strand concordance
	\n\n";

	exit(0);
}

sub doSvm($$$$$$);

print "sequence : $bwt.fa\n";
print "reads(1) : $Fq[0]\n";
print "reads(2) : $Fq[1]\n";
print "mod dir  : $mod\n";
print "genes    : $gff_ref\n";
print "\n";


$NOW = time;

$now_string = localtime;
$bwt = abs_path($bwt);


print "$now_string:   ";
print "UnSplicer paired-end v$VERSION\n";

$now_string = localtime;
print "$now_string:   ";
print "Checking configurations\n";

##
## input argument checking
##

$intron_min = $intron_min || 20;
$intron_max = $intron_max || 50000;

$segment_length = $segment_length||25 ; ## length of segments 18-25
$clean = $clean || 0;
$cov = $cov || 0;
$V = $V || 2;
$M = $M || 5;
$report_read_name = $report_read_name || 0;
$canonical_flag = $canonical_flag || 0;

$OUT_DIR = $OUT_DIR || "unsplicer_out";

$base_dir = abs_path($OUT_DIR);

print "OUT_DIR = $OUT_DIR\n";
print "base_dir = $base_dir\n";

if(-d $OUT_DIR){
}
else{
	mkdir $OUT_DIR or die "failed to mkdir $OUT_DIR\n";
}
#chdir $OUT_DIR;

if($segment_length > 25){
	$segment_length = 25;
}
if($segment_length < 18){
	$segment_length = 18;
}


$P = $P || 1;

if(! -s "$bwt.fa"){
	print "Reference genome $bwt.fa not found!\n";
	exit 0;
}
if(! -s "$bwt.1.ebwt"){
	print "Bowtie index not found!\n";
	exit 0;
}
#if (! -s "$gff_ref") {
#	print "GeneMark-ES predictions not found: $gff_ref\n";
#	exit 0;
#}
if (! -s "$mod") {
	print "GeneMark-ES model parameter directory not found: $mod\n";
	exit 0;
}
if($clean == 0){
	print "\t\tTemp files will not be deleted\n";
	if($report_read_name == 1){
		print "\t\tWill print original read names instead of template names\n";
	}
}

$name1 = "temp1";
$name2 = "temp2";
$name = "temp";

#open(RUNLOG,">>RUNLOG");
@Fq_list_1 = split(/,/,$Fq[0]);
@Fq_list_2 = split(/,/,$Fq[1]);

@abs_Fq_list_1 = ();
@abs_Fq_list_2 = ();

for $i (0 .. $#Fq_list_1){
	push @abs_Fq_list_1, abs_path($Fq_list_1[$i]);
}
for $i (0 .. $#Fq_list_2){
	push @abs_Fq_list_2, abs_path($Fq_list_2[$i]);
}
$abs_Fq_1 = join(",", @abs_Fq_list_1);
$abs_Fq_2 = join(",", @abs_Fq_list_2);

open($in,"<$Fq_list_1[0]");
$C=<$in>;
$C=<$in>;
chomp($C);
$read_length = length($C);
close($in);

$now_string = localtime;
print "$now_string:   ";
if($read_length < 36){
	print "Read length $read_length is less than 36bp can not be handled\n";
	exit(0);
}
if($read_length < 50){
	print "Read length is $read_length and segment length is forced to be",int($read_length/2),"\n";
	$segment_length = int($read_length/2);
}
else{
	print "Read length is $read_length and segment length is $segment_length\n";
}


$Fq[1] = abs_path($Fq[1]);
chdir $OUT_DIR;
open(RUNLOG,">>RUNLOG");
mySystem("$Bin/checkChr.pl $bwt.fa","Checking reference file");

$pm=new Parallel::ForkManager($P);
for ($i=1;$i<=2;$i++){
	my $pid = $pm->start and next;
	if($i == 1){
		mySystem("$Bin/readID_pair.pl $abs_Fq_1 1 > truesight_name_link","Preparing reads");
	}
	else{
		mySystem("$Bin/readID_pair.pl $abs_Fq_2 2");
	}
	$pm->finish;
}
$pm->wait_all_children;

##
## full-length read alignment
##

mySystem("$Bin/bowtie --quiet --sam --sam-nohead -k 30 -v 2 -p $P --un $name1.un.fq $bwt truesight_pair_temp.1.fastq $name1 2>>run.log","Full alignment on left reads");
system("rm truesight_pair_temp.1.fastq");


mySystem("$Bin/bowtie --quiet --sam --sam-nohead -k 30 -v 2 -p $P --un $name2.un.fq $bwt truesight_pair_temp.2.fastq $name2 2>>run.log","Full alignment on right reads");
system("rm truesight_pair_temp.2.fastq");

##
## 'anchor and extend' gapped alignment of reads unmapped above
##

$pm=new Parallel::ForkManager($P);
for ($i=1;$i<=2;$i++){
	my $pid = $pm->start and next;
	if($i == 1){
		mySystem("awk {'FS = \"\\t\" ; if(\$2 != 4) print \$0}' $name1 $name2 | sort -t '	' -k 3,3 -k 4,4n > $name.sam.sort; rm $name1 $name2");# sort fully aligned reads based on position
		mySystem("sort -t '/' -k 1,1n $name.sam.sort > $name.sam.sort_by_name");
	}
	else{
		mySystem("$Bin/centralPlan.pl $segment_length $P $bwt 2","Distribute unmapped reads");
	}
	$pm->finish;
}
$pm->wait_all_children;


mySystem("$Bin/getUnPairFullAli.pl temp.un.fq $name.sam.sort_by_name | $Bin/finalLink.pl $read_length  > EX_FULL");


open(my $in,"<temp_data/list");
my @list = <$in>;
chomp(@list);
my $total_job = scalar @list;
$sparse = 0;
if(substr($list[0],0,-2) eq "temp_data/temp.un"){
	$sparse = 1;
}

$C=0;
for (my $i=0; $i<scalar @list; $i++){
	$C++;
	my $id = $i+1;
	open($in,"<N");
	$N_frag = <$in>;
	chomp($N_frag);
	if($sparse == 1){
		mySystem("$Bin/GappedAlignment4sparse $N_frag $segment_length $list[$i].bwt $M 8 $V $canonical_flag $bwt.fa $list[$i].fq $P $intron_max $intron_min $read_length > $id.log","Gapped alignment on $id of $total_job subsets");
	}
	else{
		mySystem("$Bin/GappedAlignment $N_frag $segment_length $list[$i].bwt $M 8 $V $canonical_flag $bwt.fa $list[$i].fq $P $intron_max $intron_min $read_length > $id.log","Gapped alignment on $id of $total_job subsets");
	}
	system("mv solid solid.$id; mv non_solid non_solid.$id; mv multi multi.$id; mv new_full.sam new_full.sam.$id; rm new_full.sam.$id");
	$solid .= "solid.$id ";
	$non_solid .= "non_solid.$id ";
	$multi .= "multi.$id ";
}

system("rm $multi");
mySystem("cat $solid $non_solid | sort -t '/' -k 1,1n | $Bin/solid_filter_mem.pl > solid_temp; rm solid.*; rm non_solid.*");# use / as separater
# remove gapped alignments inconsistent with fully aligned sibling
mySystem("cat solid_temp temp_file.sam | sort -t '/' -k 1,1n | $Bin/rmUnPairGapAli_new.pl > solid");


##
## re-alignment
##

## remap pipeline
mySystem("$Bin/realign.pl -a solid -e $filter_pair -p $P -r $bwt -l $read_length -v $V -s $segment_length");

##
## classification step
##

## calculate features for each candidate splice junction
mySystem("$Bin/junctionEva.pl alignments all_junctions");
#
mySystem("$Bin/gapMark -strand -fs -scr $mod all_junctions $bwt.fa $gff_ref all");

## split cano and noncano
mySystem("$Bin/splitCano.pl all_feat $bwt.fa");	# now we have: all_feat_cano, all_feat_noncano

## get coverage features
mySystem("$Bin/getCov $read_length temp.sam.sort all_feat");

my $CLASS_START = time;

$cwd = getcwd;
mySystem("mkdir $cwd/svm");
mySystem("mkdir $cwd/svm/grid");
mySystem("cp cov_diffs svm");
chdir("$cwd/svm");

## classifier
my $tron_score_files = doSvm("$bwt.fa", $mod, $gff_ref, $P, $read_length, $class_options);

$class_duration = time -$CLASS_START;
printf("\n\nSVM running time: %02d:%02d:%02d\n\n", int($class_duration / 3600), int(($class_duration % 3600) / 60),int($class_duration % 60));
## end of classification

# back to working dir
chdir("$cwd");

##
## assemble output files
##

$pm=new Parallel::ForkManager($P);
for ($i=1;$i<=2;$i++){
	my $pid = $pm->start and next;
	if($i == 1){
		mySystem("$Bin/updateReadScore.pl svm/fj_cano alignments > solid_score");
	}
	if($i == 2){
		mySystem("$Bin/updateReadScore.pl svm/fj_noncano alignments > non_cano_score");
	}
	$pm->finish;

}
$pm->wait_all_children;


mySystem("cat svm/fj_cano svm/fj_noncano > junctions_score");

mySystem("cat solid_score non_cano_score > Gapped.sam_4");

mySystem("$Bin/presentJunc_us.pl PJ_canoflag junctions_score Gapped.sam_4 $bwt.fa temp_file.sam"); ##MAY 25

$now_string = localtime;
print "$now_string:   ";
print "Final Output: Generating bam file\n";

mySystem("$Bin/ID2name_pair.pl truesight_name_link GapAli.sam > GapAli.sam_; mv GapAli.sam_ GapAli.sam");
mySystem("$Bin/ID2name_pair.pl truesight_name_link temp.sam.sort_by_name > temp.sam.sort");

mySystem("cat head GapAli.sam temp.sam.sort_by_name | samtools view -Sb - | samtools sort - alignment");

if($clean == 1){
	`rm -r temp_data`;
	system("rm -r cov* FULL.sam solid* non* temp* REMAP* alignments all* multi* remap_sorted.sam RUNLOG svm PJ* N Gapped* *.log junctions_score head EX_FULL chr_num truesight*");
}

$NOW = time -$NOW;
printf("\n\nTotal running time: %02d:%02d:%02d\n\n", int($NOW / 3600), int(($NOW % 3600) / 60),int($NOW % 60));
##
## Done.



##
## SVM
##

sub doSvm($$$$$$)
{
	my ($seq, $mod, $gff_ref, $P, $read_length, $class_options) = @_;
	
	my $cwd = getcwd;
	chdir("$cwd/svm");

	print "seq = $seq\n";
	print "mod = $mod\n";
	print "gff = $gff_ref\n";
	print "base_dir = $base_dir\n";
	print "read_length = $read_length\n";
	print "P = $P\n";
	
	# baseline run to create training set and test set
	$cmd = "$Bin/classifier -debug -verb -fset $class_options -mod_dir $mod $read_length $base_dir/all_feat_cano $base_dir/all_feat_cano $seq > svmlog";
	mySystem( $cmd );
	# scale test set
	mySystem( "$Bin/svm-scale -s training.range feat_test_set > feat_test_set.scaled" );
	## grid search
	my %params;
	&gridSearch(\%params, $P, $seq, $mod, $gff_ref);

	# canonical predictions
	print "grid search done, now classifying with (c,g) = ($params{\"c\"},$params{\"gamma\"})\n";
	
	chdir("$cwd/svm");

	print "seq = $seq\n";
	print "mod = $mod\n";
	print "gff = $gff_ref\n";
	print "P = $P\n";

	print "cwd = $cwd\n";

	mySystem("$Bin/classifier -svm_c $params{\"c\"} -svm_gamma $params{\"gamma\"} -debug -verb -fset $class_options -mod_dir $mod $read_length $base_dir/all_feat_cano $base_dir/all_feat_cano $seq > svmlog_cano");
	mySystem("mv final_junctions fj_cano");
	# noncanonical predictions
	mySystem("$Bin/classifier -nonc -svm_c $params{\"c\"} -svm_gamma $params{\"gamma\"} -debug -verb -fset $class_options -mod_dir $mod $read_length $base_dir/all_feat_noncano $base_dir/all_feat_noncano $seq > svmlog_nonc");
	mySystem("mv final_junctions fj_noncano");
}

##
## grid search
##

sub gridSearch($$$$$)
{
	my ($P) = $_[1];
	my ($seq, $mod, $gff_ref) = ($_[2], $_[3], $_[4]);
	
	my $cwd = getcwd;
	print "gridSearch: CWD = $cwd\n";
	chdir("$cwd/grid");

	mySystem("$Bin/svmTrainTest.pl $P > train.log");
	mySystem("$Bin/svmPredTest.pl -p $gff_ref -s $seq -md $mod > pred.log");

	${$_[0]}{"gamma"} = 0.;
	${$_[0]}{"c"} = 0.;
	
	mySystem("$Bin/getBestCG.pl");
	
	open(CG,"<cg");
	$c = <CG>;
	chomp($c);
	$g = <CG>;
	chomp($g);
	close(CG);
	
	print "grid search finds best (c,g) = ($c,$g)\n";

	${_[0]}{"gamma"} = $g;
	${_[0]}{"c"} = $c;
	
	# go back to parent dir
	chdir($cwd);
	
	#return \%params;
}


##
## system calls
##

sub mySystem{
	my $mySystemCommand = $_[0];
	if(defined $_[1]){
		if($_[1] =~ /^0$/){
			system( $mySystemCommand );
			return;
		}
		$now_string = localtime;
		print "$now_string:   ";
		print $_[1],"\n";
		print RUNLOG "$now_string:   ";
		print RUNLOG $_[1],"\n";
	}

	my $returnCode = system( $mySystemCommand );
	if ( $returnCode != 0 ) {
		$now_string = localtime;
		print RUNLOG "$now_string:	Failed executing [$mySystemCommand]\n";
		die "$now_string:	Failed executing [$mySystemCommand]\n";
	}
}
