
#include "options.h"
#include <stdlib.h>
#include <iostream>
#include <string>


bool cOptions::parseOptions (int argc, const char** argv)
{
	if (argc < 3) {
		std::cout << std::endl;
		std::cout << "Usage: " << PROG_NAME << " [options] <rlen> <jcn_train> <jcn_pred> <ref_seq>" << std::endl;
		std::cout << " Where" << std::endl;
		std::cout << "  <rlen>              Read length" << std::endl;
		std::cout << "  <jcn_train>         File containing candidate junctions, from which a training set will be derived" << std::endl;
		std::cout << "  <jcn_pred>          File containing candidate junctions, upon which predictions will be made" << std::endl;
		std::cout << "  <ref_seq>           DNA sequence (FASTA)" << std::endl;
		std::cout << " Options:" << std::endl;
		std::cout << "  -verb               Verbose (for debugging)" << std::endl;
		std::cout << "  -heur               Predict based on simple heuristic rules, ignoring models" << std::endl;
		//std::cout << "  -train_only         Just find positive and negative sets, and train SVM parameters, then quit" << std::endl;
		//std::cout << "  -ext <file>         Externally provided training set <file>" << std::endl;
		//std::cout << "  -p_delta <p>        Threshold to use to select positive training set (prob > 1-p) and negative set (prob < p)" << std::endl;
		std::cout << "  -nonc               Input junctions are all noncanonical" << std::endl;
		std::cout << "  -soft               Assume shallow sequencing when creating training set" << std::endl;
		std::cout << "  -fset [csvldoeamht]   Use this set of features (all are used by default). Choose any subset from:" << std::endl;
		std::cout << "            c: cod potential" << std::endl;
		std::cout << "            s: splice score" << std::endl;
		std::cout << "            v: coverage delta" << std::endl;
		std::cout << "            l: intron length" << std::endl;
		std::cout << "            d: mapping depth" << std::endl;
		std::cout << "            o: min alignment overhang" << std::endl;
		std::cout << "            e: entropy" << std::endl;
		std::cout << "            a: ambiguity" << std::endl;
		std::cout << "            m: alignment mismatches" << std::endl;
		std::cout << "            h: frame shift error" << std::endl;
		std::cout << "            t: strand error" << std::endl;
		//std::cout << "  -fnot [csvldoeamht]   Use all EXCEPT this set of features. Choose any subset from those listed above" << std::endl;
		std::cout << "  -mod_dir <dir>      Directory (folder) containing Markov model parameter files" << std::endl;
		//std::cout << "  -multi <file>       Also classify <file> containing ambiguous (multiple candidates) gapped alignments" << std::endl;
		//std::cout << "  -cod_flank m        length of intron flanking sequences to use for coding potential score (default = 81)" << std::endl;
		//std::cout << "  -order n            order of donor and acceptor site Markov models (default = 0)" << std::endl;
		std::cout << "  -svm_gamma g        parameter gamma for SVM classifier (controls width of Gaussian kernel)" << std::endl;
		std::cout << "  -svm_c c            parameter C for SVM classifier (controls error cost)" << std::endl;
		//std::cout << "  -cov n              coverage threshold for positive set selection, req min of n alignments (default = 20)" << std::endl;
		//std::cout << "  -pt_len n           length cutoff used for pass-through mode (-no_class) (default = 10,000)" << std::endl;
		std::cout << std::endl;
		std::cout << "\nVersion: " << PROG_VER << std::endl;
		std::cout << std::endl;
		std::cout << "This program carries out training and classification of gapped alignments" << std::endl;
		std::cout << " by using a Gaussian kernal SVM (SVMLIB)." << std::endl;
		std::cout << std::endl;
		return true;
	}

	//Parse argv for required options
	int argCount = 0;
	string setNot;

	for (int i=1; i<argc; ++i) {
		std::string opt = (char*)argv[i];
		if (opt.at(0)=='-')
		{
			if (opt == "-verb") {
				m_bVerbose = true;
			}
			else if (opt == "-debug") {
				m_bDebug = true;
			}
			else if (opt == "-heur") {
				m_bHeuristic = true;
			}
			else if (opt == "-brief") {
				m_bBrief = true;
			}
			else if (opt == "-ext") {
				m_extTrainFile = (char*)argv[++i];
			}
			else if (opt == "-p_delta") {
				m_p_delta = atof( (char*)argv[++i] );
			}
			else if (opt == "-nonc") {
				m_bNoncano = true;
			}
			else if (opt == "-soft") {
				m_bSoft = true;
			}
			else if (opt == "-multi") {
				m_multiFile = (char*)argv[++i];
			}
			else if (opt == "-fset") {
				m_featureSet = (char*)argv[++i];
				featSet2Mask();
			}
			//else if (opt == "-fnot") {
			//	setNot = (char*)argv[++i];
			//	featSet2Mask();
			//	m_uiFeatureMask = !m_uiFeatureMask | m_uiSetMask;
			//}
			else if (opt == "-mod_dir") {
				m_modInDir = (char*)argv[++i];
			}
			else if (opt == "-svm_gamma") {
				m_svmGamma = (char*)argv[++i];
			}
			else if (opt == "-svm_c") {
				m_svmC = (char*)argv[++i];
			}
			else if (opt == "-cov") {
				m_nCovThresh = atoi( (char*)argv[++i] );
			}
			else if (opt == "-pt_len") {
				m_nPassThruLength = atoi( (char*)argv[++i] );
			}
			else {
				std::cout << "unknown option: " << opt << std::endl;
				return true;
			}
		}
		else
		{
			if (argCount==0)
			{
				m_nReadLength = atoi( opt.c_str() );
				++argCount;
			}
			else if (argCount==1)
			{
				m_solidTrainFile = opt;
				++argCount;
			}
			else if (argCount==2)
			{
				m_solidPredFile = opt;
				++argCount;
			}
			else if (argCount==3)
			{
				m_seqFile = opt;
				++argCount;
			}
			//else if (argCount==3)
			//{
			//	m_nFragments = atoi( opt.c_str() );
			//	++argCount;
			//}
			//else if (argCount==4)
			//{
			//	m_nReadLength = atoi( opt.c_str() );
			//	++argCount;
			//}
			//else if (argCount==5)
			//{
			//	m_nSegmentLength = atoi( opt.c_str() );
			//	++argCount;
			//}
			else {
				std::cout << "Too many arguments, starting at: " << opt << std::endl;
				return true;
			}
		}
	}

	if (argCount != 4) {
		std::cout << "argCount = " << argCount << std::endl;
		return true;		// incorrect number of arguments
	}

	return false;
}

void cOptions::featSet2Mask()
{
	// set domain: csvldoeam
	m_uiFeatureMask = 0x0;

	for (int i=0; i<(int)m_featureSet.length(); ++i)
	{
		switch (m_featureSet.at(i)) {
			case 'c': case 'C':
				m_uiFeatureMask |= _FEAT_CODING;
				break;
			case 's': case 'S':
					m_uiFeatureMask |= _FEAT_SPLICE_DON;
					m_uiFeatureMask |= _FEAT_SPLICE_ACC;
				break;
			case 'v': case 'V':
				m_uiFeatureMask |= _FEAT_DIFF_COV;
				break;
			case 'l': case 'L':
				m_uiFeatureMask |= _FEAT_LENGTH;
				break;
			case 'd': case 'D':
				m_uiFeatureMask |= _FEAT_DEPTH;
				break;
			case 'o': case 'O':
				m_uiFeatureMask |= _FEAT_OVERHANG;
				break;
			case 'e': case 'E':
				m_uiFeatureMask |= _FEAT_ENTROPY;
				break;
			case 'a': case 'A':
				m_uiFeatureMask |= _FEAT_AMBIGUITY;
				break;
			case 'm': case 'M':
				m_uiFeatureMask |= _FEAT_MISMATCH;
				break;
			case 'h': case 'H':
				m_uiFeatureMask |= _FEAT_FRAMESHIFT_ERROR;
				break;
			case 't': case 'T':
				m_uiFeatureMask |= _FEAT_STRAND_ERROR;
				break;
		};
	}

	// report settings
	int ix(0);
	std::cout << "Feature settings:" << std::endl;
	if (m_uiFeatureMask & _FEAT_SPLICE_DON)	{ std::cout << "  " << ++ix << ":" << "splice_don"; }
	if (m_uiFeatureMask & _FEAT_SPLICE_ACC)	{ std::cout << "  " << ++ix << ":" << "splice_acc"; }
	if (m_uiFeatureMask & _FEAT_CODING)		{ std::cout << "  " << ++ix << ":" << "coding"; }
	if (m_uiFeatureMask & _FEAT_DIFF_COV)	{ std::cout << "  " << ++ix << ":" << "diff_cov"; }
	if (m_uiFeatureMask & _FEAT_LENGTH)		{ std::cout << "  " << ++ix << ":" << "length"; }
	if (m_uiFeatureMask & _FEAT_DEPTH)		{ std::cout << "  " << ++ix << ":" << "depth"; }
	if (m_uiFeatureMask & _FEAT_OVERHANG)	{ std::cout << "  " << ++ix << ":" << "overhang"; }
	if (m_uiFeatureMask & _FEAT_ENTROPY)	{ std::cout << "  " << ++ix << ":" << "entropy"; }
	if (m_uiFeatureMask & _FEAT_AMBIGUITY)	{ std::cout << "  " << ++ix << ":" << "ambiguity"; }
	if (m_uiFeatureMask & _FEAT_MISMATCH)	{ std::cout << "  " << ++ix << ":" << "mismatches"; }
	if (m_uiFeatureMask & _FEAT_FRAMESHIFT_ERROR)	{ std::cout << "  " << ++ix << ":" << "frameshift_err"; }
	if (m_uiFeatureMask & _FEAT_STRAND_ERROR)	{ std::cout << "  " << ++ix << ":" << "strand_err"; }
	std::cout << std::endl;

	//m_uiFeatureMask &= m_uiSetMask;
}
