// gapFilter.cpp : Defines the entry point for the console application.
//

#include "targetver.h"
#include "options.h"
#include "model.h"
#include "sseq.h"
#include "filebuf.h"
#include "multi.h"
#include "geneSet.h"
#include "time.h"
#include <string>
#include <vector>


#ifdef _WINDOWS_
	// these includes are necessary to support directory existence checks in windows
	#include <direct.h>
	#include <io.h>
#else
	#include <dirent.h>
#endif

using std::string;
using std::vector;

// struct for managing file sets
typedef struct tagFileRec {
	string strName;
	int size;
} sFileRec;

// globals
vector<cModel> heur_models;
char alphabet[256];
int startGc;

bool bScoreSites(false);

bool fileRecLessThan (sFileRec fri, sFileRec frj) { return fri.strName.compare(frj.strName) < 0; }

extern void reset();

bool getFilesInDir(vector<sFileRec>& files, string dir);
bool getFilesInDir(vector<sFileRec>& files, string dir, string strNameTag);
void setAlphabet( char* ptr );
bool initScore(string& dir);

void markKeepersByCoding( vector<sIntronSet>& sets, bool bMultiFile );
void markKeepersByStrand( vector<sIntronSet>& sets, bool bMultiFile );
void markKeepersByFrame( vector<sIntronSet>& sets, bool bMultiFile );
void markKeepersBySplice( vector<sIntronSet>& sets, cOptions& opts, bool bMultiFile );

bool isCoding(sIntron& tron);
bool isPartialCoding(sIntron& tron);
int getFiltersHigh(cOptions& opts);
int getFiltersMed(cOptions& opts);
int getFiltersLow(cOptions& opts);
int getFiltersGenic(cOptions& opts);
int getFiltersNonGenic(cOptions& opts);

double getMax(double lf0, double lf1, double lf2);

int countDonAcc( int& nDon, int& nAcc, vector<sIntron>& introns );

void getIntrons(vector<sIntronSet>& multiSets);
void markGenic(cMulti& multi, cGeneSet& genes);
bool isGenic(cGene& gene, cGeneSet& set, sIntronSet tronSet);


cOptions opts;
SSeq seq;

int spliceLow(0), spliceMed(0), spliceHigh(0), spliceNoncano(0);


int main(int argc, const char* argv[])
{
	if (opts.parseOptions(argc,argv)) {
		if (argc > 1) {
			std::cout << PROG_NAME << ": Unacceptable argument(s) given.\n";
			return 0;
		}
		return 0;
	}

	if (opts.m_bVerbose) {
		std::cout << "model dir: " << opts.m_modInDir << std::endl;
		std::cout << "sequence: " << opts.m_seqInFile << std::endl;
		std::cout << "ref: " << opts.m_refInFile << std::endl;
		std::cout << "filters: " << std::endl;
		if (opts.m_bStrand) std::cout << "-strand" << std::endl;
		if (opts.m_bSpliceSites) std::cout << "-splice" << std::endl;
		if (opts.m_bFrameShift) std::cout << "-fs" << std::endl;
		if (opts.m_bGenic) std::cout << "genic is TRUE" << std::endl;
		else std::cout << "genic is FALSE" << std::endl;
		if (opts.m_bMultiFile) std::cout << "multi is TRUE" << std::endl;
		else std::cout << "multi is FALSE" << std::endl;
	}

	time_t time_0 = time(NULL);
	time_t td;

	if (!opts.m_bHeur) {
		if (initScore(opts.m_modInDir)) {
			std::cout << PROG_NAME << ": Aborting, file error." << std::endl;
			return 0;
		}
	}

	startGc = opts.m_nGcStart;

	// check flanking sequence length
	if (opts.m_nCodFlank < 6) {
		std::cout << PROG_NAME << ": Minimum intron flanking seq length less than the minimum allowed (6)! Automatically set to 6..." << std::endl;
		opts.m_nCodFlank = 6;
	}

	cGeneSet genes;

	cMulti multi;
	multi.m_nAlnSegments = opts.m_nSegments;
	multi.n_nSegmentLength = opts.m_nSegLength;

	// load intron sets
	std::cout << PROG_NAME << ": Gapped alignment candidate file: " << opts.m_multiInFile << std::endl;
	if (opts.m_bMultiFile)
	{
		if (multi.load( opts.m_multiInFile ) )
		{
			std::cout << PROG_NAME << ": Unable to read input file: " << opts.m_multiInFile << std::endl;
			return 0;
		}
		td = time(NULL);
		std::cout << "[" << td-time_0 << "] " << PROG_NAME << ": " << multi.m_sets.size() << " multi sets loaded" << std::endl;
	}
	else if (opts.m_bScoreFile)
	{
		std::cout << "Reading SCORE intron file " << opts.m_multiInFile << std::endl;
		if (multi.loadScoreFile( opts.m_multiInFile )) {
			std::cout << "Error reading score file: " << opts.m_multiInFile << std::endl;
			return true;
		}
	}
	else
	{
		if (multi.loadSam( opts.m_multiInFile ) )
		{
			std::cout << PROG_NAME << ": Unable to read input file: " << opts.m_multiInFile << std::endl;
			return 0;
		}
		td = time(NULL);
		std::cout << "[" << td-time_0 << "] " << PROG_NAME << ": " << multi.m_sets.size() << " gapped alignments loaded" << std::endl;
	}
	std::cout << PROG_NAME << ": Sorting sets..." << std::endl;
	multi.sortSets();

	// load DNA sequence
	td = time(NULL);
	std::cout << "[" << td-time_0 << "] " << PROG_NAME << ": Loading sequence file: " << opts.m_seqInFile << std::endl;
	seq.m_bSafeFormat = true;
	seq.SetAlphabet();
	if (!seq.load( opts.m_seqInFile, 0 ) )
	{
		std::cout << PROG_NAME << ": Unable to read input file: " << opts.m_seqInFile << std::endl;
		return 0;
	}

	td = time(NULL);
	genes.setPrefix( opts.m_prefix );
	if (!opts.m_bHeur)
	{
		std::cout << "[" << td-time_0 << "] " << PROG_NAME << ": Loading reference annotation GFF file: " << opts.m_refInFile << std::endl;
		if (opts.m_bGtf)
		{
			if (genes.loadGtf( opts.m_refInFile ))
			{
				std::cout << PROG_NAME << ": Unable to read input file: " << opts.m_refInFile << std::endl;
				return 1;
			}	}
		else
		{
			if (genes.loadGff( opts.m_refInFile ))
			{
				std::cout << PROG_NAME << ": Unable to read input file: " << opts.m_refInFile << std::endl;
				return 1;
			}
		}
	}

	string fname;
	int ct(0);

	string defline;
	string seqName;

	cSite site;

	int nSeq = seq.getNumContigs();
	int i;

	vector<sIntronSet> intron_sets;
	//vector<sIntron> ref_introns;

	cMulti multiOut;

	td = time(NULL);
	for (i=0; i<nSeq; ++i)
	{
		// contig name
		seq.getDefline( defline, i );
		seqName = seq.getNameTok( defline );

		// get gapped alignment candidates
		intron_sets = multi.getIntronSets( seqName, i );
		if (opts.m_bVerbose) std::cout << intron_sets.size() << " intron sets found in " << seqName << std::endl;

		if (intron_sets.empty()) {
			std:: cout << PROG_NAME << ": intron_sets is empty for seq: " << seqName << std::endl;
			continue;
		}

		if (opts.m_bHeur) {
			multiOut.add( intron_sets );
			continue;
		}

		if (opts.m_bMultiFile)
			getIntrons(intron_sets);
	
		// get raw info for every alignment candidate
		site.initGappedAlignSets( intron_sets, &seq, i, opts.m_bMultiFile );
		site.getStrand( &seq, i, 0, 0 );
		site.setGappedAlignSets( intron_sets );

		//site.recordGappedAlignData();
		site.clearData();

		// filter out likely false positives
		if (opts.m_bStrand || opts.m_bFrameShift) {
			if (opts.m_bVerbose) std::cout << "--marking Fp..." << std::endl;
			genes.markIntronFp( intron_sets );
		}

		// coding status
		if (opts.m_bCoding) {
			if (opts.m_bVerbose) std::cout << "--marking coding indication..." << std::endl;
			markKeepersByCoding( intron_sets, opts.m_bMultiFile );
		}

		// strand conflict
		if (opts.m_bStrand) {
			if (opts.m_bVerbose) std::cout << "--marking strand conflicts..." << std::endl;
			markKeepersByStrand( intron_sets, opts.m_bMultiFile );
		}

		// frame shift
		if (opts.m_bFrameShift) {
			if (opts.m_bVerbose) std::cout << "--marking frame shift conflicts..." << std::endl;
			markKeepersByFrame( intron_sets, opts.m_bMultiFile );
		}

		// low splice site score
		if (opts.m_bSpliceSites) {
			if (opts.m_bVerbose) std::cout << "--marking by splice site score..." << std::endl;
			markKeepersBySplice( intron_sets, opts, opts.m_bMultiFile );
		}

		multiOut.add( intron_sets );
	}

	if (opts.m_bHeur)
		goto WRITE_OUTPUT;

	td = time(NULL);
	markGenic(multiOut, genes);

	if (opts.m_bSpliceSites) {
		std::cout << "spliceLow = " << spliceLow << std::endl;
		std::cout << "spliceMed = " << spliceMed << std::endl;
		std::cout << "spliceHigh = " << spliceHigh << std::endl;
		std::cout << "non-canonical = " << spliceNoncano << std::endl;
	}

	td = time(NULL);

WRITE_OUTPUT:

	int filters;
	string file = opts.m_multiOutFile;
	if (opts.m_bMultiFile)
	{
		file.append("_multi_feat");
		std::cout << PROG_NAME << ": Saving output file: " << file << std::endl;
		filters = _ALN_PROPERTY_NONCANONICAL;
		multiOut.save(file, filters);
	}
	else {
		file.append("_feat");
		if (opts.m_bHeur)
			multiOut.padZeros(file);
		else
			multiOut.saveScore(file);

		if (0)
		{
			// write genic and non-genic files
			file = opts.m_multiOutFile;
			file.append("_genic_P0");
			filters = getFiltersGenic(opts) | getFiltersHigh(opts) | _ALN_PROPERTY_NONCANONICAL;
			multiOut.saveGenicSam(file, filters);

			file = opts.m_multiOutFile;
			file.append("_genic_N0");
			filters = getFiltersGenic(opts) | getFiltersLow(opts) | _ALN_PROPERTY_NONCANONICAL;
			multiOut.saveGenicSam(file, filters);

			file = opts.m_multiOutFile;
			file.append("_non_P0");
			filters = getFiltersNonGenic(opts) | getFiltersHigh(opts) | _ALN_PROPERTY_NONCANONICAL;
			multiOut.saveNonSam(file, filters);

			file = opts.m_multiOutFile;
			file.append("_non_N0");
			filters = getFiltersNonGenic(opts) | getFiltersLow(opts) | _ALN_PROPERTY_NONCANONICAL;
			multiOut.saveNonSam(file, filters);

			file = opts.m_multiOutFile;
			file.append("_noncano");
			filters = _ALN_PROPERTY_NONCANONICAL;
			multiOut.saveSamN(file, filters);
		}
	}

	td = time(NULL);
	std::cout << "[" << td-time_0 << "] " << PROG_NAME ": done." << std::endl;

	return 0;
}

//
// -----------------------------------------------------------------------------
//

int getIx(int gc)
{
	int ix(0);

	if (opts.m_bHuman) {
		if (gc <= 40)
			ix = 0;
		else if (gc > 40 && gc <= 49)
			ix = 1;
		else
			ix = 2;
	}
	else {
		ix = gc - 20;
	}

	// constrain ix to lie withing array bounds
	int nSize = (int) heur_models.size();
	if (ix < 0) ix = 0;
	else if (ix >= nSize) ix = nSize-1;

	return ix;
}

int getGc(int ibeg, int iend, int icontig)
{
	int b, e;
	if (ibeg < iend) { b = ibeg; e = iend; }
	else { b = iend; e = ibeg; }

	int len = seq.getLength(icontig);
	if (len == -1)
		return -1;


	int dL = b - 1000 >= 0 ? 1000 : b;
	int dR = e + 1000 >= len ? len - e - 1 : 1000;

	string seqL, seqR;
	seq.getStretch( seqL, icontig, b - dL, dL, true );
	seq.getStretch( seqR, icontig, e, dR, true );

	int i, nGC(0);

	for (i=0; i<(int)seqL.length(); ++i)
	{
		if (seqL[i]=='C' || seqL[i]=='G') ++nGC;
	}

	for (i=0; i<(int)seqR.length(); ++i)
	{
		if (seqR[i]=='C' || seqR[i]=='G') ++nGC;
	}

	int L = seqL.length() + seqR.length();
	double lfGC = 100. * double(nGC) / double(L);

	nGC = int(lfGC);

	// limit to interval [20,70]
	if (nGC < 20)
		nGC = 20;
	if (nGC > 70)
		nGC = 70;

	return nGC;
}

int getFiltersGenic(cOptions& opts)
{
	int nFilters(0);
	if (opts.m_bFrameShift) nFilters |= _ALN_PROPERTY_FRAME_SHIFT;
	if (opts.m_bStrand) nFilters |= _ALN_PROPERTY_STRAND_CONFLICT;
	return nFilters;
}

int getFiltersNonGenic(cOptions& opts)
{
	int nFilters(0);
	if (opts.m_bStrand) nFilters |= _ALN_PROPERTY_STRAND_CONFLICT;
	return nFilters;
}


int getFiltersHigh(cOptions& opts)
{
	int nFilters(0);
	if (opts.m_bFrameShift) nFilters |= _ALN_PROPERTY_FRAME_SHIFT;
	if (opts.m_bSpliceSites) nFilters |= (_ALN_PROPERTY_LOW_SPLICE_SCORE | _ALN_PROPERTY_MED_SPLICE_SCORE);
	if (opts.m_bStrand) nFilters |= _ALN_PROPERTY_STRAND_CONFLICT;
	return nFilters;
}

int getFiltersMed(cOptions& opts)
{
	int nFilters(0);
	if (opts.m_bFrameShift) nFilters |= _ALN_PROPERTY_FRAME_SHIFT;
	if (opts.m_bSpliceSites) nFilters |= (_ALN_PROPERTY_LOW_SPLICE_SCORE | _ALN_PROPERTY_HIGH_SPLICE_SCORE);
	if (opts.m_bStrand) nFilters |= _ALN_PROPERTY_STRAND_CONFLICT;
	return nFilters;
}

int getFiltersLow(cOptions& opts)
{
	int nFilters(0);
	if (opts.m_bFrameShift) nFilters |= _ALN_PROPERTY_FRAME_SHIFT;
	if (opts.m_bSpliceSites) nFilters |= (_ALN_PROPERTY_MED_SPLICE_SCORE | _ALN_PROPERTY_HIGH_SPLICE_SCORE);
	if (opts.m_bStrand) nFilters |= _ALN_PROPERTY_STRAND_CONFLICT;
	return nFilters;
}

void getIntrons(vector<sIntronSet>& multiSets)
{	// set intron struct inside set candidate structures
	sIntron intron;
	sIntronCand *pcand;

	intron.extr_L = intron.extr_R = intron.extr_I = 0;
	intron.nFpL = intron.nFpR = -1;

	for (int j=0; j<(int)multiSets.size(); ++j)
	{
		for (int i=0; i<(int)multiSets[j].set.size(); ++i)
		{
			pcand = &multiSets[j].set[i];
			// ixIntron will be a proxy for file position in the list of candidates
			multiSets[j].set[i].intron.ixIntron = i;
			// beg, end
			multiSets[j].set[i].intron.ibeg = pcand->multi_info[_MULTI_BEG] + pcand->multi_info[_MULTI_ALN_L];
			multiSets[j].set[i].intron.iend = multiSets[j].set[i].intron.ibeg +  pcand->multi_info[_MULTI_GAP] - 1;
		}
	}
}

void markGenic(cMulti& multi, cGeneSet& genes)
{
	cGene gene;
	bool bGenic(true);

	for (int i=0; i<(int)multi.m_sets.size(); ++i)
	{
		if (multi.throwOut(i))
			continue;

		if (!isGenic(gene, genes, multi.m_sets[i]))
			continue;

		multi.setGenic(i);
	}
}

//void splitGenic(cMulti& multi, cGeneSet& genes)
//{
//	cGene gene;
//
//	for (int i=0; i<(int)multi.m_sets.size(); ++i)
//	{
//		if (multi.throwOut(i))
//			continue;
//
//		if (splitByGenic(genes, multi.m_sets[i]))
//			continue;
//	}
//}

bool splitByGenic(sIntronSet& genicSet, sIntronSet& nonSet, cGeneSet& genes, sIntronSet inputSet)
{
	nonSet.set.clear();
	genicSet.set.clear();

	cGene gene;
	sIntronCand cand;
	bool bFoundGene(false);

	for (int i=0; i<(int)inputSet.set.size(); ++i)
	{
		cand = inputSet.set[i];
		if (!cand.bKeep) continue;
		if (genes.findGene(gene, inputSet.seq, cand.intron.ibeg)) continue;
		bFoundGene = true;
		break;
	}

	if (!bFoundGene) {		// add all candidates to non-genic set
	}


	return false;
}

bool isGenic(cGene& gene, cGeneSet& set, sIntronSet tronSet)
{	// return true if all candidates in set ix have both boundaries inside the CDS of a gene

	sIntronCand cand;
	bool bFoundGene(false);

	for (int i=0; i<(int)tronSet.set.size(); ++i)
	{
		cand = tronSet.set[i];
		if (!cand.bKeep) continue;
		if (set.findGene(gene, tronSet.seq, cand.intron.ibeg)) continue;
		bFoundGene = true;
		break;
	}

	if (!bFoundGene) return false;

	if (opts.m_bVerbose) std::cout << "gene found for " << cand.intron.strId << "[" << cand.intron.ibeg << "," << cand.intron.iend << "]" << std::endl;

	for (int i=0; i<(int)tronSet.set.size(); ++i)
	{
		cand = tronSet.set[i];
		//if (cand.nStatus == _ALN_PROPERTY_NONCANONICAL) continue;

		if (cand.intron.ibeg < gene.getCdsBeg() || cand.intron.iend > gene.getCdsEnd())
			return false;
	}

	// all candidate boundaries lie inside CDS of same gene
	return true;
}

bool initScore(string& dir)
{
	vector<sFileRec> files;

	if (getFilesInDir(files, dir))
		return true;

	heur_models.clear();
	cModel mod;
	cSite site;

	mod.init( &site );

	string fullFile;

	for (int i=0; i<(int)files.size(); ++i)
	{
		fullFile = dir + SLASH;
		fullFile.append( files[i].strName );

		std::cout << "Loading model file: " << fullFile << std::endl;

		if (mod.load( fullFile ))
		{
			std::cout << "Error loading file: " << files[i].strName << std::endl;
			return true;
		}

		heur_models.push_back( mod );
		reset();
		//heur_models[i].gc = ++gc;
	}

	// set alphabet
	setAlphabet( alphabet );

	return false;
}

void markKeepersByCoding( vector<sIntronSet>& sets, bool bMultiFile )
{
	for (int i=0; i<(int)sets.size(); ++i)
	{
		for (int j=0; j<(int)sets[i].set.size(); ++j)
		{
			if (isCoding(sets[i].set[j].intron))
				sets[i].set[j].nStatus |= _ALN_PROPERTY_CODING;
			else if (isPartialCoding(sets[i].set[j].intron))
				sets[i].set[j].nStatus |= _ALN_PROPERTY_PARTIAL_CODING;
		}
	}
}

void markKeepersByStrand( vector<sIntronSet>& sets, bool bMultiFile )
{
	for (int i=0; i<(int)sets.size(); ++i)
	{
		for (int j=0; j<(int)sets[i].set.size(); ++j)
		{
			if ( bMultiFile && sets[i].set[j].multi_info[_MULTI_CANO]==0 ) {
				sets[i].set[j].nStatus |= _ALN_PROPERTY_NONCANONICAL;
			}

			if (sets[i].set[j].intron.iJcnType==INTRON_BAD) {
				sets[i].set[j].nStatus |= _ALN_PROPERTY_NONCANONICAL;
				//sets[i].set[j].bKeep = false;
			}

			if (sets[i].set[j].intron.nFpL==_INTRON_FP_CDS_WRONG_STRAND ||
				sets[i].set[j].intron.nFpR==_INTRON_FP_CDS_WRONG_STRAND)
				sets[i].set[j].nStatus |= _ALN_PROPERTY_STRAND_CONFLICT;
				//sets[i].set[j].bKeep = false;
		}
	}
}

void markKeepersByFrame( vector<sIntronSet>& sets, bool bMultiFile )
{
	sIntron intron;
	for (int i=0; i<(int)sets.size(); ++i)
	{
		for (int j=0; j<(int)sets[i].set.size(); ++j)
		{
			if ( bMultiFile && sets[i].set[j].multi_info[_MULTI_CANO]==0 )
				continue;

			intron = sets[i].set[j].intron;

			if (intron.iJcnType==INTRON_BAD) {
				//sets[i].set[j].bKeep = false;
				sets[i].set[j].nStatus |= _ALN_PROPERTY_NONCANONICAL;
				continue;
			}

			if (isCoding(intron) && intron.nFpL!=intron.nFpR) {
				sets[i].set[j].nStatus |= _ALN_PROPERTY_FRAME_SHIFT;
			}
		}
	}
}

void markKeepersBySplice( vector<sIntronSet>& sets, cOptions& opts, bool bMultiFile )
{
	int ix, gc, ctLow(0), ctMed(0), ctHigh(0), ctNoncano(0);
	double logDon0, logDon1, logDon2, logAcc0, logAcc1, logAcc2;
	double lDon, lAcc;
	bool bNoncano(false);

	sIntron intron;

	FILE *pf = fopen("splice_scores.txt","w");

	for (int i=0; i<(int)sets.size(); ++i)
	{
		for (int j=0; j<(int)sets[i].set.size(); ++j)
		{
			if ( bMultiFile && sets[i].set[j].multi_info[_MULTI_CANO]==0 ) {
				sets[i].set[j].nStatus |= (_ALN_PROPERTY_NONCANONICAL | _ALN_PROPERTY_LOW_SPLICE_SCORE);
				continue;
			}

			intron = sets[i].set[j].intron;

			gc = getGc(intron.ibeg, intron.iend, intron.nContig);
			ix = getIx(gc);

			heur_models[ix].scoreDonSite3( logDon0, logDon1, logDon2, intron.don, intron.iJcnType );
			heur_models[ix].scoreAccSite3( logAcc0, logAcc1, logAcc2, intron.acc, intron.iJcnType );

			//lDon = getMax(logDon0, logDon1, logDon2);
			//lAcc = getMax(logAcc0, logAcc1, logAcc2);
			lDon = logDon0;
			lAcc = logAcc0;

			if (opts.m_bSortBySum)
			{
				if ((lDon + lAcc) < (opts.m_lfDonThresholdLow + opts.m_lfAccThresholdLow)) {
					sets[i].set[j].nStatus |= _ALN_PROPERTY_LOW_SPLICE_SCORE;
					++ctLow;
				}
				else if ((lDon + lAcc) >= (opts.m_lfDonThresholdHigh + opts.m_lfAccThresholdHigh)) {
					sets[i].set[j].nStatus |= _ALN_PROPERTY_HIGH_SPLICE_SCORE;
					++ctHigh;
				}
				else {
					sets[i].set[j].nStatus |= _ALN_PROPERTY_MED_SPLICE_SCORE;
					++ctMed;
				}
			}
			else
			{
				bNoncano = false;
				if (sets[i].set[j].intron.iJcnType != INTRON_GT_AG) {
					sets[i].set[j].nStatus |= _ALN_PROPERTY_NONCANONICAL;
					bNoncano = true;
					++ctNoncano;
				}

				if (!bNoncano) {
					if ( lDon < opts.m_lfDonThresholdLow || lAcc < opts.m_lfAccThresholdLow) {
						sets[i].set[j].nStatus |= _ALN_PROPERTY_LOW_SPLICE_SCORE;
						++ctLow;
					}
					else if (lDon >= opts.m_lfDonThresholdHigh && lAcc >= opts.m_lfAccThresholdHigh) {
						sets[i].set[j].nStatus |= _ALN_PROPERTY_HIGH_SPLICE_SCORE;
						++ctHigh;
					}
					else {
						sets[i].set[j].nStatus |= _ALN_PROPERTY_MED_SPLICE_SCORE;
						++ctMed;
					}
				}
			}

			//if ( (sets[i].set[j].nStatus & _ALN_PROPERTY_LOW_SPLICE_SCORE == 0) &&
			//	 (sets[i].set[j].nStatus & _ALN_PROPERTY_HIGH_SPLICE_SCORE == 0))
			//{
			//	sets[i].set[j].nStatus |= _ALN_PROPERTY_MED_SPLICE_SCORE;
			//	++ctMed;
			//}

			int icano = bNoncano? 1 : 0;
			fprintf(pf,"%f\t%f\t%d\n", lDon, lAcc, icano);
		}
	}

	fclose(pf);

	spliceLow += ctLow;
	spliceMed += ctMed;
	spliceHigh += ctHigh;
	spliceHigh += ctHigh;
	spliceNoncano += ctNoncano;
}

bool isCoding(sIntron& tron)
{
	bool bL = tron.nFpL==_INTRON_FP_CDS_FR_0 || tron.nFpL==_INTRON_FP_CDS_FR_1 || tron.nFpL==_INTRON_FP_CDS_FR_2;
	bool bR = tron.nFpR==_INTRON_FP_CDS_FR_0 || tron.nFpR==_INTRON_FP_CDS_FR_1 || tron.nFpR==_INTRON_FP_CDS_FR_2;
	return bL && bR;
}

bool isPartialCoding(sIntron& tron)
{
	bool bL = tron.nFpL==_INTRON_FP_CDS_FR_0 || tron.nFpL==_INTRON_FP_CDS_FR_1 || tron.nFpL==_INTRON_FP_CDS_FR_2;
	bool bR = tron.nFpR==_INTRON_FP_CDS_FR_0 || tron.nFpR==_INTRON_FP_CDS_FR_1 || tron.nFpR==_INTRON_FP_CDS_FR_2;
	return bL || bR;
}

double getMax(double lf0, double lf1, double lf2)
{
	double lfM;
	int ix(-1);

	if (lf0 > lf1) {
		lfM = lf0; ix = 0;
	}
	else {
		lfM = lf1; ix = 1;
	}

	if (lfM < lf2) {
		lfM = lf2; ix = 2;
	}

	return lfM;
}

void setAlphabet( char* ptr )
{
	memset( ptr, 4, 256 );

	*(ptr + 'a' ) = 0;
	*(ptr + 'A' ) = 0;
	*(ptr + 'c' ) = 1;
	*(ptr + 'C' ) = 1;
	*(ptr + 'g' ) = 2;
	*(ptr + 'G' ) = 2;
	*(ptr + 't' ) = 3;
	*(ptr + 'T' ) = 3;
	*(ptr + 'u' ) = 3;
	*(ptr + 'U' ) = 3;
}

bool getFilesInDir(vector<sFileRec>& files, string dir)
{
	string allfiles = "*";
    return getFilesInDir(files, dir, allfiles);
}

bool getFilesInDir(vector<sFileRec>& files, string dir, string strNameTag)
{
	files.clear();

	sFileRec file_record;
	string fname;

#ifdef _WINDOWS_
	// windows directory interrogation library

	int ANY_FILE =  _A_NORMAL | _A_RDONLY | _A_HIDDEN | _A_SYSTEM | _A_ARCH;

	struct _finddatai64_t data;
	// First create the filename that will be use to initialize the find.
	// "*.*" are wild card characters that tells the find function to return a 
	// list of all the files and directories.  You can limit this if you wish
	// to just file with specific extensions, for example "*.txt".  If you do that
	// then finder will not return any directory names.

	// start the finder -- on error _findfirsti64() will return -1, otherwise if no
	// error it returns a handle greater than -1.

	//if (dir.at(dir.length()-1)==SLASH)
	//	fname = dir + wildcard;
	//else
	//{
	//	fname = dir + SLASH;
	//	fname += wildcard;
	//}
	fname = dir + "\\*.*";

	long h = _findfirsti64(fname.c_str(),&data);
	if(h >= 0)
	{
		do {
			if( (data.attrib & ANY_FILE) )
			{
				// its a file
				fname = data.name;
				if (strNameTag.compare("*")==0 || strNameTag.empty())
				{
					file_record.strName = data.name;
					file_record.size = data.size;
					files.push_back( file_record );
				}
				else if (fname.find(strNameTag.c_str(),0)!=string::npos)
				{
					file_record.strName = data.name;
					file_record.size = data.size;
					files.push_back( file_record );
				}
			}
			else
			{
				// ignore
			}
		} while ( _findnexti64(h,&data) == 0);
		// close the find handle.  
		_findclose(h);

	}

#else
	// unix directory interrogation library
	#include <sys/types.h>
	#include <sys/stat.h>

    DIR *dp;
    struct dirent *dirp;
    if((dp  = opendir(dir.c_str())) == NULL) {
		return true;
    }

	string str;
    while ((dirp = readdir(dp)) != NULL)
	{
		str = string(dirp->d_name);

		if (str.size()>0)
		{
			if (str.at(0)!='.')
			{
				struct stat st;
				size_t size;
				if ( !stat( dirp->d_name, &st ) )
					size = st.st_size;

				if (strNameTag.compare("*")==0 || strNameTag.empty())
				{
					file_record.strName = string(dirp->d_name);
					file_record.size = (int) size;
					files.push_back( file_record );
				}
				else if (str.find(strNameTag.c_str(),0)!=string::npos)
				{
					file_record.strName = string(dirp->d_name);
					file_record.size = (int) size;
					files.push_back( file_record );
				}
			}
		}
    }
    closedir(dp);

#endif

	sort (files.begin(), files.end(), fileRecLessThan);

    return false;
}


/*
	This fcn counts all unique instances of donor and acceptor site
	coordinates in the provided array of introns.
*/
int countDonAcc( int& nDon, int& nAcc, vector<sIntron>& introns )
{
	nDon = nAcc = 0;
	vector<int> vidon, viacc;
	int idon, iacc;
	int nRefUtr(0);

	for (int i=0; i<(int)introns.size(); ++i)
	{
		if (introns[i].iStrand == -1)
		{	// negative strand
			idon = introns[i].iend;
			iacc = introns[i].ibeg;
		}
		else
		{	// positive strand
			idon = introns[i].ibeg;
			iacc = introns[i].iend;
		}

		vidon.push_back( idon );
		viacc.push_back( iacc );

		if (introns[i].iframe == -1)
			++nRefUtr;
	}

	vector<int>::iterator it;

	// count unique donors
	it = unique( vidon.begin(), vidon.end() );
	nDon = it - vidon.begin();
	// count unique acceptors
	it = unique( viacc.begin(), viacc.end() );
	nAcc = it - viacc.begin();

	return nRefUtr;
}

