/*
	geneType.h
*/

#pragma once

#include <vector>
#include <string>
#include <map>
using std::map;
using std::vector;
using std::string;

// gene types
enum { SINGLE_ISOFORM = 0, MULTIPLE_ISOFORM, ALL_ISOFORMS };

// exon types
enum { EXON_TYPE_SINGLE, EXON_TYPE_INITIAL, EXON_TYPE_INTERNAL, EXON_TYPE_TERMINAL };

// feature types
enum { FEATURE_INTERGENIC, FEATURE_INTRON, FEATURE_CDS,
       FEATURE_EXON, FEATURE_EXON_INI, FEATURE_EXON_INT, FEATURE_EXON_TER,
       FEATURE_START, FEATURE_STOP, FEATURE_DON, FEATURE_ACC };

// intron type (types given by Tophat)
enum { INTRON_GT_AG = 0, INTRON_GC_AG, INTRON_AT_AC, INTRON_NON, INTRON_BAD };

// true intron types
enum { _INTRON_CDS_FR_0, _INTRON_CDS_FR_1, _INTRON_CDS_FR_2, _INTRON_UTR };

// intron false positive types (one value for left end, another value for right end)
enum { _INTRON_FP_TRUE = -1, _INTRON_FP_CDS_FR_0, _INTRON_FP_CDS_FR_1, _INTRON_FP_CDS_FR_2, _INTRON_FP_UTR, _INTRON_FP_INTRON, _INTRON_FP_INTERG, _INTRON_FP_CDS_WRONG_STRAND };

//enum { EXON_OVERALL = 0, EXON_TRANSCRIPT, EXON_PEPTIDE, EXON_ORTHO, EXON_VISTA };
//enum { GENE_OVERALL_STARS = 0, GENE_OVERALL_CLASS, GENE_TRANSCRIPT, GENE_PEPTIDE, GENE_ORTHO, GENE_VISTA };

enum { _MULTI_PART = 0, _MULTI_BEG, _MULTI_ALN_L, _MULTI_GAP, _MULTI_ALN_R, _MULTI_MIS_L, _MULTI_MIS_R, _MULTI_CANO, _MULTI_ORI };


typedef struct tagFeatureQuality
{
	int n;
	bool bDon, bAcc;
} featureQuality;

// quality structures
// (exon)
typedef struct tagExonQuality
{
	int nConfidence;
	featureQuality nTranscript;
	featureQuality nProtein;
	featureQuality nOrthology;
	featureQuality nVista;
} exonQuality;
// (gene)
typedef struct tagGeneQuality
{
	int nStars;
	int nConfidence;
	int nTranscript;
	int nProtein;
	int nOrthology;
	int nVista;
} geneQuality;

// coding segment
typedef struct tagSegment
{
	int ibeg;
	int iend;
	unsigned int extr_L;
	unsigned int extr_R;
	int iframe;
	int nOrigins;
	int level;
} segment;

// special regions
typedef struct tagSpecialRegion
{
	string			contig;
	int				contig_id;
	vector<segment> boundaries;
} specialRegion;

// introns for gapped alignments
typedef struct tagIntron
{
	string strId;
	int nContig;
	int nIntronsInGene;		// how many introns in this gene (isoform)?
	int ixIntron;			// which intron in the gene is this (index)
	int posAtg, posStp;		// position of intron starting point, relative to CDS start and stop
	int ibeg;
	int iend;
	int iframe;			// {0, 1, 2, 3 (non-coding), -1 (unknown)}
	int flankLeft;
	int flankRight;
	int iStrand;
	int gc;
	int iJcnType;			// GT_AG, GC_AG, etc
	int iIntronType;		// CDS (0,1,2), UTR
	int nCoverage;
	string don;
	string acc;
	vector<int> info;
	unsigned int extr_L;
	unsigned int extr_R;
	unsigned int extr_I;
	int dDon, dAcc;			// distance to nearest reference boundary, sign indicates direction (neg==upsream)
							//  DEF: pRef - pQuery
	int nFpL, nFpR;			// classification of left and right side of a false positive
	vector<string> feat;

	int lenExonL, lenExonR;
} sIntron;


typedef struct tagExon
{
	string strId;
	int nContig;
	int ibeg;
	int iend;
	int iframe;			// from of first nt in exon (-1 means unknown, 3=non-coding at 5' boundary)
	bool bStrand;
	unsigned int extr_L;
	unsigned int extr_R;
	unsigned int extr_E;
} sExon;

class potJunc
{
	public:
		string SAM;
		map<int,vector<vector<int> > > SAM_info;			// key is part of segment
		map<int,vector<vector<double> > > score_info;
};


typedef struct tagIntronCand
{
	vector<int> multi_info;

	sIntron intron;
	double donScore, accScore;
	double fs;
	int nConflict;
	int nStatus;
	bool bKeep;
} sIntronCand;

typedef struct tagIntronSet
{
	string SAM_line;
	string seq;
	int nContig;
	bool bGenic;
	vector<sIntronCand> set;
} sIntronSet;

// eukaryotic gene
typedef struct tagCodingRegion {
	vector<segment>	boundaries;
	vector<int>		nType;
	int				iphase;			// generally non-zero for partial genes without start codon
	int				nContig;
	bool			bStrand;
	bool			bStatus;
	bool			bPartial;
	bool			bCanonical;
	int				nId;
	int				nStars;
	double			pctGC;
	geneQuality		quality;
	string			strSeq;
	string			strGeneId;
	string			strMrnaId;
	string			strComment;
} codingRegion;

typedef struct tagTranscript {
	vector<segment>	boundaries;
	vector<int>		nType;
	int				iphase;			// generally non-zero for partial genes without start codon
	int				nContig;
	bool			bStrand;
	//bool			bStatus;
	//bool			bPartial;
	//bool			bCanonical;
	string			strGeneId;
	string			strMrnaId;
	//string			strComment;
} sTranscript;


// prokaryotic gene
struct codingRegionProk {
	string  strDefline;
	int		ibeg;
	int		iend;
	bool	bStrand;
};

typedef struct tagSeqFasta {
	string defline;
	string seq;
} seqFasta;

typedef struct tagBpRecord {
	string	seqIntronTerm;	// nt sequence over a region covering the end of each intron
	int		istart;			// coordinate in the intron sequence corresponding to the beginning of seqIntronTerm
	int		lenIntron;		// length of the complete original intron
} bpRecord;


class cHmmArg
{
public:
	cHmmArg()
	{
		bGetPartialGenes = false;
		nGffOffset = 0;
		nMinGeneLength = 0;
	};
	cHmmArg( const cHmmArg& arg )
	{
		genes			= arg.genes;
		strDefline		= arg.strDefline;
		pchSeq			= arg.pchSeq;
		badSegments		= arg.badSegments;
		strModelFile	= arg.strModelFile;
		strOutFile		= arg.strOutFile;
		strGffFile		= arg.strGffFile;
		nGffOffset		= arg.nGffOffset;
		outputTypes		= arg.outputTypes;
		nContigId		= arg.nContigId;
		nMinGeneLength	= arg.nMinGeneLength;
		nInteriorMargin = arg.nInteriorMargin;
		bGetPartialGenes = arg.bGetPartialGenes;
		bGetBadGenes	= arg.bGetBadGenes;
	}
	cHmmArg operator=( const cHmmArg& arg )
	{
		genes			= arg.genes;
		strDefline		= arg.strDefline;
		pchSeq			= arg.pchSeq;
		badSegments		= arg.badSegments;
		strModelFile	= arg.strModelFile;
		strOutFile		= arg.strOutFile;
		strGffFile		= arg.strGffFile;
		nGffOffset		= arg.nGffOffset;
		outputTypes		= arg.outputTypes;
		nContigId		= arg.nContigId;
		nMinGeneLength	= arg.nMinGeneLength;
		nInteriorMargin = arg.nInteriorMargin;
		bGetPartialGenes = arg.bGetPartialGenes;
		bGetBadGenes	= arg.bGetBadGenes;

		return (*this);
	}
	~cHmmArg()	{ /* NOP */ };

	vector<codingRegion> genes;
	string strDefline;
	char *pchSeq;
	specialRegion badSegments;
	string strModelFile;
	string strOutFile;
	string strGffFile;
	int nGffOffset;
	vector<string> outputTypes;
	int nContigId;
	int nMinGeneLength;
	int nInteriorMargin;
	bool bGetPartialGenes;
	bool bGetBadGenes;
};

typedef struct { int iseq; int ioffset; string name; string origName; } seqMap;


typedef struct tagPathCount
{
	int ix;
	long n;
} pathCount;

typedef struct tagPathRecord
{
	string name;
	int nGeneLength;
	long nSplices;
} pathRecord;

typedef struct tagSpliceNode
{
	int ix;		// coordinate
	long n;		// helpful counting parameter
	int nType;	// donor or acceptor?
	vector<int> gtEdges, agEdges;
} spliceNode;

enum { SPLICE_JCN_DON = 0, SPLICE_JCN_ACC };

typedef struct tagGffRecord
{
	string	strName;
	string	strSource;
	string	strFeature;
	int		ibeg;
	int		iend;
	string	strScore;
	char	chStrand;
	string	strFrame;
	string	strRemainder;

	string	strId;
	string	strNameTag;
	string	strParent;
	string	strComment;
	int		iFamily;
	bool	bDeleted;
	bool	bShortFeature;

//	geneQuality conGene;
//	exonQuality conExon;
} gffRecord;



//
// EST-PIPELINE output file processing (very similar to GFF, but better organization)
//
typedef struct tagEstExon
{
	int ibeg;
	int iend;
	int iframe;			// from of first nt in exon (-1 means unknown, 3=non-coding at 5' boundary)
} sEstExon;

typedef struct tagEstIsoform
{
	string strId;
	int transcriptId;
	bool bStrand;
	int contigId;

	int ibeg;
	int iend;

	vector<sEstExon> exons;
} sEstIsoform;


typedef struct tagEstGene
{
	int geneId;
	string name;

	int ibeg;
	int iend;

	int cdsStart, cdsStop;
	bool bTestSet;

	int ixb, ixe;
	int absL, absR;

	bool bProfile;

	vector<sEstIsoform> isoforms;
} sEstGene;
