
#include "targetver.h"
#include "options.h"
#include <stdlib.h>
#include <iostream>
#include <string>


bool cOptions::parseOptions (int argc, const char** argv)
{
	if (argc < 6) {
		std::cout << std::endl;
		std::cout << "Usage: " << PROG_NAME << " [options] <dir> <introns> <seq> <ref> <multi_out>" << std::endl;
		std::cout << " Where" << std::endl;
		std::cout << "  <dir>               Directory (folder) containing Genemark-E HMM model parameter files" << std::endl;
		std::cout << "  <introns>           Input gapped alignment candidates (SAM format)" << std::endl;
		std::cout << "  <seq>               DNA sequence (FASTA)" << std::endl;
		std::cout << "  <ref>               Reference annotation (GFF)" << std::endl;
		std::cout << "  <multi_out>         Output gapped alignment candidates (after filters are applied)" << std::endl;
		std::cout << std::endl;
		std::cout << " Options:" << std::endl;
		std::cout << "  -prefix <pref>      Prepend <pref> to sequence names in ref annotation (GFF) file" << std::endl;
		std::cout << "  -gtf                Reference annotation <ref> is in GTF format" << std::endl;
		std::cout << "  -multi              Input file <introns> is in TrueSight multi-candidate format" << std::endl;
		std::cout << "  -scr                Input file <introns> is in TrueSight score format (e.g., final_junctions)" << std::endl;
		std::cout << "  -human              introns and DNA sequence are from human genome" << std::endl;
		std::cout << "  -genic              output only genic introns" << std::endl;
		std::cout << "  -nongenic           output only non-genic introns" << std::endl;
		std::cout << "  -strand             Filter out alignments having a strand conflict with gene reference or predictions" << std::endl;
		std::cout << "  -fs                 Filter out alignments which appear to have coding sequence on both sides, but in different reading frame" << std::endl;
		std::cout << "  -splice             Filter by splice site scores" << std::endl;
		std::cout << "  -cod_flank m        m = length of intron flanking sequences to score with protein coding models (default = 81)" << std::endl;
		std::cout << "  -nseg n             n = number of segments each read has been divided into (by TrueSight)" << std::endl;
		std::cout << "  -slen l             l = length of the segments" << std::endl;
		std::cout << "  -sum                partition introns by (don + acc) score, instead of separately" << std::endl;
		//std::cout << "  -don_low t          t = threshold for negative set candidacy (default=8.5)" << std::endl;
		//std::cout << "  -acc_low t          t = threshold for negative set candidacy (default=19.5)" << std::endl;
		//std::cout << "  -don_high t         t = threshold for positive set candidacy (default=11)" << std::endl;
		//std::cout << "  -acc_high t         t = threshold for positive set candidacy (default=21)" << std::endl;
		std::cout << "  -off1 n             Shift input aln boundary (ORG or SAM) to the right by n nt (negative n shifts to left)" << std::endl;
		std::cout << "  -off2 n             Shift input aln boundary (ORG or SAM) to the right by n nt (negative n shifts to left)" << std::endl;
		std::cout << std::endl;
		std::cout << "\nVersion: " << PROG_VER << std::endl;
		std::cout << std::endl;
		std::cout << "This program takes a HMM model(s), multi alignment file, and reference sequence, and removes," << std::endl;
		std::cout << " all alignment candidates which fail one or more designated filtering criteria: strand conflict," << std::endl;
		std::cout << " frame shift, and splice site score." << std::endl;
		std::cout << std::endl;
		return true;
	}

	//Parse argv for required options
	int ifileCount = 0;

	for (int i=1; i<argc; ++i) {
		std::string opt = (char*)argv[i];
		if (opt.at(0)=='-')
		{
			if (opt == "-verb") {
				m_bVerbose = true;
			}
			else if (opt == "-heur") {
				m_bHeur = true;
			}
			else if (opt == "-gtf") {
				m_bGtf = true;
			}
			else if (opt == "-prefix") {
				m_prefix = (char*)argv[++i];
			}
			else if (opt == "-multi") {
				m_bMultiFile = true;
			}
			else if (opt == "-scr") {
				m_bScoreFile = true;
			}
			else if (opt == "-human") {
				m_bHuman = true;
			}
			else if (opt == "-genic") {
				m_bGenic = true;
			}
			else if (opt == "-nongenic") {
				m_bNonGenic = true;
			}
			else if (opt == "-cod_flank") {
				m_nCodFlank = atoi( (char*)argv[++i] );
			}
			else if (opt == "-off1") {
				m_nOff1 = atoi( (char*)argv[++i] );
			}
			else if (opt == "-off2") {
				m_nOff2 = atoi( (char*)argv[++i] );
			}
			else if (opt == "-splice") {
				m_bSpliceSites = true;
			}
			else if (opt == "-fs") {
				m_bFrameShift = true;
			}
			else if (opt == "-strand") {
				m_bStrand = true;
			}
			else if (opt == "-nseg") {
				m_nSegments = atoi( (char*)argv[++i] );
			}
			else if (opt == "-slen") {
				m_nSegLength = atoi( (char*)argv[++i] );
			}
			else if (opt == "-sum") {
				m_bSortBySum = true;
			}
			else if (opt == "-don_low") {
				m_lfDonThresholdLow = atof( (char*)argv[++i] );
			}
			else if (opt == "-acc_low") {
				m_lfAccThresholdLow = atof( (char*)argv[++i] );
			}
			else if (opt == "-don_high") {
				m_lfDonThresholdHigh = atof( (char*)argv[++i] );
			}
			else if (opt == "-acc_high") {
				m_lfAccThresholdHigh = atof( (char*)argv[++i] );
			}
			else
				return true;
		}
		else
		{
			if (ifileCount==0)
			{
				m_modInDir = opt;
				++ifileCount;
			}
			else if (ifileCount==1)
			{
				m_multiInFile = opt;
				++ifileCount;
			}
			else if (ifileCount==2)
			{
				m_seqInFile = opt;
				++ifileCount;
			}
			else if (ifileCount==3)
			{
				m_refInFile = opt;
				++ifileCount;
			}
			else if (ifileCount==4)
			{
				m_multiOutFile = opt;
				++ifileCount;
			}
			else
				return true;
		}
	}

	if (ifileCount != 5)
		return true;		// incorrect number of arguments

	return false;
}
