
#include "options.h"
#include "geneSet.h"
#include "model.h"
#include "locus.h"
#include "junc.h"
#include "sseq.h"
#include "metrics_sr.h"
#include "multi.h"
#include "defines.h"
#include <string>
#include <vector>

using std::string;
using std::vector;


#ifdef _WINDOWS_
	// these includes are necessary to support directory existence checks in windows
	#include <direct.h>
	#include <io.h>
#else
	#include <dirent.h>
#endif

// struct for managing file sets
typedef struct tagFileRec {
	string strName;
	int size;
} sFileRec;

// globals
//////////////////////////////////////////////////////////////////
vector<cModel> heur_models;
char alphabet[256];
int startGc;


// prototypes
//////////////////////////////////////////////////////////////////
bool initScore(string& dir);
void setAlphabet( char* ptr );
bool getFilesInDir(vector<sFileRec>& files, string dir);
bool getFilesInDir(vector<sFileRec>& files, string dir, string strNameTag);
bool loadIntrons( int& nFormat, SSeq& seq, cOptions& opts );
bool loadReference( int& nFormat, cOptions& opts );
void getQueryIntrons( vector<sIntron>& introns, int nFormat, int i, string seqName );
void getRefIntrons( vector<sIntron>& introns, int nFormat, int i, string seqName );
bool loadCalls( cOptions& opts );
bool scoreQuery(int nRefFormat, int nQueryFormat, SSeq& seq, cOptions& opts);
void scoreIntronFlanks(cOptions& opts, SSeq *pseq, bool bMark);
void splitSeq( vector<string>& seqs, vector<string>& idseq, vector<string>& ids, vector<int>& ibegs, vector<int>& iends,
			    vector<int>& iJcnTypes, vector<char>& strands, vector<int>& gcs, char *buf );
void splitSeq( vector<string>& seqs, vector<string>& idseq, vector<string>& ids, vector<int>& ibegs, vector<int>& iends, vector<int>& iJcnTypes,
					  vector<char>& strands, vector<int>& gcs, vector<int>& iFrames, vector<int>& iMark_5s, vector<int>& iMark_3s, vector<int>& iMark_Is,
					  vector<int>& nCoverage, vector<int>& nIntronsPerGene, vector<int>& indxIntron, vector<int>& dxAtg, vector<int>& dxStp,
					  vector<int>& fpL, vector<int>& fpR, vector<int>& dDon, vector<int>& dAcc, vector<string>& feat, char *buf );
void splitSeq( vector<string>& seqs5, vector<string>& inner5seqs, vector<string>& inner3seqs, vector<string>& seqs3, char *buf );
void splitSeq( vector<string>& seqs, char *buf );
int countDonAcc( int& nDon, int& nAcc, vector<sIntron>& introns );

extern void reset();

bool fileRecLessThan (sFileRec fri, sFileRec frj) { return fri.strName.compare(frj.strName) < 0; }


enum { _FORMAT_UNKNOWN = -1, _FORMAT_JUNC, _FORMAT_SCORE, _FORMAT_ORG, _FORMAT_SAM, _FORMAT_GTF, _FORMAT_GFF, _FORMAT_SQL, _FORMAT_MULTI };

cOrigin origin;
cLoci loci, lociQuery;			// loci is only used when -metrics and -oiref options are given
cJunc junc, juncQuery;
cMulti multiQuery;
cGeneSet set;
cGeneSet geneCalls;

// main
//////////////////////////////////////////////////////////////////
int main(int argc, const char* argv[])
{
	cOptions opts;
	if (opts.parseOptions(argc,argv)) {
		if (argc > 1) {
			std::cout << "Unacceptable argument(s) given.\n";
			return 0;
		}
		return 0;
	}

	// set offsets
	lociQuery.m_nOffset1 = opts.m_nOff1;
	lociQuery.m_nOffset2 = opts.m_nOff2;
	juncQuery.m_nOffset1 = opts.m_nOff1;
	juncQuery.m_nOffset2 = opts.m_nOff2;
	multiQuery.m_nOffset1 = opts.m_nOff1;
	multiQuery.m_nOffset2 = opts.m_nOff2;

	if (opts.m_bJuncTabOnly)
		juncQuery.m_delim = " \t";

	// load HMM model parameter files
	//
	if (initScore(opts.m_modInDir)) {
		std::cout << "Aborting, file error." << std::endl;
		return 0;
	}

	// load DNA sequence
	//
	SSeq seq;
	seq.setSafeFormat(true);
	std::cout << "Loading sequence file: " << opts.m_seqInFile << std::endl;
	if (!seq.load( opts.m_seqInFile, 0 ) )
	{
		std::cout << "Unable to read input file: " << opts.m_seqInFile << std::endl;
		return 0;
	}

	// load query introns
	//
	int nQueryFormat;
	if (loadIntrons( nQueryFormat, seq, opts ))
		return 0;

	// load reference (may be either genes or introns)
	//
	int nRefFormat;
	bool bMetrics = !opts.m_gffInFile.empty();
	if (bMetrics && loadReference( nRefFormat, opts ))
		return 0;

	if (loadCalls( opts ))
		return 0;

	if (scoreQuery(nRefFormat, nQueryFormat, seq, opts))
		return 0;

	string strOut = "ref_stats.txt";
	set.writeSummary(opts.m_gffInFile);

	return 0;
}


//
//////////////////////////////////////////////////////////////////
bool initScore(string& dir)
{
	vector<sFileRec> files;

	if (getFilesInDir(files, dir))
		return true;

	heur_models.clear();
	cModel mod;
	cSite site;

	mod.init( &site );

	string fullFile;

	for (int i=0; i<(int)files.size(); ++i)
	{
		fullFile = dir + SLASH;
		fullFile.append( files[i].strName );

		std::cout << "Loading model file: " << fullFile << std::endl;

		if (mod.load( fullFile ))
		{
			std::cout << "Error loading file: " << files[i].strName << std::endl;
			return true;
		}

		heur_models.push_back( mod );
		reset();
		//heur_models[i].gc = ++gc;
	}

	// set alphabet
	setAlphabet( alphabet );

	return false;
}

//
//////////////////////////////////////////////////////////////////
void createContigLoci(cLoci& loci, SSeq& seq)
{
	int nseq = seq.getNumContigs();
	locusRec lr;
	string defline;
	int i;

	for (i=0; i<nseq; ++i)
	{
		lr.ibeg = 1;
		lr.iend = seq.getLength( i );
		seq.getDefline( defline, i );
		lr.name = seq.getNameTok( defline );

		loci.addLocus( lr );
	}
}

//
//////////////////////////////////////////////////////////////////
void setAlphabet( char* ptr )
{
	memset( ptr, 4, 256 );

	*(ptr + 'a' ) = 0;
	*(ptr + 'A' ) = 0;
	*(ptr + 'c' ) = 1;
	*(ptr + 'C' ) = 1;
	*(ptr + 'g' ) = 2;
	*(ptr + 'G' ) = 2;
	*(ptr + 't' ) = 3;
	*(ptr + 'T' ) = 3;
	*(ptr + 'u' ) = 3;
	*(ptr + 'U' ) = 3;
}

//
//////////////////////////////////////////////////////////////////
bool getFilesInDir(vector<sFileRec>& files, string dir)
{
	string allfiles = "*";
    return getFilesInDir(files, dir, allfiles);
}

//
//////////////////////////////////////////////////////////////////
bool getFilesInDir(vector<sFileRec>& files, string dir, string strNameTag)
{
	files.clear();

	sFileRec file_record;
	string fname;

#ifdef _WINDOWS_
	// windows directory interrogation library

	int ANY_FILE =  _A_NORMAL | _A_RDONLY | _A_HIDDEN | _A_SYSTEM | _A_ARCH;

	struct _finddatai64_t data;
	// First create the filename that will be use to initialize the find.
	// "*.*" are wild card characters that tells the find function to return a 
	// list of all the files and directories.  You can limit this if you wish
	// to just file with specific extensions, for example "*.txt".  If you do that
	// then finder will not return any directory names.

	// start the finder -- on error _findfirsti64() will return -1, otherwise if no
	// error it returns a handle greater than -1.

	//if (dir.at(dir.length()-1)==SLASH)
	//	fname = dir + wildcard;
	//else
	//{
	//	fname = dir + SLASH;
	//	fname += wildcard;
	//}
	fname = dir + "\\*.*";

	long h = _findfirsti64(fname.c_str(),&data);
	if(h >= 0)
	{
		do {
			if( (data.attrib & ANY_FILE) )
			{
				// its a file
				fname = data.name;
				if (strNameTag.compare("*")==0 || strNameTag.empty())
				{
					file_record.strName = data.name;
					file_record.size = (int) data.size;
					files.push_back( file_record );
				}
				else if (fname.find(strNameTag.c_str(),0)!=string::npos)
				{
					file_record.strName = data.name;
					file_record.size = (int) data.size;
					files.push_back( file_record );
				}
			}
			else
			{
				// ignore
			}
		} while ( _findnexti64(h,&data) == 0);
		// close the find handle.  
		_findclose(h);

	}

#else
	// unix directory interrogation library
	#include <sys/types.h>
	#include <sys/stat.h>

    DIR *dp;
    struct dirent *dirp;
    if((dp  = opendir(dir.c_str())) == NULL) {
		return true;
    }

	string str;
    while ((dirp = readdir(dp)) != NULL)
	{
		str = string(dirp->d_name);

		if (str.size()>0)
		{
			if (str.at(0)!='.')
			{
				struct stat st;
				size_t size;
				if ( !stat( dirp->d_name, &st ) )
					size = st.st_size;

				if (strNameTag.compare("*")==0 || strNameTag.empty())
				{
					file_record.strName = string(dirp->d_name);
					file_record.size = (int) size;
					files.push_back( file_record );
				}
				else if (str.find(strNameTag.c_str(),0)!=string::npos)
				{
					file_record.strName = string(dirp->d_name);
					file_record.size = (int) size;
					files.push_back( file_record );
				}
			}
		}
    }
    closedir(dp);

#endif

	sort (files.begin(), files.end(), fileRecLessThan);

    return false;
}

//
//////////////////////////////////////////////////////////////////
bool loadIntrons( int& nFormat, SSeq& seq, cOptions& opts )
{
	nFormat = _FORMAT_UNKNOWN;

	string fname;
	bool bSaveOi(false);
	int ct(0), nio;

	if ( !opts.m_orgInFile.empty() )
	{
		nFormat = _FORMAT_ORG;
		std::cout << "Loading origin file: " << opts.m_orgInFile << "..." << std::endl;

		if ( lociQuery.load(opts.m_orgInFile) ) {
			std::cout << "Error reading origin file: " << opts.m_orgInFile << std::endl;
			return true;
		}

		if (opts.m_bClean)
		{
			// remove redundant introns (must be sorted first!)
			std::cout << "Cleaning..." << std::endl;
			lociQuery.deleteOe();			// remove exonic read alignments
			lociQuery.cleanIntrons();
		}

		nio = lociQuery.ctIntrons();
		printf("%d gapped alignments loaded (TOTAL).\n", nio);
	}
	else if ( !opts.m_samInFile.empty() )	// read SAM file
	{
		nFormat = _FORMAT_SAM;
		std::cout << "Loading short read alignments and compiling origins into loci..." << std::endl;
		createContigLoci(lociQuery, seq);
		while (!origin.IsDone())
		{
			// load some short read alignments
			if ( origin.loadSamBlock(opts.m_samInFile, opts.m_nLinesPerBlock) ) {
				std::cout << "Error loading SAM file, aborting.\n";
				return true;
			}

			// add them to the appropriate loci
			lociQuery.addOrigins( origin.m_aligns );

			nio = lociQuery.ctIntrons();
			printf("%d gapped alignments loaded.\n", nio);

			// clear origins
			origin.reset();
		}

		nio = lociQuery.ctIntrons();
		printf("%d gapped alignments loaded (TOTAL).\n", nio);

		if (opts.m_bClean)
		{
			// remove redundant introns (must be sorted first!)
			std::cout << "Cleaning..." << std::endl;
			lociQuery.deleteOe();			// remove exonic read alignments
			lociQuery.cleanIntrons();

			bSaveOi = true;
		}

		fname = "origins.org";
		std::cout << "Saving origins to file: " << fname << std::endl;
		lociQuery.writeOrigins( fname );

	}
	else if ( !opts.m_juncInFile.empty() )
	{
		nFormat = _FORMAT_JUNC;
		std::cout << "Reading JUNC intron file " << opts.m_juncInFile << std::endl;
		if (juncQuery.loadFile( opts.m_juncInFile )) {
			std::cout << "Error reading junc file: " << opts.m_juncInFile << std::endl;
			return true;
		}
	}
	else if ( !opts.m_scoreInFile.empty() )
	{
		nFormat = _FORMAT_SCORE;
		std::cout << "Reading SCORE intron file " << opts.m_scoreInFile << std::endl;
		juncQuery.m_scoreThresh = opts.m_scoreThresh;
		if (juncQuery.loadScoreFile( opts.m_scoreInFile )) {
			std::cout << "Error reading score file: " << opts.m_scoreInFile << std::endl;
			return true;
		}
		juncQuery.m_bScoreFmt = true;
	}
	else if ( !opts.m_multiInFile.empty() )
	{
		nFormat = _FORMAT_MULTI;
		std::cout << "Reading MULTI intron file " << opts.m_multiInFile << std::endl;
		if (multiQuery.load( opts.m_multiInFile )) {
			std::cout << "Error reading multi file: " << opts.m_multiInFile << std::endl;
			return true;
		}
	}
	else {
		std::cout << "Missing query intron file! Aborting..." << std::endl;
		return true;
	}

	return false;
}

//
//////////////////////////////////////////////////////////////////
bool loadReference( int& nFormat, cOptions& opts )
{
	nFormat = _FORMAT_UNKNOWN;

	if (opts.m_bOiFormat) {
		nFormat = _FORMAT_ORG;
		std::cout << "Loading reference origin file: " << opts.m_gffInFile << "..." << std::endl;
		if ( loci.load(opts.m_gffInFile) ) {
			std::cout << "Error reading origin file: " << opts.m_gffInFile << std::endl;
			return true;
		}
	}
	else if (opts.m_bJunc) {
		nFormat = _FORMAT_JUNC;
		std::cout << "Loading reference junc file: " << opts.m_gffInFile << "..." << std::endl;
		junc.m_bGetUtrStatus = true;
		junc.m_bStrandCol = opts.m_bJuncStrand;
		junc.m_refPrefix = opts.m_refPrefix;
		if ( junc.loadFile(opts.m_gffInFile) ) {
			std::cout << "Error reading junc file: " << opts.m_gffInFile << std::endl;
			return true;
		}
	}
	else if (opts.m_bSqlFormat) {
		nFormat = _FORMAT_SQL;
		std::cout << "Loading reference origin file: " << opts.m_gffInFile << "..." << std::endl;
		if ( set.loadTable(opts.m_gffInFile) ) {
			std::cout << "Error reading origin file: " << opts.m_gffInFile << std::endl;
			return true;
		}
	}
	else if (opts.m_bGtfFormat) {
		nFormat = _FORMAT_GTF;
		std::cout << "Loading reference annotation GTF file: " << opts.m_gffInFile << std::endl;
		if (set.loadGtf(opts.m_gffInFile))
		{
			std::cout << "Unable to read input file: " << opts.m_gffInFile << std::endl;
			return true;
		}
	}
	else {
		nFormat = _FORMAT_GFF;
		std::cout << "Loading reference annotation GFF file: " << opts.m_gffInFile << std::endl;
		set.setPrefix( opts.m_refPrefix );
		if (set.loadGff(opts.m_gffInFile))
		{
			std::cout << "Unable to read input file: " << opts.m_gffInFile << std::endl;
			return true;
		}
		//gffRef.setPrefix( opts.m_refPrefix );
	}

	return false;
}

//
//////////////////////////////////////////////////////////////////
bool loadCalls( cOptions& opts )
{
	if (!opts.m_gffCallsFile.empty()) {
		std::cout << "Loading gene calls in GFF file: " << opts.m_gffCallsFile << std::endl;
		if (geneCalls.loadGff(opts.m_gffCallsFile))
		{
			std::cout << "Unable to read input file: " << opts.m_gffCallsFile << std::endl;
			return true;
		}
	}

	return false;
}

//
//////////////////////////////////////////////////////////////////
void getQueryIntrons( vector<sIntron>& introns, int nFormat, int i, string seqName )
{
	switch (nFormat)
	{
	case _FORMAT_JUNC: case _FORMAT_SCORE:
		introns = juncQuery.getIntrons( seqName ); break;
	case _FORMAT_MULTI:
		introns = multiQuery.getIntrons( seqName, i ); break;
	default:
		introns = lociQuery.getIntrons( i, seqName );
	};
}

//
//////////////////////////////////////////////////////////////////
void getRefIntrons( vector<sIntron>& introns, int nFormat, int i, string seqName )
{
	//if (opts.m_bRnaFormat) {
	//	contig_genes = gffRef.getRnas( i, seq, false );
	//	gffRef.cr2int( ref_introns, contig_genes );
	//}

	switch (nFormat)
	{
	case _FORMAT_ORG:
		introns = loci.getIntrons( i, seqName ); break;
	case _FORMAT_JUNC:
		introns = junc.getIntrons( seqName ); break;
	default:
		introns = set.getRichIntrons( seqName );
	};
}

//
//////////////////////////////////////////////////////////////////
bool scoreQuery(int nRefFormat, int nQueryFormat, SSeq& seq, cOptions& opts)
{
	string defline;
	string seqName;

	startGc = opts.m_nGcStart;

	cMetricsSR metrics;
	metrics.reset();

	cSite site;
	site.setMark( true );

	int nSeq = seq.getNumContigs();

	int nDonContigQ(0), nAccContigQ(0), nDonQuery(0), nAccQuery(0), nUtrQuery(0);
	int nDonContig(0), nAccContig(0), nDonRef(0), nAccRef(0), nUtrRef(0);
	int nIntronRefCount = 0;
	int nQueryCount = 0;

	vector<sIntron> que_introns, ref_introns;
	vector<codingRegion> contig_genes;

	FILE *pfo, *pfr;

	pfo = fopen("org.txt","w");
	pfr = fopen("ref.txt","w");
	int i, j;
	bool bGff(false);
	bool bMetrics = !opts.m_gffInFile.empty();

	for (i=0; i<nSeq; ++i)
	{
		// contig name
		seq.getDefline( defline, i );
		seqName = seq.getNameTok( defline );

		// get query introns
		getQueryIntrons( que_introns, nQueryFormat, i, seqName );

		if (que_introns.empty())
			printf("que_introns is empty for seq: %s\n", seqName.c_str());

		if (bMetrics)
		{
			getRefIntrons( ref_introns, nRefFormat, i, seqName );

			if (nRefFormat == _FORMAT_JUNC) {
				site.getStrand( ref_introns, &seq, i, opts.m_nOff1, opts.m_nOff2 );
			}

			// get GFF annotated introns
			if (opts.m_bVerbose)
				printf("getting introns... for seq %d (%s)\n", i, seqName.c_str());

			site.getStrand( que_introns, &seq, i, opts.m_nOff1, opts.m_nOff2 );

			if (ref_introns.empty())
				printf("ref_introns is empty for seq: %s\n", seqName.c_str());

			nUtrRef += countDonAcc( nDonContig, nAccContig, ref_introns );
			nDonRef += nDonContig;
			nAccRef += nAccContig;
			nIntronRefCount += ref_introns.size();

			nUtrQuery += countDonAcc( nDonContigQ, nAccContigQ, que_introns );
			nDonQuery += nDonContigQ;
			nAccQuery += nAccContigQ;
			nQueryCount += que_introns.size();

			//std::cout << "seqName = " << seqName << ", que_introns.size() = " << que_introns.size() << std::endl;

			metrics.m_chr = seqName;

			if ( metrics.markAgreement(ref_introns, que_introns, 0) )
			{
				std::cout << "Problem with intron query comparison." << std::endl;
				return true;
			}

			if ( metrics.markAgreement(que_introns, ref_introns, 1) )
			{
				std::cout << "Problem with intron query comparison." << std::endl;
				return true;
			}

			if ( metrics.getNN(ref_introns, que_introns) )
			{
				std::cout << "Problem with NN determination." << std::endl;
				return true;
			}

			for (j=0; j<(int)que_introns.size(); ++j)
			{
				fprintf(pfo,"%s\t%d\t%d\t%d\t%d\t%d\t%f\t%d\t%d\t%d\n", seqName.c_str(), que_introns[j].ibeg, que_introns[j].iend, que_introns[j].iframe,
					                                    que_introns[j].dDon, que_introns[j].dAcc, que_introns[j].score,
														que_introns[j].lenExonL, que_introns[j].lenExonR, que_introns[j].extr_I);
			}
			for (j=0; j<(int)ref_introns.size(); ++j)
			{
				fprintf(pfr,"%s\t%d\t%d\t%d\t%d\n", ref_introns[j].strId.c_str(), ref_introns[j].iStrand, ref_introns[j].ibeg, ref_introns[j].iend, ref_introns[j].iframe);
			}

			if (nRefFormat==_FORMAT_GFF || nRefFormat==_FORMAT_SQL)
			{	// categorize false positives
				set.markIntronFp( que_introns, seqName );
			}
			else if (!opts.m_gffCallsFile.empty())
			{
				geneCalls.markIntronFp( que_introns, seqName );
			}
		}

		if (!opts.m_bMetricsOnly)
		{
			site.initOrigins( que_introns, &seq, i );

			//std::cout << "Extracting gapped alignment features from " << seqName << "..." << std::endl;
			if ( site.getGappedAlignSitesHet(i, opts.m_nCodFlank, false) )
			{
				std::cout << "Problem with site extraction." << std::endl;
				return true;
			}

			site.recordGappedAlignData();
			site.clearData();
		}

	}

	fclose(pfo);
	fclose(pfr);

	// calculate prediction metrics
	string fout = "metrics.txt";
	if (bMetrics)
	{
		std::cout << "Writing performance metrics file..." << std::endl;
		metrics.writeMetrics( fout, nIntronRefCount, nDonRef, nAccRef, nQueryCount, nDonQuery, nAccQuery, nUtrRef );
	}

	if (!opts.m_bMetricsOnly)
	{
		// score submodels
		//
		std::cout << "Scoring intron alignments against models..." << std::endl;
		scoreIntronFlanks(opts, &seq, bMetrics);
	}

	return false;
}

//
//////////////////////////////////////////////////////////////////
void scoreIntronFlanks(cOptions& opts, SSeq *pseq, bool bMark)
{
	int i, ix, gc, gc_int;
	vector<string> intronSeq, flank5Seq, flank3Seq, inner5seq, inner3seq, idseq, ids;
	vector<char> strands;
	vector<int> ibegs, iends, ijcntypes, iFrame, iMark5, iMark3, iMarkI, fpL, fpR, dDon, dAcc, gcs;
	vector<int> nCoverage, nIntronsInGene, ixIntron, posAtg, posStp, flankGC;
	vector<double> flank5Score0, flank5Score1, flank5Score2, flank3Score0, flank3Score1, flank3Score2;
	vector<double> intron5Score0, intron5Score1, intron5Score2, intron3Score0, intron3Score1, intron3Score2;
	vector<double> intron0, intron1, intron2;
	// background model scores
	vector<double> intron5Back, flank5Back, intron3Back, flank3Back;
	vector<double> intronBack;
	vector<string> feat;

	string fname = opts.m_scoreOutFile;
	bool bHuman = opts.m_bHuman;
	int gc_bias = opts.m_gcBias;
	bool bScoreSites = opts.m_bSpliceSites;

	string str;

	std::cout << "loading introns.txt file" << std::endl;

	// intron sequence
	FileBuf f;
	f.Load( "introns.txt" );
	std::cout << "splitting sequences" << std::endl;
	if (bMark)
		splitSeq( intronSeq, idseq, ids, ibegs, iends, ijcntypes, strands, gcs, iFrame, iMark5, iMark3, iMarkI,
		          nCoverage, nIntronsInGene, ixIntron, posAtg, posStp, fpL, fpR, dDon, dAcc, feat, f.begin );
	else
		splitSeq( intronSeq, idseq, ids, ibegs, iends, ijcntypes, strands, gcs, f.begin );
	f.Delete();

	// 5' and 3' intron flanking sequence
	std::cout << "loading flank.txt file" << std::endl;
	f.Load( "flank.txt" );
	splitSeq( flank5Seq, inner5seq, inner3seq, flank3Seq, f.begin );
	f.Delete();

	vector<string> donSeqs, accSeqs;

	// donor and acceptor sites
	std::cout << "loading donor and acceptor sites" << std::endl;
	f.Load( "don_site.txt" );
	splitSeq( donSeqs, f.begin );
	f.Delete();

	f.Load( "acc_site.txt" );
	splitSeq( accSeqs, f.begin );
	f.Delete();

	std::cout << "finished loading splice sites" << std::endl;

	// intron sequence scoring
	//
	double logScore;
	double logScore0, logScore1, logScore2;
	double logScore0_5, logScore1_5, logScore2_5;
	double logScore0_3, logScore1_3, logScore2_3;
	char alphabet[256];
	cMatrixMod tmpMat;
	tmpMat.SetAlphabet( alphabet );
	int lnIntron;
	int order;

	double gcFlank5, gcFlank3;
	int gc_5_3;

	str.insert( str.begin(), 96, 'A' );
	str.replace( 93, 1, 1, 'T');

	int gcIntron;

	double logDon0, logDon1, logDon2, logAcc0, logAcc1, logAcc2;
	vector<double> donScore0, donScore1, donScore2, accScore0, accScore1, accScore2;


	std::cout << "calculating scores" << std::endl;

	// intron scores
	for (i=0; i<(int)flank5Seq.size(); ++i)
	{
//		std::cout << "i = " << i << std::endl;

		gc = gcs[i] + gc_bias;

		if (bHuman) {
			if (gc <= 40)
				ix = 0;
			else if (gc > 40 && gc <= 49)
				ix = 1;
			else
				ix = 2;
		}
		else {
			ix = gc - startGc;
		}

		if (ix < 0) ix = 0;
		if (ix >= (int)heur_models.size()) ix = heur_models.size() - 1;

		order = heur_models[ix].m_order;

		gcIntron = (int) ( pseq->GetGC( intronSeq[i] ) + 0.5 );
		gc_int = gcIntron - startGc;
		if (gc_int < 0) gc_int = 0;
		if (gc_int >= (int)heur_models.size()) gc_int = heur_models.size() - 1;

		// full intron sequence scores
		if ( heur_models[ix].scoreMarkov3( logScore0, order, 1, intronSeq[i].c_str(), alphabet ) ) { printf("intronSeq 1\n"); break; }
		if ( heur_models[ix].scoreMarkov3( logScore1, order, 2, intronSeq[i].c_str(), alphabet ) ) { printf("intronSeq 2\n"); break; }
		if ( heur_models[ix].scoreMarkov3( logScore2, order, 3, intronSeq[i].c_str(), alphabet ) ) { printf("intronSeq 3\n"); break; }

		intron0.push_back( logScore0_5 );
		intron1.push_back( logScore1_5 );
		intron2.push_back( logScore2_5 );

		// intron flanking sequence
		// 5'
		if ( heur_models[ix].scoreMarkov3( logScore0_5, order, 1, flank5Seq[i].c_str(), alphabet ) ) { printf("flank5Seq 1\n"); break; }
		if ( heur_models[ix].scoreMarkov3( logScore1_5, order, 2, flank5Seq[i].c_str(), alphabet ) ) { printf("flank5Seq 2\n"); break; }
		if ( heur_models[ix].scoreMarkov3( logScore2_5, order, 3, flank5Seq[i].c_str(), alphabet ) ) { printf("flank5Seq 3\n"); break; }

		// 3'
		if ( heur_models[ix].scoreMarkov3( logScore0_3, order, 1, flank3Seq[i].c_str(), alphabet ) ) { printf("flank3Seq 1\n"); break; }
		if ( heur_models[ix].scoreMarkov3( logScore1_3, order, 2, flank3Seq[i].c_str(), alphabet ) ) { printf("flank3Seq 2\n"); break; }
		if ( heur_models[ix].scoreMarkov3( logScore2_3, order, 3, flank3Seq[i].c_str(), alphabet ) ) { printf("flank3Seq 3\n"); break; }

		flank5Score0.push_back( logScore0_5 );
		flank5Score1.push_back( logScore1_5 );
		flank5Score2.push_back( logScore2_5 );

		flank3Score0.push_back( logScore0_3 );
		flank3Score1.push_back( logScore1_3 );
		flank3Score2.push_back( logScore2_3 );

		//// intron terminal sequence scores
		//// 5'
		//if ( heur_models[ix].scoreMarkov3( logScore0, order, 1, inner5seq[i].c_str(), alphabet ) ) { printf("inner5seq 1\n"); break; }
		//if ( heur_models[ix].scoreMarkov3( logScore1, order, 2, inner5seq[i].c_str(), alphabet ) ) { printf("inner5seq 2\n"); break; }
		//if ( heur_models[ix].scoreMarkov3( logScore2, order, 3, inner5seq[i].c_str(), alphabet ) ) { printf("inner5seq 3\n"); break; }

		//intron5Score0.push_back( logScore0 );
		//intron5Score1.push_back( logScore1 );
		//intron5Score2.push_back( logScore2 );

		//// intron terminal sequence scores
		//// 3'
		//if ( heur_models[ix].scoreMarkov3( logScore0, order, 1, inner3seq[i].c_str(), alphabet ) ) { printf("inner3seq 1\n"); break; }
		//if ( heur_models[ix].scoreMarkov3( logScore1, order, 2, inner3seq[i].c_str(), alphabet ) ) { printf("inner3seq 2\n"); break; }
		//if ( heur_models[ix].scoreMarkov3( logScore2, order, 3, inner3seq[i].c_str(), alphabet ) ) { printf("inner3seq 3\n"); break; }

		//intron3Score0.push_back( logScore0 );
		//intron3Score1.push_back( logScore1 );
		//intron3Score2.push_back( logScore2 );

		//if ( heur_models[ix].orfScoring( pfOrf, i+1, order, orfLen, iorfFrame, orfScore, flank5Seq[i].c_str(), flank3Seq[i].c_str(), alphabet ) )
		//	{ printf("orfScore\n"); break; }

		gcFlank5 = pseq->GetGC( flank5Seq[i] );
		gcFlank3 = pseq->GetGC( flank3Seq[i] );
		gc_5_3 = int( (gcFlank5+gcFlank3)/2. + .5 );

		flankGC.push_back( gc_5_3 );

		//orf_.push_back( orfScore[0] );
		//orf_C.push_back( orfScore[1] );
		//orf_CC.push_back( orfScore[2] );

		//if ( heur_models[gc_int].orfIntronScoring( tronScore, order, intronSeq[i].c_str(), alphabet ) )
		//	{ printf("orfIntronScoring\n"); break; }

		if (bScoreSites)
		{
			heur_models[ix].scoreDonSite3( logDon0, logDon1, logDon2, donSeqs[i], ijcntypes[i] );
			heur_models[ix].scoreAccSite3( logAcc0, logAcc1, logAcc2, accSeqs[i], ijcntypes[i] );

			donScore0.push_back(logDon0);
			donScore1.push_back(logDon1);
			donScore2.push_back(logDon2);

			accScore0.push_back(logAcc0);
			accScore1.push_back(logAcc1);
			accScore2.push_back(logAcc2);
		}

		//orfLens.push_back( orfLen[0] );
		//orfLensC.push_back( orfLen[1] );
		//orfLensCC.push_back( orfLen[2] );
		//orfFrame.push_back( iorfFrame );
		//orfIntr.push_back( tronScore );

		// get background model scores
		heur_models[ix].scoreMarkovBack( logScore, order, flank5Seq[i].c_str(), alphabet );	flank5Back.push_back( logScore );
		heur_models[ix].scoreMarkovBack( logScore, order, flank3Seq[i].c_str(), alphabet );	flank3Back.push_back( logScore );
		heur_models[ix].scoreMarkovBack( logScore, order, inner5seq[i].c_str(), alphabet );	intron5Back.push_back( logScore );
		heur_models[ix].scoreMarkovBack( logScore, order, inner3seq[i].c_str(), alphabet );	intron3Back.push_back( logScore );
		heur_models[ix].scoreMarkovBack( logScore, order, intronSeq[i].c_str(), alphabet );	intronBack.push_back( logScore );
	}

	std::cout << "writing score file" << std::endl;

	FILE *pF_score = fopen( fname.c_str(), "w" );
	string strtype;

	fprintf(pF_score, "# ID_seq\tID_intr\tbeg\tend\tstrand\tjcn\tflank5_0\tflank5_1\tflank5_2\tflank5back\t");
	fprintf(pF_score, "flank3_0\tflank3_1\tflank3_2\tflank3_back\tintron0\tintron1\tintron2\tintronBack");
	if (bMark)
		fprintf(pF_score, "\tFlank_GC\tframe\t5p\t3p\tann\tFPL\tFPR\tCov\tdDon\tdAcc\tIpG\tixIntron\tposAtg\tposStp\tfeat");
	else
		fprintf(pF_score, "\tFlank_GC\tORF_length\tORF_frame\tORF_score\tORF_intr");
	
	if (bScoreSites) {
		fprintf(pF_score, "\tdon0\tdon1\tdon2\tacc0\tacc1\tacc2");
	}
	fprintf(pF_score, "\n");

	for (i=0; i<(int)flank5Seq.size(); ++i)
	{
		switch (ijcntypes[i])
		{
		case INTRON_GT_AG:
			strtype = "GT_AG";
			break;
		case INTRON_GC_AG:
			strtype = "GC_AG";
			break;
		case INTRON_AT_AC:
			strtype = "AT_AC";
			break;
		default:
			strtype = "?";
		};

		lnIntron = intronSeq[i].length() + 20 + 27;
		fprintf( pF_score, "%s\t%s\t%d\t%d\t%c\t%s\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e",
			idseq[i].c_str(), ids[i].c_str(), ibegs[i], iends[i],
			strands[i], strtype.c_str(),
			flank5Score0[i], flank5Score1[i], flank5Score2[i], flank5Back[i],
			flank3Score0[i], flank3Score1[i], flank3Score2[i], flank3Back[i],
			intron0[i], intron1[i], intron2[i], intronBack[i] );

		if (bMark) {
			fprintf( pF_score, "\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d", flankGC[i], iFrame[i], iMark5[i], iMark3[i], iMarkI[i], fpL[i], fpR[i],
				nCoverage[i], dDon[i], dAcc[i], nIntronsInGene[i], ixIntron[i], posAtg[i], posStp[i] );
			
			// features carried over from input score (junction) file
			if (!opts.m_scoreInFile.empty() && juncQuery.m_nFeat > 0) {
				fprintf( pF_score, "\t%s", feat[i].c_str() );
			}
			else
				fprintf( pF_score, "\t|" );
		}

		if (bScoreSites)
			fprintf( pF_score, "\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e", donScore0[i], donScore1[i], donScore2[i], accScore0[i], accScore1[i], accScore2[i]);

		fprintf( pF_score, "\n" );
	}

	fclose( pF_score );
}

//
//////////////////////////////////////////////////////////////////
void splitSeq( vector<string>& seqs, vector<string>& idseq, vector<string>& ids, vector<int>& ibegs, vector<int>& iends,
			    vector<int>& iJcnTypes, vector<char>& strands, vector<int>& gcs, char *buf )
{
	string seq, strId, str_nums, s;
	char *pch, pchId[128], pchSeq[128], chStrand;
	size_t ix0;
	int ibeg, iend, ijcntype, gc;

	seqs.clear();
	ids.clear();
	strands.clear();
	iJcnTypes.clear();

	pch = strtok( buf, "\n" );
	while (pch != NULL)
	{
		seq = pch;
		ix0 = seq.find_first_of( "\t" );
		if (ix0!=string::npos)
		{
			str_nums = seq.substr(ix0);
			if ( sscanf(str_nums.c_str(), "%s %s %d %d %d %c", pchSeq, pchId, &ibeg, &iend, &ijcntype, &chStrand, &gc) != 6 )
				return;		// error
		}

		strId = pchId;
		seqs.push_back( seq );
		ids.push_back( strId );
		idseq.push_back( pchSeq );
		ibegs.push_back( ibeg );
		iends.push_back( iend );
		iJcnTypes.push_back( ijcntype );
		strands.push_back( chStrand );
		gcs.push_back( gc );

		pch = strtok( NULL, "\n" );
	}
}

void splitSeq( vector<string>& seqs, vector<string>& idseq, vector<string>& ids, vector<int>& ibegs, vector<int>& iends, vector<int>& iJcnTypes,
					  vector<char>& strands, vector<int>& gcs, vector<int>& iFrames, vector<int>& iMark_5s, vector<int>& iMark_3s, vector<int>& iMark_Is,
					  vector<int>& nCoverage, vector<int>& nIntronsPerGene, vector<int>& indxIntron, vector<int>& dxAtg, vector<int>& dxStp,
					  vector<int>& fpL, vector<int>& fpR, vector<int>& dDon, vector<int>& dAcc, vector<string>& feat, char *buf )
{
	string seq, strId, str_nums, s;
	char *pch, pchId[128], pchSeq[128], pchFeat[256], chStrand;
	size_t ix0;
	int ibeg, iend, ijcntype, iframe, imark5, imark3, imarkI, gc;
	int nCopies, nIntronsInGene , ixIntron, posAtg, posStp, ifpL, ifpR, ddon, dacc;

	seqs.clear();
	ids.clear();
	strands.clear();
	iJcnTypes.clear();
	iFrames.clear();
	iMark_5s.clear();
	iMark_3s.clear();
	iMark_Is.clear();
	fpL.clear();
	fpR.clear();
	dDon.clear();
	dAcc.clear();
	feat.clear();

	pch = strtok( buf, "\n" );

	while (pch != NULL)
	{
		seq = pch;
		ix0 = seq.find_first_of( "\t" );
		if (ix0!=string::npos)
		{
			str_nums = seq.substr(ix0, string::npos);
			if ( sscanf(str_nums.c_str(), "%s %s %d %d %d %c %d %d %d %d %d %d %d %d %d %d %d %d %d %d %s", pchSeq, pchId, &ibeg, &iend,
				&ijcntype, &chStrand, &gc, &iframe, &imark5, &imark3, &imarkI, &nCopies, &nIntronsInGene , &ixIntron,
				&posAtg, &posStp, &ifpL, &ifpR, &ddon, &dacc, pchFeat) != 21 )
				return;		// error
		}
		//sprintf( pch, "\t%s\t%s\t%d\t%d\t%d\t%c\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d", name.c_str(), intron.strId.c_str(), istart, istop, intron.iJcnType,
		//	chStrand, iframe, mark5, mark3, markI,
		//	intron.nCoverage, intron.nIntronsInGene, intron.ixIntron, intron.posAtg, intron.posStp );

		strId = pchId;
		seqs.push_back( seq );
		ids.push_back( strId );
		idseq.push_back( pchSeq );
		ibegs.push_back( ibeg );
		iends.push_back( iend );
		iJcnTypes.push_back( ijcntype );
		strands.push_back( chStrand );
		gcs.push_back( gc );
		iFrames.push_back( iframe );
		iMark_5s.push_back( imark5 );
		iMark_3s.push_back( imark3 );
		iMark_Is.push_back( imarkI );
		fpL.push_back( ifpL );
		fpR.push_back( ifpR );
		dDon.push_back( ddon );
		dAcc.push_back( dacc );
		feat.push_back( pchFeat );

		nCoverage.push_back( nCopies );
		nIntronsPerGene.push_back( nIntronsInGene );
		indxIntron.push_back( ixIntron );
		dxAtg.push_back( posAtg );
		dxStp.push_back( posStp );

		pch = strtok( NULL, "\n" );
	}
}

void splitSeq( vector<string>& seqs5, vector<string>& inner5seqs, vector<string>& inner3seqs, vector<string>& seqs3, char *buf )
{
	string seq, seq5, seq3, inn5, inn3, s;
	size_t ix0;
	char *pch;

	seqs5.clear();
	seqs3.clear();

	int ct(0);

	pch = strtok( buf, "\n" );
	while (pch != NULL)
	{
		seq = pch;

		ix0 = seq.find_first_of( "\t" );
		if (ix0!=string::npos)
		{
			seq5 = seq.substr(0,ix0);
			//seq3 = seq.substr(ix0+1, string::npos);
			seq = seq.substr(ix0+1, string::npos);
		}
		ix0 = seq.find_first_of( "\t" );
		if (ix0!=string::npos)
		{
			inn5 = seq.substr(0,ix0);
			seq = seq.substr(ix0+1, string::npos);
		}
		ix0 = seq.find_first_of( "\t" );
		if (ix0!=string::npos)
		{
			inn3 = seq.substr(0,ix0);
			seq = seq.substr(ix0+1, string::npos);
		}
		seq3 = seq;

		seqs5.push_back( seq5 );
		inner5seqs.push_back( inn5 );
		inner3seqs.push_back( inn3 );
		seqs3.push_back( seq3 );

		pch = strtok( NULL, "\n" );
	}
}

void splitSeq( vector<string>& seqs, char *buf )
{
	seqs.clear();

	char pchWhite[] = " \t\r\n";
	string seq;
	size_t istart, istop;

	char *pch = strtok( buf, "\n" );
	while (pch != NULL)
	{
		seq = pch;

		istart = seq.find_first_not_of(pchWhite);
		if (istart!=string::npos)
			seq = seq.substr(istart,string::npos);
		istop = seq.find_last_not_of(pchWhite);
		if (istop!=string::npos)
			seq = seq.substr(0,istop+1);

		seqs.push_back( seq );
		pch = strtok( NULL, "\n" );
	}
}

//
//////////////////////////////////////////////////////////////////
int countDonAcc( int& nDon, int& nAcc, vector<sIntron>& introns )
{
	nDon = nAcc = 0;
	vector<int> vidon, viacc;
	int idon, iacc;
	int nRefUtr(0);

	for (int i=0; i<(int)introns.size(); ++i)
	{
		if (introns[i].iStrand == -1)
		{	// negative strand
			idon = introns[i].iend;
			iacc = introns[i].ibeg;
		}
		else
		{	// positive strand
			idon = introns[i].ibeg;
			iacc = introns[i].iend;
		}

		vidon.push_back( idon );
		viacc.push_back( iacc );

		if (introns[i].iframe == -1)
			++nRefUtr;
	}

	vector<int>::iterator it;

	// count unique donors
	it = unique( vidon.begin(), vidon.end() );
	nDon = it - vidon.begin();
	// count unique acceptors
	it = unique( viacc.begin(), viacc.end() );
	nAcc = it - viacc.begin();

	return nRefUtr;
}
