
#include "targetver.h"
#include "options.h"
#include <stdlib.h>
#include <iostream>
#include <string>


bool cOptions::parseOptions (int argc, const char** argv)
{
	if (argc < 5) {
		std::cout << std::endl;
		std::cout << "Usage: " << PROG_NAME << " [options] [-org <org> | -sam <sam> | -jnc <junc> | -scr <score>] <dir> <seq> <score_out>" << std::endl;
		std::cout << " Where" << std::endl;
		std::cout << "  -org <org>          ORG file: all origins within each locus" << std::endl;
		std::cout << "  -sam <sam>          SAM file: all read alignments to DNA sequence" << std::endl;
		std::cout << "  -jnc <junc>         JUNC file: query introns are in .junc format" << std::endl;
		std::cout << "  -scr <score>        SCORE file: query introns are in .score format (e.g., final_junctions in TrueSight)" << std::endl;
		std::cout << "  <dir>               Directory (folder) containing Genemark-E HMM model parameter files" << std::endl;
		std::cout << "  <seq>               DNA sequence" << std::endl;
		std::cout << "  <score_out>         Output scores for each intron implied by all gapped alignments" << std::endl;
		std::cout << " Options:" << std::endl;
		std::cout << "  -human              Use human genome supervised models, with includes 3 different models (GC<=41, 41<GC<=49, 49<GC)" << std::endl;
		std::cout << "  -cod_flank m        m = length of intron flanking sequences to score with protein coding models (default = 81)" << std::endl;
		std::cout << "  -gc_bias b          b = GC bias (%) added to local GC measurement for selection of MM model" << std::endl;
		std::cout << "  -bs n               Blocksize (SAM file only): read input SAM file in increments of n lines at a time (default=1000000)" << std::endl;
		std::cout << "  -metrics <gff>      Calculate Sn/Sp metrics, using file <gff> as 'truth'" << std::endl;
		std::cout << "  -calls <gff>        Gene predictions to use if reference introns are in JUNC or ORG format" << std::endl;
		std::cout << "  -no_seq             Do a metrics-only run" << std::endl;
		//std::cout << "  -mrna               Special mRNA-only GFF format (e.g., honey bee annotation)" << std::endl;
		std::cout << "  -gtf                Reference is in GTF format (e.g., GeneMark.hmm output)" << std::endl;
		std::cout << "  -oiref              Reference is in ORG format (introns)" << std::endl;
		std::cout << "  -junc               Reference is in JUNC format (introns): 'chrom.intronStart.exonStart  nOrigins'" << std::endl;
		std::cout << "  -junctabs            Only use TAB as delimiter for JUNC file" << std::endl;
		std::cout << "  -sql                Reference is in a SQL format commonly used by human genome project (16 columns)" << std::endl;
		std::cout << "    -ccds             ccds sql format (15 columns)" << std::endl;
		std::cout << "  -jstrand            Reference is in strand-JUNC format (introns): 'chrom  intronStart  intronStop  chStrand'" << std::endl;
		std::cout << "  -prefix s           Add prefix s to names of reference introns to match sequence names" << std::endl;
		std::cout << "  -clean              Remove redundant introns from input ORG or SAM file" << std::endl;
		std::cout << "  -off1 n             Shift input aln boundary (ORG or SAM) to the right by n nt (negative n shifts to left)" << std::endl;
		std::cout << "  -off2 n             Shift input aln boundary (ORG or SAM) to the right by n nt (negative n shifts to left)" << std::endl;
		std::cout << "  -sql                Reference is in a SQL format commonly used by human genome project (16 columns)" << std::endl;
		std::cout << "  -splice             Include splice site scores" << std::endl;
		std::cout << "  -score_thresh t     If query introns are in -scr format, only count those having score >= t" << std::endl;
		std::cout << std::endl;
		std::cout << "\nVersion: " << PROG_VER << std::endl;
		std::cout << std::endl;
		std::cout << "This program takes RNA-seq read alignment input file (SAM, JUNC, or ORG format)," << std::endl;
		std::cout << " a folder of Genemark-E HMM model parameter files (.0mtx), and DNA sequence (FASTA), and scores" << std::endl;
		std::cout << " each read alignment which spans a candidate intron. Likelihood scores are calculated based" << std::endl;
		std::cout << " flanking 5' and 3' sequence (3 periodic coding sequence model, and non-coding sequence model)." << std::endl;
		std::cout << " If -metrics option is given, then intron agreement with the reference is evaluated (TP or FP)," << std::endl;
		std::cout << " and each FP is classified as being coding (incl. frame), UTR, intron, or intergenic." << std::endl;
		std::cout << std::endl;
		return true;
	}

	//Parse argv for required options
	int ifileCount = 0;

	for (int i=1; i<argc; ++i) {
		std::string opt = (char*)argv[i];
		if (opt.at(0)=='-')
		{
			if (opt == "-so") {
				m_orgOutFile = (char*)argv[++i];
			}
			else if (opt == "-cod_flank") {
				m_nCodFlank = atoi( (char*)argv[++i] );
			}
			else if (opt == "-org") {
				m_orgInFile = (char*)argv[++i];
			}
			else if (opt == "-sam") {
				m_samInFile = (char*)argv[++i];
			}
			else if (opt == "-jnc") {
				m_juncInFile = (char*)argv[++i];
			}
			else if (opt == "-scr") {
				m_scoreInFile = (char*)argv[++i];
			}
			else if (opt == "-multi") {
				m_multiInFile = (char*)argv[++i];
			}
			else if (opt == "-calls") {
				m_gffCallsFile = (char*)argv[++i];
			}
			else if (opt == "-prefix") {
				m_refPrefix = (char*)argv[++i];
			}
			else if (opt == "-metrics") {
				m_gffInFile = (char*)argv[++i];
			}
			else if (opt == "-gc_start") {
				m_nGcStart = atoi( (char*)argv[++i] );
			}
			else if (opt == "-gc_bias") {
				m_gcBias = atoi( (char*)argv[++i] );
			}
			else if (opt == "-human") {
				m_bHuman = true;
			}
			//else if (opt == "-mrna") {
			//	m_bRnaFormat = true;
			//}
			else if (opt == "-gtf") {
				m_bGtfFormat = true;
			}
			else if (opt == "-oiref") {
				m_bOiFormat = true;
			}
			else if (opt == "-junc") {
				m_bJunc = true;
			}
			else if (opt == "-junctabs") {
				m_bJuncTabOnly = true;
			}
			else if (opt == "-jstrand") {
				m_bJuncStrand = true;
			}
			else if (opt == "-sql") {
				m_bSqlFormat = true;
			}
			else if (opt == "-ccds") {
				m_bCcdsSqlFormat = true;
			}
			else if (opt == "-clean") {
				m_bClean = true;
			}
			else if (opt == "-off1") {
				m_nOff1 = atoi( (char*)argv[++i] );
			}
			else if (opt == "-off2") {
				m_nOff2 = atoi( (char*)argv[++i] );
			}
			else if (opt == "-no_seq") {
				m_bMetricsOnly = true;
			}
			else if (opt == "-splice") {
				m_bSpliceSites = true;
			}
			else if (opt == "-score_thresh") {
				m_scoreThresh = atof( (char*)argv[++i] );
			}
			else
				return true;
		}
		else
		{
			if (ifileCount==0)
			{
				m_modInDir = opt;
				++ifileCount;
			}
			else if (ifileCount==1)
			{
				m_seqInFile = opt;
				++ifileCount;
			}
			else if (ifileCount==2)
			{
				m_scoreOutFile = opt;
				++ifileCount;
			}
			else
				return true;
		}
	}

	if (ifileCount != 3)
		return true;		// incorrect number of arguments

	return false;
}
