
#pragma once

#include "gene.h"
#include "align.h"
#include "structs.h"
#include "geneType.h"
#include <map>
#include <fstream>
#include <iostream>
#include <algorithm>

using std::map;

class cGeneSet
{
public:
	cGeneSet() { m_strGeneKey = "gene"; m_iFindPos = 0; m_uid = 0; };
	~cGeneSet() { };

	bool loadGff(string& fname);
	bool loadGtf(string& fname);
	bool loadTable(string& fname);
	bool applyNameMap(string& fname, int nColFrom, int nColTo);
	void indexBySeq();

	void setPrefix(string& pref) { m_chrPrefix = pref; }

	//loadGtf(string& fname);

	bool writeSummary(string& title);
	bool writeGff(string& fname);

	vector<sIntron> getIntrons();
	vector<sIntron> getIntrons(string& seqName);
	vector<sIntron> getRichIntrons(string& seqName);
	bool findIntron(cInterval& query, int& fpL, int& fpR, vector<cInterval>& ref);
	void markIntronFp(sIntron& intron, string& seq);
	void markIntronFp(vector<sIntron>& introns, string& seqName);
	void markIntronFp(vector<sIntronSet>& multi);
	void markGenicIntron( int& nType, int nPos, vector<cGene>::iterator it, vector<cInterval>& refTrons, int nOff );

	void sort(void);
	bool findGene( cGene& gene, string& seq, int ipos );

private:
	bool loadNameMap(map<string,string>& nameMap, string& fname, int nColFrom, int nColTo);
	bool parseMapLine( string& strFrom, string& strTo, string& line, int nColFrom, int nColTo );
	bool parseGffLine( string& line );
	bool parseGtfLine( string& line );
	bool parseTableLine( string& line );
	bool getTag( string& tag, string& desc, string tagName );
	bool getGtfTag( string& tag, string desc, string tagName );
	bool getTags( vector<string>& tags, string& desc, string tagName );
	bool recs2feat(void);
	bool recs2featGtf(void);
	vector<int> findGeneRecs();
	vector<int> findAlignRecs();
	vector<int> findChildren(int ix);
	void addIsoform(void);
	void addIsoform(int ix_parent, vector<int>& vix_children);
	void addAlignment(int ix_parent, vector<int>& vix_children);
	void addGene(void);

	vector<string> getSequenceNames();

	void csv2vstring(vector<string>& labels, string& csv);
	bool isOneOf(string& query, vector<string>& set);

	vector<sIntron> interval2intron(vector<cInterval>& ctrons_in, int ix);
	vector<sIntron> interval2intron(vector<cRichInterval>& ctrons_in, int ix);
	sIntron interval2intron(cInterval& ctron);
	sIntron interval2intron(cRichInterval& ctron);

	int findBySeq(string& seq);

	vector<cGene>::iterator findGene( vector<cGene>::iterator it_beg, vector<cGene>::iterator it_end, string& seq, int ipos );

private:
	vector<cGene> m_genes;
	std::map<string,int> m_seqMap;		// index points to first gene in each chr (seq)

	cGene m_temp_gene;
	cAlign m_temp_align;
	vector<gffStruct> m_temp_records;
	cIsoform m_temp_isoform;

	string m_chrPrefix;

	int m_iFindPos;

	int m_uid;

	string m_strGeneKey;
};


void cGeneSet::sort()
{
	std::sort( m_genes.begin(), m_genes.end() );
}

bool cGeneSet::applyNameMap(string& fname, int nColFrom, int nColTo)
{
	map<string,string> nameMap;
	bool bWarn(false);

	if (loadNameMap(nameMap, fname, nColFrom, nColTo))
		return true;

	string from;
	for (int i=0; i<(int)m_genes.size(); ++i)
	{
		if (from!=m_genes[i].getSeq())
			from = m_genes[i].getSeq();

		if (nameMap.find(from)==nameMap.end() && bWarn) {
			std::cout << "Warning: no name map for " << from << std::endl;
			bWarn = true;
		}
		else {
			m_genes[i].setSeq( nameMap[from] );
			bWarn = false;
		}
	}

	return false;
}

bool cGeneSet::loadNameMap(map<string,string>& nameMap, string& fname, int nColFrom, int nColTo)
{
	std::ifstream data_file( fname.c_str() );

	if ( !data_file.is_open() )
	{
		std::cout << "error, opening file " << fname << "\n"; 
		return true;
	}

	string line, strFrom, strTo;
	nameMap.clear();

	while( !data_file.eof() )
	{
		getline( data_file, line );
		if ( parseMapLine( strFrom, strTo, line, nColFrom, nColTo ) )
			std::cout << "warning, file " << fname << " line ignored : " << line << "\n";

		nameMap[strFrom] = strTo;
	}

	data_file.close();

	return false;
}

bool cGeneSet::parseMapLine( string& strFrom, string& strTo, string& line, int nColFrom, int nColTo )
{
	std::basic_string <char>::size_type L, R;
	
	L = line.find_first_not_of( " \t" );
	
	// skip white space and comments
	if (( L == -1 )||( line[L] == '#' ))
		return false;

	vector<string> cols;
	int nDone = std::max(nColFrom, nColTo);

	R = line.find_first_of( " \t", L );

	for( int i = 0 ; (i < nDone)&&(L != -1); ++i )
	{
		cols.push_back( line.substr( L, R - L ) );

		L = line.find_first_not_of( " \t", R );
		R = line.find_first_of( " \t", L );
	}

	// all remaining text
	//cols.push_back( line.substr( L, string::npos ) );

	if ( cols.size() != nDone )
	{
		std::cout << "Data format error: " << line << std::endl;
		return true;
	}

	strFrom = cols[nColFrom-1];
	strTo = cols[nColTo-1];

	return false;
}

bool cGeneSet::loadGff(string& fname)
{
	std::ifstream data_file( fname.c_str() );

	if ( !data_file.is_open() )
	{
		std::cout << "error, opening file " << fname << "\n"; 
		return true;
	}

	string line;
	m_temp_records.clear();

	while( !data_file.eof() )
	{
		getline( data_file, line );
		if ( parseGffLine( line ) )
			std::cout << "warning, file " << fname << " line ignored : " << line << "\n";
	}

	data_file.close();

	// finish up last gene locus
	recs2feat();
	m_temp_records.clear();

	std::sort( m_genes.begin(), m_genes.end() );

	indexBySeq();

	return false;
}

void cGeneSet::indexBySeq()
{
	m_seqMap.clear();
	if (m_genes.empty()) return;

	string lastSeq = m_genes[0].getSeq();
	string seq = lastSeq;

	m_seqMap[seq] = 0;

	for (int i=1; i<(int)m_genes.size(); ++i)
	{
		seq = m_genes[i].getSeq();
		if (seq != lastSeq) {
			lastSeq = seq;
			m_seqMap[seq] = i;
		}
	}
}

bool cGeneSet::parseGffLine( string& line )
{
	gffStruct rec;
	std::basic_string <char>::size_type L, R;
	
	L = line.find_first_not_of( " \t" );
	
	// skip white space and comments
	if (( L == -1 )||( line[L] == '#' ))
		return false;

	vector< string > gff;
	gff.reserve(9);

	R = line.find_first_of( " \t", L );

	for( int i = 0 ; (i < 8)&&(L != -1); ++i )
	{
		gff.push_back( line.substr( L, R - L ) );

		L = line.find_first_not_of( " \t", R );
		R = line.find_first_of( " \t", L );
	}

	// all remaining text
	gff.push_back( line.substr( L, string::npos ) );

	if ( gff.size() != 9 )
	{
		std::cout << "Data format error: " << line << std::endl;
		return true;
	}

	rec.strName = m_chrPrefix + gff[0];			// sequence name
	rec.strSource = gff[1];			// source
	rec.strFeature = gff[2];		// feature
	rec.ibeg = atoi( gff[3].c_str() );		// start coordinate
	rec.iend  = atoi( gff[4].c_str() );		// stop coordinate
	rec.strScore = gff[5];
	rec.chStrand = gff[6].at(0);
	rec.strFrame = gff[7];			// frame
	rec.strRemainder = gff[8];		// transcript id

	rec.iFamily = -1;
	rec.bDeleted = false;
	rec.bShortFeature = false;

	getTag( rec.strNameTag, rec.strRemainder, string("Name") );
	getTags( rec.vstrParent, rec.strRemainder, string("Parent") );
	getTag( rec.strId, rec.strRemainder, string("ID") );

	if (( rec.ibeg > rec.iend )||( rec.ibeg < 0 ))	//||( rec.iend > size ))
	{
		std::cout << "Data format error: " << line << std::endl;
		return true;
	}

	if (( rec.chStrand != '+' )&&( rec.chStrand != '-' )&&( rec.chStrand != '.'))
	{
		std::cout << "Data format error: " << line << std::endl;
		return true;
	}

	if (rec.strFeature.compare(m_strGeneKey)==0) {
		recs2feat();
		m_temp_records.clear();
	}

	// save new record
	m_temp_records.push_back( rec );

	return false;
}

bool cGeneSet::getTag( string& tag, string& desc, string tagName )
{
	tag.clear();
	size_t ix = desc.find(tagName,0);
	size_t ixe;
	size_t offset = tagName.length() + 1;

	if (ix!=string::npos)
	{
		ixe = desc.find_first_of(";",ix);
		if (ixe==string::npos)
			tag = desc.substr(ix+offset, string::npos);
		else
			tag = desc.substr(ix+offset, ixe-(ix+offset));
	}

	return false;
}

bool cGeneSet::getTags( vector<string>& tags, string& desc, string tagName )
{
	tags.clear();
	size_t ix = desc.find(tagName,0);
	size_t ixe;
	size_t offset = tagName.length() + 1;
	string tag;

	if (ix!=string::npos)
	{
		ixe = desc.find_first_of(";",ix);
		if (ixe==string::npos)
			tag = desc.substr(ix+offset, string::npos);
		else
			tag = desc.substr(ix+offset, ixe-(ix+offset));

		csv2vstring(tags, tag);
	}

	return false;
}

bool cGeneSet::loadGtf(string& fname)
{
	std::ifstream data_file( fname.c_str() );

	if ( !data_file.is_open() )
	{
		std::cout << "error, opening file " << fname << "\n"; 
		return true;
	}

	string line;
	m_temp_records.clear();

	m_strGeneKey = string("");

	while( !data_file.eof() )
	{
		getline( data_file, line );
		if ( parseGtfLine( line ) )
			std::cout << "warning, file " << fname << " line ignored : " << line << "\n";
	}

	data_file.close();

	// finish up last gene locus
	recs2featGtf();
	m_temp_records.clear();

	std::sort( m_genes.begin(), m_genes.end() );

	indexBySeq();

	return false;
}

bool cGeneSet::parseGtfLine( string& line )
{
	gffStruct rec;
	std::basic_string <char>::size_type L, R;
	
	L = line.find_first_not_of( " \t" );
	
	// skip white space and comments
	if (( L == -1 )||( line[L] == '#' ))
		return false;

	vector< string > gtf;
	gtf.reserve(9);

	R = line.find_first_of( " \t", L );

	for( int i = 0 ; (i < 8)&&(L != -1); ++i )
	{
		gtf.push_back( line.substr( L, R - L ) );

		L = line.find_first_not_of( " \t", R );
		R = line.find_first_of( " \t", L );
	}

	// all remaining text
	gtf.push_back( line.substr( L, string::npos ) );

	if ( gtf.size() != 9 )
	{
		std::cout << "Data format error: " << line << std::endl;
		return true;
	}

	rec.strName = gtf[0];			// sequence name
	rec.strSource = gtf[1];			// source
	rec.strFeature = gtf[2];		// feature
	rec.ibeg = atoi( gtf[3].c_str() );		// start coordinate
	rec.iend  = atoi( gtf[4].c_str() );		// stop coordinate
	rec.strScore = gtf[5];
	rec.chStrand = gtf[6].at(0);
	rec.strFrame = gtf[7];			// frame
	rec.strRemainder = gtf[8];		// transcript id

	rec.iFamily = -1;
	rec.bDeleted = false;
	rec.bShortFeature = false;

	getGtfTag( rec.strId, rec.strRemainder, string("gene_id") );
	getGtfTag( rec.strNameTag, rec.strRemainder, string("transcript_id") );

	if (( rec.ibeg > rec.iend )||( rec.ibeg < 0 ))	//||( rec.iend > size ))
	{
		std::cout << "Data format error: " << line << std::endl;
		return true;
	}

	if (( rec.chStrand != '+' )&&( rec.chStrand != '-' )&&( rec.chStrand != '.'))
	{
		std::cout << "Data format error: " << line << std::endl;
		return true;
	}

	if (m_strGeneKey.empty())
		m_strGeneKey = rec.strNameTag;
	else {
		if (rec.strNameTag != m_strGeneKey) {
			recs2featGtf();
			m_temp_records.clear();
			m_strGeneKey = rec.strNameTag;
		}
	}

	// save new record
	m_temp_records.push_back( rec );

	return false;
}

bool cGeneSet::getGtfTag( string& tag, string desc, string tagName )
{
	tag.clear();
	size_t ix = desc.find(tagName,0);
	size_t ixbq, ixeq;		// begin quote and end quote coords
	size_t offset = tagName.length() + 1;

	if (ix!=string::npos)
	{
		ixbq = desc.find_first_of("\"",ix);
		if (ixbq==string::npos)
			return true;
		else
		{
			ixeq = desc.find_first_of("\"",ixbq+1);
			tag = desc.substr(ixbq+1, ixeq-ixbq-1);
		}
	}

	return false;
}

bool cGeneSet::loadTable(string& fname)
{
	std::ifstream data_file( fname.c_str() );

	if ( !data_file.is_open() )
	{
		std::cout << "error, opening file " << fname << "\n"; 
		return true;
	}

	string line;
	m_temp_records.clear();

	m_strGeneKey = string("");

	while( !data_file.eof() )
	{
		getline( data_file, line );
		if ( parseTableLine( line ) )
			std::cout << "warning, file " << fname << " line ignored : " << line << "\n";
	}

	data_file.close();
	std::sort( m_genes.begin(), m_genes.end() );

	return false;
}

bool cGeneSet::parseTableLine( string& line )
{
	gffStruct rec;
	std::basic_string <char>::size_type L, R;
	
	L = line.find_first_not_of( " \t" );
	
	// skip white space and comments
	if (( L == -1 )||( line[L] == '#' ))
		return false;

	vector< string > cols;
	int nCols(11), i;
	cols.reserve(nCols);

	R = line.find_first_of( " \t", L );

	for( i = 0 ; (i < nCols-1)&&(L != -1); ++i )
	{
		cols.push_back( line.substr( L, R - L ) );

		L = line.find_first_not_of( " \t", R );
		R = line.find_first_of( " \t", L );
	}

	// all remaining text
	cols.push_back( line.substr( L, string::npos ) );

	if ( (int)cols.size() < nCols-1 )
	{
		std::cout << "Data format error: " << line << std::endl;
		return true;
	}

	int nBeg, nEnd;
	nBeg = atoi( cols[3].c_str() );
	nEnd = atoi( cols[4].c_str() );
	if (!m_temp_gene.isOverlapping(nBeg,nEnd))
	{
		m_genes.push_back( m_temp_gene );
		m_temp_gene.clear();
	}

	m_temp_isoform.clear();

	m_temp_isoform.setId( cols[0] );
	m_temp_isoform.setStrand( cols[2].at(0) );

	int nExons = atoi( cols[7].c_str() );

	vector<string> strBegs, strEnds;
	csv2vstring( strBegs, cols[8] );
	csv2vstring( strEnds, cols[9] );

	for (i=0; i<nExons; ++i)
	{
		nBeg = atoi( strBegs[i].c_str() );
		nEnd = atoi( strEnds[i].c_str() );
		m_temp_isoform.addTrans( nBeg, nEnd, _INTERVAL_EXON_NF );
	}

	int nCdsBeg = atoi( cols[5].c_str() );
	int nCdsEnd = atoi( cols[6].c_str() );
	m_temp_isoform.deriveCoding( nCdsBeg, nCdsEnd );
	m_temp_isoform.setLimits();

	m_temp_gene.addIsoform( m_temp_isoform );
	m_temp_gene.setSeq( cols[1] );
	m_temp_gene.setId( cols[0] );
	m_temp_gene.setLimits();

	return false;
}

bool cGeneSet::recs2feat()
{
	int i, j;
	// get all unique IDs
	//

	// find gene feature(s) in this set
	// (level 0)
	vector<int> gix = findGeneRecs();

	vector<int> rix, aix, fix;
	for (i=0; i<(int)gix.size(); ++i)
	{
		// find all features pointing to gene rec (e.g., mRNA)
		// (level 1)
		m_temp_gene.clear();
		rix = findChildren(gix[i]);

		// find all features pointing to level 1 features
		// (level 2)
		for (j=0; j<(int)rix.size(); ++j)
		{
			fix = findChildren(rix[j]);
			if (fix.empty()) {
				addIsoform(gix[i], rix);	// 1-level isoform (e.g., gene->exons)
				break;
			}
			else
				addIsoform(rix[j], fix);	// 2-level isoform (e.g., gene->mRNA->exons
		}

		// find all features pointing to alignment rec (e.g., match_part)
		//
		aix = findAlignRecs();
		for (j=0; j<(int)aix.size(); ++j)
		{
			m_temp_align.clear();
			fix = findChildren(aix[j]);
			addAlignment(aix[j], fix);
		}

		// add gene
		int flag(0);
		m_temp_gene.setLimits(m_temp_records[gix[i]].ibeg, m_temp_records[gix[i]].iend);
		if (m_temp_gene.getBeg()==-1 || m_temp_gene.getEnd()==-1)
			flag = 1;

		m_temp_gene.setId( m_temp_records[gix[i]].strId );
		m_temp_gene.setSeq( m_temp_records[gix[i]].strName );
		addGene();
	}

	return false;
}

bool cGeneSet::recs2featGtf()
{
	++m_uid;
	if (m_uid % 100 == 0)
		std::cout << m_uid << " genes loaded" << std::endl;

	addIsoform();

	m_temp_gene.setId( m_temp_records[0].strId );
	m_temp_gene.setSeq( m_temp_records[0].strName );

	addGene();

	m_temp_gene.clear();

	return false;
}

void cGeneSet::addGene()
{
	m_genes.push_back( m_temp_gene );
}

void cGeneSet::addIsoform()
{
	m_temp_isoform.clear();
	if (m_temp_records.empty())
		return;

	m_temp_isoform.setId( m_temp_records[0].strId );
	m_temp_isoform.setStrand( m_temp_records[0].chStrand );

	for (int i=0; i<(int)m_temp_records.size(); ++i)
	{
		if (m_temp_records[i].strFeature.compare("exon")==0)
			m_temp_isoform.addTrans( m_temp_records[i].ibeg, m_temp_records[i].iend, m_temp_records[i].strFrame );
		else if (m_temp_records[i].strFeature.compare("CDS")==0)
			m_temp_isoform.addCoding( m_temp_records[i].ibeg, m_temp_records[i].iend, m_temp_records[i].strFrame );
	}

	m_temp_isoform.setLimits();

	m_temp_gene.addIsoform( m_temp_isoform );
	m_temp_gene.setLimits();
}

void cGeneSet::addIsoform(int ix_parent, vector<int>& vix_children)
{
	m_temp_isoform.clear();

	m_temp_isoform.setId( m_temp_records[ix_parent].strId );
	m_temp_isoform.setStrand( m_temp_records[ix_parent].chStrand );

	int ix;
	for (int i=0; i<(int)vix_children.size(); ++i)
	{
		ix = vix_children[i];
		if (m_temp_records[ix].strFeature.compare("exon")==0)
			m_temp_isoform.addTrans( m_temp_records[ix].ibeg, m_temp_records[ix].iend, m_temp_records[ix].strFrame );
		else if (m_temp_records[ix].strFeature.compare("CDS")==0)
			m_temp_isoform.addCoding( m_temp_records[ix].ibeg, m_temp_records[ix].iend, m_temp_records[ix].strFrame );
	}

	m_temp_isoform.setLimits();

	m_temp_gene.addIsoform( m_temp_isoform );
}

void cGeneSet::addAlignment(int ix_parent, vector<int>& vix_children)
{
	m_temp_align.clear();

	m_temp_align.setId( m_temp_records[ix_parent].strId );
	m_temp_align.setStrand( m_temp_records[ix_parent].chStrand );

	int ix;
	for (int i=0; i<(int)vix_children.size(); ++i)
	{
		ix = vix_children[i];
		if (m_temp_records[ix].strFeature.compare("match_part")==0)
			m_temp_align.addMatch( m_temp_records[ix].ibeg, m_temp_records[ix].iend );
	}

	m_temp_align.setLimits();

	m_temp_gene.addAlignment( m_temp_align );
}

vector<int> cGeneSet::findGeneRecs()
{
	vector<int> gix;

	for (int i=0; i<(int)m_temp_records.size(); ++i)
	{
		if (m_temp_records[i].strFeature.compare("gene")==0)
			gix.push_back(i);
	}

	return gix;
}

vector<int> cGeneSet::findAlignRecs()
{
	vector<int> gix;

	for (int i=0; i<(int)m_temp_records.size(); ++i)
	{
		if (m_temp_records[i].strFeature.compare("match")==0)
			gix.push_back(i);
	}

	return gix;
}

vector<int> cGeneSet::findChildren(int ix)
{
	vector<int> result_ix;

	if (ix<0 || ix>(int)m_temp_records.size())
		return result_ix;

	string strId = m_temp_records[ix].strId;

	for (int i=0; i<(int)m_temp_records.size(); ++i)
	{
		if ( isOneOf(strId, m_temp_records[i].vstrParent) )
			result_ix.push_back(i);
	}

	return result_ix;
}

void cGeneSet::csv2vstring(vector<string>& labels, string& csv)
{
	labels.clear();
	string s;

	size_t start = 0;
	size_t pos = csv.find_first_of(',');

	while (pos != string::npos)
	{
		s = csv.substr(start,pos-start);
		labels.push_back( s );
		start = pos + 1;
		pos = csv.find_first_of(',',start);
	}
	s = csv.substr(start,pos-start);
	labels.push_back( s );
}

bool cGeneSet::isOneOf(string& query, vector<string>& set)
{
	for (int i=0; i<(int)set.size(); ++i)
	{
		if (set[i].compare(query)==0)
			return true;
	}
	return false;
}

bool cGeneSet::writeGff(string& fname)
{
	std::filebuf fb;
	fb.open( fname.c_str(), std::ios::out );
	std::ostream os(&fb);

	if ( !fb.is_open() )
	{
		std::cout << PROG_NAME << " : error, opening file " << fname << std::endl;; 
		return true;
	}

	for (int i=0; i<(int)m_genes.size(); ++i)
		m_genes[i].writeGff( os );

	fb.close();

	return false;
}

bool cGeneSet::writeSummary(string& title)
{
	int nIso(0), nMultiIso(0), nExons(0), nCds(0), nIntrons(0), nSingleExons(0), nSingleCds(0);
	int lenCds(0), lenExons(0), lenIntrons(0);
	int ne, nte, nce;

	for (int i=0; i<(int)m_genes.size(); ++i)
	{
		nIso += m_genes[i].numel();
		nMultiIso += m_genes[i].numel()==1 ? 0 : 1;

		nte = m_genes[i].ctTransExons();
		nce = m_genes[i].ctCdsExons();
		ne = nte > nce ? nte : nce;
		if (ne==1)
			++nSingleExons;
		nExons += ne;

		if (nce==1)
			++nSingleCds;
		nCds += nce;

		nIntrons += m_genes[i].ctIntrons();

		lenExons += m_genes[i].lenGene();
		lenCds += m_genes[i].lenCds();
		lenIntrons += m_genes[i].lenIntrons();
	}

	std::cout << "-------- for file: " << title << " ----------" << std::endl;
	std::cout << "Genes :                               " << m_genes.size() << std::endl;
	std::cout << "Isoforms (total) :                    " << nIso << std::endl;
	std::cout << "Multi-isoform genes :                 " << nMultiIso << std::endl;
	std::cout << "Isoforms per gene :                   " << double(nIso)/double(m_genes.size()) << std::endl;
	std::cout << "Single exon genes :                   " << nSingleExons << std::endl;
	std::cout << "Single CDS exon genes :               " << nSingleCds << std::endl;
	std::cout << "--------- results below peratain to longest isoform ---------" << std::endl;
	std::cout << "Exons :                               " << nExons << std::endl;
	std::cout << "Exons per gene :                      " << double(nExons)/double(m_genes.size()) << std::endl;
	std::cout << "CDS exons :                           " << nCds << std::endl;
	std::cout << "CDS exons per gene :                  " << double(nCds)/double(m_genes.size()) << std::endl;
	std::cout << "Introns :                             " << nIntrons << std::endl;
	std::cout << "Introns per gene :                    " << double(nIntrons)/double(m_genes.size()) << std::endl;
	std::cout << "Introns per gene (multi-exon genes) : " << double(nIntrons)/double(m_genes.size()-nSingleExons) << std::endl;
	std::cout << "------------------------------------------------------------" << std::endl;
	std::cout << "Avg. gene length :                    " << double(lenExons)/double(m_genes.size()) << std::endl;
	std::cout << "Avg. CDS length :                     " << double(lenCds)/double(m_genes.size()) << std::endl;
	std::cout << "Avg. intron length :                  " << double(lenIntrons)/double(nIntrons) << std::endl;
	std::cout << std::endl;

	return false;
}

vector<sIntron> cGeneSet::interval2intron(vector<cRichInterval>& ctrons_in, int ix)
{
	vector<sIntron> strons_out;
	sIntron stron;

	int istrand = m_genes[ix].getStrand()=='+' ? 1 : -1;

	for (int i=0; i<(int)ctrons_in.size(); ++i) {
		stron = interval2intron(ctrons_in[i]);
		stron.iStrand = istrand;
		stron.strId = m_genes[ix].getSeq();

		stron.iIntronType = 0;
		if ( stron.ibeg>m_genes[ix].getCdsBeg() && stron.iend<m_genes[ix].getCdsEnd() )
			stron.iIntronType = 1;

		strons_out.push_back(stron);
	}

	return strons_out;
}

vector<sIntron> cGeneSet::interval2intron(vector<cInterval>& ctrons_in, int ix)
{
	vector<sIntron> strons_out;
	sIntron stron;

	int istrand = m_genes[ix].getStrand()=='+' ? 1 : -1;

	for (int i=0; i<(int)ctrons_in.size(); ++i) {
		stron = interval2intron(ctrons_in[i]);
		stron.iStrand = istrand;
		stron.strId = m_genes[ix].getSeq();

		stron.iIntronType = 0;
		if ( stron.ibeg>m_genes[ix].getCdsBeg() && stron.iend<m_genes[ix].getCdsEnd() )
			stron.iIntronType = 1;

		strons_out.push_back(stron);
	}

	return strons_out;
}

sIntron cGeneSet::interval2intron(cInterval& ctron)
{
	sIntron tron;

	tron.extr_L = tron.extr_R = tron.extr_I = 0;
	tron.ibeg = ctron.first();
	tron.iend = ctron.second();
	tron.iframe = -2;
	
	switch (ctron.nType) {
		case _INTERVAL_INTRON_0:
			tron.iframe = 0;
			break;
		case _INTERVAL_INTRON_1:
			tron.iframe = 1;
			break;
		case _INTERVAL_INTRON_2:
			tron.iframe = 2;
			break;
		case _INTERVAL_INTRON_NF:
			tron.iframe = -1;
	};
	return tron;
}

sIntron cGeneSet::interval2intron(cRichInterval& ctron)
{
	sIntron tron;

	tron.extr_L = tron.extr_R = tron.extr_I = 0;
	tron.ibeg = ctron.first();
	tron.iend = ctron.second();
	tron.iframe = -2;
	tron.lenExonL = ctron.m_lenExonL;
	tron.lenExonR = ctron.m_lenExonR;
	
	switch (ctron.nType) {
		case _INTERVAL_INTRON_0:
			tron.iframe = 0;
			break;
		case _INTERVAL_INTRON_1:
			tron.iframe = 1;
			break;
		case _INTERVAL_INTRON_2:
			tron.iframe = 2;
			break;
		case _INTERVAL_INTRON_NF:
			tron.iframe = -1;
	};
	return tron;
}

vector<sIntron> cGeneSet::getIntrons(string& seqName)
{
	vector<sIntron> introns, si_trons;
	vector<cInterval> ci_trons;

	int ix = findBySeq(seqName);
	if (ix == -1)
		return introns;

	int num = (int)m_genes.size();
	
	while (ix<num && m_genes[ix].getSeq()==seqName)
	{
		ci_trons = m_genes[ix].getIntrons();
		si_trons = interval2intron( ci_trons, ix );
		introns.insert( introns.end(), si_trons.begin(), si_trons.end() );
		++ix;
	}

	return introns;
}

vector<sIntron> cGeneSet::getRichIntrons(string& seqName)
{
	vector<sIntron> introns, si_trons;
	vector<cRichInterval> ci_trons;

	int ix = findBySeq(seqName);
	if (ix == -1)
		return introns;

	int num = (int)m_genes.size();
	
	while (ix<num && m_genes[ix].getSeq()==seqName)
	{
		ci_trons = m_genes[ix].getRichIntrons();
		si_trons = interval2intron( ci_trons, ix );
		introns.insert( introns.end(), si_trons.begin(), si_trons.end() );
		++ix;
	}

	return introns;
}

vector<string> cGeneSet::getSequenceNames()
{
	vector<string> names;
	string name;

	for (int i=0; i<(int)m_genes.size(); ++i)
	{
		name = m_genes[i].getSeq();
		names.push_back( name );
	}

	vector<string>::iterator it = std::unique( names.begin(), names.end() );
	names.resize( it - names.begin() );
	return names;
}

vector<sIntron> cGeneSet::getIntrons()
{
	vector<sIntron> introns, si_trons;
	vector<cInterval> ci_trons;

	vector<string> seqNames = getSequenceNames();

	int num = (int)seqNames.size();
	
	for (int i=0; i<num; ++i)
	{
		si_trons = getIntrons( seqNames[i] );
		introns.insert( introns.end(), si_trons.begin(), si_trons.end() );
	}

	return introns;
}

int cGeneSet::findBySeq(string& seq)
{
	int i;
	for (i=m_iFindPos; i<(int)m_genes.size(); ++i)
	{
		if (m_genes[i].getSeq()==seq) {
			m_iFindPos = i;
			return i;
		}
	}

	for (i=0; i<m_iFindPos; ++i)
	{
		if (m_genes[i].getSeq()==seq) {
			m_iFindPos = i;
			return i;
		}
	}

	return -1;		// did not find
}

bool cGeneSet::findGene( cGene& gene, string& seq, int ipos )
{
	if (m_seqMap.find(seq)==m_seqMap.end())
		return true;

	int ixStart = m_seqMap[seq];

	//std::cout << "found seqMap for " << seq << " at " << ipos << ". ixStart=" << ixStart << std::endl;

	vector<cGene>::iterator it = findGene( m_genes.begin()+ixStart, m_genes.end(), seq, ipos );
	// return true if not found
	if (it==m_genes.end()) return true;
	gene = *it;
	return false;
}

vector<cGene>::iterator cGeneSet::findGene( vector<cGene>::iterator it_beg, vector<cGene>::iterator it_end, string& seq, int ipos )
{
	vector<cGene>::iterator it = it_beg, it_last;

	while (it != it_end)
	{
		if (it->getSeq()!=seq) {
			++it;
			it_beg = it;			// move beg to current gene
			continue;
		}

		//std::cout << "beginning of gene = " << it->getBeg() << ", ipos = " << ipos << std::endl;

		if (it->getBeg() > ipos)	// we passed it
		{
			if (it == it_beg)			// oops, no hits
				return it_end;

			if (it_last->getBeg() < ipos)	// last gene before we passed it
				return it_last;
			else
				return it_end;			// don't know.. wtf
		}
		it_last = it;

		++it;
	}

	// no hits
	return it_end;
}

void cGeneSet::markGenicIntron( int& nType, int nPos, vector<cGene>::iterator it, vector<cInterval>& refTrons, int nOff )
{
	int i;
	int nCdsLeft = it->getCdsBeg();
	int nCdsRight = it->getCdsEnd();

	char chStrand;
	chStrand = it->getStrand();

	nType = _INTRON_FP_INTERG;
	int len, lenCds(0), left, ixLastCod(-1), fr, ix;
	bool bCds;

	if (nPos < nCdsLeft || nPos > nCdsRight)
		nType = _INTRON_FP_UTR;
	else {
		for (i=0; i<(int)refTrons.size(); ++i)
		{
			bCds = refTrons[i].first() > nCdsLeft && refTrons[i].second() < nCdsRight;

			if (bCds)
				ixLastCod = i;

			if (bCds && i > 0)
				lenCds += (refTrons[i].first() - std::max(nCdsLeft,refTrons[i-1].second()+1) );
			else if (bCds && i == 0)
				lenCds += (refTrons[i].first() - nCdsLeft);

			if (nPos > refTrons[i].first() && nPos < refTrons[i].second()) {	// nPos internal to intron
				nType = _INTRON_FP_INTRON;
				return;
			}
			else if (bCds && nPos <= refTrons[i].first()) {				// nPos internal to CDS
				// intron must be CDS, now determine which frame
				len = lenCds - (refTrons[i].first() - nPos - 1);
				fr = chStrand=='+' ? (len+nOff) % 3 : 2 - ((len+nOff) % 3);
				switch (fr) {
					case 0:
						nType = _INTRON_FP_CDS_FR_0; break;
					case 1:
						nType = _INTRON_FP_CDS_FR_1; break;
					case 2:
						nType = _INTRON_FP_CDS_FR_2; break;
					default:
						nType = _INTRON_FP_UTR;	// can happen in odd cases
				};
				return;
			}
			else if (!bCds && nPos <= refTrons[i].first()) {
				// intron is inside UTR
				nType = _INTRON_FP_UTR;
				return;
			}
		}

		// nPos is to the right of all introns

		if (nPos <= nCdsRight) {
			ix = ixLastCod;
			if ( !refTrons.empty() && ix==-1 ) {
				nType = _INTRON_FP_UTR;
				return;
			}
			left = refTrons.empty() ? nCdsLeft : refTrons[ix].second() + 1;
			lenCds += (nPos - left + 1);
			fr = chStrand=='+' ? (lenCds+nOff) % 3 : 2 - ((lenCds+nOff) % 3);
			switch (fr) {
				case 0:
					nType = _INTRON_FP_CDS_FR_0; break;
				case 1:
					nType = _INTRON_FP_CDS_FR_1; break;
				case 2:
					nType = _INTRON_FP_CDS_FR_2; break;
				default:
					nType = _INTRON_FP_UTR;	// can happen in odd cases
			};
			return;
		}
		else {
			nType = _INTRON_FP_UTR;
			return;
		}
	}
}

//enum { _INTERVAL_UNDEFINED = -1, _INTERVAL_EXON_0, _INTERVAL_EXON_1, _INTERVAL_EXON_2, _INTERVAL_EXON_NF,
//		_INTERVAL_INTRON_0, _INTERVAL_INTRON_1, _INTERVAL_INTRON_2, _INTERVAL_INTRON_NF };

bool cGeneSet::findIntron(cInterval& query, int& fpL, int& fpR, vector<cInterval>& ref)
{
	for (int i=0; i<(int)ref.size(); ++i)
	{
		if (query==ref[i]) {
			switch (ref[i].nType) {
				case _INTERVAL_INTRON_0:
					fpL = fpR = _INTRON_FP_CDS_FR_0; break;
				case _INTERVAL_INTRON_1:
					fpL = fpR = _INTRON_FP_CDS_FR_1; break;
				case _INTERVAL_INTRON_2:
					fpL = fpR = _INTRON_FP_CDS_FR_2; break;
				default: // ( _INTERVAL_INTRON_NF )
					fpL = fpR = _INTRON_FP_UTR;
			};
			return true;
		}
	}

	return false;
}

void cGeneSet::markIntronFp(sIntron& intron, string& seq)
{
	vector<cGene>::iterator itLast = m_genes.end(), itf, it = m_genes.begin();
	vector<cInterval> refTrons;
	cInterval que;
	char chStrand;

	int flag(0);

	int ixStart = m_seqMap[seq];
	it += ixStart;

	if (seq.compare("2L")==0)
		flag = 1;

	itf = findGene( it, m_genes.end(), seq, intron.ibeg );

	if (itf != m_genes.end()) {	// genic, now check if inside CDS, UTR, or intron
		chStrand = itf->getStrand();
		if (itLast != itf) {
			itLast = itf;
			refTrons = itf->getIntrons();
		}

		que = cInterval(intron.ibeg, intron.iend);

		if (findIntron( que, intron.nFpL, intron.nFpR, refTrons )) {
			return;
		}

		if (chStrand=='+' && intron.iStrand==1)
			markGenicIntron( intron.nFpL, intron.ibeg, itf, refTrons, 0 );
		else if (chStrand=='-' && intron.iStrand==-1)
			markGenicIntron( intron.nFpL, intron.ibeg, itf, refTrons, -1 );
		else if (chStrand=='*' || intron.iStrand==0)
			intron.nFpL = _INTRON_FP_UTR;
		else if ( intron.ibeg>=itf->getBeg() && intron.ibeg<=itf->getEnd() )
			intron.nFpL = _INTRON_FP_CDS_WRONG_STRAND;
		else
			intron.nFpL = _INTRON_FP_INTERG;
	}
	else
		intron.nFpL = _INTRON_FP_INTERG;

	itf = findGene( it, m_genes.end(), seq, intron.iend );

	if (itf != m_genes.end()) {	// genic, now check if inside CDS, UTR, or intron
		chStrand = itf->getStrand();
		if (itLast != itf) {
			itLast = itf;
			refTrons = itf->getIntrons();
		}

		if (chStrand=='+' && intron.iStrand==1)
			markGenicIntron( intron.nFpR, intron.iend, itf, refTrons, 1 );
		else if (chStrand=='-' && intron.iStrand==-1)
			markGenicIntron( intron.nFpR, intron.iend, itf, refTrons, 0 );
		else if (chStrand=='*' || intron.iStrand==0)
			intron.nFpR = _INTRON_FP_UTR;
		else if ( intron.iend>=itf->getBeg() && intron.iend<=itf->getEnd() )
			intron.nFpR = _INTRON_FP_CDS_WRONG_STRAND;
		else
			intron.nFpR = _INTRON_FP_INTERG;
	}
	else
		intron.nFpR = _INTRON_FP_INTERG;
}

void cGeneSet::markIntronFp(vector<sIntron>& introns, string& seq)
{	// for all introns not found in genesI (.extr_I = 0), find location of boundaries:
	//  1/ genic: 1a/ CDS (& frame), 2a/ intron, 3a/ UTR, 3/ intergenic

	for (int i=0; i<(int)introns.size(); ++i)
		markIntronFp( introns[i], seq );

}

void cGeneSet::markIntronFp(vector<sIntronSet>& multi)
{
	for (int i=0; i<(int)multi.size(); ++i)
	{
		for (int j=0; j<(int)multi[i].set.size(); ++j)
			markIntronFp( multi[i].set[j].intron, multi[i].seq );
	}
}
