#include "junc.h"
#include "targetver.h"
#include <iostream>
#include <fstream>
#include <algorithm>

extern bool intronLessThan2(sIntron i1, sIntron i2);
extern bool intronEqualTo(sIntron i1, sIntron i2);

bool cJunc::loadFile( string& fname )
{
	std::ifstream data_file( fname.c_str() );

	m_nContig = 0;

	if ( !data_file.is_open() )
	{
		std::cout << PROG_NAME << " : error, opening file " << fname << "\n"; 
		return true;
	}

	string line;

	while( !data_file.eof() )
	{
		getline( data_file, line );

		if (data_file.fail())
			continue;

		if ( parseDataLine( line ) )
			std::cout << PROG_NAME << " : warning, file " << fname << " line ignored : " << line << "\n";
	}

	data_file.close();


	std::sort( m_introns.begin(), m_introns.end(), intronLessThan2 );
	assignContigs();

	return false;
}

bool cJunc::loadScoreFile( string& fname )
{
	std::ifstream data_file( fname.c_str() );

	m_nContig = 0;

	if ( !data_file.is_open() )
	{
		std::cout << PROG_NAME << " : error, opening file " << fname << "\n"; 
		return true;
	}

	string line;

	while( !data_file.eof() )
	{
		getline( data_file, line );

		if (data_file.fail())
			continue;

		if ( parseScoreLine( line ) )
			std::cout << PROG_NAME << " : warning, file " << fname << " line ignored : " << line << "\n";
	}

	data_file.close();


	std::sort( m_introns.begin(), m_introns.end(), intronLessThan2 );
	assignContigs();

	return false;
}

//// introns for gapped alignments
//typedef struct tagIntron
//{
//	string strId;
//	int nContig;
//	int nIntronsInGene;		// how many introns in this gene (isoform)?
//	int ixIntron;			// which intron in the gene is this (index)
//	int posAtg, posStp;		// position of intron starting point, relative to CDS start and stop
//	int ibeg;
//	int iend;
//	int iframe;			// {0, 1, 2, 3 (non-coding), -1 (unknown)}
//	int flankLeft;
//	int flankRight;
//	int iStrand;
//	int gc;
//	int iJcnType;
//	unsigned int extr_L;
//	unsigned int extr_R;
//	unsigned int extr_I;
//} sIntron;

bool cJunc::parseDataLine( string& line )
{
	sIntron intron;
	std::basic_string <char>::size_type L, R;
	
	L = line.find_first_not_of( m_delim );
	
	// skip white space and comments
	if (( L == -1 )||( line[L] == '#' ))
		return false;

	vector< string > cols;
	cols.reserve(16);
	int i;

	R = line.find_first_of( m_delim, L );

	for( i = 0 ; L != -1; ++i )
	{
		cols.push_back( line.substr( L, R - L ) );

		L = line.find_first_not_of( m_delim, R );
		R = line.find_first_of( m_delim, L );
	}

	bool b4thCol = cols.size()==4;

	if ( cols.size() < 3 )
	{
		std::cout << "Data format error: " << line << std::endl;
		return true;
	}

	intron.extr_I = intron.extr_L = intron.extr_R = 0;
	intron.iframe = -2;
	int id;
	char ch;
	if (b4thCol) {
		if (m_bStrandCol) {
			ch = cols[3].at(0);
			intron.iStrand = ch=='+'? 1 : -1;
		}
		else {
			id = atoi( cols[3].c_str() );
			if (id==0 && m_bGetUtrStatus) intron.iframe = 0;
			else if (m_bGetUtrStatus) intron.iframe = -1;
		}
	}

	intron.strId = m_refPrefix + cols[0];			// sequence name
	intron.ibeg = atoi( cols[1].c_str() ) + 1;
	intron.iend = atoi( cols[2].c_str() );
	intron.iStrand = 0;		// unknown for the time being
	intron.nCoverage = 0;

	intron.nFpL = intron.nFpR = -99;

	intron.ibeg += m_nOffset1;
	intron.iend += m_nOffset2;

	// save new record
	m_introns.push_back( intron );

	return false;
};

bool cJunc::parseScoreLine( string& line )
{
	sIntron intron;
	std::basic_string <char>::size_type L, R;
	
	L = line.find_first_not_of( " \t" );
	
	// skip white space and comments
	if (( L == -1 )||( line[L] == '#' ))
		return false;

	vector< string > cols;
	cols.reserve(16);
	int i;

	R = line.find_first_of( " \t", L );

	for( i = 0 ; L != -1; ++i )
	{
		cols.push_back( line.substr( L, R - L ) );

		L = line.find_first_not_of( " \t", R );
		R = line.find_first_of( " \t", L );
	}

	if ( cols.size() < 4 )
	{
		std::cout << "Data format error: " << line << std::endl;
		return true;
	}

	intron.extr_I = intron.extr_L = intron.extr_R = 0;
	intron.iframe = -2;

	intron.strId = m_refPrefix + cols[0];			// sequence name
	intron.ibeg = atoi( cols[1].c_str() ) + 1;
	intron.iend = atoi( cols[2].c_str() );
	intron.iStrand = 0;		// unknown for the time being
	intron.nCoverage = 0;
	intron.lenExonL = 0;
	intron.lenExonR = 0;
	intron.score = atof( cols[cols.size()-1].c_str() );

	intron.nFpL = intron.nFpR = -99;

	intron.ibeg += m_nOffset1;
	intron.iend += m_nOffset2;

	// all other colums will be features
	if (cols.size() >= 4)
		intron.feat.insert( intron.feat.begin(), cols.begin()+3, cols.end() );

	m_nFeat = cols.size() - 3;

	// save new record
	m_introns.push_back( intron );

	return false;
};

void cJunc::assignContigs()
{
	m_nContig = -1;
	m_chrNames.clear();

	for (int i=0; i<(int)m_introns.size(); ++i)
	{
		if (findName(m_introns[i].strId))
			m_introns[i].nContig = m_nContig;
		else {
			m_introns[i].nContig = ++m_nContig;
			m_chrNames.push_back( m_introns[i].strId );
		}
	}
}

bool cJunc::findName(string &name)
{
	if (m_chrNames.empty())
		return false;

	for (int i=m_chrNames.size()-1; i>=0; --i)
	{
		if (name.compare( m_chrNames[i] ) == 0)
			return true;		// found it
	}

	return false;
}

vector<sIntron> cJunc::getIntrons(int nContig, string& seqName)
{
	vector<sIntron> introns = getIntrons(seqName);

	for (int i=0; i<(int)introns.size(); ++i)
		introns[i].nContig = nContig;

	return introns;
}

vector<sIntron> cJunc::getIntrons(string& seqName)
{
	// return all "unique" introns
	vector<sIntron> introns;
	int i, ifirst(-1), ilast(-1);

	for (i=0; i<(int)m_introns.size(); ++i)
	{
		if (i==0)
			std::cout << "cJunc id=" << m_introns[i].strId << ", query id=" << seqName << std::endl;

		if (m_introns[i].strId.compare(seqName)==0)
		{
			if (m_bScoreFmt) {
				if (m_introns[i].score >= m_scoreThresh) introns.push_back( m_introns[i] );
			}
			else {
				introns.push_back( m_introns[i] );
			}
		}
	}

	//printf("%d ref %s introns BEFORE sort/unique\n", introns.size(), seqName.c_str());

	std::sort( introns.begin(), introns.end(), intronLessThan2 );
	vector<sIntron>::iterator it = std::unique( introns.begin(), introns.end(), intronEqualTo );
	introns.resize( it - introns.begin() );

	//printf("%d ref %s introns AFTER sort/unique\n", introns.size(), seqName.c_str());

	return introns;
}
