
#include "origin.h"
#include "targetver.h"
#include <algorithm>
#include <math.h>

bool originRecLessThan (alignRec ri, alignRec rj)
{
	if (ri.istart == rj.istart)
	{
		return (ri.pieces.size() > rj.pieces.size());
	}
	else
		return (ri.istart < rj.istart);
}

bool cOrigin::loadSam(string& fname)
{
	std::ifstream data_file( fname.c_str() );

	if ( !data_file.is_open() )
	{
		std::cout << PROG_NAME << " : error, opening file " << fname << "\n"; 
		return true;
	}

	string line;

	while( !data_file.eof() )
	{
		getline( data_file, line );
		if ( parseDataLine( line ) )
			std::cout << PROG_NAME << " : warning, file " << fname << " line ignored : " << line << "\n";
	}

	data_file.close();

	return false;
}

void cOrigin::reset()
{
	m_aligns.clear();
}

bool cOrigin::loadSamBlock(string& fname, int nLines)
{
	if (m_bDone)
		return false;

	std::ifstream data_file( fname.c_str(), std::ifstream::binary );

	if ( !data_file.is_open() )
	{
		std::cout << PROG_NAME << " : error, opening file " << fname << "\n"; 
		return true;
	}

	string line;
	int lineCt(0);

	data_file.seekg( m_streamPos );

	while( !data_file.eof() )
	{
		getline( data_file, line );
		if ( parseDataLine( line ) )
			std::cout << PROG_NAME << " : warning, file " << fname << " line ignored : " << line << "\n";

		++lineCt;
		if (lineCt >= nLines) {
			m_streamPos = data_file.tellg();
			break;
		}
	}

	if (data_file.eof())
		m_bDone = true;

	data_file.close();

	return false;
}

bool cOrigin::parseDataLine( string& line )
{
	alignRec rec;
	//gffLine line_rec;
	std::basic_string <char>::size_type L, R;

	if (line.empty())
		return false;

	if (line.at(0) == '@')
		return false;
	
	L = line.find_first_not_of( " \t" );
	
	// skip white space and comments
	if (( L == -1 )||( line[L] == '#' ))
		return false;

	vector< string > cols;
	cols.reserve(12);

	R = line.find_first_of( " \t", L );

	for( int i = 0 ; (i < 11)&&(L != -1); ++i )
	{
		cols.push_back( line.substr( L, R - L ) );

		L = line.find_first_not_of( " \t", R );
		R = line.find_first_of( " \t", L );
	}

	// all remaining text
	if (L != string::npos)
		cols.push_back( line.substr( L, string::npos ) );

	//if ( cols.size() != 12 )
	//{
	//	std::cout << "Data format error: " << line << std::endl;
	//	return true;
	//}

	rec.qname = cols[0];			// sequence name
	rec.flag = atoi( cols[1].c_str() );
	rec.rname = cols[2];
	rec.istart = atoi( cols[3].c_str() );		// start coordinate
	rec.mapqual = atoi( cols[4].c_str() );
	rec.cigar = cols[5];

	procCigar( rec );
	if (!rec.bGap)
		return false;
	
	rec.seq = cols[9];
	if (cols.size() > 10)
		rec.qual = cols[10];

	setScore( rec );

	// save new record
	m_aligns.push_back( rec );

	return false;
};

void cOrigin::Sort()
{
	sort( m_aligns.begin(), m_aligns.end(), originRecLessThan );
}

void cOrigin::procCigar(alignRec& rec)
{
	// determine rec.pieces : the regions of the query sequence which cover DNA nt
	//
	size_t pos = rec.cigar.find_first_of(m_cigarChars);
	size_t plast;
	if (pos == string::npos)
		return;

	string s = rec.cigar.substr(0,pos);
	int ilast(-1), n, len(0);
	piece p;
	bool bGap(false);
	
	char ch;

	while (pos != string::npos)
	{
		ch = rec.cigar.at(pos);
		
		switch (ch)
		{
		case 'M': case '=':
			p.start = ilast + 1;
			p.length = atoi( s.c_str() );
			len += p.length;
			ilast = p.start + p.length;
			rec.pieces.push_back( p );
			break;

		case 'I':		// insertion to the reference
			n = atoi( s.c_str() );
			len += n;
			break;

		case 'D':		// deletion from the reference
			break;

		case 'N':		// skipped region from the reference (intron)
			n = atoi( s.c_str() );
			ilast += n - 1;
			bGap = true;
			break;

		case 'S': case 'H':		// soft clipping and hard clipping
			break;

		case 'P':		// padding (silent deletion from padded reference)
			break;

		case 'X':		// sequence mismatch
			p.start = ilast + 1;
			p.length = atoi( s.c_str() );
			len += p.length;
			rec.pieces.push_back( p );
			break;
		};

		plast = pos;
		pos = rec.cigar.find_first_of(m_cigarChars, plast + 1);

		s = rec.cigar.substr(plast + 1,pos);

	}	// while pos != string::npos

	rec.nLength = len;
	rec.bGap = bGap;
}

void cOrigin::setScore(alignRec& rec)
{
	vector<double> vd;
	int i, iq;
	double lf, lscore(0.);

	for (i=0; i<(int)rec.qual.length(); ++i)
	{
		iq = rec.qual.at(i) - 33;
		lf = -10. * log10( double(iq) );
		lscore += lf;
	}

	// normalize by length
	rec.score = lscore / double(rec.qual.length());
}
