
#include "metrics_sr.h"
#include <stdio.h>
#include <algorithm>
#include <limits>

// for gene sorting
bool intronLessThan (sIntron i1, sIntron i2)
{
	if (i1.nContig==i2.nContig)
	{
		if (i1.ibeg==i2.ibeg)
			return (i1.iend < i2.iend);
		else
			return (i1.ibeg < i2.ibeg);
	}
	else
		return i1.nContig < i2.nContig;
}

bool intronLessThan2 (sIntron i1, sIntron i2)
{
	if (i1.strId.compare(i2.strId)==0)
	{
		if (i1.ibeg==i2.ibeg)
			return (i1.iend < i2.iend);
		else
			return (i1.ibeg < i2.ibeg);
	}
	else
		return i1.strId.compare(i2.strId) < 0;
}

bool intronLessThan3 (sIntron i1, sIntron i2)
{
	if (i1.ibeg==i2.ibeg)
		return (i1.iend < i2.iend);
	else
		return (i1.ibeg < i2.ibeg);
}

bool intronEqualTo(sIntron i1, sIntron i2)
{
	return (i1.nContig==i2.nContig && i1.ibeg==i2.ibeg && i1.iend==i2.iend);
}

bool intronEqualTo2(sIntron i1, sIntron i2)
{
	return (i1.nContig==i2.nContig && i1.ibeg==i2.ibeg && i1.iend==i2.iend);
}

// assumes introns are on same contig
bool donLessThan(sIntron i1, sIntron i2)
{
	int don1 = i1.iStrand < 0 ? i1.iend : i1.ibeg;
	int don2 = i2.iStrand < 0 ? i2.iend : i2.ibeg;

	return don1 < don2;
}
// assumes introns are on same contig
bool accLessThan(sIntron i1, sIntron i2)
{
	int acc1 = i1.iStrand < 0 ? i1.ibeg : i1.iend;
	int acc2 = i2.iStrand < 0 ? i2.ibeg : i2.iend;

	return acc1 < acc2;
}

void cMetricsSR::reset()
{
	nDon = nAcc = nIntron = 0;
	nDonP = nAccP = nIntronP = 0;
	nDonOk = nAccOk = nIntronOk = 0;
	nUtrIntronOk = 0;
	nRefIntrons = 0;

	nDonPred = nAccPred = nIntronPred = nUtrPred = 0;

	nDonMiss = nAccMiss = 0;
	nDonFalse = nAccFalse = 0;
	nIntronMiss = nFalseIntron = 0;
}

bool cMetricsSR::markAgreement( vector<sIntron>& ref_introns, vector<sIntron>& query_introns, int iType )
	// modify ref_genes boundary properties which agree with the query
{
	// 1. argument checking
	if (ref_introns.empty() || query_introns.empty())
		return false;

	// 2. sort both intron dbs
	sort( ref_introns.begin(), ref_introns.end(), intronLessThan3 );
	sort( query_introns.begin(), query_introns.end(), intronLessThan3 );

	int nIntrons = ref_introns.size();
	int ixQuery = 0, ii = 0, icount;
	int nQuery = query_introns.size();
	bool bMiss;

	int nDon0 = nDonOk;
	int nAcc0 = nAccOk;
	int nIntron0 = nIntronOk;
	int nTot(0);
	
	// 3. find all boundary + feature matches with query
	// NOTE: this assumes that NO query intron features are self-overlapping
	for (ii=0; ii<nIntrons; ++ii)
	{
		bMiss = false;

		// get to first overlapping intron (if any)
		while ( query_introns[ixQuery].iend < ref_introns[ii].ibeg )
		{
			++ixQuery;
			if (ixQuery>=nQuery) {
				if (iType==0) { nIntronMiss += nIntrons - (ii+1); }
				else { nFalseIntron += nIntrons - (ii+1); }

				goto end_e;
			}
		}

		icount = 0;

		while ( query_introns[ixQuery+icount].ibeg < ref_introns[ii].iend )
		{
			if (checkOverlap(ref_introns[ii], query_introns[ixQuery+icount]))
			{
				if (iType==0)
					markIntronPh(ref_introns[ii], query_introns[ixQuery+icount]);
				else
					markIntron(ref_introns[ii], query_introns[ixQuery+icount]);
			}

			++icount;

			if (ixQuery+icount >= nQuery)
				break;
		}

		bMiss = ref_introns[ii].extr_L == 0 && ref_introns[ii].extr_R == 0;

		if (iType==0)
		{
			if (bMiss)
				++nIntronMiss;
		}
		else
		{
			if (bMiss)
				++nFalseIntron;
		}

		// should be through with ref_introns[ii].. so score it
		scoreIntron( ref_introns[ii], iType );
	}

end_e:
	nTot = (int) query_introns.size();

	if (iType == 0)
		printf("%s (%d, %d, %d/%d) don, acc, full matches\n", ref_introns[0].strId.c_str(), nDonOk-nDon0, nAccOk-nAcc0, nIntronOk-nIntron0, nTot );

	return false;
}

bool cMetricsSR::getNN( vector<sIntron>& ref_introns, vector<sIntron>& query_introns )
	// modify query_introns to indicate the nearest ref donor (upstream and downstream)
	//  and acceptor (upstream and downstream)
{
	// argument checking
	if (ref_introns.empty() || query_introns.empty())
		return false;

	vector<sIntron> tmp_que = query_introns;
	vector<sIntron> tmp_ref = ref_introns;

	std::sort( tmp_que.begin(), tmp_que.end(), donLessThan );
	std::sort( tmp_ref.begin(), tmp_ref.end(), donLessThan );

	donNN( tmp_ref, tmp_que );

	std::sort( tmp_que.begin(), tmp_que.end(), accLessThan );
	std::sort( tmp_ref.begin(), tmp_ref.end(), accLessThan );

	accNN( tmp_ref, tmp_que );

	// set output
	std::sort( tmp_que.begin(), tmp_que.end(), intronLessThan );
	query_introns = tmp_que;
	
	return false;
}

bool cMetricsSR::resetDonDist( vector<sIntron>& query_introns )
{
	for (int i=0; i<(int)query_introns.size(); ++i)
	{
		query_introns[i].dDon = std::numeric_limits<int>::max();
	}

	return false;
}

bool cMetricsSR::resetAccDist( vector<sIntron>& query_introns )
{
	for (int i=0; i<(int)query_introns.size(); ++i)
	{
		query_introns[i].dAcc = std::numeric_limits<int>::max();
	}

	return false;
}

bool cMetricsSR::donNN( vector<sIntron>& ref_introns, vector<sIntron>& query_introns )
{
	int nRef = (int) ref_introns.size();
	int nQuery = (int) query_introns.size();
	int i, j(-1);
	int pRef(0), pLastRef(0), pQue;
	int dUp, dDown;
	int flag(0);

	if (ref_introns.empty()) {
		resetDonDist(query_introns);
		return false;
	}

	int limitL;
	int limitR;

	switch (query_introns.front().iStrand)
	{
	case 0: case 1:
		limitL = ref_introns.front().ibeg;
		break;
	default:
		limitL = ref_introns.front().iStrand < 0 ? ref_introns.front().iend : ref_introns.front().ibeg;
	};

	switch (query_introns.back().iStrand)
	{
	case 0: case 1:
		limitR = ref_introns.back().iend;
		break;
	default:
		limitR = ref_introns.back().iStrand < 0 ? ref_introns.back().ibeg : ref_introns.back().iend;
	};

	for (i=0; i<nQuery; ++i)
	{
		pQue = query_introns[i].iStrand<0 ? query_introns[i].iend : query_introns[i].ibeg;

		if (pLastRef==183630)
			flag = 1;

		pRef = nextDon( pRef, pLastRef, pQue, j, ref_introns );

		if (pQue > limitL && pQue <= limitR) {
			dUp = pLastRef - pQue;
			dDown = pRef - pQue;
			query_introns[i].dDon = abs(dUp) > abs(dDown) ? dDown : dUp;
			if ( abs(query_introns[i].dDon) > 50000 )
				flag = 1;
		}
		else if (pQue <= limitL) {
			query_introns[i].dDon = pRef - pQue;
		}
		else if (pQue > limitR) {
			query_introns[i].dDon = pLastRef - pQue;
		}

	}

	return false;
}

bool cMetricsSR::accNN( vector<sIntron>& ref_introns, vector<sIntron>& query_introns )
{
	int nRef = (int) ref_introns.size();
	int nQuery = (int) query_introns.size();
	int i, j(-1);
	int pRef(0), pLastRef(0), pQue;
	int dUp, dDown;

	if (ref_introns.empty()) {
		resetAccDist(query_introns);
		return false;
	}

	int limitL = query_introns.front().iStrand<0 ? ref_introns.front().iend : ref_introns.front().ibeg;
	int limitR = query_introns.back().iStrand<0 ? ref_introns.back().iend : ref_introns.back().ibeg;

	for (i=0; i<nQuery; ++i)
	{
		pQue = query_introns[i].iStrand<0 ? query_introns[i].ibeg : query_introns[i].iend;

		pRef = nextAcc( pRef, pLastRef, pQue, j, ref_introns );

		if (pQue > limitL && pQue <= limitR) {
			dUp = pLastRef - pQue;
			dDown = pRef - pQue;
			query_introns[i].dAcc = abs(dUp) > abs(dDown) ? dDown : dUp;
		}
		else if (pQue <= limitL) {
			query_introns[i].dAcc = pRef - pQue;
		}
		else if (pQue > limitR) {
			query_introns[i].dAcc = pLastRef - pQue;
		}

	}

	return false;
}

int cMetricsSR::nextDon( int curDon, int& lastDon, int pos, int& ix, vector<sIntron>& introns )
{
	int sz = (int) introns.size();
	int next;
	if (ix < 0)
		next = lastDon;
	else
		next = introns[ix].iStrand < 0 ? introns[ix].iend : introns[ix].ibeg;

	int flag(0), ct(0);
	int ixStart = ix;

	while (next <= pos)
	{
		if (ix < sz-1) {
			++ix;
			++ct;
			lastDon = next;
			next = introns[ix].iStrand < 0 ? introns[ix].iend : introns[ix].ibeg;
			if ( (next-lastDon)>50000 )
				flag=1;
		}
		else return introns.back().iStrand < 0 ? introns.back().iend : introns.back().ibeg;
	}
	return next;
}

int cMetricsSR::nextAcc( int curAcc, int& lastAcc, int pos, int& ix, vector<sIntron>& introns )
{
	int sz = (int) introns.size();
	int next;
	if (ix < 0)
		next = lastAcc;
	else
		next = introns[ix].iStrand < 0 ? introns[ix].ibeg : introns[ix].iend;

	while (next <= pos)
	{
		if (ix < sz-1) {
			++ix;
			lastAcc = next;
			next = introns[ix].iStrand < 0 ? introns[ix].ibeg : introns[ix].iend;
		}
		else return introns.back().iStrand < 0 ? introns.back().ibeg : introns.back().iend;
	}
	return next;
}

bool cMetricsSR::checkOverlap( sIntron& intron1, sIntron& intron2 )
// check if whole gene regions have any overlap (does not check CDS)
{
	if (intron1.ibeg > intron2.iend)
		return false;

	if (intron1.iend < intron2.ibeg)
		return false;

	return true;
}

bool cMetricsSR::markIntron( sIntron& intron, sIntron& query_intron )
// mark intron's boundary features if they match query
{
	if (intron.ibeg == query_intron.ibeg) {
		intron.extr_L = 1;
		intron.lenExonL = query_intron.lenExonL;
	}

	if (intron.iend == query_intron.iend) {
		intron.extr_R = 1;
		intron.lenExonR = query_intron.lenExonR;
	}

	if (intron.ibeg == query_intron.ibeg &&
		intron.iend == query_intron.iend)
		intron.extr_I = 1;


	return false;
}

bool cMetricsSR::markIntronPh( sIntron& intron, sIntron& query_intron )
// mark intron's boundary features if they match query
//  Also: in case of full match, copy ref intron frame over to query intron
{
	if (intron.ibeg == query_intron.ibeg)
		intron.extr_L = 1;

	if (intron.iend == query_intron.iend)
		intron.extr_R = 1;

	if (intron.ibeg == query_intron.ibeg &&
		intron.iend == query_intron.iend)
	{
		intron.extr_I = 1;
		query_intron.iframe = intron.iframe;
	}

	return false;
}

void cMetricsSR::scoreIntron( sIntron& intron, int iType )
{
	bool bDon(false);
	bool bAcc(false);

	if (iType == 0)
	{
		if (intron.extr_L != 0)
		{
			if (intron.iStrand == -1)
				//nAccOk += 1;
				bAcc = true;
			else
				//nDonOk += 1;
				bDon = true;
		}

		if (intron.extr_R != 0)
		{
			if (intron.iStrand == -1)
				//nDonOk += 1;
				bDon = true;
			else
				//nAccOk += 1;
				bAcc = true;
		}

		if (bDon && !findDon(intron))
			++nDonOk;
		
		if (bAcc && !findAcc(intron))
			++nAccOk;

		if (intron.extr_I != 0 && !findIntron(intron)) {
			++nIntronOk;
			if (intron.iframe == -1)
				++nUtrIntronOk;
		}

	}
	else
	{
		if (intron.extr_L != 0)
		{
			if (intron.iStrand == -1)
				nAccPred += 1;
			else
				nDonPred += 1;
		}

		if (intron.extr_R != 0)
		{
			if (intron.iStrand == -1)
				nDonPred += 1;
			else
				nAccPred += 1;
		}

		if (intron.extr_I != 0) {
			nIntronPred += 1;
			if (intron.iframe == -1)
				++nUtrPred;
		}
	}

	// update cache
	m_intronCache.push_back(intron);
	if ((int)m_intronCache.size() > m_cacheLen)
		m_intronCache.erase( m_intronCache.begin() );
}
		//metrics.writeMetrics( fout, nIntronRefCount, nDonRef, nAccRef, nQueryCount, nDonQuery, nAccQuery, nUtrRef );

void cMetricsSR::writeMetrics( string fname, int nRefIntrons, int nDonRef, int nAccRef, int nPredIntrons, int nDonQ, int nAccQ, int nUtr )
{
	// write results to file
	FILE *pf = fopen( fname.c_str(), "w" );

	//fprintf( pf, "nDonOk =               %d\n", nDonOk );
	//fprintf( pf, "nDonRef =              %d\n", nDonRef );
	//fprintf( pf, "nDonQ =                %d\n", nDonQ );
	//fprintf( pf, "nAccOk =               %d\n", nAccOk );
	//fprintf( pf, "nAccRef =              %d\n", nAccRef );
	//fprintf( pf, "nAccQ =                %d\n", nAccQ );
	fprintf( pf, "- - - - - - - - - - - -\n" );
	fprintf( pf, "Don     Sn =      %5.2f\n", double(nDonOk)/double(nDonRef)*100. );
	fprintf( pf, "        Sp =      %5.2f\n", double(nDonOk)/double(nDonQ)*100. );
	fprintf( pf, "- - - - - - - - - - - -\n" );
	fprintf( pf, "Acc     Sn =      %5.2f\n", double(nAccOk)/double(nAccRef)*100. );
	fprintf( pf, "        Sp =      %5.2f\n", double(nAccOk)/double(nAccQ)*100. );
	fprintf( pf, "- - - - - - - - - - - -\n" );
	fprintf( pf, "Intron  Sn =      %5.2f\n", double(nIntronOk)/double(nRefIntrons)*100. );
	fprintf( pf, "        Sp =      %5.2f\n", double(nIntronOk)/double(nPredIntrons)*100. );
	fprintf( pf, "- - - - - - - - - - - -\n" );
	fprintf( pf, "Number of reference introns = %d\n", nRefIntrons );
	fprintf( pf, "   ref. UTR introns  =  %d\n", nUtr );
	fprintf( pf, "   ref. CDS introns  =  %d\n", nRefIntrons - nUtr );
	fprintf( pf, "Number of introns implied by alignment = %d\n", nPredIntrons );
	fprintf( pf, "   correct UTR introns  =  %d\n", nUtrPred );
	fprintf( pf, "   correct CDS introns  =  %d\n", nIntronOk - nUtrPred );
	fprintf( pf, "   correct introns (total)  =  %d\n", nIntronOk );
	fprintf( pf, "   incorrect introns  =  %d\n", nPredIntrons - nIntronOk );
	fprintf( pf, "Missed introns =  %d\n", nIntronMiss );
	fprintf( pf, "Fully unmatched introns  =  %d\n", nFalseIntron );
	fprintf( pf, "\n" );
	fprintf( pf, "Correct donors  =  %d\n", nDonOk );
	fprintf( pf, "Correct acceptors  =  %d\n", nAccOk );

	fclose( pf );
}

bool cMetricsSR::findDon(sIntron& intron)
{
	int i;

	if (intron.iStrand == -1)
	{
		for (i=m_intronCache.size()-1; i>=0; --i)
		{
			if (m_intronCache[i].iend==intron.iend)
				return true;
		}
	}
	else
	{	// positive strand
		for (i=m_intronCache.size()-1; i>=0; --i)
		{
			if (m_intronCache[i].ibeg==intron.ibeg)
				return true;
		}
	}

	return false;
}

bool cMetricsSR::findAcc(sIntron& intron)
{
	int i;

	if (intron.iStrand == -1)
	{
		for (i=m_intronCache.size()-1; i>=0; --i)
		{
			if (m_intronCache[i].ibeg==intron.ibeg)
				return true;
		}
	}
	else
	{	// positive strand
		for (i=m_intronCache.size()-1; i>=0; --i)
		{
			if (m_intronCache[i].iend==intron.iend)
				return true;
		}
	}

	return false;
}

bool cMetricsSR::findIntron(sIntron& intron)
{
	int i;

	for (i=m_intronCache.size()-1; i>=0; --i)
	{
		if (m_intronCache[i].ibeg==intron.ibeg && m_intronCache[i].iend==intron.iend)
			return true;
	}

	return false;
}
