//**************************************************
//  file:     matrix.cpp
//  project:  ehmm
//**************************************************

#include <stdio.h>
#include <memory.h>
#include <math.h>

#include "matrixmod.h"
//#include "common.h"

//================================================
// size = 4 ^ ( model order + 1 )
//
// order   0   1   2    3     4     5
//  size   4  16  64  256  1024  4096
//================================================

//--------------------------------------------------
// P(dir)(  | e1 | e2 | e3 | ...           | en |  )
// P(rev)(  | e1 | e2 | e3 | ...           | en |  )
//
// P(dir)( N ) != P(rev)( N ) == P(dir)( N' )
// calc N' from N 
//
//          | en |           ... | e3 | e2 | e1 |
//             |                    |    |    |
// P(dir)(  | en'|           ... | e3'| e2'| e1'|  )
//
// P(rev) == P(dir)
//
// to convert element in complement 
// element - { A C G T }
// switch( element ) {         | if elements order are fixed
//   case A: element = T;      | A C G T
//   case C: element = G;      | than 3 - element ==
//   case G: element = C;      | T G C A
//   case T: element = A;      | 
// }                           | 
//--------------------------------------------------
void cMatrixMod::DoRevFromDir( double* dir, double* rev , int order)
{
	int n = 0;
	int calc;
	int temp;
	const int mask = 3;				// mask == 0000 0011

	int size = 4<<(2*order);

	while ( n < size ) {

		temp = n;
		calc = 0;

		for ( int count = 0; count <= order; ++count ) {

			calc <<= 2;
			calc +=  3 - ( temp & mask);
			temp >>= 2;
		}

		rev[n] = dir[calc];
		++n;
	}
};
//--------------------------------------------------
// sum of   hiAbs( - - ... - T )
//          hiAbs( - - ... - G )
//          hiAbs( - - ... - C )
//          hiAbs( - - ... - A )
//          ------------------
//   Prob   loAbs( - - ... - )
//--------------------------------------------------
void cMatrixMod::DoHiToLo( double* hi, double* lo , int hiOrder )
{
	double* loopEnd = hi + ( 4 << ( 2*hiOrder ) );

	while ( hi < loopEnd ) {

		*lo = hi[A] + hi[C] + hi[G] + hi[T];
		++lo;
		hi += 4;
	}
};
//--------------------------------------------------
//  P( e1 before e2 e3 ... en ) = P( e1 e2 e3 ... en ) / P( e2 e3 ... en )
//
//	P( e2 e3 ... en ) = sum of
//         P( T e2 e3 ... en )
//         ...
//         P( G e2 e3 ... en )
//         ...
//         P( C e2 e3 ... en )
//         ...                     1/4 of matrix
//         P( A e2 e3 ... en )
//--------------------------------------------------
void cMatrixMod::DoAbsToFirst( double* abs, double* rel, int order )
{
	int size = 4<<(2*order);
	double* loopEnd = abs + size/4;
	double temp;

	const int leftA = ( A << (2*order) );
	const int leftC = ( C << (2*order) );
	const int leftG = ( G << (2*order) );
	const int leftT = ( T << (2*order) );

	while ( abs < loopEnd ) {

		temp = abs[ leftA ] + abs[ leftC ] + abs[ leftG ] + abs[ leftT ];

		if ( temp != 0 ) {
			rel[ leftA ] = abs[ leftA ] / temp;
			rel[ leftC ] = abs[ leftC ] / temp;
			rel[ leftG ] = abs[ leftG ] / temp;
			rel[ leftT ] = abs[ leftT ] / temp;
		} else {
			rel[ leftA ] = 0;
			rel[ leftC ] = 0;
			rel[ leftG ] = 0;
			rel[ leftT ] = 0;
		}

		++abs;
		++rel;
	}
};
//--------------------------------------------------
bool cMatrixMod::AllocArray( double** data, int size )
{
	if ( *data ) delete [] *data;
	*data = new double [ size ];
	if ( !*data ) {
		printf( "error, out of memory\n" );
		return false;
	}
	memset( *data, 0, size*sizeof(double) );

	return true;
};
//--------------------------------------------------
bool cMatrixMod::RevFromDir( double* dir, double** rev , int order )
{
	if ( !AllocArray( rev, (4<<(2*order)) + 1 ) )
		return false;

	DoRevFromDir( dir, *rev , order );

	(*rev)[4<<(2*order)] = dir[4<<(2*order)];

	return true;
};
//--------------------------------------------------
bool cMatrixMod::AbsToFirst( double* abs, double** rel, int order )
{
	if ( !AllocArray( rel, (4<<(2*order)) + 1 ) )
		return false;

	DoAbsToFirst( abs, *rel , order );

	(*rel)[4<<(2*order)] = abs[4<<(2*order)];

	return true;
};
//--------------------------------------------------
void cMatrixMod::Log( double* ptr , int size )
{
	for ( int i = 0; i < size; ++i )
	{
		if ( ptr[i] < 0 )
			i=i;

		if ( ptr[i] )
			ptr[i] = log( ptr[i] );
		else
			ptr[i] = log0;
	}
};
//--------------------------------------------------
void cMatrixMod::Exp( double* ptr , int size )
{
	for ( int i = 0; i < size; ++i )
	{
		if (ptr[i]==log0)
			ptr[i] = 0.;
		else
			ptr[i] = exp( ptr[i] );
	}
}

//--------------------------------------------------
bool cMatrixMod::HiToLo( double* hi, double** lo, int hiOrder )
{
	if ( hiOrder < 1 )
	{
		printf( "error in calculation of low order matrix\n" );
		return false;
	}
	if ( !AllocArray( lo, (4<<(2*(hiOrder-1))) + 1 ) )
		return false;

	DoHiToLo( hi, *lo, hiOrder );

	(*lo)[4<<(2*(hiOrder - 1))] = hi[4<<(2*hiOrder)];

	return true;
};
//--------------------------------------------------


//----------------------------------------------
// a - 0 c - 1 g - 2 t - 3
//----------------------------------------------
void cMatrixMod::SetAlphabet( char* ptr )
{
	memset( ptr, 4, 256 );

	*(ptr + 'a' ) = 0;
	*(ptr + 'A' ) = 0;
	*(ptr + 'c' ) = 1;
	*(ptr + 'C' ) = 1;
	*(ptr + 'g' ) = 2;
	*(ptr + 'G' ) = 2;
	*(ptr + 't' ) = 3;
	*(ptr + 'T' ) = 3;
	*(ptr + 'u' ) = 3;
	*(ptr + 'U' ) = 3;
};
//----------------------------------------------
// no frames
//----------------------------------------------
int cMatrixMod::Calc0( double* mat, const char* buf, char* alphabet, int order, int fsize, bool bAddCompl )
{
	int nDataCount = 0;
	int count = 0;
	int i = 0, j;
	int new_letter;
	int letter_shift;
	int adr = 0, adr_rc = 0;
	int adr0 = 0;
	int msize = 4<<(2*order);
	int mask = (4<<(2*order)) - 1;
	int letter_mask = 0x3;
	int compl_letter_mask = 0x3;

	char chLetter;

	for (j=0; j<order; ++j) {
		compl_letter_mask = compl_letter_mask<<(2*order);
	}


	int NNN = 0;

	while ( nDataCount < fsize  ) {

		if ( isalpha(*(buf+nDataCount)) != 0 ) {

			chLetter = buf[nDataCount];
			new_letter = alphabet[chLetter];

			NNN = (NNN<<2);
			NNN &= mask;
			if ( new_letter == 4 )
				NNN += 3;

			if ( i < order ) {
			
				adr = (adr<<2);
				adr += new_letter;

				if ( new_letter != 4 )
				{
					// i should always equal zero here
					for (j=0; j<4; ++j) {
						adr0 = j;
						adr0 = (adr0<<2);
						adr0 += new_letter;
						assert(adr0>=0 && adr0<msize);
						//{
						//	printf("adr0 = %d, msize = %d, letter = %s\n", adr0, msize, chLetter);
						//}
						++mat[adr0];
					}
					++count;
				}

			} else {

				// direct address
				adr = (adr<<2);
				adr += new_letter;
				adr &= mask;

				// reverse complement address
				letter_shift = (letter_mask & (~new_letter))<<(2*order);
				adr_rc = (adr_rc>>2);
				adr_rc += letter_shift;			// set msb position to letter_shift

				if ( !NNN ) {
					assert( (adr+msize*i)>=0 && (adr+msize*i)<(msize*fsize) );
					++mat[adr + msize*i];
					++count;
					if (bAddCompl) {
						assert(adr_rc>=0 && adr_rc<(msize*fsize));
						++mat[adr_rc];
						++count;
					}
				}
			}
			++i;
		}
		else {
			//chLetter = buf[nDataCount];
			//printf("buf=%s, isalpha==0 for: %c, nDataCount=%d\n", buf, chLetter, nDataCount);
		}
		++nDataCount;
	}

	return count;
}

int cMatrixMod::Calc( double* mat, const char* buf, char* alphabet, int order, int fr, int fsize, bool bAddCompl )
{
	int nDataCount = 0;
	int count = 0;
	int i = 0;
	int new_letter;
	int letter_shift;
	int adr = 0, adr_rc = 0;
	int msize = 4<<(2*order);
	int mask = msize - 1;
	int letter_mask = 0x3;
	int compl_letter_mask = 0x3;

	for (int j=0; j<order; ++j) {
		compl_letter_mask = compl_letter_mask<<(2*order);
	}

	int next = fr + order - 1;

	int NNN = 0;

	while ( nDataCount < fsize  ) {

		if ( isalpha(*(buf+nDataCount)) != 0 ) {

			new_letter = alphabet[*(buf+nDataCount)];

			NNN = (NNN<<2);
			NNN &= mask;
			if ( new_letter == 4 )
				NNN += 3;

			if ( i < order ) {
			
				adr = (adr<<2);
				adr += new_letter;
			} else {

				// direct address
				adr = (adr<<2);
				adr += new_letter;
				adr &= mask;

				// reverse complement address
				letter_shift = (letter_mask & (~new_letter))<<(2*order);
				adr_rc = (adr_rc>>2);
				adr_rc += letter_shift;			// set msb position to letter_shift

				if ( i == next ) {

					next += 3;

					if ( !NNN ) {
						++mat[adr];
						if (bAddCompl) {
							assert(adr_rc>=0 && adr_rc<msize);
							++mat[adr_rc];
						}
					}
				}
			}
			++i;
		}
		++nDataCount;
	}

	count = 0;

	for ( i=0; i<(msize); ++i )
        {
		count += *(mat+i);
	}

	return count;
}

int cMatrixMod::Calc( double* mat, char* buf, char* alphabet, int order, int fr, int fsize, bool bAddCompl )
{
	// cast the buffer ptr to const
	const char *cpBuf = buf;
	return Calc( mat, cpBuf, alphabet, order, fr, fsize, bAddCompl );
}

void cMatrixMod::Calc( double *pMat, vector<string>& dataVect, int order, int flankLeft, int flankRight )
{
	// pMat is 4^order-by-len matrix

	int msize = 4<<(2*order);
	char alphabet[256];
	int i, irow, icol, j;
	int len = (int)dataVect[0].size();

	int iflag = 0;
	if (order>0)
		iflag = 1;

	// baseline pseudocount
	double lfpc = 0.01;

	// initialize pMat (pseudocount)
	for (i=0; i<msize*len; ++i)
		pMat[i] = 0.;

	// storage for zero-order model nt frequency vs position data
	double *pMat0 = new double[4*len];
	// initialize pMat0 (pseudocount)
	for (i=0; i<4*len; ++i)
		pMat0[i] = 0.;

	// convert char alphabet to positive integers 0 - 3 (ACGT), or 4 (N)
	SetAlphabet( alphabet );

	// counting
	int nCount = 0, nCount0 = 0;

	for (irow=0; irow<(int)dataVect.size(); ++irow)
	{
		// accumulate counts in two-dimension matrix pMat
		nCount += Calc0( pMat, dataVect[irow].c_str(), alphabet, order, len, false );
		if (order>0) {
			// collect zero-order model data
			nCount0 += Calc0( pMat0, dataVect[irow].c_str(), alphabet, 0, len, false );
		}
	}

	// normalize pMat0 by nCount0
	for (i=0; i<4*len; ++i) {
		pMat0[i] /= double(dataVect.size());
	}

	// normalize pMat0 by nCount0
	for (i=0; i<4*len; ++i) {
		pMat0[i] /= double(dataVect.size());
	}

	for (i=0; i<msize*len; ++i) {

		icol = (int) floor( double(i)/double(msize) );
		bool bCanonicalCol = !( icol<flankLeft || (icol+1)>(len-flankRight) );

		// normalize pMat by count
		if (pMat[i]>0)
			pMat[i] /= double(dataVect.size());
		else if ( !bCanonicalCol )
			pMat[i] = lfpc;
		else
			pMat[i] = 0.;

		if (order>0 && i>=msize) {		// normalize by zero-order nt frequencies
			// normalize by 0-order nt frequencies
			j = (int) ( double(i)/(double(msize)/4.) - double(msize)/4. );
			assert(j>=0 && j<(4*len));
			assert(i>=0 && i<(msize*len));
			if (pMat0[j]==0) {
				pMat[i] = 0.;
			}
			else {
				pMat[i] /= pMat0[j];
			}
		}
	}
	
	delete []pMat0;

}

void cMatrixMod::Calc( double *pMat, vector<string>& dataVect, int order, int len, int flankLeft, int flankRight )
{
	// pMat is 4^order-by-len matrix

	int msize = 4<<(2*order);
	char alphabet[256];
	int i, irow, icol, j;

	int iflag = 0;
	if (order>0)
		iflag = 1;

	// baseline pseudocount
	double lfpc = 0.01;

	// initialize pMat (pseudocount)
	for (i=0; i<msize*len; ++i)
		pMat[i] = 0.;

	// storage for zero-order model nt frequency vs position data
	double *pMat0 = new double[4*len];
	// initialize pMat0 (pseudocount)
	for (i=0; i<4*len; ++i)
		pMat0[i] = 0.;

	// convert char alphabet to positive integers 0 - 3 (ACGT), or 4 (N)
	SetAlphabet( alphabet );

	// counting
	int nCount = 0, nCount0 = 0;

	for (irow=0; irow<(int)dataVect.size(); ++irow)
	{
		// accumulate counts in two-dimension matrix pMat
		nCount += Calc0( pMat, dataVect[irow].c_str(), alphabet, order, len, false );
		if (order>0) {
			// collect zero-order model data
			nCount0 += Calc0( pMat0, dataVect[irow].c_str(), alphabet, 0, len, false );
		}
	}

	// normalize pMat0 by nCount0
	for (i=0; i<4*len; ++i) {
		pMat0[i] /= double(dataVect.size());
	}

	for (i=0; i<msize*len; ++i) {

		icol = (int) floor( double(i)/double(msize) );
		bool bCanonicalCol = !( icol<flankLeft || (icol+1)>(len-flankRight) );

		// normalize pMat by count
		if (pMat[i]>0)
			pMat[i] /= double(dataVect.size());
		else if ( !bCanonicalCol )
			pMat[i] = lfpc;
		else
			pMat[i] = 0.;

		if (bCanonicalCol && pMat[i]>0.)
			pMat[i] = 1.;

		if (!bCanonicalCol && icol==0 && order>0)
			pMat[i] *= pMat0[i/4];

		if (order>0 && i>=msize) {		// normalize by zero-order nt frequencies
			// normalize by 0-order nt frequencies
			j = (int) ( double(i)/(double(msize)/4.) - double(msize)/4. );
			assert(j>=0 && j<(4*len));
			assert(i>=0 && i<(msize*len));
			if (pMat0[j]==0) {
				pMat[i] = 0.;
			}
			//else {
			//	pMat[i] /= pMat0[j];
			//}
		}
	}
	
	delete []pMat0;

}

//----------------------------------------------
char* cMatrixMod::MakeIndex( int k, int order, char* ptr )
{
	char letter[4] = { 'A','C','G','T' };
	int i;

	for ( i = 0; i <= order; ++i ) {

		ptr[order - i] = letter[ k&3];
		k >>= 2;
	}
	ptr[i] = '\0';

	return ptr;
};
//----------------------------------------------
