
#include "model.h"
#include "targetver.h"
#include <limits>
#include <iostream>

char UAA[21]  = "ACDEFGHIKLMNPQRSTVWY";

extern bool loadModel( string& fname );
extern bool getOrganism( string& name );
extern bool getOrder( int& order );
extern bool getMarkovCod( cMarkovCod& m );
extern bool getMarkovNorm( cMarkovCod& mc, cMarkovNon& mn );
extern bool getMarkovIntr( cMarkovIntr& m );
extern bool getMarkovNon( cMarkovNon& m );
extern bool getSite( double **pval, int& order, int& width, int& margin, string& keyName );
extern bool getLength( double **pval, int *pmin, int *pmax, string keyName );
extern bool getPhaseProbs( double pIni[3], double pTer[3], double pInt[9] );
extern bool getTransitionProbs( double& psg, double& pie );
extern bool intronLengthConstant();

extern bool loadModel_2( string& fname );
extern bool getOrganism_2( string& name );
extern bool getOrder_2( int& order );
extern bool getMarkovCod_2( cMarkovCod& m );
extern bool getMarkovNorm_2( cMarkovCod& mc, cMarkovNon& mn );
extern bool getMarkovIntr_2( cMarkovIntr& m );
extern bool getMarkovNon_2( cMarkovNon& m );
extern bool getSite_2( double **pval, int& order, int& width, int& margin, string& keyName );
extern bool getLength_2( double **pval, int *pmin, int *pmax, string keyName );


void cModel::init(cSite *psite)
{
	m_pSite = psite;

	m_b3phIntron = false;

	// separate intron and non-coding 5th order models?
	m_bSeparateIntronModel_5 = false;

	m_bVerbose = true;

	m_nIter = 0;

	m_pDonMotif = m_pAccMotif = NULL;
	m_pATGMotif = m_pTAAMotif = m_pTAGMotif = m_pTGAMotif = NULL;
	m_pShortAccMotif = m_pBpMotif = NULL;

	m_bExtr = false;
}

void cModel::setVerbose(bool bVerbose)
{
	m_bVerbose = bVerbose;
}

void cModel::report(string str)
{
	if (m_bVerbose)
		printf("%s\n",str.c_str());
}


string cModel::strInt2Str( string strPrefix, int i )
{
	char pch[32];
	sprintf( pch, "%d", i );

	string strOut = strPrefix;
	strOut.append( pch );

	return strOut;
}

void cModel::setDebugDir( string strDebugOutDir )
{
	m_debugOutDir = strDebugOutDir + SLASH;
}


void cModel::setFileNames()
{
	m_startSiteFile = "start_site.txt";
	m_stopSiteFile = "stop_site.txt";
	m_donSiteFile = "don_site.txt";
	m_accSiteFile = "acc_site.txt";
	m_phaseFile = "splice_phase.txt";
	m_codingFile = "coding.txt";
	m_intronFile = "introns.txt";
	m_flankFile = "flank.txt";
	m_nonCodingFile = "non_coding.txt";
}



void cModel::getModelParams(void)
{
	m_iorderAtg = m_pSite->m_iorderAtg;
	m_iorderTaa = m_pSite->m_iorderTaa;
	m_iorderTag = m_pSite->m_iorderTag;
	m_iorderTga = m_pSite->m_iorderTga;
	m_iorderDon0 = m_pSite->m_iorderDon0;
	m_iorderDon1 = m_pSite->m_iorderDon1;
	m_iorderDon2 = m_pSite->m_iorderDon2;
	m_iorderAcc0 = m_pSite->m_iorderAcc0;
	m_iorderAcc1 = m_pSite->m_iorderAcc1;
	m_iorderAcc2 = m_pSite->m_iorderAcc2;

	m_iorderShortAcc0 = m_pSite->m_iorderShortAcc0;
	m_iorderShortAcc1 = m_pSite->m_iorderShortAcc1;
	m_iorderShortAcc2 = m_pSite->m_iorderShortAcc2;

	m_pModelMatAtg = m_pSite->m_pModelMatAtg;
	m_pModelMatTaa = m_pSite->m_pModelMatTaa;
	m_pModelMatTag = m_pSite->m_pModelMatTag;
	m_pModelMatTga = m_pSite->m_pModelMatTga;
	m_pModelMatDon0 = m_pSite->m_pModelMatDon0;
	m_pModelMatDon1 = m_pSite->m_pModelMatDon1;
	m_pModelMatDon2 = m_pSite->m_pModelMatDon2;
	m_pModelMatAcc0 = m_pSite->m_pModelMatAcc0;
	m_pModelMatAcc1 = m_pSite->m_pModelMatAcc1;
	m_pModelMatAcc2 = m_pSite->m_pModelMatAcc2;

	m_startFlankLeft = m_pSite->m_startFlankLeft;
	m_startFlankRight = m_pSite->m_startFlankRight;
	m_donFlankLeft = m_pSite->m_donFlankLeft;
	m_donFlankRight = m_pSite->m_donFlankRight;
	m_accFlankLeft = m_pSite->m_accFlankLeft;
	m_accFlankRight = m_pSite->m_accFlankRight;
	m_termFlankLeft = m_pSite->m_termFlankLeft;
	m_termFlankRight = m_pSite->m_termFlankRight;

	m_shortAccFlankLeft = m_pSite->m_shortAccFlankLeft;
	m_shortAccFlankRight = m_pSite->m_shortAccFlankRight;
	m_intronTermLen = m_pSite->m_intronTermLen;
	m_intronTermFlankRight = m_pSite->m_intronTermFlankRight;
}

bool cModel::createMkModel()
{
	report(". . . . . . . . . .");
	FileBuf f;
	string strFile;

	report("Creating coding markov models");

	// generate coding sequence models
	m_markovCod_0.clear();
	m_markovCod_1.clear();
	m_markovCod_2.clear();
	m_markovCod_5.clear();

	strFile = m_codingFile;
	f.Load( strFile.c_str() );
	if (!m_markovCod_0.MkMarkov( 0, f.begin ))	{ f.Delete(); return true; }
	if (!m_markovCod_1.MkMarkov( 1, f.begin ))	{ f.Delete(); return true; }
	if (!m_markovCod_2.MkMarkov( 2, f.begin ))	{ f.Delete(); return true; }
	if (!m_markovCod_3.MkMarkov( 3, f.begin ))	{ f.Delete(); return true; }
	if (!m_markovCod_4.MkMarkov( 4, f.begin ))	{ f.Delete(); return true; }
	if (!m_markovCod_5.MkMarkov( 5, f.begin ))	{ f.Delete(); return true; }
	if (!m_markovCod_1n.MkMarkovNorm( m_markovCod_1, m_markovCod_0 ))	{ f.Delete(); return true; }

	//report("Codon usage....");
	//calcCodonUsage();

	return false;
}

bool cModel::createMkModel(char *pchCoding)
{
	// generate coding sequence models
	m_markovCod_0.clear();
	m_markovCod_1.clear();
	m_markovCod_5.clear();

	if (!m_markovCod_0.MkMarkov( 0, pchCoding ))	{ return true; }
	if (!m_markovCod_1.MkMarkov( 1, pchCoding ))	{ return true; }
	if (!m_markovCod_2.MkMarkov( 2, pchCoding ))	{ return true; }
	if (!m_markovCod_3.MkMarkov( 3, pchCoding ))	{ return true; }
	if (!m_markovCod_4.MkMarkov( 4, pchCoding ))	{ return true; }
	if (!m_markovCod_5.MkMarkov( 5, pchCoding ))	{ return true; }
	if (!m_markovCod_1n.MkMarkovNorm( m_markovCod_1, m_markovCod_0 ))	{ return true; }

	return false;
}


bool cModel::deriveMarkovCod_01()
{
	m_markovCod_1.m_mkModel_d1 = new double[16];
	m_markovCod_1.m_mkModel_d2 = new double[16];
	m_markovCod_1.m_mkModel_d3 = new double[16];

	m_markovCod_0.m_mkModel_d1 = new double[4];
	m_markovCod_0.m_mkModel_d2 = new double[4];
	m_markovCod_0.m_mkModel_d3 = new double[4];

	if (m_markovCod_1.m_mkModel_d1==NULL || m_markovCod_1.m_mkModel_d2==NULL || m_markovCod_1.m_mkModel_d3==NULL ||
		m_markovCod_0.m_mkModel_d1==NULL || m_markovCod_0.m_mkModel_d2==NULL || m_markovCod_0.m_mkModel_d3==NULL)
		return true;

	switch (m_order) {
		case 1:
			ReorderRightL( m_markovCod_1.m_mkModel_d1, m_markovCod_1.d1a, m_markovCod_1.m_order, 1 );
			ReorderRightL( m_markovCod_1.m_mkModel_d2, m_markovCod_1.d2a, m_markovCod_1.m_order, 1 );
			ReorderRightL( m_markovCod_1.m_mkModel_d3, m_markovCod_1.d3a, m_markovCod_1.m_order, 1 );
			break;
		case 2:
			ReorderRightL( m_markovCod_1.m_mkModel_d1, m_markovCod_2.d1a, m_markovCod_2.m_order, 1 );
			ReorderRightL( m_markovCod_1.m_mkModel_d2, m_markovCod_2.d2a, m_markovCod_2.m_order, 1 );
			ReorderRightL( m_markovCod_1.m_mkModel_d3, m_markovCod_2.d3a, m_markovCod_2.m_order, 1 );
			break;
		case 3:
			ReorderRightL( m_markovCod_1.m_mkModel_d1, m_markovCod_3.d1a, m_markovCod_3.m_order, 1 );
			ReorderRightL( m_markovCod_1.m_mkModel_d2, m_markovCod_3.d2a, m_markovCod_3.m_order, 1 );
			ReorderRightL( m_markovCod_1.m_mkModel_d3, m_markovCod_3.d3a, m_markovCod_3.m_order, 1 );
			break;
		case 4:
			ReorderRightL( m_markovCod_1.m_mkModel_d1, m_markovCod_4.d1a, m_markovCod_4.m_order, 1 );
			ReorderRightL( m_markovCod_1.m_mkModel_d2, m_markovCod_4.d2a, m_markovCod_4.m_order, 1 );
			ReorderRightL( m_markovCod_1.m_mkModel_d3, m_markovCod_4.d3a, m_markovCod_4.m_order, 1 );
			break;
		case 5:
			ReorderRightL( m_markovCod_1.m_mkModel_d1, m_markovCod_5.d1a, m_markovCod_5.m_order, 1 );
			ReorderRightL( m_markovCod_1.m_mkModel_d2, m_markovCod_5.d2a, m_markovCod_5.m_order, 1 );
			ReorderRightL( m_markovCod_1.m_mkModel_d3, m_markovCod_5.d3a, m_markovCod_5.m_order, 1 );
			break;
		default:
			return true;
	};


	ReorderRight( m_markovCod_0.m_mkModel_d1, m_markovCod_1.m_mkModel_d1, 1, 0 );
	ReorderRight( m_markovCod_0.m_mkModel_d2, m_markovCod_1.m_mkModel_d2, 1, 0 );
	ReorderRight( m_markovCod_0.m_mkModel_d3, m_markovCod_1.m_mkModel_d3, 1, 0 );

	m_markovCod_1.d1a = new double[17]; memcpy( m_markovCod_1.d1a, m_markovCod_1.m_mkModel_d1, 16*sizeof(double) );
	m_markovCod_1.d2a = new double[17]; memcpy( m_markovCod_1.d2a, m_markovCod_1.m_mkModel_d2, 16*sizeof(double) );
	m_markovCod_1.d3a = new double[17]; memcpy( m_markovCod_1.d3a, m_markovCod_1.m_mkModel_d3, 16*sizeof(double) );

	m_markovCod_0.d1a = new double[5]; memcpy( m_markovCod_0.d1a, m_markovCod_0.m_mkModel_d1, 4*sizeof(double) );
	m_markovCod_0.d2a = new double[5]; memcpy( m_markovCod_0.d2a, m_markovCod_0.m_mkModel_d2, 4*sizeof(double) );
	m_markovCod_0.d3a = new double[5]; memcpy( m_markovCod_0.d3a, m_markovCod_0.m_mkModel_d3, 4*sizeof(double) );

	m_markovCod_1.m_order = 1;
	m_markovCod_0.m_order = 0;

	m_markovCod_1.Set( 1, 0.99 );
	m_markovCod_0.Set( 0, 0.99 );

	m_markovCod_1n = m_markovCod_1;

	return false;
}

bool cModel::deriveMarkovNon_01()
{
	m_markovNon_1.m_mkModel = new double[16];
	m_markovNon_0.m_mkModel = new double[4];
	if (m_markovNon_1.m_mkModel==NULL || m_markovNon_0.m_mkModel==NULL)
		return true;

	switch (m_order) {
		case 1:
			ReorderRightL( m_markovNon_1.m_mkModel, m_markovNon_1.na, m_markovNon_1.m_order, 1 );
			break;
		case 2:
			ReorderRightL( m_markovNon_1.m_mkModel, m_markovNon_2.na, m_markovNon_2.m_order, 1 );
			break;
		case 3:
			ReorderRightL( m_markovNon_1.m_mkModel, m_markovNon_3.na, m_markovNon_3.m_order, 1 );
			break;
		case 4:
			ReorderRightL( m_markovNon_1.m_mkModel, m_markovNon_4.na, m_markovNon_4.m_order, 1 );
			break;
		case 5:
			ReorderRightL( m_markovNon_1.m_mkModel, m_markovNon_5.na, m_markovNon_5.m_order, 1 );
			break;
		default:
			return true;
	};

	ReorderRight( m_markovNon_0.m_mkModel, m_markovNon_1.m_mkModel, 1, 0 );

	m_markovNon_1.na = new double[16+1]; memcpy( m_markovNon_1.na, m_markovNon_1.m_mkModel, 16*sizeof(double) );
	m_markovNon_0.na = new double[4+1]; memcpy( m_markovNon_0.na, m_markovNon_0.m_mkModel, 4*sizeof(double) );

	m_markovNon_1.m_order = 1;
	m_markovNon_0.m_order = 0;

	m_markovNon_1.Set( 1, 0.99 );
	m_markovNon_0.Set( 0, 0.99 );

	m_markovNon_1n = m_markovNon_1;

	return false;
}

bool cModel::log2linear( double *pData, int size )
{	// for markov models
	for (int i=0; i<size; ++i)
		pData[i] = exp( pData[i] );

	return false;
}

bool cModel::log2linear( double *pData, int order, int width )
{	// for site models
	int size = ( 4 << 2*order ) * width;

	for (int i=0; i<size; ++i)
		pData[i] = exp( pData[i] );

	return false;
}

bool cModel::rel2abs( double *pData, int order, int width )
{
	int size = ( 4 << 2*order );

	int ipos, j, ix, im;
	double *m0;
	double sum;

	m0 = new double[4];
	memset( m0, 0, 4*sizeof(double) );

	for (ipos=0; ipos<width; ++ipos)
	{
		ix = ipos*width;
		ReorderRight( m0, pData+ix, 1, 0 );
		// normalize m0
		sum = 0.;
		for (j=0; j<4; ++j)
			sum += m0[j];

		for (j=0; j<4; ++j)
			m0[j] /= sum;

		// de-normalize pData
		for (j=0; j<size; ++j)
		{
			im = (int) j / 4;
			pData[ix+j] *= m0[im];
		}
	}

	delete []m0;

	return false;
}

void cModel::invNormLog( double* ptr, int size )
{
	int i;

	for ( i = 0; i < size; ++i )
		ptr[i] = exp( ptr[i] ) / double(size);
};

bool cModel::ln2hist( intDoubleHist& hist, double *pln, int min, int max )
{
	if (pln==NULL)
		return true;

	hist.clear();
	intDouble id;

	for (int i=min; i<max; ++i)
	{
		id.i = i;
		id.y = exp( pln[i] );
		hist.push_back( id );
	}
	id.i = max;
	hist.push_back( id );

	return false;
}

bool cModel::siteSite(double **pval, int order, int width, double *pData)
{
	int height = 4<<(2*order);
	int size = width * height;
	*pval = new double[size];
	if (*pval == NULL)
		return true;		// memory failure

	memcpy( *pval, pData, size*sizeof(double) );

	return false;
}

bool cModel::copyDon(const cModel& model)
{
	int width = 2 + model.m_donFlankLeft + model.m_donFlankRight;

	m_donFlankLeft = model.m_donFlankLeft;
	m_donFlankRight = model.m_donFlankRight;

	if (m_pModelMatDon0 != NULL) { delete []m_pModelMatDon0; m_pModelMatDon0 = NULL; }
	m_iorderDon0 = model.m_iorderDon0;
	if (siteSite(&m_pModelMatDon0, model.m_iorderDon0, width, model.m_pModelMatDon0)) return true;

	if (m_pModelMatDon1 != NULL) { delete []m_pModelMatDon1; m_pModelMatDon1 = NULL; }
	m_iorderDon1 = model.m_iorderDon1;
	if (siteSite(&m_pModelMatDon1, model.m_iorderDon1, width, model.m_pModelMatDon1)) return true;

	if (m_pModelMatDon2 != NULL) { delete []m_pModelMatDon2; m_pModelMatDon2 = NULL; }
	m_iorderDon2 = model.m_iorderDon2;
	if (siteSite(&m_pModelMatDon2, model.m_iorderDon2, width, model.m_pModelMatDon2)) return true;

	return false;	// all good
}

bool cModel::copyAcc(const cModel& model)
{
	int width = 2 + model.m_accFlankLeft + model.m_accFlankRight;

	m_accFlankLeft = model.m_accFlankLeft;
	m_accFlankRight = model.m_accFlankRight;

	if (m_pModelMatAcc0 != NULL) { delete []m_pModelMatAcc0; m_pModelMatAcc0 = NULL; }
	m_iorderAcc0 = model.m_iorderAcc0;
	if (siteSite(&m_pModelMatAcc0, model.m_iorderAcc0, width, model.m_pModelMatAcc0)) return true;

	if (m_pModelMatAcc1 != NULL) { delete []m_pModelMatAcc1; m_pModelMatAcc1 = NULL; }
	m_iorderAcc1 = model.m_iorderAcc1;
	if (siteSite(&m_pModelMatAcc1, model.m_iorderAcc1, width, model.m_pModelMatAcc1)) return true;

	if (m_pModelMatAcc2 != NULL) { delete []m_pModelMatAcc2; m_pModelMatAcc2 = NULL; }
	m_iorderAcc2 = model.m_iorderAcc2;
	if (siteSite(&m_pModelMatAcc2, model.m_iorderAcc2, width, model.m_pModelMatAcc2)) return true;

	return false;	// all good
}

bool cModel::load( string& fname )
{
	if ( loadModel(fname) ) return true;

	if ( getOrder(m_order) ) return true;

	// markov models
	//
	switch (m_order) {
		case 0:
			if ( getMarkovCod(m_markovCod_0) ) return true;
			if ( getMarkovNon(m_markovNon_0) ) return true;
			break;
		case 1:
			if ( getMarkovCod(m_markovCod_1) ) return true;
			if ( getMarkovNon(m_markovNon_1) ) return true;
			break;
		case 2:
			if ( getMarkovCod(m_markovCod_2) ) return true;
			if ( getMarkovNon(m_markovNon_2) ) return true;
			break;
		case 3:
			if ( getMarkovCod(m_markovCod_3) ) return true;
			if ( getMarkovNon(m_markovNon_3) ) return true;
			break;
		case 4:
			if ( getMarkovCod(m_markovCod_4) ) return true;
			if ( getMarkovNon(m_markovNon_4) ) return true;
			break;
		case 5:
			if ( getMarkovCod(m_markovCod_5) ) return true;
			if ( getMarkovNon(m_markovNon_5) ) return true;
			break;
		default:
			return true;
	};


	if ( deriveMarkovCod_01() ) return true;		// derive 0 and 1st order models
	if ( deriveMarkovNon_01() ) return true;
	if ( !m_markovCod_1n.SetMkFromAbs() ) return true;
	if ( !m_markovNon_1n.SetMkFromAbs() ) return true;

	// sites
	//
	int nSiteWidth, nMargin;
	string hdr;
	// ATG
	hdr = "INI";
	if ( getSite( &m_pModelMatAtg, m_iorderAtg, nSiteWidth, nMargin, hdr ) ) return true;
	m_startFlankLeft = nMargin;
	m_startFlankRight = nSiteWidth - nMargin - 3;
	log2linear( m_pModelMatAtg, m_iorderAtg, nSiteWidth );

	// TAA
	hdr = "TERM_TAA";
	if ( getSite( &m_pModelMatTaa, m_iorderTaa, nSiteWidth, nMargin, hdr ) ) return true;
	m_termFlankLeft = nMargin;
	m_termFlankRight = nSiteWidth - nMargin - 3;
	log2linear( m_pModelMatTaa, m_iorderTaa, nSiteWidth );
	// TAG
	hdr = "TERM_TAG";
	if ( getSite( &m_pModelMatTag, m_iorderTag, nSiteWidth, nMargin, hdr ) ) return true;
	log2linear( m_pModelMatTag, m_iorderTag, nSiteWidth );
	// TGA
	hdr = "TERM_TGA";
	if ( getSite( &m_pModelMatTga, m_iorderTga, nSiteWidth, nMargin, hdr ) ) return true;
	log2linear( m_pModelMatTga, m_iorderTga, nSiteWidth );

	// DON 0
	hdr = "DONOR_0";
	if ( getSite( &m_pModelMatDon0, m_iorderDon0, nSiteWidth, nMargin, hdr ) ) return true;
	m_donFlankLeft = nMargin;
	m_donFlankRight = nSiteWidth - nMargin - 2;
	log2linear( m_pModelMatDon0, m_iorderDon0, nSiteWidth );
	// DON 1
	hdr = "DONOR_1";
	if ( getSite( &m_pModelMatDon1, m_iorderDon1, nSiteWidth, nMargin, hdr ) ) return true;
	log2linear( m_pModelMatDon1, m_iorderDon1, nSiteWidth );
	// DON 2
	hdr = "DONOR_2";
	if ( getSite( &m_pModelMatDon2, m_iorderDon2, nSiteWidth, nMargin, hdr ) ) return true;
	log2linear( m_pModelMatDon2, m_iorderDon2, nSiteWidth );

	// ACC 0
	hdr = "ACCEPTOR_0";
	if ( getSite( &m_pModelMatAcc0, m_iorderAcc0, nSiteWidth, nMargin, hdr ) ) return true;
	m_accFlankLeft = nMargin;
	m_accFlankRight = nSiteWidth - nMargin - 2;
	log2linear( m_pModelMatAcc0, m_iorderAcc0, nSiteWidth );
	// ACC 1
	hdr = "ACCEPTOR_1";
	if ( getSite( &m_pModelMatAcc1, m_iorderAcc1, nSiteWidth, nMargin, hdr ) ) return true;
	log2linear( m_pModelMatAcc1, m_iorderAcc1, nSiteWidth );
	// ACC 2
	hdr = "ACCEPTOR_2";
	if ( getSite( &m_pModelMatAcc2, m_iorderAcc2, nSiteWidth, nMargin, hdr ) ) return true;
	log2linear( m_pModelMatAcc2, m_iorderAcc2, nSiteWidth );

	m_pSite->m_donFlankLeft = m_donFlankLeft;
	m_pSite->m_donFlankRight = m_donFlankRight;
	m_pSite->m_accFlankLeft = m_accFlankLeft;
	m_pSite->m_accFlankRight = m_accFlankRight;
	m_pSite->m_startFlankLeft = m_startFlankLeft;
	m_pSite->m_startFlankRight = m_startFlankRight;
	m_pSite->m_termFlankLeft = m_termFlankLeft;
	m_pSite->m_termFlankRight = m_termFlankRight;

	// length histograms
	string strSingle = "SINGLE";
	string strExon = "EXON";
	string strIni = "INITIAL";
	string strTer = "TERMINAL";
	string strIntron = "INTRON";
	string strInterg = "INTERGENIC";

	m_pSingleExon = m_pExon = m_pIniExon = m_pTerExon = m_pIntron = m_pInterg = NULL;

	if (getLength( &m_pSingleExon, &m_lenMinSingleExon, &m_lenMaxSingleExon, strSingle )) return true;
	if (getLength( &m_pExon, &m_lenMinExon, &m_lenMaxExon, strExon )) return true;
	if (getLength( &m_pIniExon, &m_lenMinIniExon, &m_lenMaxIniExon, strIni )) return true;
	if (getLength( &m_pTerExon, &m_lenMinTerExon, &m_lenMaxIniExon, strTer )) return true;
	if (getLength( &m_pIntron, &m_lenMinIntron, &m_lenMaxIntron, strIntron )) return true;
	if (getLength( &m_pInterg, &m_lenMinInterg, &m_lenMaxInterg, strInterg )) return true;

	//intDoubleHist m_distrExonSingLen, m_distrExonIniLen, m_distrExonTermLen;
	//intDoubleHist m_distrExonIntLen;
	//intDoubleHist m_distrIntronLen;
	//intDoubleHist m_distrIntergenicLen;
	ln2hist( m_distrExonSingLen,	m_pSingleExon,	m_lenMinSingleExon,	m_lenMaxSingleExon );
	ln2hist( m_distrExonIniLen,		m_pIniExon,		m_lenMinIniExon,	m_lenMaxIniExon );
	ln2hist( m_distrExonTermLen,	m_pTerExon,		m_lenMinTerExon,	m_lenMaxIniExon );
	ln2hist( m_distrExonIntLen,		m_pExon,		m_lenMinExon,		m_lenMaxExon );
	ln2hist( m_distrIntronLen,		m_pIntron,		m_lenMinIntron,		m_lenMaxIntron );
	ln2hist( m_distrIntergenicLen,	m_pInterg,		m_lenMinInterg,		m_lenMaxInterg );

	m_bEsPlusIntronLD = !intronLengthConstant();

	double pIni[3], pTer[3], pInt[9];
	double lf1, lf2;

	if ( getPhaseProbs( pIni, pTer, pInt ) ) return true;
	if ( getTransitionProbs( lf1, lf2 ) ) return true;

	m_lfToSingle = exp( lf1 );
	m_probToInternalExon = exp( lf2 );

	invNormLog( pIni, 3 );
	m_probPhase0 = pIni[0];
	m_probPhase1 = pIni[1];
	m_probPhase2 = pIni[2];

	return false;
}


//
// load model into globalModel_2
//

bool cModel::load_2( string& fname )
{
	// load model, without re-normalization... good for heuristic models
	//

	if ( loadModel_2(fname) ) return true;

	if ( getOrder_2(m_order) ) return true;

	// markov models
	//
	switch (m_order) {
		case 0:
			if ( getMarkovCod_2(m_markovCod_0) ) return true;
			if ( getMarkovNon_2(m_markovNon_0) ) return true;
			break;
		case 1:
			if ( getMarkovCod_2(m_markovCod_1) ) return true;
			if ( getMarkovNon_2(m_markovNon_1) ) return true;
			break;
		case 2:
			if ( getMarkovCod_2(m_markovCod_2) ) return true;
			if ( getMarkovNon_2(m_markovNon_2) ) return true;
			break;
		case 3:
			if ( getMarkovCod_2(m_markovCod_3) ) return true;
			if ( getMarkovNon_2(m_markovNon_3) ) return true;
			break;
		case 4:
			if ( getMarkovCod_2(m_markovCod_4) ) return true;
			if ( getMarkovNon_2(m_markovNon_4) ) return true;
			break;
		case 5:
			if ( getMarkovCod_2(m_markovCod_5) ) return true;
			if ( getMarkovNon_2(m_markovNon_5) ) return true;
			break;
		default:
			return true;
	};

	//if ( getMarkovNorm_2(m_markovCod_1n,m_markovNon_1n) ) return true;

	if ( deriveMarkovCod_01() ) return true;		// derive 0 and 1st order models
	if ( deriveMarkovNon_01() ) return true;
	//if (!m_markovCod_1n.MkMarkovNorm( m_markovCod_1, m_markovCod_0 )) return true;
	if ( !m_markovCod_1n.SetMkFromAbs() ) return true;
	if ( !m_markovNon_1n.SetMkFromAbs() ) return true;

	// sites
	//
	// ATG
	int nSiteWidth, nMargin;
	string hdr;
	hdr = "INI";
	if ( getSite_2( &m_pModelMatAtg, m_iorderAtg, nSiteWidth, nMargin, hdr ) ) return true;
	m_startFlankLeft = nMargin;
	m_startFlankRight = nSiteWidth - nMargin - 3;
	log2linear( m_pModelMatAtg, m_iorderAtg, nSiteWidth );

	// TAA
	hdr = "TERM_TAA";
	if ( getSite_2( &m_pModelMatTaa, m_iorderTaa, nSiteWidth, nMargin, hdr ) ) return true;
	m_termFlankLeft = nMargin;
	m_termFlankRight = nSiteWidth - nMargin - 3;
	log2linear( m_pModelMatTaa, m_iorderTaa, nSiteWidth );
	// TAG
	hdr = "TERM_TAG";
	if ( getSite_2( &m_pModelMatTag, m_iorderTag, nSiteWidth, nMargin, hdr ) ) return true;
	log2linear( m_pModelMatTag, m_iorderTag, nSiteWidth );
	// TGA
	hdr = "TERM_TGA";
	if ( getSite_2( &m_pModelMatTga, m_iorderTga, nSiteWidth, nMargin, hdr ) ) return true;
	log2linear( m_pModelMatTga, m_iorderTga, nSiteWidth );

	// DON 0
	hdr = "DONOR_0";
	if ( getSite_2( &m_pModelMatDon0, m_iorderDon0, nSiteWidth, nMargin, hdr ) ) return true;
	m_donFlankLeft = nMargin;
	m_donFlankRight = nSiteWidth - nMargin - 2;
	m_pSite->m_donFlankLeft = m_donFlankLeft;
	m_pSite->m_donFlankRight = m_donFlankRight;
	log2linear( m_pModelMatDon0, m_iorderDon0, nSiteWidth );
	// DON 1
	hdr = "DONOR_1";
	if ( getSite_2( &m_pModelMatDon1, m_iorderDon1, nSiteWidth, nMargin, hdr ) ) return true;
	log2linear( m_pModelMatDon1, m_iorderDon1, nSiteWidth );
	// DON 2
	hdr = "DONOR_2";
	if ( getSite_2( &m_pModelMatDon2, m_iorderDon2, nSiteWidth, nMargin, hdr ) ) return true;
	log2linear( m_pModelMatDon2, m_iorderDon2, nSiteWidth );

	// ACC 0
	hdr = "ACCEPTOR_0";
	if ( getSite_2( &m_pModelMatAcc0, m_iorderAcc0, nSiteWidth, nMargin, hdr ) ) return true;
	m_accFlankLeft = nMargin;
	m_accFlankRight = nSiteWidth - nMargin - 2;
	m_pSite->m_accFlankLeft = m_accFlankLeft;
	m_pSite->m_accFlankRight = m_accFlankRight;
	log2linear( m_pModelMatAcc0, m_iorderAcc0, nSiteWidth );
	// ACC 1
	hdr = "ACCEPTOR_1";
	if ( getSite_2( &m_pModelMatAcc1, m_iorderAcc1, nSiteWidth, nMargin, hdr ) ) return true;
	log2linear( m_pModelMatAcc1, m_iorderAcc1, nSiteWidth );
	// ACC 2
	hdr = "ACCEPTOR_2";
	if ( getSite_2( &m_pModelMatAcc2, m_iorderAcc2, nSiteWidth, nMargin, hdr ) ) return true;
	log2linear( m_pModelMatAcc2, m_iorderAcc2, nSiteWidth );


	m_pSite->m_donFlankLeft = m_donFlankLeft;
	m_pSite->m_donFlankRight = m_donFlankRight;
	m_pSite->m_accFlankLeft = m_accFlankLeft;
	m_pSite->m_accFlankRight = m_accFlankRight;
	m_pSite->m_startFlankLeft = m_startFlankLeft;
	m_pSite->m_startFlankRight = m_startFlankRight;
	m_pSite->m_termFlankLeft = m_termFlankLeft;
	m_pSite->m_termFlankRight = m_termFlankRight;

	// length histograms
	string strSingle = "SINGLE";
	string strExon = "EXON";
	string strIni = "INITIAL";
	string strTer = "TERMINAL";
	string strIntron = "INTRON";
	string strInterg = "INTERGENIC";

	m_pSingleExon = m_pExon = m_pIniExon = m_pTerExon = m_pIntron = m_pInterg = NULL;

	if (getLength_2( &m_pSingleExon, &m_lenMinSingleExon, &m_lenMaxSingleExon, strSingle )) return true;
	if (getLength_2( &m_pExon, &m_lenMinExon, &m_lenMaxExon, strExon )) return true;
	if (getLength_2( &m_pIniExon, &m_lenMinIniExon, &m_lenMaxIniExon, strIni )) return true;
	if (getLength_2( &m_pTerExon, &m_lenMinTerExon, &m_lenMaxIniExon, strTer )) return true;
	if (getLength_2( &m_pIntron, &m_lenMinIntron, &m_lenMaxIntron, strIntron )) return true;
	if (getLength_2( &m_pInterg, &m_lenMinInterg, &m_lenMaxInterg, strInterg )) return true;

	//ln2hist( m_distrExonSingLen,	m_pSingleExon,	m_lenMinSingleExon,	m_lenMaxSingleExon );
	//ln2hist( m_distrExonIniLen,		m_pIniExon,		m_lenMinIniExon,	m_lenMaxIniExon );
	//ln2hist( m_distrExonTermLen,	m_pTerExon,		m_lenMinTerExon,	m_lenMaxIniExon );
	//ln2hist( m_distrExonIntLen,		m_pExon,		m_lenMinExon,		m_lenMaxExon );
	ln2hist( m_distrIntronLen,		m_pIntron,		m_lenMinIntron,		m_lenMaxIntron );
	//ln2hist( m_distrIntergenicLen,	m_pInterg,		m_lenMinInterg,		m_lenMaxInterg );

	double pIni[3], pTer[3], pInt[9];
	double lf1, lf2;

	if ( getPhaseProbs( pIni, pTer, pInt ) ) return true;
	if ( getTransitionProbs( lf1, lf2 ) ) return true;

	m_lfToSingle = exp( lf1 );
	m_probToInternalExon = exp( lf2 );

	invNormLog( pIni, 3 );
	m_probPhase0 = pIni[0];
	m_probPhase1 = pIni[1];
	m_probPhase2 = pIni[2];

	return false;
}

void cModel::normalizeDonAcc()
{
	int msize, nCols;
	double *pModelOld(NULL);

	// Don 0
	//
	msize = 4<<(2*m_iorderDon0);
	nCols = m_pSite->getDonSiteLength();

	if (pModelOld!=NULL)
		delete []pModelOld;

	pModelOld = new double[msize*nCols];
	memcpy( pModelOld, m_pModelMatDon0, msize*nCols*sizeof(double) );
	if (m_iorderDon0==0)
		normalizeModelDon( m_pModelMatDon0, pModelOld, 0, m_iorderDon0, &m_markovCod_0, &m_markovNon_0 );
	else if (m_iorderDon0==1)
		normalizeModelDon( m_pModelMatDon0, pModelOld, 0, m_iorderDon0, &m_markovCod_1n, &m_markovNon_1n );

	if (pModelOld!=NULL) {
		delete []pModelOld;
		pModelOld = NULL;
	}
	//rel2abs( m_pModelMatDon0, m_iorderDon0, nSiteWidth );

	// Don 1
	//
	msize = 4<<(2*m_iorderDon1);
	nCols = m_pSite->getDonSiteLength();

	if (pModelOld!=NULL)
		delete []pModelOld;

	pModelOld = new double[msize*nCols];
	memcpy( pModelOld, m_pModelMatDon1, msize*nCols*sizeof(double) );
	if (m_iorderDon1==0)
		normalizeModelDon( m_pModelMatDon1, pModelOld, 1, m_iorderDon1, &m_markovCod_0, &m_markovNon_0 );
	else if (m_iorderDon1==1)
		normalizeModelDon( m_pModelMatDon1, pModelOld, 1, m_iorderDon1, &m_markovCod_1n, &m_markovNon_1n );

	if (pModelOld!=NULL) {
		delete []pModelOld;
		pModelOld = NULL;
	}
	//rel2abs( m_pModelMatDon1, m_iorderDon1, nSiteWidth );

	// Don 2
	//
	msize = 4<<(2*m_iorderDon2);
	nCols = m_pSite->getDonSiteLength();

	if (pModelOld!=NULL)
		delete []pModelOld;

	pModelOld = new double[msize*nCols];
	memcpy( pModelOld, m_pModelMatDon2, msize*nCols*sizeof(double) );
	if (m_iorderDon2==0)
		normalizeModelDon( m_pModelMatDon2, pModelOld, 2, m_iorderDon2, &m_markovCod_0, &m_markovNon_0 );
	else if (m_iorderDon2==1)
		normalizeModelDon( m_pModelMatDon2, pModelOld, 2, m_iorderDon2, &m_markovCod_1n, &m_markovNon_1n );

	if (pModelOld!=NULL) {
		delete []pModelOld;
		pModelOld = NULL;
	}
	//rel2abs( m_pModelMatDon2, m_iorderDon2, nSiteWidth );

	// Acc 0
	//
	msize = 4<<(2*m_iorderAcc0);
	nCols = m_pSite->getAccSiteLength();

	if (pModelOld!=NULL)
		delete []pModelOld;

	pModelOld = new double[msize*nCols];
	memcpy( pModelOld, m_pModelMatAcc0, msize*nCols*sizeof(double) );
	if (m_iorderAcc0==0)
		normalizeModelAcc( m_pModelMatAcc0, pModelOld, 0, m_iorderAcc0, &m_markovCod_0, &m_markovNon_0 );
	else if (m_iorderAcc0==1)
		normalizeModelAcc( m_pModelMatAcc0, pModelOld, 0, m_iorderAcc0, &m_markovCod_1n, &m_markovNon_1n );

	if (pModelOld!=NULL) {
		delete []pModelOld;
		pModelOld = NULL;
	}
	//rel2abs( m_pModelMatAcc0, m_iorderAcc0, nSiteWidth );

	// Acc 1
	//
	msize = 4<<(2*m_iorderAcc1);
	nCols = m_pSite->getAccSiteLength();

	if (pModelOld!=NULL)
		delete []pModelOld;

	pModelOld = new double[msize*nCols];
	memcpy( pModelOld, m_pModelMatAcc1, msize*nCols*sizeof(double) );
	if (m_iorderAcc1==0)
		normalizeModelAcc( m_pModelMatAcc1, pModelOld, 1, m_iorderAcc1, &m_markovCod_0, &m_markovNon_0 );
	else if (m_iorderAcc1==1)
		normalizeModelAcc( m_pModelMatAcc1, pModelOld, 1, m_iorderAcc1, &m_markovCod_1n, &m_markovNon_1n );

	if (pModelOld!=NULL) {
		delete []pModelOld;
		pModelOld = NULL;
	}
	//rel2abs( m_pModelMatAcc1, m_iorderAcc1, nSiteWidth );

	// Acc 2
	//
	msize = 4<<(2*m_iorderAcc2);
	nCols = m_pSite->getAccSiteLength();

	if (pModelOld!=NULL)
		delete []pModelOld;

	pModelOld = new double[msize*nCols];
	memcpy( pModelOld, m_pModelMatAcc2, msize*nCols*sizeof(double) );
	if (m_iorderAcc2==0)
		normalizeModelAcc( m_pModelMatAcc2, pModelOld, 2, m_iorderAcc2, &m_markovCod_0, &m_markovNon_0 );
	else if (m_iorderAcc2==1)
		normalizeModelAcc( m_pModelMatAcc2, pModelOld, 2, m_iorderAcc2, &m_markovCod_1n, &m_markovNon_1n );

	if (pModelOld!=NULL) {
		delete []pModelOld;
		pModelOld = NULL;
	}
	//rel2abs( m_pModelMatAcc2, m_iorderAcc2, nSiteWidth );
}

void cModel::unNormalizeDonAcc()
{
	int msize, nCols;
	double *pModelOld(NULL);

	// Don 0
	//
	msize = 4<<(2*m_iorderDon0);
	nCols = m_pSite->getDonSiteLength();

	if (pModelOld!=NULL)
		delete []pModelOld;

	pModelOld = new double[msize*nCols];
	memcpy( pModelOld, m_pModelMatDon0, msize*nCols*sizeof(double) );
	if (m_iorderDon0==0)
		unNormalizeModelDon( m_pModelMatDon0, pModelOld, 0, m_iorderDon0, &m_markovCod_0, &m_markovNon_0 );
	else if (m_iorderDon0==1)
		unNormalizeModelDon( m_pModelMatDon0, pModelOld, 0, m_iorderDon0, &m_markovCod_1n, &m_markovNon_1n );

	if (pModelOld!=NULL) {
		delete []pModelOld;
		pModelOld = NULL;
	}
	//rel2abs( m_pModelMatDon0, m_iorderDon0, nSiteWidth );

	// Don 1
	//
	msize = 4<<(2*m_iorderDon1);
	nCols = m_pSite->getDonSiteLength();

	if (pModelOld!=NULL)
		delete []pModelOld;

	pModelOld = new double[msize*nCols];
	memcpy( pModelOld, m_pModelMatDon1, msize*nCols*sizeof(double) );
	if (m_iorderDon1==0)
		unNormalizeModelDon( m_pModelMatDon1, pModelOld, 1, m_iorderDon1, &m_markovCod_0, &m_markovNon_0 );
	else if (m_iorderDon1==1)
		unNormalizeModelDon( m_pModelMatDon1, pModelOld, 1, m_iorderDon1, &m_markovCod_1n, &m_markovNon_1n );

	if (pModelOld!=NULL) {
		delete []pModelOld;
		pModelOld = NULL;
	}
	//rel2abs( m_pModelMatDon1, m_iorderDon1, nSiteWidth );

	// Don 2
	//
	msize = 4<<(2*m_iorderDon2);
	nCols = m_pSite->getDonSiteLength();

	if (pModelOld!=NULL)
		delete []pModelOld;

	pModelOld = new double[msize*nCols];
	memcpy( pModelOld, m_pModelMatDon2, msize*nCols*sizeof(double) );
	if (m_iorderDon2==0)
		unNormalizeModelDon( m_pModelMatDon2, pModelOld, 2, m_iorderDon2, &m_markovCod_0, &m_markovNon_0 );
	else if (m_iorderDon2==1)
		unNormalizeModelDon( m_pModelMatDon2, pModelOld, 2, m_iorderDon2, &m_markovCod_1n, &m_markovNon_1n );

	if (pModelOld!=NULL) {
		delete []pModelOld;
		pModelOld = NULL;
	}
	//rel2abs( m_pModelMatDon2, m_iorderDon2, nSiteWidth );

	// Acc 0
	//
	msize = 4<<(2*m_iorderAcc0);
	nCols = m_pSite->getAccSiteLength();

	if (pModelOld!=NULL)
		delete []pModelOld;

	pModelOld = new double[msize*nCols];
	memcpy( pModelOld, m_pModelMatAcc0, msize*nCols*sizeof(double) );
	if (m_iorderAcc0==0)
		unNormalizeModelAcc( m_pModelMatAcc0, pModelOld, 0, m_iorderAcc0, &m_markovCod_0, &m_markovNon_0 );
	else if (m_iorderAcc0==1)
		unNormalizeModelAcc( m_pModelMatAcc0, pModelOld, 0, m_iorderAcc0, &m_markovCod_1n, &m_markovNon_1n );

	if (pModelOld!=NULL) {
		delete []pModelOld;
		pModelOld = NULL;
	}
	//rel2abs( m_pModelMatAcc0, m_iorderAcc0, nSiteWidth );

	// Acc 1
	//
	msize = 4<<(2*m_iorderAcc1);
	nCols = m_pSite->getAccSiteLength();

	if (pModelOld!=NULL)
		delete []pModelOld;

	pModelOld = new double[msize*nCols];
	memcpy( pModelOld, m_pModelMatAcc1, msize*nCols*sizeof(double) );
	if (m_iorderAcc1==0)
		unNormalizeModelAcc( m_pModelMatAcc1, pModelOld, 1, m_iorderAcc1, &m_markovCod_0, &m_markovNon_0 );
	else if (m_iorderAcc1==1)
		unNormalizeModelAcc( m_pModelMatAcc1, pModelOld, 1, m_iorderAcc1, &m_markovCod_1n, &m_markovNon_1n );

	if (pModelOld!=NULL) {
		delete []pModelOld;
		pModelOld = NULL;
	}
	//rel2abs( m_pModelMatAcc1, m_iorderAcc1, nSiteWidth );

	// Acc 2
	//
	msize = 4<<(2*m_iorderAcc2);
	nCols = m_pSite->getAccSiteLength();

	if (pModelOld!=NULL)
		delete []pModelOld;

	pModelOld = new double[msize*nCols];
	memcpy( pModelOld, m_pModelMatAcc2, msize*nCols*sizeof(double) );
	if (m_iorderAcc2==0)
		unNormalizeModelAcc( m_pModelMatAcc2, pModelOld, 2, m_iorderAcc2, &m_markovCod_0, &m_markovNon_0 );
	else if (m_iorderAcc2==1)
		unNormalizeModelAcc( m_pModelMatAcc2, pModelOld, 2, m_iorderAcc2, &m_markovCod_1n, &m_markovNon_1n );

	if (pModelOld!=NULL) {
		delete []pModelOld;
		pModelOld = NULL;
	}
	//rel2abs( m_pModelMatAcc2, m_iorderAcc2, nSiteWidth );
}

void cModel::scoreMarkovModels()
{
	vector<double> codingScores, intronScores, nonCodingScores;
	vector<string> codingSeq, intronSeq, nonCodingSeq, codingMrnaIds, intronMrnaIds, intergMrnaIds;
	vector<int> codingMarks_5, codingMarks_3, intronMarks_5, intronMarks_3, nonCodingMarks_5, nonCodingMarks_3;
	vector<int> codingStarts, codingStops, intronStarts, intronStops, ncStarts, ncStops;
	vector<char> codingStrands, intronStrands, ncStrands;		// ncStrands is a dummy here

	FileBuf f;

	// coding sequence
	f.Load( m_codingFile.c_str() );
	splitSeq( codingSeq, codingMrnaIds, codingMarks_5, codingMarks_3, codingStarts, codingStops, codingStrands, f.begin );
	f.Delete();

	// intron sequence
	f.Load( m_intronFile.c_str() );
	splitSeq( intronSeq, intronMrnaIds, intronMarks_5, intronMarks_3, intronStarts, intronStops, intronStrands, f.begin );
	f.Delete();

	// non-coding sequence
	f.Load( m_nonCodingFile.c_str() );
	splitSeq( nonCodingSeq, intergMrnaIds, nonCodingMarks_5, nonCodingMarks_3, ncStarts, ncStops, ncStrands, f.begin );
	f.Delete();

	// scoring
	//
	double logScore, lfLen;
	char alphabet[256];
	setAlphabet( alphabet );
	int i;

	// coding scores
	for (i=0; i<(int)codingSeq.size(); ++i)
	{
		scoreMarkov( logScore, m_markovCod_5.d1, 5, codingSeq[i].c_str(), alphabet );
		lfLen = (double) codingSeq[i].length();
		//codingScores.push_back( logScore/(lfLen-5.) );
		codingScores.push_back( logScore );
	}

	// intron scores
	for (i=0; i<(int)intronSeq.size(); ++i)
	{
		scoreMarkov( logScore, m_markovIntr_5.d, 5, intronSeq[i].c_str(), alphabet );
		lfLen = (double) intronSeq[i].length();
		//nonCodingScores.push_back( logScore/(lfLen-5.) );
		intronScores.push_back( logScore );
	}

	// non-coding scores
	for (i=0; i<(int)nonCodingSeq.size(); ++i)
	{
		scoreMarkov( logScore, m_markovNon_5.n, 5, nonCodingSeq[i].c_str(), alphabet );
		lfLen = (double) nonCodingSeq[i].length();
		//nonCodingScores.push_back( logScore/(lfLen-5.) );
		nonCodingScores.push_back( logScore );
	}

	// record results
	FILE *pF_cod_score = fopen( "cod_score.txt", "w" );
	FILE *pF_intron_score = fopen( "intron_score.txt", "w" );
	FILE *pF_non_score = fopen( "non_score.txt", "w" );

	// coding
	fprintf(pF_cod_score, "#feat\tID\tstart\tstop\tscore\tstrand\tmark5\tmark3\n");
	for (i=0; i<(int)codingScores.size(); ++i)
	{
		fprintf( pF_cod_score, "cds\t%s\t%d\t%d\t%f\t%c\t%d\t%d\n", codingMrnaIds[i].c_str(), codingStarts[i], codingStops[i], codingScores[i], codingStrands[i], codingMarks_5[i], codingMarks_3[i] );
	}

	// intron
	fprintf(pF_intron_score, "#feat\tID\tstart\tstop\tscore\tstrand\tmark5\tmark3\n");
	for (i=0; i<(int)intronScores.size(); ++i)
	{
		fprintf( pF_intron_score, "intron\t%s\t%d\t%d\t%f\t%c\t%d\t%d\n", intronMrnaIds[i].c_str(), intronStarts[i], intronStops[i], intronScores[i], intronStrands[i], intronMarks_5[i], intronMarks_3[i] );
	}

	// intergenic
	fprintf(pF_non_score, "#feat\tID\tstart\tstop\tscore\tstrand\tmark5\tmark3\n");
	for (i=0; i<(int)nonCodingScores.size(); ++i)
	{
		fprintf( pF_non_score, "interg\t%s\t%d\t%d\t%f\t%c\t%d\t%d\n", intergMrnaIds[i].c_str(), ncStarts[i], ncStops[i], nonCodingScores[i], ncStrands[i], nonCodingMarks_5[i], nonCodingMarks_3[i] );
	}

	fclose( pF_cod_score );
	fclose( pF_intron_score );
	fclose( pF_non_score );
}

void cModel::scoreMarkovModelsTrans()
{
	vector<double> intronScores, nonCodingScores;
	vector<string> intronSeq, nonCodingSeq, intronMrnaIds, intergMrnaIds;
	vector<int> intronMarks_5, intronMarks_3, nonCodingMarks_5, nonCodingMarks_3;
	vector<int> intronStarts, intronStops, ncStarts, ncStops;
	vector<char> intronStrands, ncStrands;		// ncStrands is a dummy here

	FileBuf f;

	// intron sequence
	f.Load( m_intronFile.c_str() );
	splitSeq( intronSeq, intronMrnaIds, intronMarks_5, intronMarks_3, intronStarts, intronStops, intronStrands, f.begin );
	f.Delete();

	// non-coding sequence
	f.Load( m_nonCodingFile.c_str() );
	splitSeq( nonCodingSeq, intergMrnaIds, nonCodingMarks_5, nonCodingMarks_3, ncStarts, ncStops, ncStrands, f.begin );
	f.Delete();

	// scoring
	//
	double logScore, lfLen;
	char alphabet[256];
	cMatrixMod tmpMat;
	tmpMat.SetAlphabet( alphabet );
	int i;

	// intron scores
	for (i=0; i<(int)intronSeq.size(); ++i)
	{
		scoreMarkov( logScore, m_markovIntr_5.d, 5, intronSeq[i].c_str(), alphabet );
		lfLen = (double) intronSeq[i].length();
		//intronScores.push_back( logScore/(lfLen-5.) );
		intronScores.push_back( logScore );
	}

	// non-coding scores
	for (i=0; i<(int)nonCodingSeq.size(); ++i)
	{
		scoreMarkov( logScore, m_markovNon_5.n, 5, nonCodingSeq[i].c_str(), alphabet );
		lfLen = (double) nonCodingSeq[i].length();
		//nonCodingScores.push_back( logScore/(lfLen-5.) );
		nonCodingScores.push_back( logScore );
	}

	// record results
	FILE *pF_non_score = fopen( "non_score.txt", "w" );
	FILE *pF_intron_score = fopen( "intron_score.txt", "w" );

	fprintf(pF_non_score, "#feat\tID\tstart\tstop\tscore\tstrand\tmark5\tmark3\n");
	for (i=0; i<(int)nonCodingScores.size(); ++i)
	{
		fprintf( pF_non_score, "interg\t%s\t%d\t%d\t%f\t.\t%d\t%d\n", intergMrnaIds[i].c_str(), ncStarts[i], ncStops[i], nonCodingScores[i], nonCodingMarks_5[i], nonCodingMarks_3[i] );
	}

	fprintf(pF_intron_score, "#feat\tID\tstart\tstop\tscore\tstrand\tmark5\tmark3\n");
	for (i=0; i<(int)intronScores.size(); ++i)
	{
		fprintf( pF_intron_score, "intron\t%s\t%d\t%d\t%f\t%c\t%d\t%d\n", intronMrnaIds[i].c_str(), intronStarts[i], intronStops[i], intronScores[i], intronStrands[i], intronMarks_5[i], intronMarks_3[i] );
	}

	fclose( pF_non_score );
	fclose( pF_intron_score );
}

void cModel::splitSeq( vector<string>& seqs, vector<string>& mrnaIds, vector<int>& iMark_5s, vector<int>& iMark_3s, vector<int>& starts, vector<int>& stops, vector<char>& strands, char *buf )
{
	string seq, str_nums, s;
	int iMark0, iMark1, istart, istop, ifeature;
	char *pch, pchMrna[128], chStrand;
	stPos ix0;

	seqs.clear();
	iMark_5s.clear();
	iMark_3s.clear();
	starts.clear();
	stops.clear();
	strands.clear();

	pch = strtok( buf, "\n" );
	while (pch != NULL)
	{
		seq = pch;
		ix0 = seq.find_first_of( "\t" );
		if (ix0!=string::npos)
		{
			str_nums = seq.substr(ix0);
			if ( sscanf(str_nums.c_str(), "%s %d %d %d %d %c %d", pchMrna, &iMark0, &iMark1, &istart, &istop, &chStrand, &ifeature) != 7 )
				return;		// error
		}
		else
		{
			iMark0 = iMark1 = -1;
		}

		seqs.push_back( seq );
		mrnaIds.push_back( pchMrna );
		iMark_5s.push_back( iMark0 );
		iMark_3s.push_back( iMark1 );
		starts.push_back( istart );
		stops.push_back( istop );
		strands.push_back( chStrand );

		pch = strtok( NULL, "\n" );
	}
}

void cModel::splitSeq( vector<string>& seqs, vector<string>& idseq, vector<string>& ids, vector<int>& ibegs, vector<int>& iends, vector<int>& iJcnTypes, vector<char>& strands, char *buf )
{
	string seq, strId, str_nums, s;
	char *pch, pchId[128], pchSeq[128], chStrand;
	size_t ix0;
	int ibeg, iend, ijcntype;

	seqs.clear();
	ids.clear();
	strands.clear();
	iJcnTypes.clear();

	pch = strtok( buf, "\n" );
	while (pch != NULL)
	{
		seq = pch;
		ix0 = seq.find_first_of( "\t" );
		if (ix0!=string::npos)
		{
			str_nums = seq.substr(ix0);
			if ( sscanf(str_nums.c_str(), "%s %s %d %d %d %c", pchSeq, pchId, &ibeg, &iend, &ijcntype, &chStrand) != 6 )
				return;		// error
		}

		strId = pchId;
		seqs.push_back( seq );
		ids.push_back( strId );
		idseq.push_back( pchSeq );
		ibegs.push_back( ibeg );
		iends.push_back( iend );
		iJcnTypes.push_back( ijcntype );
		strands.push_back( chStrand );

		pch = strtok( NULL, "\n" );
	}
}

void cModel::splitSeq( vector<string>& seqs, vector<string>& idseq, vector<string>& ids, vector<int>& ibegs, vector<int>& iends, vector<int>& iJcnTypes,
					  vector<char>& strands, vector<int>& iFrames, vector<int>& iMark_5s, vector<int>& iMark_3s, vector<int>& iMark_Is,
					  vector<int>& nIntrons, vector<int>& indxIntron, vector<int>& dxAtg, vector<int>& dxStp, char *buf )
{
	string seq, strId, str_nums, s;
	char *pch, pchId[128], pchSeq[128], chStrand;
	size_t ix0;
	int ibeg, iend, ijcntype, iframe, imark5, imark3, imarkI;
	int nIntronsInGene , ixIntron, posAtg, posStp;

	seqs.clear();
	ids.clear();
	strands.clear();
	iJcnTypes.clear();
	iFrames.clear();
	iMark_5s.clear();
	iMark_3s.clear();
	iMark_Is.clear();

	pch = strtok( buf, "\n" );
	while (pch != NULL)
	{
		seq = pch;
		ix0 = seq.find_first_of( "\t" );
		if (ix0!=string::npos)
		{
			str_nums = seq.substr(ix0);
			if ( sscanf(str_nums.c_str(), "%s %s %d %d %d %c %d %d %d %d %d %d %d %d", pchSeq, pchId, &ibeg, &iend,
				&ijcntype, &chStrand, &iframe, &imark5, &imark3, &imarkI, &nIntronsInGene , &ixIntron, &posAtg, &posStp) != 14 )
				return;		// error
		}

		strId = pchId;
		seqs.push_back( seq );
		ids.push_back( strId );
		idseq.push_back( pchSeq );
		ibegs.push_back( ibeg );
		iends.push_back( iend );
		iJcnTypes.push_back( ijcntype );
		strands.push_back( chStrand );
		iFrames.push_back( iframe );
		iMark_5s.push_back( imark5 );
		iMark_3s.push_back( imark3 );
		iMark_Is.push_back( imarkI );

		nIntrons.push_back( nIntronsInGene );
		indxIntron.push_back( ixIntron );
		dxAtg.push_back( posAtg );
		dxStp.push_back( posStp );

		pch = strtok( NULL, "\n" );
	}
}

void cModel::splitSeq( vector<string>& seqs5, vector<string>& seqs3, char *buf )
{
	string seq, seq5, seq3, s;
	size_t ix0;
	char *pch;

	seqs5.clear();
	seqs3.clear();

	pch = strtok( buf, "\n" );
	while (pch != NULL)
	{
		seq = pch;
		ix0 = seq.find_first_of( "\t" );
		if (ix0!=string::npos)
		{
			seq5 = seq.substr(0,ix0);
			seq3 = seq.substr(ix0+1, string::npos);
		}

		seqs5.push_back( seq5 );
		seqs3.push_back( seq3 );

		pch = strtok( NULL, "\n" );
	}
}

void cModel::scoreSiteModels()
{
	char pch[64], pchMrna[128], chStrand;
	int iMark, istart, istop, iframe, ifeature;
	int i;
	vector<string> startSite, taaSite, tagSite, tgaSite;
	vector<string> donSite, accSite;
	vector<string> startMrna, taaMrna, tagMrna, tgaMrna, donMrna, accMrna;

	vector<int> startMarks, taaMarks, tagMarks, tgaMarks, donMarks, accMarks;
	vector<int> donStarts, donStops, donFrame, accStarts, accStops, accFrame;
	vector<int> atgStarts, atgStops, atgFrame, taaStarts, taaStops, taaFrame;
	vector<int> tagStarts, tagStops, tagFrame, tgaStarts, tgaStops, tgaFrame;
	vector<double> startScores, taaScores, tagScores, tgaScores, donScores, accScores;

	// start
	FILE *pF_start_site = fopen( m_startSiteFile.c_str(), "r" );
	while ( fscanf(pF_start_site, "%s %s %d %d %d %c %d %d\n", pch, pchMrna, &iMark, &istart, &istop, &chStrand, &iframe, &ifeature)!=EOF )
	//while ( fscanf(pF_start_site, "%s %d %d %d\n", pch, &iMark, &geneid, &ifeature)!=EOF )
	{
		startMarks.push_back( iMark );
		startMrna.push_back( string(pchMrna) );
		atgStarts.push_back( istart );
		atgStops.push_back( istop );
		atgFrame.push_back( iframe );
		startSite.push_back( string(pch) );
	}
	fclose( pF_start_site );
	scoreSiteModel( m_pModelMatAtg, m_iorderAtg, startSite, startScores );

	// stop
	// split stop sites into TAA, TAG, and TGA
	FileBuf f;
	f.Load( m_stopSiteFile.c_str() );
	taa.clear();
	tag.clear();
	tga.clear();
	m_pSite->splitStopSites( taa, tag, tga, other_stops, f.begin );
	f.Delete();

	// taa
	string s;
	for (i=0; i<(int)taa.size(); ++i)
	{
		sscanf( taa[i].c_str(), "%s %s %d %d %d %c %d %d", pch, pchMrna, &iMark, &istart, &istop, &chStrand, &iframe, &ifeature );
		taaMrna.push_back( string(pchMrna) );
		taaStarts.push_back( istart );
		taaStops.push_back( istop );
		taaFrame.push_back( iframe );
		taaMarks.push_back( iMark );
		taaSite.push_back( string(pch) );
	}
	scoreSiteModel( m_pModelMatTaa, m_iorderTaa, taaSite, taaScores );
	// tag
	for (i=0; i<(int)tag.size(); ++i)
	{
		sscanf( tag[i].c_str(), "%s %s %d %d %d %c %d %d", pch, pchMrna, &iMark, &istart, &istop, &chStrand, &iframe, &ifeature );
		tagMrna.push_back( pchMrna );
		tagStarts.push_back( istart );
		tagStops.push_back( istop );
		tagFrame.push_back( iframe );
		tagMarks.push_back( iMark );
		tagSite.push_back( string(pch) );
	}
	scoreSiteModel( m_pModelMatTag, m_iorderTag, tagSite, tagScores );
	// tga
	for (i=0; i<(int)tga.size(); ++i)
	{
		sscanf( tga[i].c_str(), "%s %s %d %d %d %c %d %d", pch, pchMrna, &iMark, &istart, &istop, &chStrand, &iframe, &ifeature );
		tgaMrna.push_back( pchMrna );
		tgaStarts.push_back( istart );
		tgaStops.push_back( istop );
		tgaFrame.push_back( iframe );
		tgaMarks.push_back( iMark );
		tgaSite.push_back( string(pch) );
	}
	scoreSiteModel( m_pModelMatTga, m_iorderTga, tgaSite, tgaScores );

	// don
	FILE *pF_don_site = fopen( m_donSiteFile.c_str(), "r" );
	while ( fscanf(pF_don_site, "%s %s %d %d %d %c %d %d\n", pch, pchMrna, &iMark, &istart, &istop, &chStrand, &iframe, &ifeature)!=EOF )
	//while ( fscanf(pF_don_site, "%s %d %d %d\n", pch, &iMark, &geneid, &ifeature)!=EOF )
	{
		donMarks.push_back( iMark );
		donMrna.push_back( string(pchMrna) );
		donStarts.push_back( istart );
		donStops.push_back( istop );
		donFrame.push_back( iframe );
		donSite.push_back( string(pch) );
	}
	fclose( pF_don_site );
	scoreSiteModel( m_pModelMatDon0, m_iorderDon0, donSite, donScores );

	FILE *pF_acc_site = fopen( m_accSiteFile.c_str(), "r" );
	while ( fscanf(pF_don_site, "%s %s %d %d %d %c %d %d\n", pch, pchMrna, &iMark, &istart, &istop, &chStrand, &iframe, &ifeature)!=EOF )
	//while ( fscanf(pF_acc_site, "%s %d %d %d\n", pch, &iMark, &geneid, &ifeature)!=EOF )
	{
		accMarks.push_back( iMark );
		accMrna.push_back( string(pchMrna) );
		accStarts.push_back( istart );
		accStops.push_back( istop );
		accFrame.push_back( iframe );
		accSite.push_back( string(pch) );
	}
	fclose( pF_acc_site );
	scoreSiteModel( m_pModelMatAcc0, m_iorderAcc0, accSite, accScores );

	FILE *pF_start_score = fopen( "start_score.txt", "w" );
	FILE *pF_taa_score = fopen( "taa_score.txt", "w" );
	FILE *pF_tag_score = fopen( "tag_score.txt", "w" );
	FILE *pF_tga_score = fopen( "tga_score.txt", "w" );
	FILE *pF_don_score = fopen( "don_score.txt", "w" );
	FILE *pF_acc_score = fopen( "acc_score.txt", "w" );

	// start
	fprintf(pF_start_score, "#feat\tID\tstart\tstop\tscore\tmark\n");
	for (i=0; i<(int)startSite.size(); ++i)
	{
		fprintf( pF_start_score, "ATG\t%s\t%d\t%d\t%f\t%d\n", startMrna[i].c_str(), atgStarts[i], atgStops[i], startScores[i], startMarks[i] );
	}
	//stop
	fprintf(pF_taa_score, "#feat\tID\tstart\tstop\tscore\tmark\n");
	for (i=0; i<(int)taaSite.size(); ++i)
	{
		fprintf( pF_taa_score, "TAA\t%s\t%d\t%d\t%f\t%d\n", taaMrna[i].c_str(), taaStarts[i], taaStops[i], taaScores[i], taaMarks[i] );
	}
	fprintf(pF_tag_score, "#feat\tID\tstart\tstop\tscore\tmark\n");
	for (i=0; i<(int)tagSite.size(); ++i)
	{
		fprintf( pF_tag_score, "TAG\t%s\t%d\t%d\t%f\t%d\n", tagMrna[i].c_str(), tagStarts[i], tagStops[i], tagScores[i], tagMarks[i] );
	}
	fprintf(pF_tga_score, "#feat\tID\tstart\tstop\tscore\tmark\n");
	for (i=0; i<(int)tgaSite.size(); ++i)
	{
		fprintf( pF_tga_score, "TGA\t%s\t%d\t%d\t%f\t%d\n", tgaMrna[i].c_str(), tgaStarts[i], tgaStops[i], tgaScores[i], tgaMarks[i] );
	}
	// don
	fprintf(pF_don_score, "#feat\tID\tstart\tstop\tscore\tmark\n");
	for (i=0; i<(int)donSite.size(); ++i)
	{
		fprintf( pF_don_score, "don\t%s\t%d\t%d\t%f\t%d\n", donMrna[i].c_str(), donStarts[i], donStops[i], donScores[i], donMarks[i] );
	}
	// acc
	fprintf(pF_acc_score, "#feat\tID\tstart\tstop\tscore\tmark\n");
	for (i=0; i<(int)accSite.size(); ++i)
	{
		fprintf( pF_acc_score, "acc\t%s\t%d\t%d\t%f\t%d\n", accMrna[i].c_str(), accStarts[i], accStops[i], accScores[i], accMarks[i] );
	}


	fclose( pF_start_score );
	fclose( pF_taa_score );
	fclose( pF_tag_score );
	fclose( pF_tga_score );
	fclose( pF_don_score );
	fclose( pF_acc_score );
}

void cModel::scoreSiteModelsTrans()
{
	char pch[64], pchMrna[128], chStrand;
	int iMark, istart, istop, iframe, ifeature;
	int i;
	vector<string> donSite, accSite;
	vector<string> donMrna, accMrna;
	vector<int> donMarks, accMarks, donStarts, donStops, donFrame, accStarts, accStops, accFrame;
	vector<double> donScores, accScores;

	// don
	FILE *pF_don_site = fopen( m_donSiteFile.c_str(), "r" );
	while ( fscanf(pF_don_site, "%s %s %d %d %d %c %d %d\n", pch, pchMrna, &iMark, &istart, &istop, &chStrand, &iframe, &ifeature)!=EOF )
	{
		donMrna.push_back( pchMrna );
		donMarks.push_back( iMark );
		donStarts.push_back( istart );
		donStops.push_back( istop );
		donFrame.push_back( iframe );
		donSite.push_back( string(pch) );
	}
	fclose( pF_don_site );

	FILE *pF_acc_site = fopen( m_accSiteFile.c_str(), "r" );
	while ( fscanf(pF_acc_site, "%s %s %d %d %d %c %d %d\n", pch, pchMrna, &iMark, &istart, &istop, &chStrand, &iframe, &ifeature)!=EOF )
	{
		accMrna.push_back( pchMrna );
		accMarks.push_back( iMark );
		accStarts.push_back( istart );
		accStops.push_back( istop );
		accFrame.push_back( iframe );
		accSite.push_back( string(pch) );
	}
	fclose( pF_acc_site );

	// score the sites using site models
	scoreDonSites( donScores, donSite, donFrame );
	scoreAccSites( accScores, accSite, accFrame );

	FILE *pF_don_score = fopen( "don_score.txt", "w" );
	FILE *pF_acc_score = fopen( "acc_score.txt", "w" );

	// don
	fprintf(pF_don_score, "#feat\tID\tstart\tstop\tscore\tmark\n");
	for (i=0; i<(int)donSite.size(); ++i)
	{
		fprintf( pF_don_score, "don\t%s\t%d\t%d\t%f\t%d\n", donMrna[i].c_str(), donStarts[i], donStops[i], donScores[i], donMarks[i] );
	}
	// acc
	fprintf(pF_acc_score, "#feat\tID\tstart\tstop\tscore\tmark\n");
	for (i=0; i<(int)accSite.size(); ++i)
	{
		fprintf( pF_acc_score, "acc\t%s\t%d\t%d\t%f\t%d\n", accMrna[i].c_str(), accStarts[i], accStops[i], accScores[i], accMarks[i] );
	}

	fclose( pF_don_score );
	fclose( pF_acc_score );
}

void cModel::scoreIntronAlignments(string& fname, bool bMark)
{
	char pch[64];
	int i;
	double lfScore;
	vector<string> donSite, accSite;
	vector<string> intronSeq, flank5Seq, flank3Seq, idseq, ids;
	vector<char> strands;
	vector<int> ibegs, iends, ijcntypes, iFrame, iMark5, iMark3, iMarkI;
	vector<int> nIntronsInGene, ixIntron, posAtg, posStp;
	vector<double> donScores0, donScores1, donScores2, intronScores;
	vector<double> accScores0, accScores1, accScores2, lenScores;
	vector<double> flank5Score0, flank5Score1, flank5Score2, flank3Score0, flank3Score1, flank3Score2;
	// background model scores
	vector<double> donBack, accBack, flank5Back, flank3Back;

	string str;

	// don
	FILE *pF_don_site = fopen( m_donSiteFile.c_str(), "r" );
	while ( fscanf(pF_don_site, "%s\n", pch)!=EOF )
	{
		donSite.push_back( string(pch) );
	}
	fclose( pF_don_site );

	FILE *pF_acc_site = fopen( m_accSiteFile.c_str(), "r" );
	while ( fscanf(pF_acc_site, "%s\n", pch)!=EOF )
	{
		accSite.push_back( string(pch) );
	}
	fclose( pF_acc_site );

	// intron sequence
	FileBuf f;
	f.Load( m_intronFile.c_str() );
	if (bMark)
		splitSeq( intronSeq, idseq, ids, ibegs, iends, ijcntypes, strands, iFrame, iMark5, iMark3, iMarkI, nIntronsInGene, ixIntron, posAtg, posStp, f.begin );
	else
		splitSeq( intronSeq, idseq, ids, ibegs, iends, ijcntypes, strands, f.begin );
	f.Delete();

	// 5' and 3' intron flanking sequence
	f.Load( m_flankFile.c_str() );
	splitSeq( flank5Seq, flank3Seq, f.begin );
	f.Delete();

	// score the sites using site models
	scoreDonSitesLin3( donScores0, donScores1, donScores2, donSite, ijcntypes );
	scoreAccSitesLin3( accScores0, accScores1, accScores2, accSite, ijcntypes );

	// intron sequence scoring
	//
	double logScore;
	double logScore0_5, logScore1_5, logScore2_5;
	double logScore0_3, logScore1_3, logScore2_3;
	char alphabet[256];
	cMatrixMod tmpMat;
	tmpMat.SetAlphabet( alphabet );
	int lnIntron;

	// get logarithm of background model
	double pZmodel_1[16], pZmodel_0[4];
	for (i=0; i<4; ++i)
		pZmodel_0[i] = log( m_markovNon_0.m_mkModel[i] );
	for (i=0; i<16; ++i) {
		pZmodel_1[i] = log( m_markovNon_1.m_mkModel[i] / m_markovNon_0.m_mkModel[i%4] );
	}

	str.insert( str.begin(), 96, 'A' );
	str.replace( 93, 1, 1, 'T');

	// intron scores
	for (i=0; i<(int)intronSeq.size(); ++i)
	{
		if (intronSeq[i].compare("*")==0)
		{
			intronScores.push_back(-1e12);
			lenScores.push_back(-1e12);
		}
		else
		{
			// score sequence
			scoreMarkov( logScore, m_markovIntr_5.d, 5, intronSeq[i].c_str(), alphabet );
			intronScores.push_back( logScore );

			// score length
			lnIntron = (int)intronSeq[i].length() + m_pSite->m_intronMarginLeft + m_pSite->m_intronMarginRight;
			getLengthScore( lfScore, lnIntron, m_pIntron, m_lenMinIntron, m_lenMaxIntron );
			lenScores.push_back( lfScore );
		}

		// intron flanking sequence
		// 5'
		//scoreMarkov3( logScore0_5, m_markovCod_5, 5, 1, str.c_str(), alphabet );
		//scoreMarkov3( logScore1_5, m_markovCod_5, 5, 2, str.c_str(), alphabet );
		//scoreMarkov3( logScore2_5, m_markovCod_5, 5, 3, str.c_str(), alphabet );
		scoreMarkov3( logScore0_5, 5, 1, flank5Seq[i].c_str(), alphabet );
		scoreMarkov3( logScore1_5, 5, 2, flank5Seq[i].c_str(), alphabet );
		scoreMarkov3( logScore2_5, 5, 3, flank5Seq[i].c_str(), alphabet );

		// 3'
		//scoreMarkov3( logScore0_3, m_markovCod_5, 5, 1, str.c_str(), alphabet );
		//scoreMarkov3( logScore1_3, m_markovCod_5, 5, 2, str.c_str(), alphabet );
		//scoreMarkov3( logScore2_3, m_markovCod_5, 5, 3, str.c_str(), alphabet );
		scoreMarkov3( logScore0_3, 5, 1, flank3Seq[i].c_str(), alphabet );
		scoreMarkov3( logScore1_3, 5, 2, flank3Seq[i].c_str(), alphabet );
		scoreMarkov3( logScore2_3, 5, 3, flank3Seq[i].c_str(), alphabet );

		flank5Score0.push_back( logScore0_5 );
		flank5Score1.push_back( logScore1_5 );
		flank5Score2.push_back( logScore2_5 );

		flank3Score0.push_back( logScore0_3 );
		flank3Score1.push_back( logScore1_3 );
		flank3Score2.push_back( logScore2_3 );

		//int fr5, fr3;
		//checkFrame( fr5, fr3, logScore0_5, logScore1_5, logScore2_5, logScore0_3, logScore1_3, logScore2_3 );
		//if (fr5==fr3) {
		//	switch (fr5) {
		//		case 0:
		//			flank5Score.push_back( logScore0_5 );
		//			flank3Score.push_back( logScore0_3 );
		//			break;
		//		case 1:
		//			flank5Score.push_back( logScore1_5 );
		//			flank3Score.push_back( logScore1_3 );
		//			break;
		//		default:
		//			flank5Score.push_back( logScore2_5 );
		//			flank3Score.push_back( logScore2_3 );
		//	}
		//}
		//else {
		//	flank5Score.push_back( -1e9 );
		//	flank3Score.push_back( -1e9 );
		//}

		// get background model scores
		scoreMarkov( logScore, m_markovNon_5.na, m_markovNon_5.n, 5, flank5Seq[i].c_str(), alphabet );	flank5Back.push_back( logScore );
		scoreMarkov( logScore, m_markovNon_5.na, m_markovNon_5.n, 5, flank3Seq[i].c_str(), alphabet );	flank3Back.push_back( logScore );

		scoreMarkov( logScore, pZmodel_1, pZmodel_0, 1, donSite[i].c_str(), alphabet );		donBack.push_back( logScore );
		scoreMarkov( logScore, pZmodel_1, pZmodel_0, 1, accSite[i].c_str(), alphabet );		accBack.push_back( logScore );
	}

	FILE *pF_score = fopen( fname.c_str(), "w" );
	string strtype;

	fprintf(pF_score, "# ID_seq\tID_intr\tbeg\tend\tstrand\tjcn\tflank5_0\tflank5_1\tflank5_2\tflank5back\tdon_score0\tdon_score1\tdon_score2\tdon_back\tintron_score\tlen_score\tacc_score0\tacc_score1\tacc_score2\tacc_back\tflank3_0\tflank3_1\tflank3_2\tflank3_back");
	if (bMark)
		fprintf(pF_score, "\tframe\t5p\t3p\tann\tIpG\tixIntron\tposAtg\tposStp\n");
	else
		fprintf(pF_score, "\n");

	for (i=0; i<(int)intronScores.size(); ++i)
	{
		switch (ijcntypes[i])
		{
		case INTRON_GT_AG:
			strtype = "GT_AG";
			break;
		case INTRON_GC_AG:
			strtype = "GC_AG";
			break;
		case INTRON_AT_AC:
			strtype = "AT_AC";
			break;
		default:
			strtype = "?";
		};

		lnIntron = (int)intronSeq[i].length() + m_pSite->m_intronMarginLeft + m_pSite->m_intronMarginRight;
		fprintf( pF_score, "%s\t%s\t%d\t%d\t%c\t%s\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e",
			idseq[i].c_str(), ids[i].c_str(), ibegs[i], iends[i],
			strands[i], strtype.c_str(),
			flank5Score0[i], flank5Score1[i], flank5Score2[i], flank5Back[i],
			donScores0[i], donScores1[i], donScores2[i], donBack[i],
			intronScores[i], lenScores[i],
			accScores0[i], accScores1[i], accScores2[i], accBack[i],
			flank3Score0[i], flank3Score1[i], flank3Score2[i], flank3Back[i] );

		if (bMark)
			fprintf( pF_score, "\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\n", iFrame[i], iMark5[i], iMark3[i], iMarkI[i], nIntronsInGene[i], ixIntron[i], posAtg[i], posStp[i] );
		else
			fprintf( pF_score, "\n" );
	}

	fclose( pF_score );
}

void cModel::scoreIntronFlanks(string& fname, bool bMark)
{
	int i;
	vector<string> intronSeq, flank5Seq, flank3Seq, idseq, ids;
	vector<char> strands;
	vector<int> ibegs, iends, ijcntypes, iFrame, iMark5, iMark3, iMarkI;
	vector<int> nIntronsInGene, ixIntron, posAtg, posStp;
	vector<double> flank5Score0, flank5Score1, flank5Score2, flank3Score0, flank3Score1, flank3Score2;
	// background model scores
	vector<double> flank5Back, flank3Back;

	string str;

	// intron sequence
	FileBuf f;
	f.Load( m_intronFile.c_str() );
	if (bMark)
		splitSeq( intronSeq, idseq, ids, ibegs, iends, ijcntypes, strands, iFrame, iMark5, iMark3, iMarkI, nIntronsInGene, ixIntron, posAtg, posStp, f.begin );
	else
		splitSeq( intronSeq, idseq, ids, ibegs, iends, ijcntypes, strands, f.begin );
	f.Delete();

	// 5' and 3' intron flanking sequence
	f.Load( m_flankFile.c_str() );
	splitSeq( flank5Seq, flank3Seq, f.begin );
	f.Delete();

	// intron sequence scoring
	//
	double logScore;
	double logScore0_5, logScore1_5, logScore2_5;
	double logScore0_3, logScore1_3, logScore2_3;
	char alphabet[256];
	cMatrixMod tmpMat;
	tmpMat.SetAlphabet( alphabet );
	int lnIntron;

	str.insert( str.begin(), 96, 'A' );
	str.replace( 93, 1, 1, 'T');

	// intron scores
	for (i=0; i<(int)flank5Seq.size(); ++i)
	{
		// intron flanking sequence
		// 5'
		scoreMarkov3( logScore0_5, 5, 1, flank5Seq[i].c_str(), alphabet );
		scoreMarkov3( logScore1_5, 5, 2, flank5Seq[i].c_str(), alphabet );
		scoreMarkov3( logScore2_5, 5, 3, flank5Seq[i].c_str(), alphabet );

		// 3'
		scoreMarkov3( logScore0_3, 5, 1, flank3Seq[i].c_str(), alphabet );
		scoreMarkov3( logScore1_3, 5, 2, flank3Seq[i].c_str(), alphabet );
		scoreMarkov3( logScore2_3, 5, 3, flank3Seq[i].c_str(), alphabet );

		flank5Score0.push_back( logScore0_5 );
		flank5Score1.push_back( logScore1_5 );
		flank5Score2.push_back( logScore2_5 );

		flank3Score0.push_back( logScore0_3 );
		flank3Score1.push_back( logScore1_3 );
		flank3Score2.push_back( logScore2_3 );

		// get background model scores
		scoreMarkov( logScore, m_markovNon_5.na, m_markovNon_5.n, 5, flank5Seq[i].c_str(), alphabet );	flank5Back.push_back( logScore );
		scoreMarkov( logScore, m_markovNon_5.na, m_markovNon_5.n, 5, flank3Seq[i].c_str(), alphabet );	flank3Back.push_back( logScore );
	}

	FILE *pF_score = fopen( fname.c_str(), "w" );
	string strtype;

	fprintf(pF_score, "# ID_seq\tID_intr\tbeg\tend\tstrand\tjcn\tflank5_0\tflank5_1\tflank5_2\tflank5back\tflank3_0\tflank3_1\tflank3_2\tflank3_back");
	if (bMark)
		fprintf(pF_score, "\tframe\t5p\t3p\tann\tIpG\tixIntron\tposAtg\tposStp\n");
	else
		fprintf(pF_score, "\n");

	for (i=0; i<(int)flank5Seq.size(); ++i)
	{
		switch (ijcntypes[i])
		{
		case INTRON_GT_AG:
			strtype = "GT_AG";
			break;
		case INTRON_GC_AG:
			strtype = "GC_AG";
			break;
		case INTRON_AT_AC:
			strtype = "AT_AC";
			break;
		default:
			strtype = "?";
		};

		lnIntron = (int)intronSeq[i].length() + m_pSite->m_intronMarginLeft + m_pSite->m_intronMarginRight;
		fprintf( pF_score, "%s\t%s\t%d\t%d\t%c\t%s\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e\t%7.4e",
			idseq[i].c_str(), ids[i].c_str(), ibegs[i], iends[i],
			strands[i], strtype.c_str(),
			flank5Score0[i], flank5Score1[i], flank5Score2[i], flank5Back[i],
			flank3Score0[i], flank3Score1[i], flank3Score2[i], flank3Back[i] );

		if (bMark)
			fprintf( pF_score, "\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\n", iFrame[i], iMark5[i], iMark3[i], iMarkI[i], nIntronsInGene[i], ixIntron[i], posAtg[i], posStp[i] );
		else
			fprintf( pF_score, "\n" );
	}

	fclose( pF_score );
}

void cModel::checkFrame( int& fr5, int& fr3, double& logScore0_5, double& logScore1_5, double& logScore2_5,
						                     double& logScore0_3, double& logScore1_3, double& logScore2_3 )
{
	int i5, i3;

	if (logScore0_5 > logScore1_5 && logScore0_5 > logScore2_5) {
		i5 = 0;
	}
	else if (logScore1_5 > logScore0_5 && logScore1_5 > logScore2_5) {
		i5 = 1;
	}
	else { // (logScore3_5 > logScore0_5 && logScore3_5 > logScore2_5)
		i5 = 2;
	}

	if (logScore0_3 > logScore1_3 && logScore0_3 > logScore2_3) {
		i3 = 0;
	}
	else if (logScore1_3 > logScore0_3 && logScore1_3 > logScore2_3) {
		i3 = 1;
	}
	else { // (logScore3_3 > logScore0_3 && logScore3_3 > logScore2_3)
		i3 = 2;
	}
}

void cModel::ReorderLeft( double* mout, double* min, int orderIn, int orderOut )
{
	int size = 4<<(2*orderIn), i;
	double *mlast, *mcur, *ptr;
	double sum(0.);

	int newN;
	int leftA;
	int leftC;
	int leftG;
	int leftT;
	int ict;

	mlast = min;

	for (i = orderIn; i>orderOut; --i)
	{
		newN = 4<<(2*(i-1));
		mcur = new double[newN];
		ptr = mcur;

		leftA = ( A << (2*i) );
		leftC = ( C << (2*i) );
		leftG = ( G << (2*i) );
		leftT = ( T << (2*i) );

		ict = 0;

		while ( ict < newN ) {
			sum = mlast[leftA+ict] + mlast[leftC+ict] + mlast[leftG+ict] + mlast[leftT+ict];
			mcur[ict] = sum;
			++ict;
		}

		if (i != orderIn)
			delete []mlast;

		mlast = mcur;
	}

	// copy working buffer to output buffer
	memcpy( mout, mcur, newN * sizeof(double) );
	delete []mcur;
}

void cModel::ReorderRight( double* mout, double* min, int orderIn, int orderOut )
{
	int size = 4<<(2*orderIn), i;
	double *mlast, *mcur, *ptr;
	double sum(0.);

	int newN;
	int ict;

	mlast = min;

	for (i = orderIn; i>orderOut; --i)
	{
		newN = 4<<(2*(i-1));
		mcur = new double[newN];
		ptr = mcur;

		ict = 0;

		while ( ict < newN ) {
			sum = mlast[4*ict] + mlast[4*ict+1] + mlast[4*ict+2] + mlast[4*ict+3];
			mcur[ict] = sum;
			++ict;
		}

		if (i != orderIn)
			delete []mlast;

		mlast = mcur;
	}

	// copy working buffer to output buffer
	memcpy( mout, mcur, newN * sizeof(double) );
	delete []mcur;
}

void cModel::ReorderLeftL( double* mout, double* min, int orderIn, int orderOut )
{	// convert to lower order, and also from log units to linear
	int size = 4<<(2*orderIn), i;
	double *mlast, *mcur, *ptr;
	double sum(0.);

	int newN;
	int leftA;
	int leftC;
	int leftG;
	int leftT;
	int ict;

	mlast = min;

	for (i = orderIn; i>orderOut; --i)
	{
		newN = 4<<(2*(i-1));
		mcur = new double[newN];
		ptr = mcur;

		leftA = ( A << (2*i) );
		leftC = ( C << (2*i) );
		leftG = ( G << (2*i) );
		leftT = ( T << (2*i) );

		ict = 0;

		while ( ict < newN ) {
			sum = exp(mlast[leftA+ict]) + exp(mlast[leftC+ict]) + exp(mlast[leftG+ict]) + exp(mlast[leftT+ict]);
			mcur[ict] = log( sum );
			++ict;
		}

		if (i != orderIn)
			delete []mlast;

		mlast = mcur;
	}

	// copy working buffer to output buffer & convert from log to linear
	for (i=0; i<newN; ++i)
		mout[i] = exp( mcur[i] );

	delete []mcur;
}

void cModel::ReorderRightL( double* mout, double* min, int orderIn, int orderOut )
{	// convert to lower order, and also from log units to linear
	int size = 4<<(2*orderIn), i;
	double *mlast, *mcur, *ptr;
	double sum(0.);

	int newN;
	int ict;

	mlast = min;

	for (i = orderIn; i>orderOut; --i)
	{
		newN = 4<<(2*(i-1));
		mcur = new double[newN];
		ptr = mcur;

		ict = 0;

		while ( ict < newN ) {
			sum = exp(mlast[4*ict]) + exp(mlast[4*ict+1]) + exp(mlast[4*ict+2]) + exp(mlast[4*ict+3]);
			mcur[ict] = log( sum );
			++ict;
		}

		if (i != orderIn)
			delete []mlast;

		mlast = mcur;
	}

	// copy working buffer to output buffer & convert from log to linear
	for (i=0; i<newN; ++i)
		mout[i] = exp( mcur[i] );

	delete []mcur;
}

void cModel::scoreLengthModels(vector<sTranscript>& transcripts)
{
	int ig, i, lnIni, lnExon, lnTer, lnIntron;
	int nExons;
	double scoreExon, scoreIntron;

	FILE *pF_ln_score = fopen( "length_score.txt", "w" );

	for (ig=0; ig<(int)transcripts.size(); ++ig) {

		nExons = (int)transcripts[ig].boundaries.size();
		if (nExons>1) {

			lnIni = transcripts[ig].boundaries[0].iend - transcripts[ig].boundaries[0].ibeg + 1;
			getLengthScore( scoreExon, lnIni, m_pIniExon, m_lenMinIniExon, m_lenMaxIniExon );

			fprintf(pF_ln_score, "%s\t%f", transcripts[ig].strMrnaId.c_str(), scoreExon );

			for (i=1; i<nExons-2; ++i) {
				lnExon = transcripts[ig].boundaries[i].iend - transcripts[ig].boundaries[i].ibeg + 1;
				lnIntron = transcripts[ig].boundaries[i].ibeg - 1 - transcripts[ig].boundaries[i-1].iend;

				getLengthScore( scoreExon, lnExon, m_pExon, m_lenMinExon, m_lenMaxExon );
				getLengthScore( scoreIntron, lnIntron, m_pIntron, m_lenMinIntron, m_lenMaxIntron );

				fprintf(pF_ln_score, "\t%f\t%f", scoreIntron, scoreExon );
			}	// exon loop

			lnIntron = transcripts[ig].boundaries[nExons-1].ibeg - 1 - transcripts[ig].boundaries[nExons-2].iend;
			lnTer = transcripts[ig].boundaries[nExons-1].iend - transcripts[ig].boundaries[nExons-1].ibeg + 1;
		
			getLengthScore( scoreExon, lnTer, m_pTerExon, m_lenMinTerExon, m_lenMaxTerExon );
			getLengthScore( scoreIntron, lnIntron, m_pIntron, m_lenMinIntron, m_lenMaxIntron );

			fprintf(pF_ln_score, "\t%f\t%f\n", scoreIntron, scoreExon );
		}
		else {		// single exon gene
			lnExon = transcripts[ig].boundaries[0].iend - transcripts[ig].boundaries[0].ibeg + 1;
			getLengthScore( scoreExon, lnExon, m_pSingleExon, m_lenMinTerExon, m_lenMaxTerExon );
			fprintf(pF_ln_score, "%s\t%f\n", transcripts[ig].strMrnaId.c_str(), scoreExon );
		}

	}	// gene loop

	fclose( pF_ln_score );
}

//void cModel::isoformClassify(vector<codingRegion>& genes)
//{
//	if (genes.empty())
//		return;
//
//	bool bDone(false);
//	int ix(0), i, j, ixRef;
//	int nGeneCount(0), nIsoformCount(0);
//
//	vector<codingRegion> geneSiblings;
//	string sibName;
//	vector<isoformType> vitTypes;
//
//	FILE *pF = fopen("mi_classify.txt","w");
//
//	while (!bDone)
//	{
//		geneSiblings.clear();
//		sibName = genes[ix].strGeneId;
//		while (genes[ix].strGeneId.compare(sibName)==0)
//		{
//			geneSiblings.push_back( genes[ix] );
//			++ix;
//			if (ix >= genes.size()) { bDone = true; break; }
//		}
//
//		++nGeneCount;
//
//		// skip single-isoform genes
//		if (geneSiblings.size() <= 1)
//			continue;
//
//		// length of nTypes is one less than number of siblings
//		classifyGroup( ixRef, vitTypes, geneSiblings );
//
//		j = 0;
//		for (i=0; i<geneSiblings.size(); ++i)
//		{
//			if (i != ixRef) {
//				fprintf(pF, "%s\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\n", geneSiblings[j].strMrnaId.c_str(),
//					nGeneCount, j+1,
//					vitTypes[j].nTypes[0], vitTypes[j].nTypes[1], vitTypes[j].nTypes[2], vitTypes[j].nTypes[3],
//					vitTypes[j].nTypes[4], vitTypes[j].nTypes[5], vitTypes[j].nTypes[6], vitTypes[j].nTypes[7]);
//				++j;
//			}
//		}
//
//	}	// while (!bDone)
//
//	fclose( pF );
//}
//
//int cModel::getUniqueCount( vector<int>& vi )
//{
//	// helper fcn which counts number of unique entries in vi
//	vector<int> uset;		// set of unique values
//	bool bFound;
//
//	for (int i=0; i<vi.size(); ++i)
//	{
//		bFound = false;
//		for (int j=0; j<uset.size(); ++j)
//		{
//			if (vi[i]==uset[j]) { bFound = true; break; }
//		}
//		if (!bFound)
//			uset.push_back( vi[i] );
//	}
//	int iret = uset.size();
//	return iret;
//}
//
//void cModel::classifyGroup( int& ixRef, vector<isoformType>& isoTypes, vector<codingRegion>& genes )
//{
//	if (genes.size() < 2) return;
//
//	int nGenes = genes.size();
//	int nExons, nExonsRef;
//	int i, j, ie, iType;
//	int cdsLenMax(-1), len;
//
//	isoformType types;
//	isoTypes.clear();
//
//	// select reference gene (pick one with longest CDS)
//	for (i=0; i<nGenes; ++i)
//	{
//		len = getCdsLength(genes[i]);
//		if (len>cdsLenMax) {
//			cdsLenMax = len;
//			ixRef = i;
//		}
//	}
//
//	nExonsRef = genes[ixRef].boundaries.size();
//
//	RMatrix mat;
//	bool bRowHit(false), *pbColHit;
//	bool bStrand = genes[ixRef].bStrand;
//
//	// count number of CDS starts separately
//	vector<int> iStarts;
//	for (i=0; i<nGenes; ++i)
//	{
//		iStarts.push_back( genes[i].boundaries[0].ibeg );
//	}
//	bool bCheck = false;
//	
//	for (i=0; i<nGenes; ++i)
//	{
//		if (i==ixRef)
//			continue;
//
//		// reset types
//		for (j=0; j<8; ++j)
//			types.nTypes[j] = IC_NOTHING;
//
//		iType = 0;
//		nExons = genes[i].boundaries.size();
//
//		//if (!bCheck) {
//		//	bCheck = true;
//		//	types.nTypes[ IC_CDS_START ] = getUniqueCount( iStarts ) - 1;
//		//}
//		if (bStrand && genes[i].boundaries[0].ibeg != genes[ixRef].boundaries[0].ibeg ||
//			!bStrand && genes[i].boundaries[nExons-1].iend != genes[ixRef].boundaries[nExonsRef-1].iend) {
//			++types.nTypes[ IC_CDS_START ];
//		}
//
//		// gene[i], gene[ixRef] overlap analysis
//		exonOverlap( mat, genes[i], genes[ixRef] );
//
//		countRetainedIntrons( types, genes[i], genes[ixRef] );
//
//		// keep track of all ref exons overlapped by query gene
//		pbColHit = new bool[mat.columns];
//		double lf;
//		for (j=0; j<mat.columns; ++j) pbColHit[j] = false;
//
//		for (ie=0; ie<mat.rows; ++ie) {
//			bRowHit = false;
//
//			for (j=0; j<mat.columns; ++j) {
//
//				lf = mat.Get(ie,j);
//				if (lf != 0.) {
//					bRowHit = true;
//					pbColHit[j] = true;
//				}
//
//				lf = mat.Get(ie,j);
//				if ( (int) lf == 1 ) {		// partial overlap
//					if (genes[i].boundaries[ie].ibeg == genes[ixRef].boundaries[j].ibeg) {
//						if (bStrand) { ++types.nTypes[ IC_DONOR ]; } else { ++types.nTypes[ IC_ACCEPTOR ]; }
//					}
//					else if (genes[i].boundaries[ie].iend == genes[ixRef].boundaries[j].iend) {
//						if (bStrand) { ++types.nTypes[ IC_ACCEPTOR ]; } else { ++types.nTypes[ IC_DONOR ]; }
//					}
//					else {
//						++types.nTypes[ IC_DONOR ];
//						++types.nTypes[ IC_ACCEPTOR ];
//					}
//				}
//
//			}	// loop over columns (query gene exons)
//
//			if ( !bRowHit ) {	// exon ie is inserted
//				++types.nTypes[ IC_INSERTED_EXON ];
//			}
//
//		}	// loop over rows (ref gene exons)
//
//		for (j=0; j<mat.columns; ++j) {
//			if (!pbColHit[j]) ++types.nTypes[ IC_SKIPPED_EXON ];
//		}
//
//		isoTypes.push_back( types );
//
//	}	// loop over genes
//
//}
//
//bool cModel::countRetainedIntrons( isoformType& types, codingRegion& gene, codingRegion& geneRef )
//{
//	// scan geneRef for retained introns (geneRef has longest CDS, so makes sense to check there)
//	int ii, j;
//	int intronBeg, intronEnd;
//	
//	for (ii=0; ii<gene.boundaries.size()-1; ++ii)
//	{
//		intronBeg = gene.boundaries[ii].iend + 1;
//		intronEnd = gene.boundaries[ii+1].ibeg - 1;
//
//		// loop over all ref gene exons
//		for (j=0; j<geneRef.boundaries.size(); ++j)
//		{
//			// ii^th intron of gene inside j^th exon of geneRef?
//			if (intronBeg>geneRef.boundaries[j].ibeg && intronEnd<geneRef.boundaries[j].iend)
//				++types.nTypes[ IC_RETAINED_INTRON ];
//		}
//	}
//
//	return false;
//}
//
//bool cModel::exonOverlap(RMatrix& mat, codingRegion& gene1, codingRegion& gene2)
//{
//	int nExons1 = gene1.boundaries.size();
//	int nExons2 = gene2.boundaries.size();
//
//	if (mat.Create(nExons1, nExons2)) return true;
//
//	int i,j;
//	int itype;
//
//	for (i=0; i<nExons1; ++i) {
//		for (j=0; j<nExons2; ++j) {
//			checkOverlap(itype, gene1.boundaries[i], gene2.boundaries[j]);
//			mat.Set(i, j, double(itype));
//		}
//	}
//	return false;
//}
//
//void cModel::checkOverlap(int& itype, segment& seg1, segment& seg2)
//{
//	if (seg1.iend < seg2.ibeg || seg1.ibeg > seg2.iend)
//		itype = 0;		// no overlap
//	else if (seg1.ibeg == seg2.ibeg && seg1.iend == seg2.iend)
//		itype = 2;		// perfect overlap
//	else
//		itype = 1;		// partial overlap
//}
//
//int cModel::getCdsLength(codingRegion& gene)
//{
//	int len(0);
//
//	for (int ie=0; ie<gene.boundaries.size(); ++ie)
//	{
//		len += gene.boundaries[ie].iend - gene.boundaries[ie].ibeg + 1;
//	}
//	return len;
//}
//
//int cModel::getXscriptLength(sTranscript& xscript)
//{
//	int len(0);
//
//	for (int ie=0; ie<xscript.boundaries.size(); ++ie)
//	{
//		len += xscript.boundaries[ie].iend - xscript.boundaries[ie].ibeg + 1;
//	}
//	return len;
//}
//
//void cModel::isoformClassify(vector<sTranscript>& xscripts)
//{
//	if (xscripts.empty())
//		return;
//
//	bool bDone(false);
//	int ix(0), i, j, ixRef;
//
//	vector<sTranscript> geneSiblings;
//	string sibName;
//	vector<isoformType> vitTypes;
//
//	FILE *pF = fopen("mi_classify_xs.txt","w");
//
//	while (!bDone)
//	{
//		geneSiblings.clear();
//		sibName = xscripts[ix].strGeneId;
//		while (xscripts[ix].strGeneId.compare(sibName)==0)
//		{
//			geneSiblings.push_back( xscripts[ix] );
//			++ix;
//			if (ix >= xscripts.size()) { bDone = true; break; }
//		}
//
//		// skip single-isoform genes
//		if (geneSiblings.size() <= 1)
//			continue;
//
//		// length of nTypes is one less than number of siblings
//		classifyGroup( ixRef, vitTypes, geneSiblings );
//
//		j = 0;
//		for (i=0; i<geneSiblings.size(); ++i)
//		{
//			if (i != ixRef) {
//				fprintf(pF, "%s\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\n", geneSiblings[j].strMrnaId.c_str(),
//				vitTypes[j].nTypes[0], vitTypes[j].nTypes[1], vitTypes[j].nTypes[2], vitTypes[j].nTypes[3],
//				vitTypes[j].nTypes[4], vitTypes[j].nTypes[5], vitTypes[j].nTypes[6], vitTypes[j].nTypes[7]);
//				++j;
//			}
//		}
//
//	}	// while (!bDone)
//
//	fclose( pF );
//}
//
//void cModel::classifyGroup( int& ixRef, vector<isoformType>& isoTypes, vector<sTranscript>& xscripts )
//{
//	if (xscripts.size() < 2) return;
//
//	int nGenes = xscripts.size();
//	int nExons;
//	int i, j, ie, iType;
//	int lenMax(-1), len;
//
//	isoformType types;
//	isoTypes.clear();
//
//	// select reference gene (pick one with longest CDS)
//	for (i=0; i<nGenes; ++i)
//	{
//		len = getXscriptLength(xscripts[i]);
//		if (len>lenMax) {
//			lenMax = len;
//			ixRef = i;
//		}
//	}
//
//	RMatrix mat;
//	bool bRowHit(false), *pbColHit;
//	bool bStrand = xscripts[ixRef].bStrand;
//
//	for (i=0; i<nGenes; ++i)
//	{
//		if (i==ixRef)
//			continue;
//
//		// reset types
//		for (j=0; j<8; ++j)
//			types.nTypes[j] = IC_NOTHING;
//
//		iType = 0;
//		nExons = xscripts[i].boundaries.size();
//
//		if (xscripts[i].bStrand && xscripts[i].boundaries[0].ibeg != xscripts[ixRef].boundaries[0].ibeg ||
//			!xscripts[i].bStrand && xscripts[i].boundaries[nExons-1].iend != xscripts[ixRef].boundaries[nExons-1].iend)
//			++types.nTypes[ IC_CDS_START ];
//
//		// gene[i], gene[ixRef] overlap analysis
//		exonOverlap( mat, xscripts[i], xscripts[ixRef] );
//
//		// keep track of all ref exons overlapped by query gene
//		pbColHit = new bool[mat.columns];
//		double lf;
//		for (j=0; j<mat.columns; ++j) pbColHit[j] = false;
//
//		for (ie=0; ie<mat.rows; ++ie) {
//			bRowHit = false;
//
//			for (j=0; j<mat.columns; ++j) {
//
//				lf = mat.Get(ie,j);
//				if (lf != 0.) {
//					bRowHit = true;
//					pbColHit[j] = true;
//				}
//
//				lf = mat.Get(ie,j);
//				if ( (int) lf == 1 ) {		// partial overlap
//					if (xscripts[i].boundaries[ie].ibeg == xscripts[ixRef].boundaries[j].ibeg) {
//						if (bStrand) { ++types.nTypes[ IC_DONOR ]; } else { ++types.nTypes[ IC_ACCEPTOR ]; }
//					}
//					else if (xscripts[i].boundaries[ie].iend == xscripts[ixRef].boundaries[j].iend) {
//						if (bStrand) { ++types.nTypes[ IC_ACCEPTOR ]; } else { ++types.nTypes[ IC_DONOR ]; }
//					}
//					else {
//						++types.nTypes[ IC_DONOR ];
//						++types.nTypes[ IC_ACCEPTOR ];
//					}
//				}
//
//			}	// loop over columns (query gene exons)
//
//			if ( !bRowHit ) {	// exon ie is inserted
//				++types.nTypes[ IC_INSERTED_EXON ];
//			}
//
//		}	// loop over rows (ref gene exons)
//
//		for (j=0; j<mat.columns; ++j) {
//			if (!pbColHit[j]) ++types.nTypes[ IC_SKIPPED_EXON ];
//		}
//
//		isoTypes.push_back( types );
//
//	}	// loop over genes
//
//}
//
//bool cModel::exonOverlap(RMatrix& mat, sTranscript& gene1, sTranscript& gene2)
//{
//	int nExons1 = gene1.boundaries.size();
//	int nExons2 = gene2.boundaries.size();
//
//	if (mat.Create(nExons1, nExons2)) return true;
//
//	int i,j;
//	int itype;
//
//	for (i=0; i<nExons1; ++i) {
//		for (j=0; j<nExons2; ++j) {
//			checkOverlap(itype, gene1.boundaries[i], gene2.boundaries[j]);
//			mat.Set(i, j, double(itype));
//		}
//	}
//	return false;
//}

void cModel::scoreLengthModels(vector<codingRegion>& genes)
{
	int ig, i, lnIni, lnExon, lnTer, lnIntron;
	int nExons;
	double scoreExon, scoreIntron;

	FILE *pF_ln_score = fopen( "length_score.txt", "w" );

	for (ig=0; ig<(int)genes.size(); ++ig) {

		nExons = (int)genes[ig].boundaries.size();
		if (nExons>1) {

			lnIni = genes[ig].boundaries[0].iend - genes[ig].boundaries[0].ibeg + 1;
			getLengthScore( scoreExon, lnIni, m_pIniExon, m_lenMinIniExon, m_lenMaxIniExon );

			fprintf(pF_ln_score, "%s\t%f", genes[ig].strMrnaId.c_str(), scoreExon );

			for (i=1; i<nExons-1; ++i) {
				lnExon = genes[ig].boundaries[i].iend - genes[ig].boundaries[i].ibeg + 1;
				lnIntron = genes[ig].boundaries[i].ibeg - 1 - genes[ig].boundaries[i-1].iend;

				getLengthScore( scoreExon, lnExon, m_pExon, m_lenMinExon, m_lenMaxExon );
				getLengthScore( scoreIntron, lnIntron, m_pIntron, m_lenMinIntron, m_lenMaxIntron );

				fprintf(pF_ln_score, "\t%f\t%f", scoreIntron, scoreExon );
			}	// exon loop

			lnIntron = genes[ig].boundaries[nExons-1].ibeg - 1 - genes[ig].boundaries[nExons-2].iend;
			lnTer = genes[ig].boundaries[nExons-1].iend - genes[ig].boundaries[nExons-1].ibeg + 1;
		
			getLengthScore( scoreExon, lnTer, m_pTerExon, m_lenMinTerExon, m_lenMaxTerExon );
			getLengthScore( scoreIntron, lnIntron, m_pIntron, m_lenMinIntron, m_lenMaxIntron );

			fprintf(pF_ln_score, "\t%f\t%f\n", scoreIntron, scoreExon );
		}
		else {		// single exon gene
			lnExon = genes[ig].boundaries[0].iend - genes[ig].boundaries[0].ibeg + 1;
			getLengthScore( scoreExon, lnExon, m_pSingleExon, m_lenMinSingleExon, m_lenMaxSingleExon );
			fprintf(pF_ln_score, "%s\t%f\n", genes[ig].strMrnaId.c_str(), scoreExon );
		}

	}	// gene loop

	fclose( pF_ln_score );
}


bool cModel::scoreLengths( vector<double>& lenScores, vector<int>& nLengths, string& modelName )
{
	if (modelName.compare("SINGLE") == 0)
		scoreLength( lenScores, nLengths, m_pSingleExon, m_lenMinSingleExon, m_lenMaxSingleExon );
	else if (modelName.compare("INITIAL") == 0)
		scoreLength( lenScores, nLengths, m_pSingleExon, m_lenMinIniExon, m_lenMaxIniExon );
	else if (modelName.compare("EXON") == 0)
		scoreLength( lenScores, nLengths, m_pSingleExon, m_lenMinExon, m_lenMaxExon );
	else if (modelName.compare("TERMINAL") == 0)
		scoreLength( lenScores, nLengths, m_pSingleExon, m_lenMinTerExon, m_lenMaxTerExon );
	else if (modelName.compare("INTRON") == 0)
		scoreLength( lenScores, nLengths, m_pSingleExon, m_lenMinIntron, m_lenMaxIntron );
	else if (modelName.compare("INTERGENIC") == 0)
		scoreLength( lenScores, nLengths, m_pSingleExon, m_lenMinInterg, m_lenMaxInterg );
	else
		return true;

	return false;
}

void cModel::scoreLength( vector<double>& lenScores, vector<int>& nLengths, double *pLd, int min, int max )
{
	int i, n((int)nLengths.size());
	double score;
	lenScores.clear();

	for (i=0; i<n; ++i) {
		getLengthScore( score, nLengths[i], pLd, min, max );
		lenScores.push_back( score );
	}
}

void cModel::getLengthScore( double& score, int nLength, double *pLd, int min, int max )
{
	if (nLength>=min && nLength<=max)
		score = pLd[nLength];
	else
		score = -1e12;
		//score = -numeric_limits<double>::max();
}

bool cModel::scoreDonSites( vector<double>& donScores, vector<string>& donSite, vector<int>& donFrame )
{
	if (donSite.empty())
		return false;

	donScores.clear();
	double logScore;

	int i;
	int iwidth = (int)donSite[0].length();

	for (i=0; i<(int)donFrame.size(); ++i)
	{
		switch (donFrame[i]) {
			case 1:
				scoreSite( logScore, m_pModelMatDon1, m_iorderDon1, iwidth, donSite[i].c_str(), m_alphabet );
				break;

			case 2:
				scoreSite( logScore, m_pModelMatDon2, m_iorderDon2, iwidth, donSite[i].c_str(), m_alphabet );
				break;

			default:
				scoreSite( logScore, m_pModelMatDon0, m_iorderDon0, iwidth, donSite[i].c_str(), m_alphabet );
		}

		donScores.push_back( logScore );
	}
	return false;
}

bool cModel::scoreAccSites( vector<double>& accScores, vector<string>& accSite, vector<int>& accFrame )
{
	if (accSite.empty())
		return false;

	accScores.clear();
	double logScore;

	int i;
	int iwidth = (int)accSite[0].length();

	for (i=0; i<(int)accFrame.size(); ++i)
	{
		switch (accFrame[i]) {
			case 1:
				scoreSite( logScore, m_pModelMatAcc1, m_iorderAcc1, iwidth, accSite[i].c_str(), m_alphabet );
				break;

			case 2:
				scoreSite( logScore, m_pModelMatAcc2, m_iorderAcc2, iwidth, accSite[i].c_str(), m_alphabet );
				break;

			default:
				scoreSite( logScore, m_pModelMatAcc0, m_iorderAcc0, iwidth, accSite[i].c_str(), m_alphabet );
		}

		accScores.push_back( logScore );
	}
	return false;
}

bool cModel::scoreDonSite3( double& logScore0, double& logScore1, double& logScore2, string& donSite, int& iJcnType )
{
	int iwidth = (int)donSite.length();
	if (iJcnType==INTRON_GT_AG)
	{
		scoreSite( logScore0, m_pModelMatDon0, m_iorderDon0, iwidth, donSite.c_str(), m_alphabet );
		scoreSite( logScore1, m_pModelMatDon1, m_iorderDon1, iwidth, donSite.c_str(), m_alphabet );
		scoreSite( logScore2, m_pModelMatDon2, m_iorderDon2, iwidth, donSite.c_str(), m_alphabet );
	}
	else if (iJcnType==INTRON_GC_AG || iJcnType==INTRON_AT_AC)
	{
		// convert to canonical donor type and score
		string site = donSite;
		site.at(3) = 'G'; site.at(4) = 'T';
		scoreSite( logScore0, m_pModelMatDon0, m_iorderDon0, iwidth, site.c_str(), m_alphabet );
		scoreSite( logScore1, m_pModelMatDon1, m_iorderDon1, iwidth, site.c_str(), m_alphabet );
		scoreSite( logScore2, m_pModelMatDon2, m_iorderDon2, iwidth, site.c_str(), m_alphabet );
	}
	else
	{
		logScore0 = logScore1 = logScore2 = -1e12;
	}
	return false;
}

bool cModel::scoreAccSite3( double& logScore0, double& logScore1, double& logScore2, string& accSite, int& iJcnType )
{
	int iwidth = (int)accSite.length();
	if (iJcnType==INTRON_GT_AG)
	{
		scoreSite( logScore0, m_pModelMatAcc0, m_iorderAcc0, iwidth, accSite.c_str(), m_alphabet );
		scoreSite( logScore1, m_pModelMatAcc1, m_iorderAcc1, iwidth, accSite.c_str(), m_alphabet );
		scoreSite( logScore2, m_pModelMatAcc2, m_iorderAcc2, iwidth, accSite.c_str(), m_alphabet );
	}
	else if (iJcnType==INTRON_GC_AG || iJcnType==INTRON_AT_AC)
	{
		// convert to canonical donor type and score
		string site = accSite;
		site.at(18) = 'A'; site.at(19) = 'G';
		scoreSite( logScore0, m_pModelMatDon0, m_iorderDon0, iwidth, site.c_str(), m_alphabet );
		scoreSite( logScore1, m_pModelMatDon1, m_iorderDon1, iwidth, site.c_str(), m_alphabet );
		scoreSite( logScore2, m_pModelMatDon2, m_iorderDon2, iwidth, site.c_str(), m_alphabet );
	}
	else
	{
		logScore0 = logScore1 = logScore2 = -1e12;
	}
	return false;
}

bool cModel::scoreDonSites3( vector<double>& donScores0, vector<double>& donScores1, vector<double>& donScores2, vector<string>& donSite, vector<int>& iJcnTypes )
{
	if (donSite.empty())
		return false;

	donScores0.clear();
	donScores1.clear();
	donScores2.clear();
	double logScore0, logScore1, logScore2;

	int i;
	int iwidth = (int)donSite[0].length();

	for (i=0; i<(int)donSite.size(); ++i)
	{
		if (iJcnTypes[i]==INTRON_GT_AG)
		{
			scoreSite( logScore0, m_pModelMatDon0, m_iorderDon0, iwidth, donSite[i].c_str(), m_alphabet );
			scoreSite( logScore1, m_pModelMatDon1, m_iorderDon1, iwidth, donSite[i].c_str(), m_alphabet );
			scoreSite( logScore2, m_pModelMatDon2, m_iorderDon2, iwidth, donSite[i].c_str(), m_alphabet );
		}
		else
		{
			logScore0 = logScore1 = logScore2 = -1e12;
		}

		donScores0.push_back( logScore0 );
		donScores1.push_back( logScore1 );
		donScores2.push_back( logScore2 );
	}
	return false;
}

bool cModel::scoreAccSites3( vector<double>& accScores0, vector<double>& accScores1, vector<double>& accScores2, vector<string>& accSite, vector<int>& iJcnTypes )
{
	if (accSite.empty())
		return false;

	accScores0.clear();
	accScores1.clear();
	accScores2.clear();
	double logScore0, logScore1, logScore2;

	int i;
	int iwidth = (int)accSite[0].length();

	for (i=0; i<(int)accSite.size(); ++i)
	{
		if (iJcnTypes[i]==INTRON_GT_AG)
		{
			scoreSite( logScore0, m_pModelMatAcc0, m_iorderAcc0, iwidth, accSite[i].c_str(), m_alphabet );
			scoreSite( logScore1, m_pModelMatAcc1, m_iorderAcc1, iwidth, accSite[i].c_str(), m_alphabet );
			scoreSite( logScore2, m_pModelMatAcc2, m_iorderAcc2, iwidth, accSite[i].c_str(), m_alphabet );
		}
		else
		{
			logScore0 = logScore1 = logScore2 = -1e12;
		}

		accScores0.push_back( logScore0 );
		accScores1.push_back( logScore1 );
		accScores2.push_back( logScore2 );
	}
	return false;
}

bool cModel::scoreDonSitesLin3( vector<double>& donScores0, vector<double>& donScores1, vector<double>& donScores2, vector<string>& donSite, vector<int>& iJcnTypes )
{	// model is linear, so must use appropriate scoreSite fcn call
	if (donSite.empty())
		return false;

	donScores0.clear();
	donScores1.clear();
	donScores2.clear();
	double logScore0, logScore1, logScore2;

	int i;
	int iwidth = (int)donSite[0].length();

	for (i=0; i<(int)donSite.size(); ++i)
	{
		if (iJcnTypes[i]==INTRON_GT_AG)
		{
			scoreSiteLin( logScore0, m_pModelMatDon0, m_iorderDon0, iwidth, donSite[i].c_str(), m_alphabet );
			scoreSiteLin( logScore1, m_pModelMatDon1, m_iorderDon1, iwidth, donSite[i].c_str(), m_alphabet );
			scoreSiteLin( logScore2, m_pModelMatDon2, m_iorderDon2, iwidth, donSite[i].c_str(), m_alphabet );
		}
		else
		{
			logScore0 = logScore1 = logScore2 = -1e12;
		}

		donScores0.push_back( logScore0 );
		donScores1.push_back( logScore1 );
		donScores2.push_back( logScore2 );
	}
	return false;
}

bool cModel::scoreAccSitesLin3( vector<double>& accScores0, vector<double>& accScores1, vector<double>& accScores2, vector<string>& accSite, vector<int>& iJcnTypes )
{	// model is linear, so must use appropriate scoreSite fcn call
	if (accSite.empty())
		return false;

	accScores0.clear();
	accScores1.clear();
	accScores2.clear();
	double logScore0, logScore1, logScore2;

	int i;
	int iwidth = (int)accSite[0].length();

	for (i=0; i<(int)accSite.size(); ++i)
	{
		if (iJcnTypes[i]==INTRON_GT_AG)
		{
			scoreSiteLin( logScore0, m_pModelMatAcc0, m_iorderAcc0, iwidth, accSite[i].c_str(), m_alphabet );
			scoreSiteLin( logScore1, m_pModelMatAcc1, m_iorderAcc1, iwidth, accSite[i].c_str(), m_alphabet );
			scoreSiteLin( logScore2, m_pModelMatAcc2, m_iorderAcc2, iwidth, accSite[i].c_str(), m_alphabet );
		}
		else
		{
			logScore0 = logScore1 = logScore2 = -1e12;
		}

		accScores0.push_back( logScore0 );
		accScores1.push_back( logScore1 );
		accScores2.push_back( logScore2 );
	}
	return false;
}

bool cModel::scoreSiteModel( double *pModel, int iorder, vector<string>& sites, vector<double>& scores )
{
	// initialize output
	scores.clear();

	if (sites.size()==0)
		return false;

	double logScore;
	char alphabet[256];
	cMatrixMod tmpMat;
	tmpMat.SetAlphabet( alphabet );
	int iwidth = (int)sites[0].length();

	for (int is=0; is<(int)sites.size(); ++is)
	{
		scoreSite( logScore, pModel, iorder, iwidth, sites[is].c_str(), alphabet );
		scores.push_back( logScore );
	}

	return false;
}

bool cModel::scoreSite( double& score, double *mat, int order, int iwidth, const char* buf, char* alphabet )
{
	// initialize output
	score = 0.;
	double prob;

	int nDataCount = 0;
	int count = 0;
	int i = 0;
	int new_letter;
	int adr = 0;
	int adr0 = 0;
	int msize = 4<<(2*order);
	int mask = (4<<(2*order)) - 1;
	int letter_mask = 0x3;

	int NNN = 0;
	char chLetter;

	while ( nDataCount < iwidth  ) {

		if ( isalpha(*(buf+nDataCount)) != 0 ) {

			chLetter = buf[nDataCount];
			new_letter = alphabet[chLetter];

			NNN = (NNN<<2);
			NNN &= mask;
			if ( new_letter == 4 )
				NNN += 3;

			if ( i < order ) {
			
				adr = (adr<<2);
				adr += new_letter;

				if (!NNN)
				{
					prob = mat[adr];
					score += prob;
				}

				//if ( new_letter != 4 )
				//{
				//	// i should always equal zero here
				//	for (j=0; j<4; ++j) {
				//		adr0 = j;
				//		adr0 = (adr0<<2);
				//		adr0 += new_letter;
				//		assert(adr0>=0 && adr0<msize);
				//		tempScore += mat[adr0];
				//	}
				//}

			} else {

				// direct address
				adr = (adr<<2);
				adr += new_letter;
				adr &= mask;

				if ( !NNN ) {
					assert( (adr+msize*i)>=0 && (adr+msize*i)<(msize*iwidth) );
					prob = mat[adr + msize*i];
					score += prob;
					++count;
				}
			}
			++i;
		}
		++nDataCount;
	}

	return false;
}


bool cModel::scoreSiteLin( double& score, double *mat, int order, int iwidth, const char* buf, char* alphabet )
{
	// initialize output
	score = 0.;
	double prob;

	int nDataCount = 0;
	int count = 0;
	int i = 0;
	int new_letter;
	int adr = 0;
	int adr0 = 0;
	int msize = 4<<(2*order);
	int mask = (4<<(2*order)) - 1;
	int letter_mask = 0x3;

	int NNN = 0;
	char chLetter;

	double log0 = 1e-12;

	while ( nDataCount < iwidth  ) {

		if ( isalpha(*(buf+nDataCount)) != 0 ) {

			chLetter = buf[nDataCount];
			new_letter = alphabet[chLetter];

			NNN = (NNN<<2);
			NNN &= mask;
			if ( new_letter == 4 )
				NNN += 3;

			if ( i < order ) {
			
				adr = (adr<<2);
				adr += new_letter;

				if (!NNN)
				{
					prob = mat[adr];
					if (prob > 0.)
						score += log( prob );
					else
						score += log0;
				}

			} else {

				// direct address
				adr = (adr<<2);
				adr += new_letter;
				adr &= mask;

				if ( !NNN ) {
					assert( (adr+msize*i)>=0 && (adr+msize*i)<(msize*iwidth) );
					prob = mat[adr + msize*i];
					if (prob > 0.)
						score += log( prob );
					else
						score += log0;
					++count;
				}
			}
			++i;
		}
		++nDataCount;
	}

	return false;
}

bool cModel::scoreMarkovCod5( double& score, int iframe, const char* buf, char* alphabet )
{
	// first, score the initial (order+1) nt with the absolute probability model,
	//  then score the rest of the sequence using the
	int adr = 0;
	int nDataCount = 0;
	int order(5);

	int NNN = 0;
	char chLetter;
	int new_letter;
	int mask = (4<<(2*order)) - 1;

	while ( nDataCount <= order  ) {

		if ( isalpha(*(buf+nDataCount)) != 0 ) {

			chLetter = buf[nDataCount];
			new_letter = alphabet[chLetter];

			NNN = (NNN<<2);
			NNN &= mask;
			if ( new_letter == 4 )
				NNN += 3;

			adr = (adr<<2);
			adr += new_letter;
		}

		++nDataCount;
	}

	double score0;

	switch (iframe)
	{
	case 0:
		score0 = m_markovCod_5.d1a[adr];
		scoreMarkov(score, m_markovCod_5.d1, 5, buf+1, alphabet);
		break;

	case 1:
		score0 = m_markovCod_5.d2a[adr];
		scoreMarkov(score, m_markovCod_5.d2, 5, buf+1, alphabet);
		break;

	case 2:
		score0 = m_markovCod_5.d3a[adr];
		scoreMarkov(score, m_markovCod_5.d3, 5, buf+1, alphabet);
		break;

	default:
		return true;
	};

	score += score0;

	return false;
}

bool cModel::scoreMarkovCod0( double& score, int iframe, const char* buf, char* alphabet )
{
	return scoreMarkov( score, m_markovCod_0.d1, 0, buf, alphabet );
}

bool cModel::scoreMarkov( double& score, double *mat, int order, const char* buf, char* alphabet )
{
	// initialize output
	score = 0.;
	double prob;

	int nDataCount = 0;
	int count = 0;
	int i = 0;
	int new_letter;
	int adr = 0;
	int adr0 = 0;
	int msize = 4<<(2*order);
	int mask = (4<<(2*order)) - 1;
	int letter_mask = 0x3;

	int NNN = 0;
	char chLetter;
	int nLength = (int)strlen( buf );

	while ( nDataCount < nLength  ) {

		if ( isalpha(*(buf+nDataCount)) != 0 ) {

			chLetter = buf[nDataCount];
			new_letter = alphabet[chLetter];

			NNN = (NNN<<2);
			NNN &= mask;
			if ( new_letter == 4 )
				NNN += 3;

			if ( i < order ) {
			
				adr = (adr<<2);
				adr += new_letter;

			} else {

				// direct address
				adr = (adr<<2);
				adr += new_letter;
				adr &= mask;

				if ( !NNN ) {
					assert( adr>=0 && adr<msize );
					prob = mat[adr];
					score += prob;
					++count;
				}
			}
			++i;
		}
		++nDataCount;
	}

	return false;
}

bool cModel::scoreMarkov( double& score, double *mat, double *matn, int order, const char* buf, char* alphabet )
{
	// initialize output
	score = 0.;
	double prob;

	int i, a;
	int msize = 4<<(2*order);
	vector<int> adr;
	int nLength = (int)strlen( buf );

	getAddr( adr, buf, nLength, order, alphabet );

	i = nLength - 1;

	if ( !(i>=0 && i<nLength) ) {
		printf("buf='%s'\n",buf);
	}
	assert( i>=0 && i<nLength );
	a = adr[i];

	assert( a>=0 && a<msize );
	score = mat[a];

	--i;

	while ( i >= order )
	{
		assert( i>=0 && i<nLength );
		a = adr[i];

		assert( a>=0 && a<msize );
		prob = matn[a];
		score += prob;

		--i;
	}

	return false;
}

bool cModel::scoreMarkovBack( double& score, int order, const char* buf, char* alphabet )
{
	// initialize output
	score = 0.;

	double *mat, *matn;

	switch (order) {
		case 0:
			mat = m_markovNon_0.na;
			matn = m_markovNon_0.n;
			break;
		case 1:
			mat = m_markovNon_1.na;
			matn = m_markovNon_1.n;
			break;
		case 2:
			mat = m_markovNon_2.na;
			matn = m_markovNon_2.n;
			break;
		case 3:
			mat = m_markovNon_3.na;
			matn = m_markovNon_3.n;
			break;
		case 4:
			mat = m_markovNon_4.na;
			matn = m_markovNon_4.n;
			break;
		case 5:
			mat = m_markovNon_5.na;
			matn = m_markovNon_5.n;
			break;
		default:
			return true;	// error
	};

	return scoreMarkov( score, mat, matn, order, buf, alphabet );
}

bool cModel::getAddr( vector<int>& adr, const char *buf, int len, int order, char* alphabet )
{
	// Example: order = 5
	// sequence: CCAAATAA
	// adr     : {0, 0, 0, 0, 0, 2563, 2060, 48}

	adr.clear();

	char chLetter;
	int new_letter;
	int letter_mask = 0x3;
	int NNN = 0;
	int mask = (4<<(2*order)) - 1;

	int iadr = 0;

	for (int i=0; i<len; ++i)
	{
		chLetter = buf[i];
		new_letter = alphabet[chLetter];

		NNN = (NNN<<2);
		NNN &= mask;
		if ( new_letter == 4 )
			NNN += 3;

		iadr = (iadr<<2);

		//if (!NNN)
		//{
			iadr += new_letter;
			iadr &= mask;
		//}
		//else {
		//	iadr = -1;
		//}

		adr.push_back( iadr );
	}

	return false;
}

bool cModel::scoreMarkov3( double& score, int order, int fr, const char* buf, char* alphabet )
{
	switch (order) {
		case 0:
			if (m_markovCod_0.m_bRelLeft) return scoreMarkov3_Left( score, m_markovCod_0, order, fr, buf, alphabet );
			else return scoreMarkov3_Right( score, m_markovCod_0, order, fr, buf, alphabet );
		case 1:
			if (m_markovCod_1.m_bRelLeft) return scoreMarkov3_Left( score, m_markovCod_1, order, fr, buf, alphabet );
			else return scoreMarkov3_Right( score, m_markovCod_1, order, fr, buf, alphabet );
		case 2:
			if (m_markovCod_2.m_bRelLeft) return scoreMarkov3_Left( score, m_markovCod_2, order, fr, buf, alphabet );
			else return scoreMarkov3_Right( score, m_markovCod_2, order, fr, buf, alphabet );
		case 3:
			if (m_markovCod_3.m_bRelLeft) return scoreMarkov3_Left( score, m_markovCod_3, order, fr, buf, alphabet );
			else return scoreMarkov3_Right( score, m_markovCod_3, order, fr, buf, alphabet );
		case 4:
			if (m_markovCod_4.m_bRelLeft) return scoreMarkov3_Left( score, m_markovCod_4, order, fr, buf, alphabet );
			else return scoreMarkov3_Right( score, m_markovCod_4, order, fr, buf, alphabet );
		case 5:
			if (m_markovCod_5.m_bRelLeft) return scoreMarkov3_Left( score, m_markovCod_5, order, fr, buf, alphabet );
			else return scoreMarkov3_Right( score, m_markovCod_5, order, fr, buf, alphabet );
		default:
			return true;
	};

	return false;
}

bool cModel::scoreMarkov3_Left( double& score, cMarkovCod& mk, int order, int fr, const char* buf, char* alphabet )
{
	// initialize output
	score = 0.;

	int nDataCount = -1;
	int count = 0;
	int i;
	vector<int> adr;
	int a;

	int msize = 4<<(2*order);
	int mask = (4<<(2*order)) - 1;

	double *mat;
	int nLength = (int)strlen( buf );

	getAddr( adr, buf, nLength, order, alphabet );

	i = nLength - 1;

	//FILE *pf = fopen("fr_debug.txt","a");
	//fprintf(pf, "start: fr = %d\n", fr);

	switch (fr) {
		case 2:
			mat = mk.d2a; fr = 1;
			break;
		case 3:
			mat = mk.d3a; fr = 2;
			break;
		default:		// fr 1
			mat = mk.d1a; fr = 3;
	}

	if (!(i>=0 && i<nLength))
		return true;
	//assert( i>=0 && i<nLength );
	a = adr[i];

	if( a>=0 && a<msize )
		score = mat[a];

	--i;
	
	while ( i >= order )
	{
		switch (fr) {
			case 2:
				mat = mk.d2; fr = 1;
				break;
			case 3:
				mat = mk.d3; fr = 2;
				break;
			default:		// fr 1
				mat = mk.d1; fr = 3;
		}

		assert( i>=0 && i<nLength );
		a = adr[i];

		if ( a>=0 && a<msize ) {
			score += mat[a];
		}

		//fprintf(pf, "fr = %d, prob == %f, adr = %d, score = %f\n", fr, prob, a, score);

		--i;
	}

	//fclose(pf);

	return false;
}

bool cModel::scoreMarkov3_Right( double& score, cMarkovCod& mk, int order, int fr, const char* buf, char* alphabet )
{
	// initialize output
	score = 0.;

	int nDataCount = -1;
	int count = 0;
	int i;
	vector<int> adr;
	int a;

	int msize = 4<<(2*order);
	int mask = (4<<(2*order)) - 1;

	double *mat;
	int nLength = (int)strlen( buf );

	getAddr( adr, buf, nLength, order, alphabet );

	i = order;

	switch (fr) {
		case 2:
			mat = mk.d2a; fr = 3;
			break;
		case 3:
			mat = mk.d3a; fr = 1;
			break;
		default:		// fr 1
			mat = mk.d1a; fr = 2;
	}

	if (!(i>=0 && i<nLength))
		return true;
	a = adr[i];

	if( a>=0 && a<msize )
		score = mat[a];

	++i;
	
	while ( i < nLength )
	{
		switch (fr) {
			case 2:
				mat = mk.d2; fr = 3;
				break;
			case 3:
				mat = mk.d3; fr = 1;
				break;
			default:		// fr 1
				mat = mk.d1; fr = 2;
		}

		assert( i>=0 && i<nLength );
		a = adr[i];

		if ( a>=0 && a<msize ) {
			score += mat[a];
		}

		++i;
	}

	return false;
}

bool cModel::scoreMarkovNon( double& score, int order, const char* buf, char* alphabet )
{
	switch (order) {
		case 0:
			if (m_markovCod_0.m_bRelLeft) return scoreMarkovNon_Left( score, m_markovNon_0, 0, buf, alphabet );
			else return scoreMarkovNon_Right( score, m_markovNon_0, 0, buf, alphabet );
		case 1:
			if (m_markovCod_1.m_bRelLeft) return scoreMarkovNon_Left( score, m_markovNon_1, 1, buf, alphabet );
			else return scoreMarkovNon_Right( score, m_markovNon_1, 1, buf, alphabet );
		case 2:
			if (m_markovCod_2.m_bRelLeft) return scoreMarkovNon_Left( score, m_markovNon_2, 2, buf, alphabet );
			else return scoreMarkovNon_Right( score, m_markovNon_2, 2, buf, alphabet );
		case 3:
			if (m_markovCod_3.m_bRelLeft) return scoreMarkovNon_Left( score, m_markovNon_3, 3, buf, alphabet );
			else return scoreMarkovNon_Right( score, m_markovNon_3, 3, buf, alphabet );
		case 4:
			if (m_markovCod_4.m_bRelLeft) return scoreMarkovNon_Left( score, m_markovNon_4, 4, buf, alphabet );
			else return scoreMarkovNon_Right( score, m_markovNon_4, 4, buf, alphabet );
		case 5:
			if (m_markovCod_5.m_bRelLeft) return scoreMarkovNon_Left( score, m_markovNon_5, 5, buf, alphabet );
			else return scoreMarkovNon_Right( score, m_markovNon_5, 5, buf, alphabet );
		default:
			return true;
	};

	return false;
}

bool cModel::scoreMarkovNon_Left( double& score, cMarkovNon& mk, int order, const char* buf, char* alphabet )
{
	// initialize output
	score = 0.;

	int nDataCount = -1;
	int count = 0;
	int i;
	vector<int> adr;
	int a;

	int msize = 4<<(2*order);
	int mask = (4<<(2*order)) - 1;

	double *mat;
	int nLength = (int)strlen( buf );

	getAddr( adr, buf, nLength, order, alphabet );

	i = nLength - 1;
	mat = mk.na;

	if (!(i>=0 && i<nLength))
		return true;

	a = adr[i];

	if( a>=0 && a<msize )
		score = mat[a];

	--i;
	mat = mk.n;
	
	while ( i >= order )
	{
		assert( i>=0 && i<nLength );
		a = adr[i];

		if ( a>=0 && a<msize ) {
			score += mat[a];
		}

		--i;
	}

	return false;
}

bool cModel::scoreMarkovNon_Right( double& score, cMarkovNon& mk, int order, const char* buf, char* alphabet )
{
	// initialize output
	score = 0.;

	int nDataCount = -1;
	int count = 0;
	int i;
	vector<int> adr;
	int a;

	int msize = 4<<(2*order);
	int mask = (4<<(2*order)) - 1;

	double *mat;
	int nLength = (int)strlen( buf );

	getAddr( adr, buf, nLength, order, alphabet );

	i = order;
	mat = mk.na;

	if (!(i>=0 && i<nLength))
		return true;

	a = adr[i];

	if( a>=0 && a<msize )
		score = mat[a];

	++i;
	mat = mk.n;
	
	while ( i < nLength )
	{
		assert( i>=0 && i<nLength );
		a = adr[i];

		if ( a>=0 && a<msize ) {
			score += mat[a];
		}

		++i;
	}

	return false;
}

bool cModel::scoreMarkov3( double& score, double *mat1, double *mat2, double *mat3, int order, int fr, const char* buf, char* alphabet )
{
	// initialize output
	score = 0.;
	double prob;

	int nDataCount = 0;
	int count = 0;
	int i = 0;
	int new_letter;
	int adr = 0;
	int adr0 = 0;
	int msize = 4<<(2*order);
	int mask = (4<<(2*order)) - 1;
	int letter_mask = 0x3;

	double *mat;

	int NNN = 0;
	char chLetter;
	int nLength = (int)strlen( buf );

	//int next = fr + order - 1;
	//int next = fr - 1;

	while ( nDataCount < nLength  ) {

		if ( isalpha(*(buf+nDataCount)) != 0 ) {

			chLetter = buf[nDataCount];
			new_letter = alphabet[chLetter];

			NNN = (NNN<<2);
			NNN &= mask;
			if ( new_letter == 4 )
				NNN += 3;

			if ( i < order ) {
			
				adr = (adr<<2);
				adr += new_letter;

			} else {

				switch (fr) {
					case 2:
						mat = mat2; fr = 3;
						break;
					case 3:
						mat = mat3; fr = 1;
						break;
					default:		// fr 1
						mat = mat1; fr = 2;
				}

				// direct address
				adr = (adr<<2);
				adr += new_letter;
				adr &= mask;

				if ( !NNN ) {
					assert( adr>=0 && adr<msize );
					prob = mat[adr];
					score += prob;
					++count;
				}

			}
			++i;
		}
		++nDataCount;
	}

	return false;
}

//
//int cMatrixMod::Calc0( double* mat, const char* buf, char* alphabet, int order, int fsize, bool bAddCompl )
//{
//	int nDataCount = 0;
//	int count = 0;
//	int i = 0, j;
//	int new_letter;
//	int letter_shift;
//	int adr = 0, adr_rc = 0;
//	int adr0 = 0;
//	int msize = 4<<(2*order);
//	int mask = (4<<(2*order)) - 1;
//	int letter_mask = 0x3;
//	int compl_letter_mask = 0x3;
//
//	char chLetter;
//
//	for (j=0; j<order; ++j) {
//		compl_letter_mask = compl_letter_mask<<(2*order);
//	}
//
//
//	int NNN = 0;
//
//	while ( nDataCount < fsize  ) {
//
//		if ( isalpha(*(buf+nDataCount)) != 0 ) {
//
//			chLetter = buf[nDataCount];
//			new_letter = alphabet[chLetter];
//
//			NNN = (NNN<<2);
//			NNN &= mask;
//			if ( new_letter == 4 )
//				NNN += 3;
//
//			if ( i < order ) {
//			
//				adr = (adr<<2);
//				adr += new_letter;
//
//				if ( new_letter != 4 )
//				{
//					// i should always equal zero here
//					for (j=0; j<4; ++j) {
//						adr0 = j;
//						adr0 = (adr0<<2);
//						adr0 += new_letter;
//						assert(adr0>=0 && adr0<msize);
//						//{
//						//	printf("adr0 = %d, msize = %d, letter = %s\n", adr0, msize, chLetter);
//						//}
//						++mat[adr0];
//					}
//				}
//
//			} else {
//
//				// direct address
//				adr = (adr<<2);
//				adr += new_letter;
//				adr &= mask;
//
//				// reverse complement address
//				letter_shift = (letter_mask & (~new_letter))<<(2*order);
//				adr_rc = (adr_rc>>2);
//				adr_rc += letter_shift;			// set msb position to letter_shift
//
//				if ( !NNN ) {
//					assert( (adr+msize*i)>=0 && (adr+msize*i)<(msize*fsize) );
//					++mat[adr + msize*i];
//					++count;
//					if (bAddCompl) {
//						assert(adr_rc>=0 && adr_rc<(msize*fsize));
//						++mat[adr_rc];
//						++count;
//					}
//				}
//			}
//			++i;
//		}
//		++nDataCount;
//	}
//
//	return count;
//}


bool cModel::calcSiteModel( double **ppModelMat, vector<string>& site, int& iorder, int nCols, int flankLeft, int flankRight )
{
	if (site.size()==0)
	{
		report("No sites found?");
		return false;
	}

	if ((int)site.size()<m_orderOneCutoff)	{ iorder = 0; }
	else								{ iorder = 1; }

	int nModelSize = 4<<(2*iorder);

	cMatrixMod tmpMat;
	ppModelMat[0] = new double[nModelSize * nCols];
	tmpMat.Calc( ppModelMat[0], site, iorder, nCols, flankLeft, flankRight );

	return false;
}

bool cModel::calcSiteModel0( double **ppModelMat, vector<string>& site, int nCols, int flankLeft, int flankRight )
{
	// zero order model

	if (site.size()==0)
	{
		report("No sites found?");
		return false;
	}

	cMatrixMod tmpMat;
	ppModelMat[0] = new double[4 * nCols];
	tmpMat.Calc( ppModelMat[0], site, 0, flankLeft, flankRight );

	return false;
}

bool cModel::normalizeModel( void *func, double *pModelMat, int iorder, int nCols,
							cMarkovCod *pCod0, cMarkovNon *pNon0, cMarkovCod *pCod1, cMarkovNon *pNon1 )
{
	int msize = 4<<(2*iorder);

	double *pModelOld = new double[msize*nCols];
	memcpy( pModelOld, pModelMat, msize*nCols*sizeof(double) );
	//if (iorder==0)
	//	(func)( pModelMat, pModelOld, iorder, pCod0, pNon0 );
	//else if (iorder==1)
	//	(func)( pModelMat, pModelOld, iorder, pCod1, pNon1 );
	//else
	//{
	//	report("Unsupported site model order > 1");
	//	reeturn true;
	//}

	if (pModelOld!=NULL) {
		delete []pModelOld;
		pModelOld = NULL;
	}

	return false;
}


void cModel::appendGenes( vector<codingRegion>& genes )
{
	m_genes.insert( m_genes.end(), genes.begin(), genes.end() );
}

void cModel::clearData()
{
	m_genes.clear();

	m_seqStarts.clear();
	m_seqTerms.clear();
	m_seqDons.clear();
	m_seqAccs.clear();
	m_splicePhase.clear();

	// coding and non-coding regions
	m_strCoding.clear();
	m_strIntrons.clear();
	m_strNonCoding.clear();
}



bool cModel::normalizeModelDon( double *pMatOut, double *pModelMatDon, int iphase, int iorderDon, cMarkovCod *pmarkovCod, cMarkovNon *pmarkovNon )
{
	int len = m_pSite->getDonSiteLength();

	if ( pModelMatDon==NULL )
		return true;			// error

	// normalize region left of donor with coding probabilities starting at phase iphase
	if (normalizeModelCod( pMatOut, pModelMatDon, 0, m_donFlankLeft-1, /*iphase*/ (iphase+2+(1-iorderDon)) % 3, iorderDon, pmarkovCod ))
		return true;

	// normalize region right of donor with non-coding probabilities
	if (normalizeModelNon( pMatOut, pModelMatDon, len-m_donFlankRight, len-1, iorderDon, pmarkovNon ))
		return true;

	int ikey, ixT = 3, icol = len-m_donFlankRight;
	if (iorderDon==1)
	{
		for (ikey=0; ikey<16; ++ikey) {
			// adjust the column immediately downstream of canonical site
			pMatOut[icol*16 + ikey] = pMatOut[icol*16 + ikey] * m_markovNon_0.m_mkModel[ixT];
			// adjust first column
			pMatOut[ikey] *= 4.;
		}
	}

	// inside donor - NOP

	return false;				// success
}

bool cModel::normalizeModelAcc( double *pMatOut, double *pModelMatAcc, int iphase, int iorderAcc, cMarkovCod *pmarkovCod, cMarkovNon *pmarkovNon )
{
	int len = m_pSite->getAccSiteLength();

	if ( pModelMatAcc==NULL )
		return true;			// error

	// normalize region left of acceptor with non-coding probabilities
	if (normalizeModelNon( pMatOut, pModelMatAcc, 0, m_accFlankLeft-1, iorderAcc, pmarkovNon ))
		return true;

	// normalize region right of acceptor with coding probabilities starting at phase iphase
	if (normalizeModelCod( pMatOut, pModelMatAcc, len-m_accFlankRight, len-1, iphase, iorderAcc, pmarkovCod ))
		return true;

	double *pCod;
	switch (iphase) {
		case 0:
			pCod = pmarkovCod->m_mkModel_d1;
			break;
		case 1:
			pCod = pmarkovCod->m_mkModel_d2;
			break;
		case 2:
			pCod = pmarkovCod->m_mkModel_d3;
			break;
	};

	int ikey, ixG = 2, icol = len-m_accFlankRight;
	if (iorderAcc==1)
	{
		for (ikey=0; ikey<16; ++ikey) {
			// adjust the column immediately downstream of canonical site
			pMatOut[icol*16 + ikey] = pMatOut[icol*16 + ikey] * pCod[ixG];
			// adjust first column
			pMatOut[ikey] *= 4.;
		}
	}

	// inside acceptor - NOP

	return false;				// success
}

bool cModel::unNormalizeModelDon( double *pMatOut, double *pModelMatDon, int iphase, int iorderDon, cMarkovCod *pmarkovCod, cMarkovNon *pmarkovNon )
{
	int len = m_pSite->getDonSiteLength();

	if ( pModelMatDon==NULL )
		return true;			// error

	// normalize region left of donor with coding probabilities starting at phase iphase

	if (unNormalizeModelCod( pMatOut, pModelMatDon, 0, m_donFlankLeft-1, (iphase+2+(1-iorderDon)) % 3, iorderDon, pmarkovCod ))
		return true;

	// normalize region right of donor with non-coding probabilities
	if (unNormalizeModelNon( pMatOut, pModelMatDon, len-m_donFlankRight, len-1, iorderDon, pmarkovNon ))
		return true;

	int ikey, ixT = 3, icol = len-m_donFlankRight;
	if (iorderDon==1)
	{
		for (ikey=0; ikey<16; ++ikey) {
			// adjust the column immediately downstream of canonical site
			pMatOut[icol*16 + ikey] = pMatOut[icol*16 + ikey] / m_markovNon_0.m_mkModel[ixT];
			// adjust first column
			pMatOut[ikey] /= 4.;
		}
	}

	// inside donor - NOP

	return false;				// success
}

bool cModel::unNormalizeModelAcc( double *pMatOut, double *pModelMatAcc, int iphase, int iorderAcc, cMarkovCod *pmarkovCod, cMarkovNon *pmarkovNon )
{
	int len = m_pSite->getAccSiteLength();

	if ( pModelMatAcc==NULL )
		return true;			// error

	// normalize region left of acceptor with non-coding probabilities
	if (unNormalizeModelNon( pMatOut, pModelMatAcc, 0, m_accFlankLeft-1, iorderAcc, pmarkovNon ))
		return true;

	// normalize region right of acceptor with coding probabilities starting at phase iphase
	if (unNormalizeModelCod( pMatOut, pModelMatAcc, len-m_accFlankRight, len-1, iphase, iorderAcc, pmarkovCod ))
		return true;

	double *pCod;
	switch (iphase) {
		case 0:
			pCod = pmarkovCod->m_mkModel_d1;
			break;
		case 1:
			pCod = pmarkovCod->m_mkModel_d2;
			break;
		case 2:
			pCod = pmarkovCod->m_mkModel_d3;
			break;
	};

	int ikey, ixG = 2, icol = len-m_accFlankRight;
	if (iorderAcc==1)
	{
		for (ikey=0; ikey<16; ++ikey) {
			// adjust the column immediately downstream of canonical site
			pMatOut[icol*16 + ikey] = pMatOut[icol*16 + ikey] / pCod[ixG];
			// adjust first column
			pMatOut[ikey] /= 4.;
		}
	}

	// inside acceptor - NOP

	return false;				// success
}

bool cModel::normalizeModelNon( double *pMatOut, double *pModelMat, int ibegin, int iend, int iorder, cMarkovNon *pmarkovNon )
{
	int i, ikey, ix;

	if (ibegin>iend)
		return false;		// do nothing

	double *pModelNon = pmarkovNon->m_mkModel;

	if ( pModelNon==NULL )
		return true;			// error

	int nKeys = 4<<(2*iorder);

	if (ibegin==0 && iorder==1)
	{
		for (ikey=0; ikey<nKeys; ++ikey) {
			ix = (int) ( fmod( double(ikey), 4 ) );
			pMatOut[ikey] = pModelMat[ikey] / m_markovNon_0.m_mkModel[ix];
		}
		++ibegin;
	}

	// left of ATG - non-coding region
	for (i=ibegin; i<=iend; ++i) {
		for (ikey=0; ikey<nKeys; ++ikey) {
			ix = i*nKeys + ikey;
//			j = (int) fmod( double(ikey)/(double(nKeys)/4.), 4. );
//			pMatOut[ix] = pModelMat[ix] / pModelNon[j];
			pMatOut[ix] = pModelMat[ix] / pModelNon[ikey];
		}
	}

	return false;				// success
}

bool cModel::unNormalizeModelNon( double *pMatOut, double *pModelMat, int ibegin, int iend, int iorder, cMarkovNon *pmarkovNon )
{
	int i, ikey, ix;

	if (ibegin>iend)
		return false;		// do nothing

	double *pModelNon = pmarkovNon->m_mkModel;

	if ( pModelNon==NULL )
		return true;			// error

	int nKeys = 4<<(2*iorder);

	if (ibegin==0 && iorder==1)
	{
		for (ikey=0; ikey<nKeys; ++ikey) {
			ix = (int) ( fmod( double(ikey), 4 ) );
			pMatOut[ikey] = pModelMat[ikey] * m_markovNon_0.m_mkModel[ix];
		}
		++ibegin;
	}

	// left of ATG - non-coding region
	for (i=ibegin; i<=iend; ++i) {
		for (ikey=0; ikey<nKeys; ++ikey) {
			ix = i*nKeys + ikey;
//			j = (int) fmod( double(ikey)/(double(nKeys)/4.), 4. );
//			pMatOut[ix] = pModelMat[ix] / pModelNon[j];
			pMatOut[ix] = pModelMat[ix] * pModelNon[ikey];
		}
	}

	return false;				// success
}

bool cModel::normalizeModelCod( double *pMatOut, double *pModelMat, int ibegin, int iend, int iphase, int iorder, cMarkovCod *pmarkovCod )
{
	int i, ikey, ix;
	int ix_begin;

	if (ibegin>iend)
		return false;		// do nothing

	double *pModelCod, *pModel0;
	
	if ( pModelMat==NULL || pmarkovCod==NULL || iphase<0 || iphase>2 )
		return true;			// error

	switch (iphase) {
		case 0:
			pModelCod = pmarkovCod->m_mkModel_d1;
			//pModel0 = m_markovCod_0.m_mkModel_d1;
			break;
		case 1:
			pModelCod = pmarkovCod->m_mkModel_d2;
			//pModel0 = m_markovCod_0.m_mkModel_d2;
			break;
		case 2:
			pModelCod = pmarkovCod->m_mkModel_d3;
			//pModel0 = m_markovCod_0.m_mkModel_d3;
			break;
	};

	switch ( (iphase+2) % 3 )
	{
		case 0:
			pModel0 = m_markovCod_0.m_mkModel_d1;
			break;
		case 1:
			pModel0 = m_markovCod_0.m_mkModel_d2;
			break;
		case 2:
			pModel0 = m_markovCod_0.m_mkModel_d3;
			break;
	};

	int nKeys = 4<<(2*iorder);

	if (iorder==1)
	{
		ix_begin = ibegin*nKeys;
		for (ikey=0; ikey<nKeys; ++ikey) {
			ix = (int) ( fmod( double(ikey), 4 ) );
			pMatOut[ix_begin+ikey] = pModelMat[ix_begin+ikey] / pModel0[ix];
		}
		++ibegin;
	}

	// normalize by coding region probabilities
	for (i=ibegin; i<=iend; ++i) {
		switch (iphase) {
			case 0:
				pModelCod = pmarkovCod->m_mkModel_d1;
				break;
			case 1:
				pModelCod = pmarkovCod->m_mkModel_d2;
				break;
			case 2:
				pModelCod = pmarkovCod->m_mkModel_d3;
				break;
		};
		for (ikey=0; ikey<nKeys; ++ikey) {
			ix = i*nKeys + ikey;
			pMatOut[ix] = pModelMat[ix] / pModelCod[ikey];
		}
		++iphase;
		if (iphase>2)
			iphase = 0;
	}

	return false;				// success
}

bool cModel::unNormalizeModelCod( double *pMatOut, double *pModelMat, int ibegin, int iend, int iphase, int iorder, cMarkovCod *pmarkovCod )
{
	int i, ikey, ix;
	int ix_begin;

	if (ibegin>iend)
		return false;		// do nothing

	double *pModelCod, *pModel0;
	
	if ( pModelMat==NULL || pmarkovCod==NULL || iphase<0 || iphase>2 )
		return true;			// error

	switch (iphase) {
		case 0:
			pModelCod = pmarkovCod->m_mkModel_d1;
			//pModel0 = m_markovCod_0.m_mkModel_d1;
			break;
		case 1:
			pModelCod = pmarkovCod->m_mkModel_d2;
			//pModel0 = m_markovCod_0.m_mkModel_d2;
			break;
		case 2:
			pModelCod = pmarkovCod->m_mkModel_d3;
			//pModel0 = m_markovCod_0.m_mkModel_d3;
			break;
	};

	switch ( (iphase+2) % 3 )
	{
		case 0:
			pModel0 = m_markovCod_0.m_mkModel_d1;
			break;
		case 1:
			pModel0 = m_markovCod_0.m_mkModel_d2;
			break;
		case 2:
			pModel0 = m_markovCod_0.m_mkModel_d3;
			break;
	};

	int nKeys = 4<<(2*iorder);

	if (iorder==1)
	{
		ix_begin = ibegin*nKeys;
		for (ikey=0; ikey<nKeys; ++ikey) {
			ix = (int) ( fmod( double(ikey), 4 ) );
			pMatOut[ix_begin+ikey] = pModelMat[ix_begin+ikey] * pModel0[ix];
		}
		++ibegin;
	}

	// normalize by coding region probabilities
	for (i=ibegin; i<=iend; ++i) {
		switch (iphase) {
			case 0:
				pModelCod = pmarkovCod->m_mkModel_d1;
				break;
			case 1:
				pModelCod = pmarkovCod->m_mkModel_d2;
				break;
			case 2:
				pModelCod = pmarkovCod->m_mkModel_d3;
				break;
		};
		for (ikey=0; ikey<nKeys; ++ikey) {
			ix = i*nKeys + ikey;
			pMatOut[ix] = pModelMat[ix] * pModelCod[ikey];
		}
		++iphase;
		if (iphase>2)
			iphase = 0;
	}

	return false;				// success
}


void cModel::createHeurSiteModels()
{
	// ATG
	if (m_pModelMatAtg != NULL)  delete []m_pModelMatAtg;
	m_pModelMatAtg = new double[4*3];
	memset( m_pModelMatAtg, 0, 4*3*sizeof(double) );
	m_pModelMatAtg[0] = m_pModelMatAtg[7] = m_pModelMatAtg[10] = 1.;

	// TAA
	if (m_pModelMatTaa != NULL)  delete []m_pModelMatTaa;
	m_pModelMatTaa = new double[4*3];
	memset( m_pModelMatTaa, 0, 4*3*sizeof(double) );
	m_pModelMatTaa[3] = m_pModelMatTaa[4] = m_pModelMatTaa[8] = 1.;

	// TAG
	if (m_pModelMatTag != NULL)  delete []m_pModelMatTag;
	m_pModelMatTag = new double[4*3];
	memset( m_pModelMatTag, 0, 4*3*sizeof(double) );
	m_pModelMatTag[3] = m_pModelMatTag[4] = m_pModelMatTag[10] = 1.;

	// TGA
	if (m_pModelMatTga != NULL)  delete []m_pModelMatTga;
	m_pModelMatTga = new double[4*3];
	memset( m_pModelMatTga, 0, 4*3*sizeof(double) );
	m_pModelMatTga[3] = m_pModelMatTga[6] = m_pModelMatTga[8] = 1.;

	// GT
	if (m_pModelMatDon0 != NULL)  delete []m_pModelMatDon0;
	if (m_pModelMatDon1 != NULL)  delete []m_pModelMatDon1;
	if (m_pModelMatDon2 != NULL)  delete []m_pModelMatDon2;
	m_pModelMatDon0 = new double[4*2];
	m_pModelMatDon1 = new double[4*2];
	m_pModelMatDon2 = new double[4*2];
	memset( m_pModelMatDon0, 0, 4*2*sizeof(double) );
	m_pModelMatDon0[2] = m_pModelMatDon0[7] = 1.;
	memcpy( m_pModelMatDon1, m_pModelMatDon0, 4*2*sizeof(double) );
	memcpy( m_pModelMatDon2, m_pModelMatDon0, 4*2*sizeof(double) );

	// AG
	if (m_pModelMatAcc0 != NULL)  delete []m_pModelMatAcc0;
	if (m_pModelMatAcc1 != NULL)  delete []m_pModelMatAcc1;
	if (m_pModelMatAcc2 != NULL)  delete []m_pModelMatAcc2;
	m_pModelMatAcc0 = new double[4*2];
	m_pModelMatAcc1 = new double[4*2];
	m_pModelMatAcc2 = new double[4*2];
	memset( m_pModelMatAcc0, 0, 4*2*sizeof(double) );
	m_pModelMatAcc0[0] = m_pModelMatAcc0[6] = 1.;
	memcpy( m_pModelMatAcc1, m_pModelMatAcc0, 4*2*sizeof(double) );
	memcpy( m_pModelMatAcc2, m_pModelMatAcc0, 4*2*sizeof(double) );
}




bool cModel::calcLenHistograms(bool bCalcHist, bool bFilter)
{
//histogram.pl --in ini_ln           --com kw  --out exon_ini_LD   --min 3    --max 10000  --smooth3
//histogram.pl --in dna.fa.good.inx  --com kw  --out exon_LD       --min 3    --max 10000  --smooth3
//histogram.pl --in ter_ln           --com kw  --out exon_ter_LD   --min 3    --max 10000  --smooth3
//histogram.pl --in cDNA_ln          --com kw  --out exon_sing_LD  --min 300  --max 10000  --scale
//histogram.pl --in dna.fa.good.int  --com k   --out intron_LD     --min 20   --max 10000
//histogram.pl --in  intg_ln         --com kw  --out interg_LD     --min 20   --max 10000	int ig, ie, len, nExons;

	intHist idistrExonSingLen, idistrExonIniLen, idistrExonTermLen, idistrIntExonLen;
	intHist idistrIntronLen, idistrIntergenicLen;

	int lenIni, lenTerm;
	int lenCoding;

	int len;

	FILE *pfCds;
	FILE *pfIniExon;
	FILE *pfTerExon;
	FILE *pfIntExon;
	FILE *pfIntron;
	FILE *pfInterg;

	if (m_bDebug)
	{
		string strCdsFile = m_debugOutDir + "cds_ln.txt";
		string strIniExonFile = m_debugOutDir + "ini_exon_ln.txt";
		string strTerExonFile = m_debugOutDir + "ter_exon_ln.txt";
		string strIntExonFile = m_debugOutDir + "int_exon_ln.txt";
		string strInronFile = m_debugOutDir + "intron_ln.txt";
		string strIntergFile = m_debugOutDir + "interg_ln.txt";

		pfCds = fopen( strCdsFile.c_str(), "w" );
		pfIniExon = fopen( strIniExonFile.c_str(), "w" );
		pfTerExon = fopen( strTerExonFile.c_str(), "w" );
		pfIntExon = fopen( strIntExonFile.c_str(), "w" );
		pfIntron = fopen( strInronFile.c_str(), "w" );
		pfInterg = fopen( strIntergFile.c_str(), "w" );
		if (pfCds==NULL || pfIniExon==NULL || pfTerExon==NULL || pfIntExon==NULL || pfIntron==NULL || pfInterg==NULL)
		{
			printf("Unable to open all files for writing length data.\n");
			return true;
		}
	}

//enum { EXON_TYPE_SINGLE, EXON_TYPE_INITIAL, EXON_TYPE_INTERNAL, EXON_TYPE_TERMINAL };

	int nExons;
	for (int ig=0; ig<(int)m_genes.size(); ++ig) {

		nExons = (int)m_genes[ig].nType.size();
		lenCoding = 0;
		lenIni = 0;
		lenTerm = 0;

		for (int ie=0; ie<nExons; ++ie) {

			// exon length
			len = m_genes[ig].boundaries[ie].iend - m_genes[ig].boundaries[ie].ibeg + 1;
			lenCoding += len;
			if (!m_genes[ig].bStatus && bFilter)
				continue;		// gene is too short... we will not allow it to contribute to exon LD stats
 
			// update exon type histogram accordingly
			switch (m_genes[ig].nType[ie]) {
				case EXON_TYPE_SINGLE:
					// NOP - hist will be updated with lenCoding below
					break;
				case EXON_TYPE_INITIAL:
					lenIni = len;
					if (m_genes[ig].boundaries[ie].extr_L==-1 || m_genes[ig].boundaries[ie].extr_R==-1) continue;
					m_hist.updateHistogram( idistrExonIniLen, lenIni );
					if (m_bDebug) { fprintf( pfIniExon, "%d\n", lenIni ); }
					// intron hist update
					if (!m_genes[ig].bStrand && ie>0) {
						len = m_genes[ig].boundaries[ie].ibeg - m_genes[ig].boundaries[ie-1].iend - 1;
						m_hist.updateHistogram( idistrIntronLen, len );
						if (m_bDebug) { fprintf( pfIntron, "%d\n", len ); }
					}
					break;
				case EXON_TYPE_INTERNAL:
					// update hist for internal exons
					if (m_genes[ig].boundaries[ie].extr_L==-1 || m_genes[ig].boundaries[ie-1].extr_R==-1) continue;
					m_hist.updateHistogram( idistrIntExonLen, len );
					if (m_bDebug) { fprintf( pfIntExon, "%d\n", len ); }
					// intron hist update
					if (ie>0) {
					len = m_genes[ig].boundaries[ie].ibeg - m_genes[ig].boundaries[ie-1].iend - 1;
						m_hist.updateHistogram( idistrIntronLen, len );
						if (m_bDebug) { fprintf( pfIntron, "%d\n", len ); }
					}
					break;
				case EXON_TYPE_TERMINAL:
					// exon hist update
					lenTerm = len;
					if (m_genes[ig].boundaries[ie].extr_L==-1 || m_genes[ig].boundaries[ie].extr_R==-1) continue;
					m_hist.updateHistogram( idistrExonTermLen, lenTerm );
					if (m_bDebug) { fprintf( pfTerExon, "%d\n", lenTerm ); }
					// intron hist update
					if (m_genes[ig].bStrand && ie>0) {
						len = m_genes[ig].boundaries[ie].ibeg - m_genes[ig].boundaries[ie-1].iend - 1;
						m_hist.updateHistogram( idistrIntronLen, len );
						if (m_bDebug) { fprintf( pfIntron, "%d\n", len ); }
					}
					break;
			};

		}	// exon loop

		m_hist.updateHistogram( idistrExonSingLen, lenCoding );
		if (m_bDebug) { fprintf( pfCds, "%d\n", lenCoding ); }

		if (!m_genes[ig].bStatus && bFilter)
			continue;		// gene is too short... we will not allow it to contribute to anything other than CDS LD

		// intergeneic length histogram
		if (ig>0 && !m_bExtr) {
			if (m_genes[ig-1].nContig == m_genes[ig].nContig)
			{
				nExons = (int)m_genes[ig-1].boundaries.size();
				len = m_genes[ig].boundaries[0].ibeg - m_genes[ig-1].boundaries[nExons-1].iend - 1;
				// avoid "negative" length jumps between contigs
				if (len>0) {
					m_hist.updateHistogram( idistrIntergenicLen, len );
					if (m_bDebug) { fprintf( pfInterg, "%d\n", len ); }
				}
			}
		}

	}	// gene loop

	if (m_bDebug)
	{
		fclose( pfCds );
		fclose( pfIniExon );
		fclose( pfTerExon );
		fclose( pfIntExon );
		fclose( pfIntron );
		fclose( pfInterg );
	}

	if (bCalcHist)
	{
		// clear old data
		m_distrExonSingLen.clear();
		m_distrExonIniLen.clear();
		m_distrExonTermLen.clear();
		m_distrExonIntLen.clear();
		m_distrIntronLen.clear();
		m_distrIntergenicLen.clear();

		doubleHist tmpHist;

		//histogram.pl --in ini_ln           --com kw  --out exon_ini_LD   --min 3    --max 10000  --smooth3
		//histogram.pl --in dna.fa.good.inx  --com kw  --out exon_LD       --min 3    --max 10000  --smooth3
		//histogram.pl --in ter_ln           --com kw  --out exon_ter_LD   --min 3    --max 10000  --smooth3
		//histogram.pl --in cDNA_ln          --com kw  --out exon_sing_LD  --min 300  --max 10000  --scale
		//histogram.pl --in dna.fa.good.int  --com k   --out intron_LD     --min 20   --max 10000
		//histogram.pl --in  intg_ln         --com kw  --out interg_LD     --min 20   --max 10000
		// conver to double and smooth the data
		string str;
		if (!idistrExonSingLen.empty())
		{
			str = m_debugOutDir + "cds";
			m_hist.getSmoothHist( m_distrExonSingLen, idistrExonSingLen, 100, 10000, 0, false, true, str );
		}
		if (!idistrExonSingLen.empty())
		{
			str = m_debugOutDir + "ini_exon";
			m_hist.getSmoothHist( m_distrExonIniLen, idistrExonIniLen, 1, 10000, 0, true, false, str );
		}
		if (!idistrExonSingLen.empty())
		{
			str = m_debugOutDir + "ter_exon";
			m_hist.getSmoothHist( m_distrExonTermLen, idistrExonTermLen, 1, 10000, 0, true, false, str );
		}
		if (!idistrExonSingLen.empty())
		{
			str = m_debugOutDir + "int_exon";
			m_hist.getSmoothHist( m_distrExonIntLen, idistrIntExonLen, 1, 10000, 0, true, false, str );
		}
		if (!idistrExonSingLen.empty())
		{
			str = m_debugOutDir + "intron";
			m_hist.getSmoothHist( m_distrIntronLen, idistrIntronLen, 20, 10000, 1, false, false, str );
		}
		if (!idistrExonSingLen.empty())
		{
			str = m_debugOutDir + "interg";
			m_hist.getSmoothHist( m_distrIntergenicLen, idistrIntergenicLen, 20, 30000, 0, false, false, str );
		}

	}

	return false;
}
