//
// cSite class definition, for site extraction and manipulation.
// Used in conjunction with HMM, which determines gene boundaries.
//

#pragma once

#include <string>
#include <vector>
#include "geneType.h"
#include "sseq.h"
#include "markov.h"
#include "defines.h"

using std::vector;
using std::string;

class cSite {
public:
	cSite() { m_psseq = NULL;
			  m_pModelMatAtg = NULL;
			  m_pModelMatTaa = m_pModelMatTag = m_pModelMatTga = NULL;
			  m_pModelMatDon0 = m_pModelMatDon1 = m_pModelMatDon2 = NULL;
			  m_pModelMatAcc0 = m_pModelMatAcc1 = m_pModelMatAcc2 = NULL;
			  m_bBranchPoint = false;
			  m_bBranchPointNext = false;

			  m_pfSiteStart = NULL, m_pfSiteStop = NULL, m_pfSiteDon = NULL, m_pfSiteAcc = NULL;
			  m_pfSplicePhase = NULL, m_pfSiteCoding = NULL, m_pfSiteIntrons = NULL, m_pfSiteNonCoding = NULL;
			  m_pfSiteShortAcc = NULL, m_pfSiteBranchPoint = NULL;

			  m_bHomNoncoding = false;		// combine intron and intergenic sequence into one non-coding model

			  m_bVerbose = false;
			  m_bFirstRun = true;

			  m_nGC = -1;

			  m_bMark = false;

			  m_bExtr = false;

			  m_bPred = false;

			  initParams();		// default params
	}
	~cSite() { 	clearModels(); }

	void initParams();

	bool m_bVerbose;
	bool m_bFirstRun;

	SSeq *m_psseq;

	int m_iorderAtg, m_iorderTaa, m_iorderTag, m_iorderTga;
	int m_iorderDon0, m_iorderDon1, m_iorderDon2;
	int m_iorderAcc0, m_iorderAcc1, m_iorderAcc2;
	int m_iorderShortAcc0, m_iorderShortAcc1, m_iorderShortAcc2;

	double *m_pModelMatAtg, *m_pModelMatTaa, *m_pModelMatTag, *m_pModelMatTga;
	double *m_pModelMatDon0, *m_pModelMatDon1, *m_pModelMatDon2;
	double *m_pModelMatAcc0, *m_pModelMatAcc1, *m_pModelMatAcc2;
	double *m_pModelMatDon, *m_pModelMatAcc;			// phase-less models (for transcriptome)

	vector<int> m_intronTypes;

	// model controls
	bool	m_bBranchPoint, m_bBranchPointNext;
	bool	m_bHomNoncoding;

	bool m_bExtr;

	int m_nGC;
	
	bool m_bMark;
	bool m_bPred;

	// site data (accumulations of individual cSite data
	vector<string> m_seqStarts, m_seqTerms, m_seqDons, m_seqAccs, m_seqShortAccs;
	//vector<string> m_seqIntronTerms;
	//vector<int> m_istartIntronTerms;
	vector<int> m_splicePhase;
	vector<bpRecord> m_seqBp;
	string m_strCoding;
	string m_strIntrons, m_strIntronFlank;
	string m_strNonCoding;

	FILE *m_pfSiteStart, *m_pfSiteStop, *m_pfSiteDon, *m_pfSiteAcc;
	FILE *m_pfSplicePhase, *m_pfSiteCoding, *m_pfSiteIntrons, *m_pfSiteFlank, *m_pfSiteNonCoding;
	FILE *m_pfSiteShortAcc, *m_pfSiteBranchPoint;

	vector<codingRegion> m_genes;
	vector<sTranscript> m_transcripts;
	vector<sIntron> m_introns;
	vector<sIntronSet> m_intronSets;
	vector<sEstGene> m_estGenes;

	string m_strOutputFileBase;

	int m_startFlankLeft, m_startFlankRight;
	int m_donFlankLeft,	m_donFlankRight;
	int m_accFlankLeft, m_accFlankRight;
	int m_termFlankLeft, m_termFlankRight;
	int m_intronMarginLeft, m_intronMarginRight;
	int m_minIntronLen;
	int m_minIntergLen;
	int m_minCodingLen;

	int m_shortAccFlankLeft, m_shortAccFlankRight;
	int m_intronTermLen, m_intronTermFlankRight, m_intronTermThresh;

	void report(string str);		// verbose status monitoring
	void init( vector<codingRegion>& genes, SSeq *psseq, int iContig );
	void initTranscripts( vector<sTranscript>& xscripts, SSeq *psseq, int iContig  );
	void initOrigins( vector<sIntron>& introns, SSeq *psseq, int iContig  );
	void initGappedAlignSets( vector<sIntronSet>& sets, SSeq *psseq, int iContig, bool bMultiFile );
	void initEstGenes( vector<sEstGene>& genes, SSeq *psseq, int iContig  );

	//void setAlignments( vector<sAln>& alns );
	void setGappedAlignSets( vector<sIntronSet>& sets );
	void addIntrons(int i, bool bMultiFile);
	void sets2introns(bool bMultiFile);

	void reset(void);

	//bool setRunState( runMatrix& runMat, int nIter );

	// site processing
	void clearModels(void);
	void getSites( vector<string>& site, char *pchBuf );
	void getBpSites( vector<bpRecord>& bpData, char *pchBuf );
	void splitStopSites( vector<string>& taa, vector<string>& tag, vector<string>& tga, vector<string>& other, char *pchBuf );

	// I-O fcns
	bool outputSites(void);
	void recordSiteData(void);
	void recordCdsData(void);
	void recordSiteDataTrans(void);
	void recordGappedAlignData(void);
	void recordEstpData(void);
	void writeBpData( FILE *pF, vector<bpRecord>& bpData );
	void writeStringArray( FILE *pF, vector<string>& strArr );
	void writeIntArray( FILE *pF, vector<int>& strArr );
	void getIntVector( vector<int>& int_vect, char *pchBuf );

	// coding/non-coding sequence extraction fcns
	bool getAllCodingSeq( string& strCod, int iSeq );
	bool getAllIntronSeq( string& strIntrons, int iSeq );
	bool getAllIntergenicSeq( string& strIntergenic, int iSeq );
	bool getAllNoncodingSeq( string& strNoncoding, int iSeq );
	bool getAllIntronSeqTrans( string& strIntrons, int iSeq );
	bool getAllIntergenicSeqTrans( string& strIntergenic, int iSeq );
	bool getAllIntronSeqSRA( string& strIntrons, string& strFlank, int iSeq, int nFlank );
	bool correctIntronStrand( int iSeq );		// assign m_introns based on terminal dinucleotides
	bool getAllIntergenicSeqEstp( string& strIntergenic, int iSeq );
	bool getAllCodingSeqEstp( string& strCod, int iSeq );
	bool getAllSpliceSites( int iSeq );

	bool getStrand( SSeq *pseq, int iSeq, int nOffset1, int nOffset2 );
	bool getStrand( vector<sIntron>& introns, SSeq *pseq, int iSeq, int nOffset1, int nOffset2 );

	bool getJcnType( int& istrand, int& iJcnType, string& s5, string& s3, int iSeq );
	bool getCanonicalStrand( int& istrand, string& s5, string& s3, int iSeq );
	bool checkPartial( int ig );

	bool getSequence(string& str, int iSeq, int iStart, int nNt, bool bStrand);
	int getContigLength(int iContig);

	bool getBoundarySites(int iSeq, bool bExtr);
	bool getBoundarySites(int);
	bool getCds(int iSeq);
	bool getBoundarySitesTrans(int iSeq);
	bool getGappedAlignSitesHet(int iSeq, int nCodFlank, bool bCanonical);
	bool getGappedAlignSites(int iSeq, int nCodFlank, bool bCanonical);
	bool getCdsBoundarySites( int iSeq );

	int getCodingLen( codingRegion& gene );
	void clearData(void);

	bool extractStarts( vector<string>& seqStarts, int iSeq, int nl, int nr );
	bool extractDon( vector<string>& seqDons, vector<int>& phaseDons, int iSeq, int nl, int nr );
	bool extractDonTrans( vector<string>& seqDons, int iSeq, int nl, int nr );
	bool extractDonSRA( vector<string>& seqDons, int iSeq, int nl, int nr, bool bCanonical );
	bool extractAcc( vector<string>& seqAccs, int iSeq, int nl, int nr );
	bool extractAccTrans( vector<string>& seqAccs, int iSeq, int nl, int nr );
	bool extractAccSRA( vector<string>& seqAccs, int iSeq, int nl, int nr, bool bCanonical );
	bool extractIntronTerm( vector<bpRecord>& bpData, int iSeq, int lenSeq, int nr_pad );
	bool extractTerm( vector<string>& seqTerm, int iSeq, int nl,int  nr );
	bool extractStartsEstp( vector<string>& seqStarts, int iSeq, int nl, int nr );
	bool extractTermEstp( vector<string>& seqTerm, int iSeq, int nl, int nr );

	bool getIntronAcc( int iSeq, int nl, int nr, bool bCanonical );
	bool getIntronDon( int iSeq, int nl, int nr, bool bCanonical );

	int getStartSiteLength(void);
	int getDonSiteLength(void);
	int getAccSiteLength(void);
	int getShortAccSiteLength(void);
	int getTermSiteLength(void);
	void setMark( bool b ) { m_bMark = b; };
	void addFeatureSuffix( string& nt, int mark5, int mark3, sTranscript& xscript, int ix, int ifeature );
	void addFeatureSuffix( string& nt, int mark5, int mark3, codingRegion& gene, int ix, int ifeature );
	void addFeatureSuffix( string& nt, int mark5, int mark3, string& name, sIntron& intron );
	void addFeatureSuffixHet( string& nt, int mark5, int mark3, string& name, sIntron& intron  );
	void addFeatureSuffixSite( string& nt, int mark, sTranscript& xscript, int ix, int ifeature );
	void addFeatureSuffixSite( string& nt, int mark, codingRegion& gene, int ix, int ifeature );


};

