//**************************************************
//	Originally part of probuild (by Alex Lomsadze),
//  refactored for GeneMark-ES 2.5 by Paul Burns,
//  and reconstituted here for this simple utility.
//
//  August 2009, Paul Burns
//**************************************************

#include "sseq.h"

char AA[65]  = "KNKNTTTTRSRSIIMIQHQHPPPPRRRRLLLLEDEDAAAAGGGGVVVV*Y*YSSSS*CWCLFLF";

SSeq::SSeq()
{
	m_bSafeFormat = false;

	m_bStartStopsOpen = false;
	m_bIntronsOpen = false;

	m_bAbbrev = false;

	m_bPrintReformat = false;

	m_lastIx = 0;
	m_nSequenceBytesWritten = 0;

	m_bForceAlphabet = true;

	SetAlphabet();
}

bool SSeq::getDefline( string& strDefline, int i )
{
	int n = (int)seqSet.size();

	if (i<0 || i>(n-1))
		return true;		// indexing error

	string strName = seqSet[i].first;
	if (strName.at(0)=='>')
		strName = strName.substr(1,string::npos);

	strDefline = strName;

	return false;		// success
}

bool SSeq::getDeflines( vector<string>& lines )
{
	int n = (int)seqSet.size();
	lines.clear();

	for (int i=0; i<n; ++i)
	{
		lines.push_back( seqSet[i].first );
	}

	return false;
}

bool SSeq::isEmpty(int iseq)
{
	if (iseq<0 || iseq>=(int)seqSet.size())
		return true;

	return seqSet[iseq].second.empty();
}

bool SSeq::getSequence( char **ppchSeq, int i )
{
	int n = (int)seqSet.size();

	if (i<0 || i>(n-1))
		return true;		// indexing error

	*ppchSeq = (char*)seqSet[i].second.c_str();

	return false;		// success
}

int SSeq::getNumContigs()
{
	return (int)seqSet.size();
}

int SSeq::getLength(int iContig)
{
	return (int)seqSet[iContig].second.size();
}

string SSeq::getDefline(int iContig)
{
	return seqSet[iContig].first;
}

void SSeq::setForceAlpha(bool bForce)
{
	m_bForceAlphabet = bForce;
}


int SSeq::findInFrameCodon( int iseq, int istart, int istop, char chStrand, string codon )
{
	// "phase" is a modulo 3 integer {0,1,2} which equals i5 % 3, i5 is the position from
	//  the 5' end of the sequence. Thus if checking is done on positive strand, i = position
	//  from sequence beginning, and if checking is done on negative strand, i = position
	//  relative to sequence end.

	int iphaseStart;
	int istartOffset = 0;

	if (codon.length() != 3)
		return -1;				// bad codon

	if (iseq<0 || iseq>(int)seqSet.size())
		return -1;				// bad iseq

	int lenSeq = getLength( iseq );
	if (istart<0 || istart>=lenSeq || istop<0 || istop>=lenSeq)
		return -1;				// bad stop or start

	// perform search
	bool bFound = false;
	int ix = -1;
	iphaseStart = (istop-istart+1) % 3;

	if (chStrand=='+')		// positive strand
	{
		// get start offset to perform in-phase codon search
		while (iphaseStart != 0)
		{
			++iphaseStart;
			++istartOffset;
			if (iphaseStart==3)
				iphaseStart = 0;
		}

		ix = istart + istartOffset;
		while (ix+3 <= istop)
		{
			if (seqSet[iseq].second.at(ix)==codon.at(0) &&
				seqSet[iseq].second.at(ix+1)==codon.at(1) &&
				seqSet[iseq].second.at(ix+2)==codon.at(2))
			{
				// codon match
				bFound = true;
				break;
			}
			ix += 3;
		}
	}
	else		// negative strand - check reverse complement
	{
		reverseComplement( codon );
		// get start offset to perform in-phase codon search
		while (iphaseStart != 0)
		{
			--iphaseStart;
			++istartOffset;
			if (iphaseStart==-1)
				iphaseStart = 2;
		}

		ix = istop - istartOffset;
		while (ix-3 >= istart)
		{
			if (seqSet[iseq].second.at(ix)==codon.at(2) &&
				seqSet[iseq].second.at(ix-1)==codon.at(1) &&
				seqSet[iseq].second.at(ix-2)==codon.at(0))
			{
				// codon match
				bFound = true;
				break;
			}
			ix -= 3;
		}
	}

	if (bFound)
		return ix;
	
	// -1 is the "not found" signal
	return -1;
}


//--------------------------------------------------//
void SSeq::SetAlphabet()
{
	codedAlphabet.assign( 256, 4 );

		codedAlphabet['a'] = 0;
		codedAlphabet['A'] = 0;

		codedAlphabet['c'] = 1;
		codedAlphabet['C'] = 1;

		codedAlphabet['g'] = 2;
		codedAlphabet['G'] = 2;

		codedAlphabet['t'] = 3;
		codedAlphabet['T'] = 3;
		codedAlphabet['u'] = 3;
		codedAlphabet['U'] = 3;

	alphabet.assign( 256, 'N' );
	alphaComplement.assign( 256, 'N' );

	if (m_bSafeFormat)
	{
		printf("format is safe...\n");
		// leave all non-ACGTU's as N's
		alphabet['a'] = 'A'; alphabet['A'] = 'A';
		alphabet['c'] = 'C'; alphabet['C'] = 'C';
		alphabet['g'] = 'G'; alphabet['G'] = 'G';
		alphabet['t'] = 'T'; alphabet['T'] = 'T';
		alphabet['u'] = 'T'; alphabet['U'] = 'T';
	}
	else
	{
		printf("format is UNsafe...\n");
		// read the nt data as-is
		unsigned char ul_offset = 0x20;		// 'a' - 'A' = 32
		for (unsigned char uch = 'A'; uch<'Z'; ++uch)
		{
			alphabet[uch] = uch;
			alphabet[uch+ul_offset] = uch+ul_offset;
		}
	}

	// convert all non-ACGT to N's for complement
	alphaComplement['a'] = 'T'; alphaComplement['A'] = 'T';
	alphaComplement['c'] = 'G'; alphaComplement['C'] = 'G';
	alphaComplement['g'] = 'C'; alphaComplement['G'] = 'C';
	alphaComplement['t'] = 'A'; alphaComplement['T'] = 'A';
	alphaComplement['u'] = 'A'; alphaComplement['U'] = 'A';
}

void SSeq::setShort(int iseq)
{
	if (iseq<0 || iseq>=(int)m_bShortFeature.size())
		return;

	m_bShortFeature[iseq] = true;
}

void SSeq::resetShort()
{
	m_bShortFeature.assign( seqSet.size(), false );
}

void SSeq::filterShortSeqs(int minLen)
{
	if (minLen==0)
		return;

	int ix = 0;
	int len;

	vector<Record>::iterator is;
	vector<Record>::iterator is_prev;
	is = is_prev = seqSet.begin();

	while ( is!= seqSet.end() )
	{
		len = (int)is->second.length();
		if (len<minLen)
		{
			if (is==seqSet.begin())
			{
				seqSet.erase( is );
				is = is_prev = seqSet.begin();
			}
			else
			{
				seqSet.erase( is );
				is = is_prev;
				++is;
			}
		}
		else
		{
			is_prev = is;
			++is;
		}
	}
}

//--------------------------------------------------//
void SSeq::FindLabel( char** begin, char** end )
{
	char* ptr = *begin;
	char* stop = *end;

	while(( ptr < stop )&&( *ptr != '>' ))
		++ptr;

	*begin = ptr;

	if(( ptr < stop )&&( *ptr == '>' ))
	{
		while(( ptr < stop )&&( *ptr != '\n' ))
			++ptr;
	}

	*end = ptr;
};
//--------------------------------------------------//
void SSeq::FindContig( char** begin, char** end )
{
	char* ptr = *begin;
	char* stop = *end;

	while(( ptr < stop )&&( *ptr != '>' ))
		++ptr;

	*end = ptr;
};
//--------------------------------------------------//
void SSeq::TextToString( string& target, char* begin, char* end )
{
	target.resize( end - begin );
	string::iterator itr = target.begin();

	for( ; begin < end; ++begin )
	{
		if( isalpha( *begin ) )
		{
			if (m_bForceAlphabet)
				*itr = alphabet[ *begin ];
			else
				*itr = *begin;
			++itr;
		}
	}
	target.resize( itr - target.begin() );
};

//--------------------------------------------------//
void SSeq::clear(void)
{
	seqSet.clear();
}

void SSeq::trimWhite( string& str )
{
	char pch_hw[] = " \t\r";
	size_t six = str.find_last_not_of( pch_hw );
	if (six==string::npos)
	{
		str.clear();
		return;
	}

	if (six<str.length()-1)
	{
		str = str.substr( 0, six+1 );
		return;
	}

	return;
}

//--------------------------------------------------//
bool SSeq::load( const string& name, int nBytesToGet )
{
	return load( name, nBytesToGet, 0 );
}

bool SSeq::load( const string& name, int nBytesToGet, int nMinLen )
{
	if (nBytesToGet==0)
		nBytesToGet = numeric_limits<int>::max();

	FileBuf f;
	if ( !f.Load( name.c_str() ) ) return false;

	Record record;
	string strLabel, strCurLabel;

	char* label_begin;
	char* label_end;
	char* seq_begin;
	char* seq_end = f.begin;
	size_t st;

	int nBytes = 0, nNewBytes, nKeep;

	while ( seq_end < f.end )
	{
		label_begin = seq_end;
		label_end = f.end;

		// get label
		FindLabel( &label_begin, &label_end );
		strLabel.assign( label_begin, label_end );
		trimWhite( strLabel );

		seq_begin = label_end;
		seq_end = f.end;

		FindContig( &seq_begin, &seq_end );

		// add new contig
		seqSet.push_back( record );
		// copy contig dna data
		TextToString( seqSet.back().second, seq_begin, seq_end );

		// set label
		if (m_bAbbrev) {
			st = strLabel.find_first_not_of(" \t");
			strLabel = strLabel.substr(st, strLabel.find_first_of(" \t",st) );		// just get first word
		}
		seqSet.back().first = strLabel;
		//printf("getting %s\n", strLabel.c_str());

		nNewBytes = (int)seqSet.back().second.size();
		if (nBytes+nNewBytes>nBytesToGet)
		{
			// load up to limit nBytesToGet and quit
			nKeep = nBytesToGet - nBytes;
			seqSet.back().second = seqSet.back().second.substr( 0, nKeep );

			m_bShortFeature.push_back( nKeep < nMinLen );
			//printf("nKeep = %d, seq length = %d\n", nKeep, seqSet.back().second.length());
			break;
		}

		m_bShortFeature.push_back( nNewBytes < nMinLen );

		nBytes += nNewBytes;

	}

	printf("%d kB loaded.\n", nBytes/1000);

	return true;
};


string SSeq::getName( string& defline )
{
	// returns the contig name string from a defline formatted as follows:
	// >scf_abc123.a.n|blah|blah-blah|...
	// or
	// scf_abc123.a.n|blah|blah-blah|...
	//
	// and returns...
	// scf_abc123.a.n

	string name;
	if (defline.empty())
		return name;

	string strSep = "|";
	size_t ix = defline.find(strSep);
	
	if (defline.at(0)=='>')
		name = defline.substr(1, ix-1);
	else
		name = defline.substr(0, ix-1);

	return name;
}

bool SSeq::getNameSuffix( string& seqName, int iseq )
{
	// returns the contig name string from a defline formatted as follows:
	// blah|blah-blah|scf_abc123.a.n
	//
	// and returns...
	// scf_abc123.a.n

	if (iseq<0 || iseq>=(int)seqSet.size())
		return true;		// bad index

	string defline = seqSet[iseq].first;
	if (defline.at(0)=='>')
		defline = defline.substr(1,string::npos);

	size_t pos_L = defline.find_last_of("|");
	if (pos_L == string::npos) {
		seqName = defline;
		return false;
	}

	char pchWhite[] = " \t";
	size_t pos_R = defline.find_first_of( pchWhite, pos_L + 1 );
	seqName = defline.substr(pos_L + 1, pos_R - pos_L - 1);

	return false;
}

string SSeq::getNameTok( string& defline )
{
	// returns the contig name string from a defline formatted as follows:
	// >scf_abc123.a.n blah blah blah...
	// or
	// scf_abc123.a.n|blah|blah-blah|...
	//
	// and returns...
	// scf_abc123.a.n

	string name;
	name = getName( defline );

	size_t ix0 = name.find_first_not_of(" \t");
	size_t ix1 = name.find_first_of(" \t",ix0+1);

	name = name.substr(ix0, ix1 - ix0);

	return name;
}

void SSeq::writeSequence( FILE *pf, string& defline, string& sequence )
{
	fprintf( pf, "%s\n", defline.c_str() );
	PrintFormated( pf, sequence, 100 );
	m_nSequenceBytesWritten += (int)sequence.size();
}

void SSeq::writeSequence( FILE *pf, int iContig, string& strExons, char& chStrand, string& other, bool bForceCanonical )
{
	if (strExons.length()<6)
		return;

	//strExons = strExons.substr(5);
	size_t ix = strExons.find_first_of(",");
	int ib = 5;
	size_t ixp, ixag;

	string strBeg, strEnd, strBegNext;
	int ibegExon, iendExon, ibegNext;
	string str, seq, gene;

	bool bCanonical = true, bCanonicalNegative = true;
	bool bNeg = false;
	int nExons = 1;

	// first (n-1) exons
	//
	while (ix != string::npos)
	{
		++nExons;

		str = strExons.substr(ib, ix-ib+1);		// str = "n..m"
		ixp = str.find_first_of(".");
		strBeg = str.substr(0,ixp);			// "n"
		strEnd = str.substr(ixp+2,str.length()-(ixp+2+1));			// "m"
		ibegExon = atoi( strBeg.c_str() );	// n
		iendExon = atoi( strEnd.c_str() );	// m

		// next exon
		ixag = strExons.find_first_of(".",ix);
		if (ixag==string::npos)	// error
		{
			std::cout << "Error extracting acceptor site" << std::endl;
			return;
		}
		strBegNext = strExons.substr(ix+1, ixag-ix-1);
		ibegNext = atoi( strBegNext.c_str() );

		if (!checkCanonicalSplice(bNeg, iContig, iendExon, ibegNext-2, chStrand))
			bCanonical = false;
		bCanonicalNegative &= bNeg;
			
		if (!getStretch( seq, iContig, ibegExon-1, iendExon-ibegExon+1, true ))	// always write positive strand
			printf("Error getting sequence for %s\n", strExons.c_str());
		else
		{
			gene.append( seq );
		}

		ib = (int)ix+1;
		ix = strExons.find_first_of(",", ix+1);
	}

	// last exon
	//
	ix = strExons.length() - 1;
	str = strExons.substr(ib, ix-ib+1);		// str = "n..m"
	ixp = str.find_first_of(".");
	strBeg = str.substr(0,ixp);			// "n"
	strEnd = str.substr(ixp+2,str.length()-(ixp+2+1));			// "m"
	ibegExon = atoi( strBeg.c_str() );	// n
	iendExon = atoi( strEnd.c_str() );	// m

	if (!getStretch( seq, iContig, ibegExon-1, iendExon-ibegExon+1, true ))	// always write positive strand
		printf("Error getting sequence for %s\n", strExons.c_str());
	else
	{
		gene.append( seq );
	}

	string defline = seqSet[iContig].first;
	defline += "|";
	defline += other;
	defline += "|";
	if ((nExons>1) && bCanonicalNegative)
	{
		// flip strand
		if (chStrand=='+')
			chStrand = '-';
		else if (chStrand=='-')
			chStrand = '+';
		defline += chStrand;
	}
	else
	{
		defline += chStrand;
	}

	defline += "|";
	defline += strExons;

	// only print "clean" genes with canonical splice sites
	if (bForceCanonical)
	{
		if ((nExons>1) && (bCanonical || bCanonicalNegative))
		{
			fprintf( pf, "%s\n", defline.c_str() );
			PrintFormated( pf, gene, 100 );
			m_nSequenceBytesWritten += (int)gene.size();
		}
	}
	else
	{
		fprintf( pf, "%s\n", defline.c_str() );
		PrintFormated( pf, gene, 100 );
		m_nSequenceBytesWritten += (int)gene.size();
	}

}

void SSeq::maskContig( int iseq, int iMaskBeg, int iMaskEnd )
{
	size_t st_n = iMaskEnd - iMaskBeg + 1;
	char ch = 'X';
	seqSet[iseq].second.replace( iMaskBeg, st_n, st_n, ch );
}

void SSeq::translate( char **pchAa, const char *pchNt, int phase )
{
	int i,j;
	int Nch = (int)strlen(pchNt);
	int Naa = (int) (Nch/3);		// truncate dangling nt's (if any)
	if (pchAa[0]!=NULL) delete []pchAa[0];
	pchAa[0] = new char[Naa+1];

	char a, b, c;
	int ia, ib, ic;

	extern char AA[65];

	for ( i = phase, j = 0; i < Nch-3 ; i += 3, ++j )
	{
		if ( (codedAlphabet[ pchNt[i] ]>3)||(codedAlphabet[ pchNt[i+1] ]>3)||(codedAlphabet[ pchNt[i+2] ]>3) )
			*(pchAa[0] + j) = 'X';
		else
		{
			*(pchAa[0] + j) = AA[ (codedAlphabet[ pchNt[i] ]<<4) + (codedAlphabet[ pchNt[i+1] ]<<2) + codedAlphabet[ pchNt[i+2] ] ];
			if (*(pchAa[0] + j)=='*')
			{
				a = pchNt[i];
				b = pchNt[i+1];
				c = pchNt[i+2];
				ia = codedAlphabet[ a ];
				ib = codedAlphabet[ b ];
				ic = codedAlphabet[ c ];
			}
		}
	}

	if (Nch==(i+3))
	{
		// last one
		if ( (codedAlphabet[ pchNt[i] ]>3)||(codedAlphabet[ pchNt[i+1] ]>3)||(codedAlphabet[ pchNt[i+2] ]>3) )
			*(pchAa[0] + j) = 'X';
		else
			*(pchAa[0] + j) = AA[ (codedAlphabet[ pchNt[i] ]<<4) + (codedAlphabet[ pchNt[i+1] ]<<2) + codedAlphabet[ pchNt[i+2] ] ];
		++j;
	}

	*(pchAa[0] + j) = '\0';
}


bool SSeq::checkCanonicalStartStop( int iContig, int geneStart, int geneStop, char chStrand, bool bWrite)
{
	string strATG, strStop;

	bool bCanonical = false;
	bool bCanonicalStop = false;

	strATG = "NNN";
	strStop = "NNN";
	if (chStrand=='-')			// negative strand
	{
		getStretch( strATG, iContig, geneStop, 3, false );
		getStretch( strStop, iContig, geneStart+2, 3, false );
	}
	else if (chStrand=='+')		// positive strand
	{
		getStretch( strATG, iContig, geneStart, 3, true );
		getStretch( strStop, iContig, geneStop-2, 3, true );
	}
	// canonical splice sites?
	bCanonicalStop = strStop.compare("TAA")==0 || strStop.compare("TAG")==0 || strStop.compare("TGA")==0;
	if (strATG.compare("ATG")==0 && (bCanonicalStop))
		bCanonical = true;

	// print debug file
	FILE *pfIntron;
	if (bWrite)
	{
		if (m_bStartStopsOpen)
			pfIntron = fopen("start_stop.txt", "a");
		else
		{
			pfIntron = fopen("start_stop.txt", "w");
			fprintf( pfIntron, "Contig\t\tstrand\t\tDon\t\tAcc\t\tcanonical\t\tneg. canonical\n" );
			m_bStartStopsOpen = true;
		}
		
		fprintf(pfIntron, "%d\t\t%c\t\t%s\t\t%s\t\t%d\n", iContig, chStrand, strATG.c_str(), strStop.c_str(),
																int(bCanonical));
		fclose( pfIntron );
	}

	return bCanonical;
}


bool SSeq::checkCanonicalSplice(bool& bCanonicalNegative, int iContig, int nIntronStart, int nIntronStop, char chStrand)
{
	return checkCanonicalSplice( bCanonicalNegative, iContig, nIntronStart, nIntronStop, chStrand, true);
}


bool SSeq::checkCanonicalSplice(bool& bCanonicalNegative, int iContig, int nIntronStart, int nIntronStop, char chStrand, bool bWrite)
{
	// Return true if splice sites are canonical on the indicated strand (chStrand).
	// If this fails, check on negative strand and set bCanonicalNegative accordingly.

	string strGT, strAG;

	bool bCanonical = false;
	bCanonicalNegative = false;

	strAG = "NN";
	strGT = "NN";
	if (chStrand=='-')
	{
		getStretch( strGT, iContig, nIntronStop, 2, false );
		getStretch( strAG, iContig, nIntronStart+1, 2, false );
	}
	else if (chStrand=='+')
	{
		getStretch( strGT, iContig, nIntronStart, 2, true );
		getStretch( strAG, iContig, nIntronStop-1, 2, true );
	}
	// canonical splice sites?
	if (strGT.compare("GT")==0 && strAG.compare("AG")==0)
		bCanonical = true;
	else if (strGT.compare("CT")==0 && strAG.compare("AC")==0)
		bCanonicalNegative = true;

	// print debug file
	FILE *pfIntron;
	if (bWrite)
	{
		if (m_bIntronsOpen)
			pfIntron = fopen("intron.txt", "a");
		else
		{
			pfIntron = fopen("intron.txt", "w");
			fprintf( pfIntron, "Contig\t\tstrand\t\tDon\t\tAcc\t\tcanonical\t\tneg. canonical\n" );
			m_bIntronsOpen = true;
		}
		
		fprintf(pfIntron, "%d\t\t%c\t\t%s\t\t%s\t\t%d\t\t%d\n", iContig, chStrand, strGT.c_str(), strAG.c_str(),
																int(bCanonical), int(bCanonicalNegative));
		fclose( pfIntron );
	}

	return bCanonical;
}

int SSeq::contigLookup( string& strName )
{
	if (seqSet.empty())
		return -1;

	int ix = 0;
	string name = seqSet[ix].first;
	if (name.at(0)=='>')
		name = name.substr(1);

	while (name.compare(strName)!=0)
	{
		++ix;
		if (ix==seqSet.size())
			return -1;

		name = seqSet[ix].first;
		if (name.at(0)=='>')
			name = name.substr(1);

		if (ix==seqSet.size())
		{
			ix = -1;
			break;
		}
	}

	return ix;
}

bool SSeq::saveFasta( string fname )
{
	return saveFasta( fname, __NO_OPERATION );
}

bool SSeq::saveFasta( string fname, int icode )
{
	std::filebuf fb;
	fb.open( fname.c_str(), std::ios::out );
	std::ostream os(&fb);

	if ( !fb.is_open() )
	{
		std::cout << "Unable to open output file " << fname << std::endl;
		return true;
	}

	for (size_t sz=0, i=0; sz<seqSet.size(); ++sz, ++i)
	{
		switch (icode)
		{
		case __NO_OPERATION:
			writeRecord( os, (int)i );
			break;
		case __FILTER_SHORT:
			if (!m_bShortFeature[i])
				writeRecord( os, (int)i );
			break;
		};
	}

	fb.close();

	return false;
}

void SSeq::writeRecord( std::ostream& os, int i )
{
	if (seqSet[i].first.at(0)=='>')
		os << seqSet[i].first << std::endl;
	else
		os << ">" << seqSet[i].first << std::endl;

	PrintFormated( os, seqSet[i].second, 100 );
}

void SSeq::writeRecord(FILE *pf, int i)
{
	if (seqSet[i].first.at(0)=='>')
		fprintf( pf, "%s\n", seqSet[i].first.c_str() );
	else
		fprintf( pf, ">%s\n", seqSet[i].first.c_str() );

	PrintFormated( pf, seqSet[i].second, 100 );
}


// return first index of sequence encountered with matching defline
int SSeq::findSequence(string defline)
{
	// strip '>' out of the defline (if present)
	if (defline.at(0)=='>') { defline = defline.substr( 1 ); }

	string strTest;

	for (int i=0; i<(int)seqSet.size(); ++i)
	{
		strTest = seqSet[m_lastIx].first;
		if (strTest.at(0)=='>') { strTest = strTest.substr( 1 ); }
		if (defline.compare(strTest)==0)
			return m_lastIx;
		++m_lastIx;
		if (m_lastIx==seqSet.size())
			m_lastIx = 0;
	}

	// did not find a match
	return -1;
}

int SSeq::findSequenceSubstr(string deflineSubstr)
{
	// strip '>' out of the defline (if present)
	if (deflineSubstr.at(0)=='>') { deflineSubstr = deflineSubstr.substr( 1 ); }

	string strTest;

	for (int i=0; i<(int)seqSet.size(); ++i)
	{
		strTest = seqSet[m_lastIx].first;
		if (strTest.find(deflineSubstr)!=string::npos)
			return m_lastIx;
		++m_lastIx;
		if (m_lastIx==seqSet.size())
			m_lastIx = 0;
	}

	// did not find a match
	return -1;
}

void SSeq::rename( int iseq, string& gene_name, string& mrna_name )
{
	if (iseq<0 || iseq>(int)seqSet.size())
		return;

	size_t bar = seqSet[iseq].first.find("|");
	if (bar == string::npos)
		return;

	string suffix = seqSet[iseq].first.substr(bar);

	// modify defline
	seqSet[iseq].first = ">";
	seqSet[iseq].first.append( gene_name );
	seqSet[iseq].first.append( " | " );
	seqSet[iseq].first.append( mrna_name );
	seqSet[iseq].first.append( " " );
	seqSet[iseq].first.append( suffix );
}

bool SSeq::boundsCheck( int iseq, int ipos, int len, bool bStrand )
{
	// return true if desired stretch will breach sequence limits
	//

	int size = (int)seqSet[iseq].second.size();

	if (bStrand && (ipos<0 || ipos+len>size))
		return false;		// index out of bounds

	if (!bStrand && (ipos-len+1<0 || ipos>=size))
		return false;		// index out of bounds

	return true;		// bounds OK
}

//--------------------------------------------------//
bool SSeq::getStretch( string& str, int iseq, int ipos, int len, bool bStrand )
{
	// bounds check
	if (!boundsCheck(iseq, ipos, len, bStrand))
		return false;		// error

	// read sequence
	str.assign(len, 'N');

	if (bStrand)
	{
		getWindow( str, ipos, ipos+len-1, seqSet[iseq].second );
	}
	else
	{
		getWindow( str, ipos-len+1, ipos, seqSet[iseq].second );
		reverseComplement( str );
	}


	return true;		// success
}

//--------------------------------------------------//
void SSeq::getWindow( string& window, int ifirst, int ilast, string& str )
{
	window = str.substr( ifirst, (ilast-ifirst+1) );
}

//--------------------------------------------------//
void SSeq::reverseComplement( string& str )
{
	int irev, ifor = 0;
	string str_copy = str;

	for (irev=(int)str.size()-1; irev>=0; --irev)
	{
		str[ifor] = alphaComplement[ str_copy[irev] ];
		ifor++;
	}
}

//--------------------------------------------------//
int SSeq::nextValidSequence( int ibeg, int nLength )
{
	for (int i=ibeg+1; i<(int)seqSet.size(); ++i)
	{
		if ((int)seqSet[i].second.size()>=nLength)
			return i;
	}

	return -1;
}

//--------------------------------------------------//
void SSeq::CountLetters( vector<int>& target, string::iterator begin, string::iterator end )
{
	while( begin < end )
	{
		target[ *begin ]++;
		++begin;
	}
};

//--------------------------------------------------//
double SSeq::CalcGC( vector<int>& counts )
{
	double GC = 100.0 * ( counts['C'] + counts['G'] );
	if ( counts['A'] + counts['C'] + counts['G'] + counts['T'] )
		GC /= (counts['A'] + counts['C'] + counts['G'] + counts['T']);
	else
		GC = 0;
	return GC;
};

double SSeq::GetGC(int iSeq, int ibeg, int iend)
{
	vector<int> nt;
	nt.resize( 256, 0 );
	vector<Record>::iterator itr = seqSet.begin() + iSeq;

	// bounds checking
	if (ibeg < 0) ibeg = 0;
	if (iend >= (int)itr->second.length()) iend = (int)itr->second.length() - 1;

	CountLetters( nt, itr->second.begin() + ibeg, itr->second.begin() + iend );
	double lfGc = CalcGC( nt );

	return lfGc;
}

double SSeq::GetGC(string& s)
{
	vector<int> nt;
	nt.resize( 256, 0 );

	CountLetters( nt, s.begin(), s.end() );
	double lfGc = CalcGC( nt );

	return lfGc;
}

//--------------------------------------------------//
string SSeq::PrintGC( int precision, string &label )
{
	vector<int> nt;
	nt.resize( 256, 0 );
	vector<Record>::iterator itr;

	for ( itr = seqSet.begin(); itr != seqSet.end(); ++itr )
		CountLetters( nt, itr->second.begin(), itr->second.end() );

	ostringstream out;
	out.setf( ios::fixed, ios::floatfield );
	out.precision( precision );
	out << label << CalcGC( nt );
	
	return out.str();
}

//--------------------------------------------------//
string SSeq::PrintStat( string label, int precision )
{
	vector<int> nt;
	nt.resize( 256, 0 );

	vector<Record>::iterator itr;

	for ( itr = seqSet.begin(); itr != seqSet.end(); ++itr )
	{
		CountLetters( nt, itr->second.begin(), itr->second.end() );
	}

	if (( seqSet.size() == 1 )&&( !seqSet.begin()->first.empty() ))
		label += (" " + seqSet.front().first);

	ostringstream out;
	out.setf( ios::fixed, ios::floatfield );
	out.precision( precision );

	int nt_count = 0;
	int i;

	for (i=0; i<(int)nt.size(); ++i)
	{
		nt_count += nt[i];
	}

//	out	<< "LABEL          " << label << '\n'
	out	<< "SEQUENCE_SIZE  " << nt_count << '\n'
		<< "SEQUENCE_ACGT  " << nt['A']+nt['C']+nt['G']+nt['T'] << '\n';

	out << "--\n";

	out << "NT_A           " << nt['A'] << '\n';
	out << "NT_C           " << nt['C'] << '\n';
	out << "NT_G           " << nt['G'] << '\n';
	out << "NT_T           " << nt['T'] << '\n';

	out << "--\n";

	for (i=0; i<(int)nt.size(); ++i)
	{
		if (nt[i]>0 && char(i)!='A' && char(i)!='C' && char(i)!='G' && char(i)!='T')
		{
			out << "NT_" << char(i) << "           " << nt[i] << '\n';
		}
	}

	out << "--\n";

	out << "GC             " << CalcGC( nt ) << '\n'
         << "RECORDS        " << seqSet.size() << '\n';

	
	// print report to stdout
	//
	report( ". . . . ." );
	if (!reportContigs())
		return false;
	report( ". . . . ." );

	return out.str();
};
//--------------------------------------------------//
bool SSeq::reportContigs()
{
	vector<int> contig, cum_contig;
	vector<Record>::iterator itr;

	int tot = 0;

	// get vector of contig lengths
	for ( itr = seqSet.begin(); itr != seqSet.end(); ++itr )
	{
		tot += (int)itr->second.size();
		contig.push_back( (int)itr->second.size() );
	}

	// sort in ascending order
	sort( contig.begin(), contig.end() );

	// (negative) cumulative sum of contig lengths
	int ni = 0;
	int n25 = (int) double(tot)/4.;
	int n50 = (int) double(tot)/2.;
	int n75 = (int) double(tot)*3./4.;
	bool b25 = false;
	bool b50 = false;
	bool b75 = false;
	for (int i=0; i<(int)contig.size(); ++i)
	{
		ni += contig[i];
		cum_contig.push_back( ni );

		// find quartile points
		if (ni>=n25 && !b25)
		{
			b25 = true;
			n25 = i;
		}
		if (ni>=n50 && !b50)
		{
			b50 = true;
			n50 = i;
		}
		if (ni>=n75 && !b75)
		{
			b75 = true;
			n75 = i;
		}
	}

	string strOut, str;
	char pch[100];
	sprintf( pch, "%d", contig.size() );
	str = pch;
	strOut = str + " contigs found in input sequence";
	report(strOut);

	// shortest
	sprintf(pch, "%d", contig[0]);
	str = pch;
	strOut = "Shortest : " + str;
	report( strOut );
	// longest
	sprintf(pch, "%d", contig[contig.size()-1]);
	str = pch;
	strOut = "Longest  : " + str;
	report( strOut );
	report("--");

	// inflection point
	if (b25 && b50 && b75)
	{
		report("Length      : Percent sequence in contigs longer than");
		// N25
		sprintf(pch, "%d       %4.1f", contig[n25], 100.*double(tot-cum_contig[n25])/double(tot) );
		strOut = pch;
		report( strOut );
		// N50
		sprintf(pch, "%d       %4.1f", contig[n50], 100.*double(tot-cum_contig[n50])/double(tot) );
		strOut = pch;
		report( strOut );
		// N75
		sprintf(pch, "%d       %4.1f", contig[n75], 100.*double(tot-cum_contig[n75])/double(tot) );
		strOut = pch;
		report( strOut );
	}

	return true;
}

//--------------------------------------------------
string SSeq::JoinAsString( const string& name, const string& separator, int i )
{
	ostringstream out;
	out << name << separator << i;
	return out.str();
};
//--------------------------------------------------
void SSeq::PrintFormated( const string& s, int per_line, ofstream& o)
{
	stPos L = 0;
	stPos R = 0;
	
	if (!per_line)
		per_line = (int)s.size();

	while( L < s.size() )
	{
		if ( L + per_line < s.size() )
			R = L + per_line;
		else
			R = s.size();

		o << s.substr( L, R - L ) << endl;
		L = R;
	}
};

void SSeq::PrintFormated( FILE* pf, string& s, int per_line)
{
	stPos L = 0;
	stPos R = 0;
	string str;

	while (L < s.size())
	{
		if ( L + per_line < s.size() )
			R = L + stPos(per_line);
		else
			R = s.size();

		str = s.substr( L, R - L );
		fprintf( pf, "%s\n", str.c_str() );

		L = R;
	}
}

void SSeq::PrintFormated( std::ostream& os, string& s, int per_line)
{
	stPos L = 0;
	stPos R = 0;
	string str;

	while (L < s.size())
	{
		if ( L + per_line < s.size() )
			R = L + stPos(per_line);
		else
			R = s.size();

		str = s.substr( L, R - L );
		os << str << std::endl;

		L = R;
	}
}

void SSeq::report(string str)
{
	if (m_bVerbose)
		printf("%s\n",str.c_str());
}

//--------------------------------------------------
/*
void SSeq::TextToCounts( vector<int>& target, char* begin, char* end )
{
	for( ; begin < end; ++begin )
		if( isalpha( *begin ) )
			target[ alphabet[ *begin ] ]++;
};
*/
//--------------------------------------------------
