//**************************************************
// file:    sseq.h
// project: General purpose DNA sequence container & operations
//          alex lomsadze, modified by Paul Burns
//**************************************************

#pragma once

#include <iostream>
#include <string>
#include <vector>
#include <sstream>
#include <fstream>
#include <utility>
#include <algorithm>

using namespace std;

enum NUCLEOTIDE { A, C, G, T, N, X };
enum STRAND { NO, DIRECT_STRAND, REVERSE_STRAND };
const double log0 = -1.0e120;


#include "filebuf.h"
#include "report.h"
#include <math.h>
#include <float.h>
#include <limits>
#include "string.h"


#define __OUTPUT_MASK		0x0001

#define __NO_OPERATION		0x0000
#define __FILTER_SHORT		0x0001

typedef pair<string,string> Record;
typedef string::size_type stPos;
typedef pair<string::iterator, string::iterator> PosPair; 

class cGff;

//--------------------------------------------------
class SSeq : private Report
{
public:

	SSeq();

	bool isEmpty(int iseq);

	void trimWhite( string& str );
	bool load( const string& name, int nBytesToGet );
	bool load( const string& name, int nBytesToGet, int nMinLen );
	bool saveFasta( string fname );
	bool saveFasta( string fname, int icode );
	void writeRecord(FILE *pf, int i);
	void writeRecord( std::ostream& os, int i );
	void writeSequence( FILE *pf, string& defline, string& sequence );
	void writeSequence( FILE *pf, int ix, string& strExons, char& chStrand, string& misc, bool bForceCanonical );
	void translate( char **pchAa, const char *pchNt, int phase );
	bool checkCanonicalStartStop( int iContig, int geneStart, int geneStop, char chStrand, bool bWrite);
	bool checkCanonicalSplice(bool& bCanonicalNegative, int iContig, int nIntronStart, int nIntronStop, char chStrand);
	bool checkCanonicalSplice(bool& bCanonicalNegative, int iContig, int nIntronStart, int nIntronStop, char chStrand, bool bWrite);
	void clear(void);

	void setSafeFormat(bool bSafe) { m_bSafeFormat = bSafe; }
	void maskContig( int iseq, int iMaskBeg, int iMaskEnd );

	void report(string str);		// verbose status monitoring
	bool m_bVerbose;		// continually print status to stdout
	bool m_bVeryVerbose;	// continually print even more status to stdout
	bool m_bInteractive;

	bool m_bSafeFormat;
	bool m_bAbbrev;
	int m_nMinLen;
	int m_nRecsWritten;

	// accessors
	bool getDefline( string&, int );
	bool getDeflines( vector<string>& lines );
	bool getSequence( char **, int );
	int getNumContigs(void);
	int contigLookup( string& strName );

	// calculate and print key statistics
	string PrintStat( string label, int precision );
	string PrintGC( int precision, string &label );

	// modification
	void filterShortSeqs(int len);
	void setForceAlpha(bool);
	void setShort(int iseq);
	void resetShort(void);

	// support for site extraction
	int getLength(int iContig);
	string getDefline(int iContig);
	bool boundsCheck( int iContig, int ipos, int len, bool bStrand );
	bool getStretch( string& str, int iContig, int ipos, int len, bool bStrand );
	void reverseComplement( string& str );

	// search
	int findInFrameCodon( int iseq, int istart, int istop, char chStrand, string codon );

	int findSequence(string defline);
	int findSequenceSubstr(string deflineSubstr);
	void rename( int iseq, string& gene_name, string& mrna_name );

	string JoinAsString( const string& name, const string& separator, int i );

	void SetAlphabet();

	string getName( string& defline );
	bool getNameSuffix( string& seqName, int iseq );
	string getNameTok( string& defline );

	double GetGC(int iSeq, int ibeg, int iend);
	double GetGC(string& s);

private:

	vector< Record > seqSet;
	vector< bool > m_bShortFeature;

	vector<int> codedAlphabet;
	vector<char> alphabet;
	vector<char> alphaComplement;

	int m_lastIx;		// index of last sequence accessed (useful for printing stuff)
	int m_nSequenceBytesWritten;

	bool m_bStartStopsOpen;
	bool m_bIntronsOpen;

	// private state storage: DNA loading options
	string m_strReformatLabel;
	string m_strDeflinePrefix;
	string m_strReformatDnaFile;
	bool m_bPrintReformat;
	bool m_bForceAlphabet;

	void FindLabel( char** begin, char** end );
	void FindContig( char** begin, char** end );
	void TextToString( string& target, char* begin, char* end );
	void CountLetters( vector<int>& target, string::iterator begin, string::iterator end );
	double CalcGC( vector<int>& counts );

	// new helper function for site extraction
	void getWindow( string& window, int ifirst, int ilast, string& str );
	int nextValidSequence( int ibeg, int nLength );

	bool reportContigs(void);

	void PrintFormated( const string& s, int per_line, ofstream& o);
	void PrintFormated( FILE* pf, string& s, int per_line);
	void PrintFormated( std::ostream& os, string& s, int per_line);
//	void TextToCounts( vector<int>& target, char* begin, char* end );
};
//--------------------------------------------------
//#endif  // __SSEQ_H__
