
#pragma once

#include "isoform.h"
#include "align.h"
#include <vector>
#include <string>
#include <algorithm>
#include <iostream>
#include "targetver.h"

using std::vector;
using std::string;


class cGene
{
public:
	cGene() { m_ibeg = m_iend = -1; };
	~cGene() { };

	// get
	string getSeq() const { return m_strSeq; }
	string getId() const { return m_strId; }
	int getBeg() const { return m_ibeg; }
	int getEnd() const { return m_iend; }
	int getCdsBeg() const { int len, imax; len = getLongestCds(imax); return imax==-1 ? -1 : m_isoforms[imax].getCdsBeg(); }
	int getCdsEnd() const { int len, imax; len = getLongestCds(imax); return imax==-1 ? -1 : m_isoforms[imax].getCdsEnd(); }
	char getStrand() const { int len, imax; len = getLongestCds(imax); return imax==-1 ? -1 : m_isoforms[imax].getStrand(); }
	vector<cInterval> getIntrons() const;
	vector<cRichInterval> getRichIntrons() const;
	
	// set
	void setSeq(string& s) { m_strSeq = s; }
	void setId(string& s) { m_strId = s; }
	void setBeg(int n) { m_ibeg = n; }
	void setEnd(int n) { m_iend = n; }

	bool operator<(const cGene gene) const {
		int ic = m_strSeq.compare( gene.getSeq() );
		if (ic==0) {
			if (m_ibeg==gene.getBeg())
				return m_iend < gene.getEnd();
			else
				return m_ibeg < gene.getBeg();
		}
		else return ic < 0;
	}
	bool operator>(const cGene gene) const {
		int ic = m_strSeq.compare( gene.getSeq() );
		if (ic==0) {
			if (m_ibeg==gene.getBeg())
				return m_iend > gene.getEnd();
			else
				return m_ibeg > gene.getBeg();
		}
		else return ic > 0;
	}

	void writeGff( std::ostream& os );

	void addIsoform(cIsoform& iso) { m_isoforms.push_back(iso); }
	void addAlignment(cAlign& aln) { m_aligns.push_back(aln); }
	void setLimits(int b, int e) { m_ibeg = b; m_iend = e; }
	void setLimits(void) {
		sort();
		if (!m_isoforms.empty()) {
			m_ibeg = m_isoforms.front().m_ibeg;
			m_iend = m_isoforms.back().m_iend;
		}
		else if (!m_aligns.empty()) {
			m_ibeg = m_aligns.front().m_ibeg;
			m_iend = m_aligns.back().m_iend;
		}
		else { m_ibeg = m_iend = -1; }
	}
	void sort(void) {
		std::sort( m_isoforms.begin(), m_isoforms.end() );
		std::sort( m_aligns.begin(), m_aligns.end() );
	}

	bool isOverlapping(int left, int right) const { return !(m_iend < left || m_ibeg > right); }

	int numel(void) const { return (int)m_isoforms.size(); }
	int ctTransExons(void) const;
	int ctCdsExons(void) const;
	int ctIntrons(void) const;

	int lenTrans(void) const;		// length of longest isoform (nt transcribed)
	int lenCds(void) const;		// length of longest isoform (nt coding)
	int lenGene() const;

	int lenIntrons(void) const;	// cumulative total length of introns in longest isoform (nt)

	int getLongestTrans() const;
	int getLongestCds() const;
	int getLongestIntrons() const;
	int getLongestTrans(int& ixMax) const;
	int getLongestCds(int& ixMax) const;

	void clear(void) {
		m_isoforms.clear(); m_aligns.clear();
		m_strId=string(""); m_strSeq=string("");
		m_ibeg = m_iend = -1;
	}


//private:
//	bool parseGffLine(string& line);

private:
	string m_strId;
	string m_strSeq;
	int m_ibeg, m_iend;
	vector<cIsoform> m_isoforms;
	vector<cAlign> m_aligns;
};


vector<cInterval> cGene::getIntrons() const
{
	vector<cInterval> vtrons, vgaps, viso;

	int i;
	int ilong;
	getLongestCds(ilong);

	for (i=0; i<(int)m_aligns.size(); ++i)
	{
		m_aligns[i].getIntrons( viso );
		vgaps.insert( vgaps.end(), viso.begin(), viso.end() );
	}
	std::sort( vgaps.begin(), vgaps.end() );
	vector<cInterval>::iterator itg = std::unique( vgaps.begin(), vgaps.end() );
	vgaps.resize( itg - vgaps.begin() );

	// assign a type to each alignment gap, where possible
	if (!m_isoforms.empty())
	{
		m_isoforms[ilong].getIntronTypes( vgaps );

		for (i=0; i<(int)m_isoforms.size(); ++i)
		{
			m_isoforms[i].getIntrons( viso );
			if (!viso.empty()) vtrons.insert( vtrons.end(), viso.begin(), viso.end() );
		}

		vtrons.insert( vtrons.end(), vgaps.begin(), vgaps.end() );
	}



	std::sort( vtrons.begin(), vtrons.end() );
	vector<cInterval>::iterator it = std::unique( vtrons.begin(), vtrons.end() );
	vtrons.resize( it - vtrons.begin() );

	return vtrons;
}

vector<cRichInterval> cGene::getRichIntrons() const
{
	vector<cRichInterval> vtrons, vgaps, viso;

	int i;
	int ilong;
	getLongestCds(ilong);

	//for (i=0; i<(int)m_aligns.size(); ++i)
	//{
	//	m_aligns[i].getIntrons( viso );
	//	vgaps.insert( vgaps.end(), viso.begin(), viso.end() );
	//}
	//std::sort( vgaps.begin(), vgaps.end() );
	//vector<cInterval>::iterator itg = std::unique( vgaps.begin(), vgaps.end() );
	//vgaps.resize( itg - vgaps.begin() );

	// assign a type to each alignment gap, where possible
	if (!m_isoforms.empty())
	{
		//m_isoforms[ilong].getIntronTypes( vgaps );

		for (i=0; i<(int)m_isoforms.size(); ++i)
		{
			m_isoforms[i].getIntrons( viso );
			if (!viso.empty()) vtrons.insert( vtrons.end(), viso.begin(), viso.end() );
		}

		vtrons.insert( vtrons.end(), vgaps.begin(), vgaps.end() );
	}

	std::sort( vtrons.begin(), vtrons.end() );
	vector<cRichInterval>::iterator it = std::unique( vtrons.begin(), vtrons.end() );
	vtrons.resize( it - vtrons.begin() );

	return vtrons;
}

void cGene::writeGff( std::ostream& os )
{
	// gene line
	os << m_strSeq << "\t"
		<< PROG_NAME << "\t"
		<< "gene" << "\t"
		<< m_ibeg << "\t"
		<< m_iend << "\t"
		<< "." << "\t"
		<< getStrand() << "\t"
		<< "." << "\t"
		<< "ID=" << m_strId << ";" << std::endl;

	// isoforms
	for (int i=0; i<(int)m_isoforms.size(); ++i)
		m_isoforms[i].writeGff( os, m_strSeq );
}

int cGene::getLongestTrans() const
{
	int imax;
	return getLongestTrans(imax);
}

int cGene::getLongestCds() const
{
	int imax;
	return getLongestCds(imax);
}

int cGene::getLongestIntrons() const
{
	int imax;
	int dc = getLongestTrans(imax);

	if (dc == -1)
		return -1;

	return m_isoforms[imax].lengthIntrons();
}

int cGene::getLongestTrans(int& ixMax) const
{
	int len, max(-1);
	ixMax = -1;

	for (int i=0; i<(int)m_isoforms.size(); ++i)
	{
		len = m_isoforms[i].lengthTrans();
		if (len > max) {
			max = len;
			ixMax = i;
		}
	}
	return max;
}

int cGene::getLongestCds(int& ixMax) const
{
	int len, max(-1);
	ixMax = -1;

	for (int i=0; i<(int)m_isoforms.size(); ++i)
	{
		len = m_isoforms[i].lengthCoding();
		if (len > max) {
			max = len;
			ixMax = i;
		}
	}
	return max;
}

int cGene::ctTransExons() const
{
	if (m_isoforms.empty())
		return -1;

	int imax;
	int len = getLongestTrans(imax);
	return m_isoforms[imax].ctExons();
}

int cGene::ctCdsExons() const
{
	if (m_isoforms.empty())
		return -1;

	int imax;
	int len = getLongestTrans(imax);
	return m_isoforms[imax].ctCds();
}

int cGene::ctIntrons() const
{
	int cte = ctTransExons();
	int cce = ctCdsExons();
	int ct = cte >= cce ? cte : cce;

	if (ct == -1)
		return ct;

	return ct - 1;
}

int cGene::lenTrans() const
{
	return getLongestTrans();
}

int cGene::lenCds() const
{
	return getLongestCds();
}

int cGene::lenGene() const
{
	int lT = lenTrans();
	int lC = lenCds();

	return lT >= lC ? lT : lC;
}

int cGene::lenIntrons() const
{
	return getLongestIntrons();
}
